package it.unimi.dsi.mg4j.index;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** This interface provides constants to be used as compression flags for the 
 * constructors of the {@link IndexReader} and {@link IndexWriter} classes.
 *
 * <P>Of course, this is an instance of the <em>constant interface antipattern</em>.
 * Blame it on Sun for having made static members access so ugly.
 *
 * @see IndexWriter
 *
 * @author Paolo Boldi 
 * @author Sebastiano Vigna 
 * @since 0.6
 */


public interface CompressionFlags {

	/** &delta; coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeDelta(int)}). */
	public static final int DELTA = 1;

	/** &gamma; coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeGamma(int)}). */
	public static final int GAMMA = 2;

	/** Golomb coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeGolomb(int,int)}). */
	public static final int GOLOMB = 3;

	/** Skewed Golomb coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeSkewedGolomb(int,int)}). */
	public static final int SKEWED_GOLOMB = 4;

	/** Arithmetic coding (see {@link it.unimi.dsi.mg4j.io.ArithmeticCoder}). */
	public static final int ARITH = 5;

	/** Interpolative coding (see {@link it.unimi.dsi.mg4j.io.InterpolativeCoding}). */
	public static final int INTERP = 6;

	/** Unary coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeUnary(int)}). */
	public static final int UNARY = 7;

	/** &zeta;<sub><var>k</var></sub> coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeZeta(int,int)}). */
	public static final int ZETA = 8;

	/** Variable-length nibble coding (see {@link it.unimi.dsi.mg4j.io.OutputBitStream#writeNibble(int)}). */
	public static final int NIBBLE = 9;

	/** A special value used to denote absence of a feature. */
	public static final int NONE = 0xFF;

	public static final String CODING_NAME[] = { "DEFAULT", "DELTA", "GAMMA", "GOLOMB", "SKEWED_GOLOMB", "ARITH", "INTERP", "UNARY", "ZETA", "NIBBLE" };

	/** The mask shift for frequency coding. */
	public static final int FREQUENCIES_SHIFT = 0;
	/** Flag: default value for frequency coding (&gamma;). */
	public static final long FREQUENCIES_DEFAULT = GAMMA << FREQUENCIES_SHIFT;
	/** Flag: write the term frequency using &gamma; coding. */
	public static final long FREQUENCIES_GAMMA = GAMMA << FREQUENCIES_SHIFT;
	/** Flag: write the term frequency using &delta; coding. */
	public static final long FREQUENCIES_DELTA = DELTA << FREQUENCIES_SHIFT;

	/** The mask shift for pointer coding. */
	public static final int POINTERS_SHIFT = 8;
	/** Flag: default value for pointer coding (Golomb). */
	public static final long POINTERS_DEFAULT = GOLOMB << POINTERS_SHIFT;
	/** Flag: write document pointers using Golomb coding with parameter {@link it.unimi.dsi.mg4j.util.Fast#golombModulus(double) Fast.golombModulus(freq/N)} (default). */
	public static final long POINTERS_GOLOMB = GOLOMB << POINTERS_SHIFT;
	/** Flag: write document pointers using &gamma; coding. */
	public static final long POINTERS_GAMMA = GAMMA << POINTERS_SHIFT;
	/** Flag: write document pointers using &delta; coding. */
	public static final long POINTERS_DELTA = DELTA << POINTERS_SHIFT;

	/** The mask shift for count coding. */
	public static final int COUNTS_SHIFT = 16;
	/** Flag: default value for count coding (&gamma;). */
	public static final long COUNTS_DEFAULT = GAMMA << COUNTS_SHIFT;
	/** Flag: write counts of occurrence lists using &gamma; coding (default). */
	public static final long COUNTS_GAMMA = GAMMA << COUNTS_SHIFT;
	/** Flag: write counts of occurrence lists using &delta; coding. */
	public static final long COUNTS_DELTA = DELTA << COUNTS_SHIFT;
	/** Flag: do not write counts (and thus, <i>a fortiori</i>, positions). */
	public static final long NO_COUNTS = NONE << COUNTS_SHIFT;

	/** The mask shift for position coding. */
	public static final int POSITIONS_SHIFT = 24;
	/** Flag: default value for position coding (Golomb). */
	public static final long POSITIONS_DEFAULT = GOLOMB << POSITIONS_SHIFT;
	/** Flag: write positions in a document using Golomb coding with parameter {@link it.unimi.dsi.mg4j.util.Fast#golombModulus(double) Fast.golombModulus(count/docSize)} (default). */
	public static final long POSITIONS_GOLOMB = GOLOMB << POSITIONS_SHIFT;
	/** Flag: write positions in a document using skewed Golomb coding with parameter {@link it.unimi.dsi.mg4j.util.Fast#golombModulus(double) Fast.golombModulus(median/docSize)}. */
	public static final long POSITIONS_SKEWED_GOLOMB = SKEWED_GOLOMB << POSITIONS_SHIFT;
	/** Flag: write positions in a document using &gamma; coding. */
	public static final long POSITIONS_GAMMA = GAMMA << POSITIONS_SHIFT;
	/** Flag: write positions in a document using &delta; coding. */
	public static final long POSITIONS_DELTA = DELTA << POSITIONS_SHIFT;
	/** Flag: write positions in a document using arithmetic coding. */
	public static final long POSITIONS_ARITH = ARITH << POSITIONS_SHIFT;
	/** Flag: write positions in a document using interpolative coding. */
	public static final long POSITIONS_INTERP = INTERP << POSITIONS_SHIFT;
	/** Flag: do not write positions. */
	public static final long NO_POSITIONS = NONE << POSITIONS_SHIFT;
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:

