package it.unimi.dsi.mg4j.index;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.Iterators;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.longs.LongArrayList;
import it.unimi.dsi.fastutil.longs.LongList;
import it.unimi.dsi.fastutil.objects.ObjectSets;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.search.IntervalIterator;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Properties;
import java.util.Set;


/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** An abstract representation of an index.
 *
 * <P>An instance of this class stores index data such as the basename, 
 * flags, etc. It allows to build easily {@linkplain IndexReader index readers}
 * and {@linkplain it.unimi.dsi.mg4j.search.DocumentIterator document iterators}
 * over the given index.
 * 
 * <P>This class provides also a {@linkplain #main(String[]) main method}
 * that can be used to dump data about an index for diagnostic purposes. 
 * 
 * @author Paolo Boldi
 * @author Sebastiano Vigna
 * @since 0.9
 */

// TODO: QueryParser should accept a Tokenizer to split words

public class Index implements CompressionFlags {

	/** The basename of this index. All file names will be stemmed from the basename. It may
	 * be <code>null</code> if this index has been built using {@link #Index(String, File, Properties, int, int, int, boolean, long, TermMap, LongList, IntList)}. */
	public final String basename;
	/** The properties of this index. It may be <code>null</code> if this index has been built using
	 * {@link #Index(String, File, Properties, int, int, int, boolean, long, TermMap, LongList, IntList)}. */
	public final Properties properties;
	/** The file containing the index. It may be <code>null</code> if this index has been built using
	 * {@link #Index(String, File, Properties, int, int, int, boolean, long, TermMap, LongList, IntList)}. */
	public final File indexFile;
	/** The number of documents of the collection. */
	public final int numberOfDocuments;
	/** The number of terms of the collection. */
	public final int numberOfTerms;
	/** The maximum number of positions in an position list, or -1 if it is not known. */
	public final int maxDocPos;
	/** The coding for frequencies. */
	public final int frequencyCoding;
	/** The coding for pointers. */
	public final int pointerCoding;
	/** The coding for counts. */
	public final int countCoding;
	/** The coding for positions. */
	public final int positionCoding;
	/** Whether this index contains counts. */
	public final boolean hasCounts;
	/** Whether this index contains positions. */
	public final boolean hasPositions;
	/** Whether this index is case sensitive. */
	public final boolean isCaseSensitive;
	/** A singleton set containing just this index. */
	public final Set singletonSet;
	/** The offset of each term, if offsets were loaded or specified at creation time, or <code>null</code>. */
	public final LongList offsets;
	/** The size of each document, or <code>null</code> if sizes are not necessary in this index. */
	public final IntList sizes;
	/** The term list for this index, or <code>null</code> if the term list was not loaded. */
	public final TermMap termMap;

	/** Produces a flag mask corresponding to a given string.
	 * 
	 * @param flagString a flag string.
	 * @return the flag mask.
	 * @throws IOException if the flag string is malformed.
	 */
	private static long string2Flags( String flagString ) throws IOException {
		long flags = 0;
		
		if ( flagString.length() != 0 ) {
			String f[] = flagString.split( "\\|" );
			for( int i = 0; i < f.length; i++ ) {
				try {
					flags |= Index.class.getField( f[ i ].trim() ).getLong( Index.class );
				}
				catch ( Exception notFound ) {
					throw new IOException( "Compression flag " + f[ i ] + " unknown." );
				}
			}
		}
		return flags;
	}

	/** Utility method to load a compressed offset file into a list.
	 *
	 * @param in the input bit stream providing the offsets (see {@link IndexWriter}).
	 * @param T the number of terms indexed.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 * @return a list of longs backed by an array; the list has
	 * an additional final element of index <code>T</code> that gives the number
	 * of bytes of the index file.
	 */

	public static LongList readOffsets( final InputBitStream in, final int T, final ProgressMeter pm ) throws java.io.IOException {
		final long offset[] = new long[ T + 1 ];
		offset[ 0 ] = in.readGamma();
		if ( pm != null ) {
			pm.expectedUpdates( T );
			pm.start( "Reading offsets..." );
			for( int i = 0; i < T; i++ ) {
				pm.update();
				offset[ i + 1 ] = in.readGamma() + offset[ i ];
			} 
			pm.done();
		}
		else for( int i = 0; i < T; i++ ) offset[ i + 1 ] = in.readGamma() + offset[ i ];
			
		return LongArrayList.wrap( offset );
	}

	/** Utility method to load a compressed size file into a list.
	 *
	 * @param in the input bit stream providing the offsets (see {@link IndexWriter}).
	 * @param N the number of documents indexed.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 * @return a list of integers backed by an array.
	 */

	public static IntList readSizes( final InputBitStream in, final int N, final ProgressMeter pm ) throws java.io.IOException {
		final int size[] = new int[ N ];
		if ( pm != null ) {
			pm.start( "Reading sizes..." ); 
			pm.expectedUpdates( N );
			for( int i = 0; i < N; i++ ) {
				pm.update();
				size[ i ] = in.readGamma();
			}
			pm.done();
		}
		else for( int i = 0; i < N; i++ ) size[ i ] = in.readGamma();
			  
		return IntArrayList.wrap( size );
	}

	/** Utility static method that loads a term map.
	 * 
	 * @param filename the name of the file containing the term map.
	 * @return the map, or <code>null</code> if the file did not exist.
	 * @throws IOException if some IOException (other than {@link FileNotFoundException}) occurred.
	 */
	public static TermMap loadTermMap( String filename ) throws IOException {
		try {
			ObjectInputStream ois = new ObjectInputStream( new FileInputStream( filename ) );
			TermMap result = (TermMap) ois.readObject();
			ois.close();
			return result;
		} catch ( FileNotFoundException e ) {
			return null;
		} catch ( ClassNotFoundException e ) {
			throw new RuntimeException( e );
		}
	}


	/** Creates a new index using the given basename.
	 * 
	 * TODO: describe what files are needed, and spend some time on properties also!
	 * 
	 * @param basename the basename of the index.
	 * @param termMap the term list for this index, or <code>null</code> if there is no term list.
	 * @param loadOffsets offsets are loaded only if this parameter is <code>true</code>.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 */
	protected Index( final CharSequence basename, final TermMap termMap, final boolean loadOffsets, final ProgressMeter pm ) throws IOException {
		
		this.basename = basename.toString();
		this.termMap = termMap;
		this.singletonSet = ObjectSets.singleton( this );
		
		indexFile = new File( this.basename + ".index" );
		if ( ! indexFile.exists() ) throw new FileNotFoundException( "Cannot find index file " + indexFile.getName() );

		String propertyFilename = this.basename + ".properties"; 
		if ( ! ( new File( propertyFilename ) ).exists() ) throw new FileNotFoundException( "Cannot open property file " + propertyFilename );	
		
		final FileInputStream propertyFile = new FileInputStream( propertyFilename );
		properties = new Properties();
		properties.load( propertyFile );
		propertyFile.close();

		numberOfDocuments = Integer.parseInt( properties.getProperty( IndexProperties.DOCUMENTS ) );
		numberOfTerms = Integer.parseInt( properties.getProperty( IndexProperties.TERMS ) );
		maxDocPos = Integer.parseInt( properties.getProperty( IndexProperties.MAXDOCPOS ) );

		long flags = string2Flags( properties.getProperty( IndexProperties.COMPRESSIONFLAGS ) );
		
		frequencyCoding = (int) (( ( flags >>> FREQUENCIES_SHIFT & 0xFF ) != 0 ? flags : FREQUENCIES_DEFAULT ) >>> FREQUENCIES_SHIFT & 0xFF);
		pointerCoding = (int) (( ( flags >>> POINTERS_SHIFT & 0xFF ) != 0 ? flags  : POINTERS_DEFAULT ) >>> POINTERS_SHIFT & 0xFF);
		countCoding = (int) (( ( flags >>> COUNTS_SHIFT & 0xFF ) != 0 ? flags : COUNTS_DEFAULT ) >>> COUNTS_SHIFT & 0xFF);
		hasCounts = countCoding != NONE;

		if ( ! hasCounts  )	{
			positionCoding = (int) (( ( flags >>> POSITIONS_SHIFT & 0xFF ) != 0 ? flags : NO_POSITIONS ) >>> POSITIONS_SHIFT & 0xFF);
			if ( positionCoding != NONE ) throw new IllegalArgumentException( "Index " + this.basename + " has positions but no counts (this can't happen)" );
		}
		else positionCoding = (int) (( ( flags >>> POSITIONS_SHIFT & 0xFF ) != 0 ? flags : POSITIONS_DEFAULT ) >>> POSITIONS_SHIFT & 0xFF);
		
		hasPositions = positionCoding != NONE;
		if ( ! hasCounts && hasPositions ) throw new IllegalArgumentException( "Index " + this.basename + " has positions but no counts (this can't happen)" );

		isCaseSensitive = Boolean.valueOf( properties.getProperty( IndexProperties.ISCASESENSITIVE ) ).booleanValue();

		if ( positionCoding == GOLOMB || positionCoding == INTERP ) sizes =  readSizes( new InputBitStream( basename + ".sizes" ), numberOfDocuments, pm );
		else sizes = null;
		
		if ( loadOffsets ) offsets = (LongList) readOffsets( new InputBitStream( basename + ".offsets" ), numberOfTerms, pm );
		else offsets = null;
	}

	
	/** Creates a new index using the given data.
	 * 
	 * <P>This constructor provides an index that is initialised exactly to the provided data.
	 * It is mainly useful for debugging and testing purposes, in the case you are creating
	 * an index file (for instance, as a memory-based bit stream) but you have no property
	 * file to use with {@link #getInstance(CharSequence)}.
	 *
	 * <P>It is usually safe to provide <code>null</code> for <code>basename</code>, <code>indexFile</code>
	 * and <code>properties</code>, but the responsibility for data consistence is up to the caller.
	 *  
	 * @param basename the basename of this index, or <code>null</code>.
	 * @param indexFile the file containing this index, or <code>null</code>.
	 * @param properties the properties of this index, or <code>null</code>.
	 * @param numberOfDocuments the number of documents in this index.
	 * @param numberOfTerms the number of terms in this index.
	 * @param maxDocPos the maximum length of an occurrence list, or -1 if it is not known.
	 * @param flags a bit mask setting the coding techniques to be used (see {@link CompressionFlags}).
	 * @param termMap the term list for this index, or <code>null</code> if there is no term list.
	 * @param offsets the offset list; may be <code>null</code> if you do not plan using {@link IndexReader#position(int)}.
	 * @param sizes the size list; may be <code>null</code> if your code does not require it.
	 */
	public Index( String basename, File indexFile, Properties properties, int numberOfDocuments,
		int numberOfTerms, int maxDocPos, boolean isCaseSensitive, long flags, TermMap termMap, LongList offsets, IntList sizes ) throws java.io.IOException {
		
		this.basename = basename;
		this.indexFile = indexFile;
		this.properties = properties;
		this.numberOfDocuments = numberOfDocuments;
		this.numberOfTerms = numberOfTerms;
		this.maxDocPos = maxDocPos;
		this.isCaseSensitive = isCaseSensitive;
		this.sizes = sizes;
		this.termMap = termMap;
		this.offsets = offsets;
		this.singletonSet = ObjectSets.singleton( this );
		
		frequencyCoding = (int) (( ( flags >>> FREQUENCIES_SHIFT & 0xFF ) != 0 ? flags : FREQUENCIES_DEFAULT ) >>> FREQUENCIES_SHIFT & 0xFF);
		pointerCoding = (int) (( ( flags >>> POINTERS_SHIFT & 0xFF ) != 0 ? flags  : POINTERS_DEFAULT ) >>> POINTERS_SHIFT & 0xFF);
		countCoding = (int) (( ( flags >>> COUNTS_SHIFT & 0xFF ) != 0 ? flags : COUNTS_DEFAULT ) >>> COUNTS_SHIFT & 0xFF);
		hasCounts = countCoding != NONE;

		if ( ! hasCounts  )	{
			positionCoding = (int) (( ( flags >>> POSITIONS_SHIFT & 0xFF ) != 0 ? flags : NO_POSITIONS ) >>> POSITIONS_SHIFT & 0xFF);
			if ( positionCoding != NONE ) throw new IllegalArgumentException( "Index " + this.basename + " has positions but no counts (this can't happen)" );
		}
		else positionCoding = (int) (( ( flags >>> POSITIONS_SHIFT & 0xFF ) != 0 ? flags : POSITIONS_DEFAULT ) >>> POSITIONS_SHIFT & 0xFF);
		
		hasPositions = positionCoding != NONE;
		if ( ! hasCounts && hasPositions ) throw new IllegalArgumentException( "Index " + this.basename + " has positions but no counts (this can't happen)" );
	}

	/** Creates a new index using the given basename, loading offsets.
	 * 
	 * @param basename the basename of the index.
	 * @param termMap the term list for this index, or <code>null</code> if there is no term list.
	 * @param loadOffsets offsets are loaded only if this parameter is <code>true</code>.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 */
	public static Index getInstance( final CharSequence basename, final TermMap termMap, final boolean loadOffsets, final ProgressMeter pm ) throws IOException {
		String propertyFilename = basename + ".properties"; 
		if ( ! ( new File( propertyFilename ) ).exists() ) throw new FileNotFoundException( "Cannot open property file " + propertyFilename );	
		
		final FileInputStream propertyFile = new FileInputStream( propertyFilename );
		final Properties properties = new Properties();
		properties.load( propertyFile );
		propertyFile.close();
		
		Constructor constructor;
		try {
			constructor =
				Class.forName( properties.getProperty( IndexProperties.INDEXCLASS, "unspecified" )).getDeclaredConstructor(
					new Class[] { CharSequence.class, TermMap.class, boolean.class, ProgressMeter.class });

			return (Index)constructor.newInstance( new Object[] { basename, termMap, Boolean.valueOf( loadOffsets ), pm } );
		} catch ( InvocationTargetException e ) {
			throw new RuntimeException( e.getCause() );
		} catch ( Exception e ) {
			System.err.println( "The Index factory was not able to locate a suitable constructor for " +
				"the index class " + properties.getProperty( "indexclass", "unspecified" ) );
			throw new RuntimeException( e );
		}
	}

	/** Creates a new index using the given basename, loading offsets. If there is a term map file (basename stemmed
	 *  with ".termmap"), it is loaded; otherwise, the term map is set to <code>null</code>.
	 * 
	 * @param basename the basename of the index.
	 * @param loadOffsets offsets are loaded only if this parameter is <code>true</code>.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 */
	public static Index getInstance( final CharSequence basename, final boolean loadOffsets, final ProgressMeter pm ) throws IOException {
		return getInstance( basename, loadTermMap( basename + ".termmap" ), loadOffsets, pm );
	}


	/** Creates a new index using the given basename, loading offsets. If there is a term map file (basename stemmed
	 *  with ".termmap"), it is loaded; otherwise, the term map is set to <code>null</code>.
	 * 
	 * @param basename the basename of the index.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 */
	public static Index getInstance( final CharSequence basename, final ProgressMeter pm ) throws IOException {
		return getInstance( basename, loadTermMap( basename + ".termmap" ), true, pm );
	}

	/** Creates a new index using the given basename. If there is a term map file (basename stemmed
	 *  with ".termmap"), it is loaded; otherwise, the term map is set to <code>null</code>.
	 * 
	 * <P>This constructor provides no progress report.
	 * 
	 * @param basename the basename of the index.
	 */
	public static Index getInstance( final CharSequence basename ) throws IOException {
		return getInstance( basename, loadTermMap( basename + ".termmap" ), true, null );
	}


	/** Creates and returns a new {@link IndexReader} based on this index. After that, you
	 *  can use the reader to read this index.
	 * 
	 * @return a new {@link IndexReader} to read this index.
	 */
	public IndexReader getReader() throws IOException {
		return getReader( InputBitStream.DEFAULT_BUFFER_SIZE );
	}

	/** Creates and returns a new {@link IndexReader} based on this index. After that, you
	 *  can use the reader to read this index.
	 * 
	 * @param bufferSize the size of the buffer to be used when opening the {@link InputBitStream} underlying the
	 * {@link IndexReader} that is going to be returned.
	 * @return a new {@link IndexReader} to read this index.
	 */
	public IndexReader getReader( int bufferSize ) throws IOException {
		return new IndexReader( this, bufferSize );
	}

	/** An iterator returning no documents based on this index. */
	private class EmptyDocumentIterator extends Iterators.EmptyIterator implements IndexIterator {
		public int document() { return -1; }
		public Set indices() { return Index.this.singletonSet; }
		public IntervalIterator intervalIterator() {throw new IllegalStateException(); }
		public Map intervalIterators() { throw new IllegalStateException(); }
		public IntervalIterator intervalIterator( Index index ) { throw new IllegalStateException(); }
		public int nextDocument() { throw new NoSuchElementException(); }
		public int skipTo( int n ) { return -1; }
		public int frequency() { throw new IllegalStateException(); }
		public int count() { throw new IllegalStateException(); }
		public IntIterator positions() { throw new IllegalStateException(); }
		public int positions(int[] positions) { throw new IllegalStateException(); }
		public int[] positionArray() { throw new IllegalStateException(); }
	}

	/** A singleton for an iterator returning no documents based on this index. */
	public final EmptyDocumentIterator emptyDocumentIterator = new EmptyDocumentIterator(); 

	
	public String toString() {
		return basename;
	}

	public static void main( String arg[] ) throws IOException {
		int ch;
		String basename;

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
		};

		Getopt g = new Getopt( "Index", arg, "hos", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: Index [OPTIONS] BASENAME" );
			System.err.println( "Loads an index." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  BASENAME    all file names will be stemmed from this name" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help        print this help screen" );
			System.err.println( "" );
			return;
		}
		
		if ( arg.length - g.getOptind() == 1 ) basename = arg[ g.getOptind() ];
		else {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}

		Index index = getInstance( basename, new ProgressMeter( 100000, "ints", System.out ) );
		
		System.out.println(" Index class:      " + index.getClass() );
		System.out.println( "Frequency coding: " + ( CODING_NAME[ index.frequencyCoding ] ) );
		System.out.println( "Pointer coding:   " + ( CODING_NAME[ index.pointerCoding ] ) );
		System.out.println( "Count coding:     " + ( index.countCoding == NONE ? "Not stored" : CODING_NAME[ index.countCoding ] ) );
		System.out.println( "Position coding:  " + ( index.positionCoding == NONE ? "Not stored" : CODING_NAME[ index.positionCoding ] ) );
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
