package it.unimi.dsi.mg4j.index;

import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.mg4j.search.DocumentIterator;

/** An iterator over an inverted list.
 *
 * <P>An index iterator scans the inverted list of an indexed term. Each
 * integer returned by the iterator is the index of a document containing the
 * term. If the index contains counts, they can be obtained after each call to
 * {@link #nextInt()} using {@link #count()}. Then, if the index contains
 * occurrences they can be obtained as an itearator using {@link
 * #positions()}, or stored into an array using {@link
 * #positions(int[])}.
 *
 * <P>Note that this interface extends {@link it.unimi.dsi.mg4j.search.DocumentIterator}.
 * The intervals returned for a document are exactly length-one intervals
 * corresponding to the positions returned by {@link #positions()}.
 * 
 */

public interface IndexIterator extends DocumentIterator {


	/** Returns the frequency, that is, the number of documents that will be returned by this iterator.
	 *
	 * @return the number of documents that will be returned by this iterator.
	 */

	public int frequency();

	/** Returns the count, that is, the number of occurrences of the term in the current document.
	 *
	 * @return the count (number of occurrences) of the term in the current document.
	 * @throws UnsupportedOperationException if this index does not contain counts.
	 */

	public int count();

	/** Returns the positions at which the term appears in the current document. 
	 *
	 * @return the positions of the current document in which the current term appears.
	 * @throws UnsupportedOperationException if this index does not contain positions.
	 */
	public IntIterator positions();

	/** Stores the positions at which the term appears in the current document in a given array. 
	 *
	 * <P>If the array is not large enough (i.e., it does not contain {@link #count()} elements),
	 * this method will return a negative number (the opposite of the count).
	 *
	 * @param positions an array that will be used to store positions.
	 * @return the {@linkplain #count() count}; it will have the sign changed if <code>positions</code> cannot
	 * hold all positions.
	 * @throws UnsupportedOperationException if this index does not contain positions.
	 */
	public int positions( int[] positions );

	/** Returns the positions at which the term appears in the current document in an array. 
	 *
	 * <P>Implementations are allowed to return the same array across different calls to this method.
	 *
	 * @return an array whose first {@linkplain #count()} elements contain the document positions.
	 * @throws UnsupportedOperationException if this index does not contain positions.
	 */
	public int[] positionArray();
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
