package it.unimi.dsi.mg4j.index;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntIterators;
import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.longs.LongList;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.mg4j.io.ArithmeticDecoder;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.InterpolativeCoding;
import it.unimi.dsi.mg4j.search.DocumentIterator;
import it.unimi.dsi.mg4j.search.Interval;
import it.unimi.dsi.mg4j.search.IntervalIterator;
import it.unimi.dsi.mg4j.search.IntervalIterators;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;

import java.io.IOException;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;


/** Provides facilities to read directly an inverted index.
 *
 *  <P>An inverted index is made by a sequence of inverted lists (one inverted list
 *  for each term). Inverted lists are made by document records. For more information
 *  on the structure of an inverted index, see the documentation of {@link it.unimi.dsi.mg4j.index}.
 * 
 *  <P>Note that {@link it.unimi.dsi.mg4j.index.Index} provides a much easier interface to
 * index access. Instead of creating an {@link IndexReader} explicitly, you can use {@link Index#getReader()}.
 * 
 *  <H2>Class usage</H2>
 *
 *  <P>The simplest way of getting data from an <code>IndexReader</code> is to ask for
 *  the {@linkplain #documents(CharSequence) documents matching a term}.
 *  
 *  <P>Otherwise, you can access the index directly. 
 *  The method {@link #readFrequency()} should be called only when the underlying
 *  stream is positioned on the start of an inverted list; in such case, it returns
 *  the frequency of the corresponding term.
 *  After that, you can read the document records sequentially: you have to read
 *  exactly as many document records as specified in the frequency, at the beginning of the list.
 *
 *  <P>For each document record, you call (possibly: many times) {@link #readDocumentPointer()} to read the document
 *  pointer, and then {@link #readDocumentPositions(int[])} (just once) to read the
 *  positions. After reading the last record, you will be positioned on the beginning of
 *  the next inverted list, if any.
 *
 *  <P>The {@link #skipTo(int)} method can be called either after or before
 *  reading a document pointer, and allows one to skip ahead in the inverted list.
 *  More precisely, assume that you are currently on a record with pointer <code>p</code>,
 *  and you call {@link #skipTo(int) skipTo(n)}; if there is a document record with pointer
 *  &ge; <code>n</code> in the current inverted list, you will be positioned on the
 *  beginning of the first such record following the current one (in particular, if
 *  <code>p</code>&ge;<code>n</code>, the reader will not move); if there is no 
 *  document record satisfying the condition, you will be positioned on the beginning
 *  of the next inverted list (or at the end of the index, if this is the last
 *  inverted list). 
 *
 *  <H2>Using offsets to position the reader</H2>
 *
 *  <P>Hence, as explained above, the index is read sequentially, one inverted list
 *  after another, one document record after another. The {@link #skipTo(int)} method
 *  can be used to jump ahead (but not backwards!) to another document record
 *  inside a single inverted list.
 *  
 *  <P>As explained in the {@link IndexWriter} documentation, an inverted index may
 *  be created with an associated offset bit stream: the offset bit stream stores
 *  the positions where each inverted list starts inside the index. You can read
 *  an offset bit stream using the {@link Index#readOffsets(InputBitStream, int, ProgressMeter)} static
 *  method: this method reads the offset bit stream and returns its content as
 *  a list.
 *
 *  <P>Such a list may be passed to the {@link IndexReader} constructor: in this case,
 *  you can use the {@link #position(int) position(t)} method to position the index reader on
 *  the beginning of the inverted list relative to a the <code>t</code>-th term.
 *  In order to do so, two conditions must be satisfied.
 *  <UL>
 *    <LI>the {@link IndexReader} must have been created with an associated offset map;
 *    <LI>the underlying bit stream must be {@link InputBitStream#position(long) positionable}.
 *  </UL>
 *  The {@link #position(int)} method may be called at any time.
 *
 *  <H2>Implementation details</H2>
 * 
 *  <P>After calling {@link #readFrequency()}, the reader is at the beginning of the
 *  first document record. We assume that this is an invariant: the reader is always
 *  at the beginning of a document record; the variable {@link #numberOfDocumentRecord} 
 *  keeps track of the number of the document record: this ranges from 0 to the
 *  frequency (when equal to the frequency, we are not really at the beginning of
 *  a document record but rather at the beginning of the next inverted list, or
 *  at the end of the file).
 *
 * @author Paolo Boldi 
 * @author Sebastiano Vigna 
 * @since 0.6
 */


public class IndexReader implements CompressionFlags {
	/** The reference index. */
	final protected Index index;
	/** The underlying input bit stream. */
	protected InputBitStream ibs;
	/** The parameter <code>b</code> for Golomb coding of pointers. */
	protected int b;
	/** The parameter <code>log2b</code> for Golomb coding of pointers; it is the most significant bit of {@link #b}. */
	protected int log2b;
	/** The current term. */
	protected int term;
	/** The current frequency. */
	protected int frequency;
	/** The current count (if this index contains counts). */
	protected int count;
	/** The number of document records that the current inverted list will contain, divided by the number of documents. */
	protected double relativeFrequency;
	/** The last document pointer in the current list, or -1 if we just read the frequency. */
	protected int currentDocument;
	/** The number of the document record we are going to read inside the current inverted list. */
	protected int numberOfDocumentRecord;
	/** This variable tracks the current state of the reader. */
	protected int state;

	/** This value of {@link #state} means that we are positioned at the start of an inverted list (just before the frequency).*/
	protected static final int BEFORE_FREQUENCY = 0;

	/** This value of {@link #state} means that we are at the start of a new document record, 
	 * unless we already read all documents (i.e., {@link #numberOfDocumentRecord} == {@link #frequency}),
	 * in which case we are at the end of the inverted list, and {@link #endOfList()} is true. */
	protected static final int BEFORE_POINTER = 1;

	/** This value of {@link #state} can be assumed only in indices that contain counts; it
	 * means that we are positioned just before the count for the current document record. */
	protected static final int BEFORE_COUNT = 2;

	/** This value of {@link #state} can be assumed only in indices that contain document positions; 
	 * it means that we are positioned just before the position list of the current document record. */
	protected static final int BEFORE_POSITIONS = 3;

	/** This is the first unused state. Subclasses may start from this value to define new states. */
	protected static final int FIRST_UNUSED_STATE = 4;

	private final static boolean ASSERTS = true;
	//private final static boolean DEBUG = false;

	/** Creates a new index reader, with the specified underlying {@link Index} and default buffer size.
	 *
	 * @param index the index.
	 * @see #IndexReader(Index, int)
	 */
	public IndexReader( final Index index ) throws java.io.IOException {
		this( index, InputBitStream.DEFAULT_BUFFER_SIZE );
	}

	/** Creates a new index reader, with the specified underlying {@link Index}.
	 *
	 * @param index the index.
	 * @param bufferSize the size of the buffer of the underlying input bit stream.
	 */
	public IndexReader( final Index index, final int bufferSize ) throws java.io.IOException {
		this( index, new InputBitStream( index.indexFile, bufferSize ) );
	}

	/** Creates a new index reader, with the specified underlying {@link Index} and input bit stream.
	 *
	 * @param index the index.
	 * @param ibs the underlying bit stream.
	 */
	public IndexReader( final Index index, final InputBitStream ibs ) throws java.io.IOException {
		this.index = index;
		this.ibs = ibs;
	}

	/** Creates a new index reader, with the specified underlying {@link InputBitStream}.
	 *
	 * @param ibs the underlying input bit stream.
	 * @param offsets the offset list; may be <code>null</code> if you do not plan using {@link #position(int)}.
	 * @param sizes the size list; may be <code>null</code> if your code does not require it.
	 * @param N the number of documents in the collection.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexReader} introduction).
	 * @deprecated Please use {@link Index#getReader()} or {@link IndexReader#IndexReader(Index)}.
	 */
	public IndexReader( InputBitStream ibs, LongList offsets, IntList sizes, int N, long flags ) throws java.io.IOException {
		this( new Index( null, null, null, N, -1, -1, true, flags, null, offsets, sizes ) );
		this.ibs = ibs;
		this.frequency = -1;
		this.state = BEFORE_FREQUENCY;

		if ( index.positionCoding == POSITIONS_ARITH ) ibs.overflow( true ); // Sets overflow if positions must be coded using an arithmetic coder
	}

	/** Creates a new index reader, with the specified underlying {@link InputBitStream}.
	 *
	 * @param ibs the underlying input bit stream.
	 * @param N the number of documents in the collection.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexReader} introduction).
	 * @deprecated Please use {@link Index#getReader()} or {@link IndexReader#IndexReader(Index)}.
	 */
	public IndexReader( InputBitStream ibs, LongList offsets, int N, long flags ) throws java.io.IOException {
		this( ibs, offsets, null, N, flags );
	}

	/** Creates a new index reader, with the specified underlying {@link InputBitStream}.
	 *
	 * @param ibs the underlying input bit stream.
	 * @param N the number of documents in the collection.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexReader} introduction).
	 * @deprecated Please use {@link Index#getReader()} or {@link IndexReader#IndexReader(Index)}.
	 */
	public IndexReader( InputBitStream ibs, int N, long flags ) throws java.io.IOException {
		this( ibs, null, null, N, flags );
	}

	

	/** Positions the index on the inverted list of a given term.
	 *
	 * This method can be called at any time.
	 *
	 * @param term a term.
	 */

	public void position( int term ) throws java.io.IOException {
		if ( index.offsets == null ) throw new IllegalStateException( "You cannot position an index without offsets" );

		ibs.position( index.offsets.getLong( term ) << 3 );
		ibs.readBits( 0 );
		this.term = term;
		state = BEFORE_FREQUENCY;
	}


	/** Returns whether there are no more document records in the current inverted list.
	 *
	 * @return whether there are no more  document records in the current inverted list.
	 */
	public boolean endOfList() {
		return numberOfDocumentRecord >= frequency - 1; // TODO: won't work at present
	}


	/** Reads and returns the frequency (number of documents where the term appears).
	 *  The underlying bit stream must be positioned on the beginning of an inverted
	 *  list, or at the real end of an inverted list (not only {@link #endOfList()} must
	 *  return true, but also all data contained in the last document record must
	 *  have been read).
	 * 
	 *  <P>After that, you are ready to read the document records.
	 *
	 *  @return the frequency.
	 *  @throws IOException if an exception is thrown by the underlying stream
	 */
	public int readFrequency() throws java.io.IOException, IllegalStateException {
		if ( state == BEFORE_POINTER && endOfList() ) { // This guarantees that it is possible to do sequential scans.
			if ( ++term >= index.numberOfTerms ) throw new IllegalStateException( "Trying to read beyond last inverted list" );
			state = BEFORE_FREQUENCY;
			ibs.align();
		}
		
		if ( state != BEFORE_FREQUENCY ) throw new IllegalStateException( "Trying to read frequency in state " + state );

		// Read the frequency
		switch( index.frequencyCoding ) {
		case GAMMA:
			frequency = ibs.readGamma() + 1; 
			break;
		case DELTA:
			frequency = ibs.readDelta() + 1;
			break;
		default:
			throw new IllegalStateException( "The required frequency coding (" + index.frequencyCoding + ") is not supported." );
		}
		
		relativeFrequency = (double)frequency / index.numberOfDocuments;

		// We compute the modulus used for pointer Golomb coding 
		if ( index.pointerCoding == GOLOMB ) {
			this.b = Fast.golombModulus( frequency / (double)index.numberOfDocuments );
			this.log2b = Fast.mostSignificantBit( b );
		}

		count = -1;
		currentDocument = -1;
		numberOfDocumentRecord = -1;
		state = BEFORE_POINTER;

		return frequency;
	}


	/** Reads a document pointer from a given {@link InputBitStream}. This
	 *  method returns the document pointer; after this, you can use the stream
	 *  to read the positions. 
	 *
	 *  @throws IOException if an exception is thrown by the underlying stream
	 */
	public int readDocumentPointer() throws java.io.IOException {
		if ( state == BEFORE_FREQUENCY ) throw new IllegalStateException( "readDocumentPointer() called before having read frequency" );
		if ( endOfList() ) throw new NoSuchElementException( "readDocumentPointer() called beyond end of list" );

		if ( index.hasCounts && state == BEFORE_COUNT ) readPositionCount();
		if ( index.hasPositions && state == BEFORE_POSITIONS ) readDocumentPositions( null );
		
		if ( state != BEFORE_POINTER ) throw new IllegalStateException( "readDocumentPointer() called in unknown state " + state );

		if ( frequency != index.numberOfDocuments ) {// We do not write pointers for everywhere occurring terms.
			switch( index.pointerCoding ) {
			case GAMMA:
				currentDocument += ibs.readGamma() + 1;
				break;
			case DELTA:
				currentDocument += ibs.readDelta() + 1;
				break;
			case GOLOMB:
				currentDocument += ibs.readGolomb( b, log2b ) + 1;
				break;
			default:
				throw new IllegalStateException( "The required pointer coding (" + index.pointerCoding + ") is not supported." );
			}
		}
		else currentDocument++;
		numberOfDocumentRecord++;
		
		if ( index.hasCounts ) {
			state = BEFORE_COUNT;
			count = -1;
		} 
		else state = BEFORE_POINTER;
		
		return currentDocument;
	}

	/** Returns the current document pointer. 
	 * @return the current document pointer, or -1 if no document pointer has been read so far.
 	 */
	
	public int currentDocumentPointer() {
		return currentDocument;
	}
	
	/** Returns true if the index is positioned just after a pointer.
	 * 
	 * @return true if the index is positioned just after a pointer.
	 */
	protected boolean justAfterPointer() {
		return ( index.hasCounts && state == BEFORE_COUNT || ! index.hasCounts && state == BEFORE_POINTER );
	}

	
	/** Reads the position count for the current document.
	 *
	 *  @return the count, that is, the number of positions of the current term in the current document.
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if you try to read a count but you have not just read a pointer.
	 *  @throws NoSuchElementException if you try to read beyond the end of an inverted list.
	 */

	public int readPositionCount() throws java.io.IOException, NoSuchElementException {
		if ( state != BEFORE_COUNT ) throw new IllegalStateException( "Trying to call readPositionCount() in state " + state );

		state = index.hasPositions ? BEFORE_POSITIONS : BEFORE_POINTER;

		switch( index.countCoding ) {
		case GAMMA: return count = ibs.readGamma() + 1;
		case DELTA: return count = ibs.readDelta() + 1;
		default: throw new IllegalStateException( "The required count coding (" + index.countCoding + ") is not supported." );
		}
	}


	/** Reads the positions of the occurrences of the current term in the current document.
	 *  
	 * <P>First the count (number of positions) is read using the specified coding (unless it has
	 * 	already been read using {@link #readPositionCount()}). Then,
	 *  the vector of positions is read using the specified coding.
	 *
	 * <P>This method can also skip the list of document positions by passing a <code>null</code> vector.
	 *
	 *  @param occ the vector of positions; this vector must be at least as large as the number of
	 *  occurrences that shall be read; if it is <code>null</code>, positions are discarded.
	 *  @return the number of positions (the number of valid entries in <code>occ</code>); if the returned value is
	 *  negative, <code>occ</code> has not enough space to store positions (the absolute returned value is the number
	 *  of positions).
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if you try to read positions without known the frequency.
	 *  @throws NoSuchElementException if you try to read beyond the end of an inverted list.
	 */

	public int readDocumentPositions( int occ[] ) throws java.io.IOException, NoSuchElementException {
		if ( state != BEFORE_POSITIONS ) throw new IllegalStateException( "Trying to call readDocumentPositions() in state " + state );

		final int docSize = index.sizes != null ? index.sizes.getInt( currentDocument ) : -1;
		if ( occ != null && count > occ.length ) return -count;

		state = BEFORE_POINTER;

		int i;
		int prev = -1;
		
		switch( index.positionCoding ) {
		case GAMMA:
			if ( occ != null ) for( i = 0; i < count; i++ ) prev = occ[ i ] = ibs.readGamma() + prev + 1;
			else for( i = 0; i < count; i++ ) ibs.readGamma();
			break;
		case DELTA:
			if ( occ != null ) for( i = 0; i < count; i++ ) prev = occ[ i ] = ibs.readDelta() + prev + 1;
			else for( i = 0; i < count; i++ ) ibs.readDelta();
			break;
		case GOLOMB:
			if ( ASSERTS ) assert index.sizes != null;
			if ( count < 3 ) {
				if ( occ != null ) for( i = 0; i < count; i++ ) occ[ i ] = ibs.readMinimalBinary( docSize );
				else for( i = 0; i < count; i++ ) ibs.readMinimalBinary( docSize );
				return count;
			}

			final int b = Fast.golombModulus( count / (double)docSize );
			final int log2b = Fast.mostSignificantBit( b );
			
			if ( occ != null )
				if ( b != 0 ) 
					for( i = 0; i < count; i++ ) 
						prev = occ[ i ] = ibs.readGolomb( b, log2b ) + prev + 1;
				else 
					for ( i = 0; i < count; i++ ) 
						occ[ i ] = i;
			else
				if ( b != 0 )
					for( i = 0; i < count; i++ ) 
						ibs.readGolomb( b, log2b );
			break;
		case SKEWED_GOLOMB:
			if ( count < 3 ) {
				if ( occ != null ) for( i = 0; i < count; i++ ) occ[ i ] = ibs.readMinimalBinary( docSize );
				else for( i = 0; i < count; i++ ) ibs.readMinimalBinary( docSize );
				return count;
			}

			final int sb = ibs.readMinimalBinary( docSize ) + 1;
			
			if ( occ != null ) for( i = 0; i < count; i++ ) prev = occ[ i ] = ibs.readSkewedGolomb( sb ) + prev + 1;
			else for( i = 0; i < count; i++ ) ibs.readSkewedGolomb( sb );
			break;
		case ARITH:
			if ( occ != null ) prev = occ[ 0 ] = ibs.readMinimalBinary( docSize );
			else ibs.readMinimalBinary( docSize );

			if ( count > 1 ) {
				ArithmeticDecoder decoder = new ArithmeticDecoder( ibs.readMinimalBinary( docSize ) );
				if ( occ != null )
					for( i = 1; i < count; i++ ) 
						prev = occ[ i ] = decoder.decode( ibs ) + prev + 1;
				else 
					for( i = 1; i < count; i++ ) decoder.decode( ibs );
				decoder.flush( ibs );
				ibs.ungetLong( decoder.getWindow(), ArithmeticDecoder.BITS );
			}
			break;

		case INTERP:
			InterpolativeCoding.read( ibs, occ, 0, count, 0, docSize - 1 );
			break;
		default:
			throw new IllegalStateException( "The required position coding (" + index.positionCoding + ") is not supported." );
		}

		return count;
	}


	/** Skips to the first document record whose document pointer is greater than or equal to the given value.
	 *
	 * <P>If this method returns true, then there is one record satisfying the condition,
	 * and the index is positioned over it. If this method returns false, then the condition is not satisfied and the index is positioned
	 * on the last document record.
	 * 
	 * <P>In any case, this method leaves the index in the same state as after a call to 
	 * {@link #readDocumentPointer()}, unless it is called when {@link #endOfList()} is true, 
	 * in which case it does nothing and returns false.
	 *
	 *  @param p a document pointer.
	 *  @return true if we found a record whose document pointer is greater than or equal to <code>p</code>.
	 */
	public boolean skipTo( int p ) throws java.io.IOException {
		if ( endOfList() ) return false;
		if ( ! justAfterPointer() ) readDocumentPointer();

		for(;;) {
			if ( currentDocument >= p ) return true;
			if ( endOfList() ) return false;
			readDocumentPointer();
		}
	}


	/** Returns the number of bits read from the last call to {@link #position(int)}.
	 *
	 * @return the number of bits read so far.
	 */
	public long readBits() {
		return ibs.readBits();
	}

	/** Closes the underlying bit stream.
	 *
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if too few records were written for the previous inverted
	 *  list.
	 */
	public void close() throws java.io.IOException, IllegalStateException {
		ibs.close();
	}
	
	private class IndexReaderDocumentIterator extends AbstractIntIterator implements IndexIterator {
		private final static boolean DEBUG = false;
		
		public IndexReaderDocumentIterator( final int term ) throws IOException {
			position( term );
			readFrequency();
		}

		public int frequency() { 
			return frequency;
		}

		public int document() { 
			if ( currentDocument < 0 ) throw new IllegalStateException();
			return currentDocument;
		}

		public int count() {
			if ( ! index.hasCounts ) throw new UnsupportedOperationException( "This index does not contain counts" );
			try {
				if ( count == -1 ) readPositionCount();
			}
			catch ( IOException e ) {
				throw new RuntimeException( e );
			}
			
			return count;
		}

		public IntIterator positions() {
			count();
			return IntIterators.wrap( positionArray(), 0, count );
		}

		public int[] positionArray() {
			final int a[] = new int[ count() ];
			positions( a );
			return a;
		}

		public int positions( int[] positions ) {
			if ( ! index.hasPositions ) throw new UnsupportedOperationException( "This index does not contain positions" );
			count();
			try {
				return readDocumentPositions( positions );
			}
			catch ( IOException e ) {
				throw new RuntimeException( e );
			}
		}

		public int skipTo( int n ) {
			try {
				if ( DEBUG ) System.out.println( this + ": skipping to " + n );
				/* If there is no skip to do, possibly because we are already at
				 * the end of the list, we just return.	 */
				if ( currentDocument >= n ) return currentDocument;
				if ( endOfList() ) return -1; 
				IndexReader.this.skipTo( n );
				if ( index.hasPositions ) intervalIterator.reset();
				return currentDocument >= n ? currentDocument : -1;
			} catch (IOException e) {
				throw new RuntimeException( e );
			}
		}

		public boolean hasNext() {
			return ! endOfList();
		}
		
		public int nextInt() {
			if ( ! hasNext() ) throw new NoSuchElementException();
			try {
				readDocumentPointer();
				if ( index.hasPositions ) intervalIterator.reset();
				return currentDocument;
			}
			catch ( IOException e ) {
				throw new RuntimeException( e );
			}
		}
		
		public int nextDocument() { return nextInt(); }
		
		public String toString() {
			return index + " [" + term + "]";
		}
		
		private class IndexIntervalIterator extends AbstractObjectIterator implements IntervalIterator {
			int[] positions = IntArrays.EMPTY_ARRAY;
			int pos = -1;			
			int count;

			public void reset() {
				pos = -1; 
				count = count();
			}

			private void ensurePositions() {
				if ( pos < 0 && count >= 0 ) {
					if ( positions( positions ) < 0 ) positions = positionArray();
					pos = 0;
				}
			}

			public boolean hasNext() {
				ensurePositions();
				return pos < count;
			}
			
			public Interval nextInterval() {
				if ( ! hasNext() ) throw new NoSuchElementException();
				return Interval.getInstance( positions[ pos++ ] );
			}
			
			public int extent() {
				return 1;
			}

			public Object next() {
				return nextInterval();
			}
			
			public String toString() {
				return IndexReader.this.index + ": #" + term + "[doc=" + currentDocument + ", count=" + count + ", pos=" + pos + "]";
			}
		};

		private final IndexIntervalIterator intervalIterator = index.hasPositions? new IndexIntervalIterator() : null; 
		private final Map singletonIntervalIterator = Reference2ReferenceMaps.singleton( IndexReader.this.index, intervalIterator );

		public Map intervalIterators() {
			return singletonIntervalIterator;
		}

		public IntervalIterator intervalIterator() {
			return intervalIterator( IndexReader.this.index );
		}

		public IntervalIterator intervalIterator( final Index index ) {
			return index == IndexReader.this.index ? intervalIterator : 
				( count > 0 ? IntervalIterators.TRUE : IntervalIterators.FALSE );
		}
		
		public Set indices() {
			return index.singletonSet;
		}
	}
	
	/** Returns a document iterator over the documents containing a term.
	 * 
	 * @param term a term.
	 */
	
	public DocumentIterator documents( final int term ) throws IOException {
		return new IndexReaderDocumentIterator( term );
	}
	
	/** Returns a document iterator over the documents containing a term; the term is
	 *  given explicitly, and the index term map is used, if present.
	 * 
	 * @param term a term (the term will be downcased if the index is case insensitive).
	 * @throws IOException if an exception occurred while accessing the index.
	 * @throws UnsupportedOperationException if the term map is not available for the underlying index.
	 */
	
	public DocumentIterator documents( final CharSequence term ) throws IOException {
		if ( index.termMap != null ) {
			final int termIndex = index.termMap.get( index.isCaseSensitive ? term : new MutableString( term ).toLowerCase() );
			return termIndex == -1 ? index.emptyDocumentIterator : documents( termIndex );
		}
		else throw new UnsupportedOperationException( "Index " + index + " has no term map" );
	}
	
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
