package it.unimi.dsi.mg4j.index;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.mg4j.io.ArithmeticCoder;
import it.unimi.dsi.mg4j.io.InterpolativeCoding;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;
import jal.ints.Sorting;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Properties;

/** Provides facilities to write an inverted index.
 *
 *  
 * <P>This class is used to write inverted lists in
 * sequential order, as follows:
 * <UL>
 *  <LI>before writing a new inverted list, the method {@link #newInvertedList(int)} should 
 *  be called, specifying the frequency;
 *  <LI>before writing a new document record, the method {@link #newDocumentRecord(int)} should
 *  be called: note that, after calling {@link #newInvertedList(int) newInvertedList(n)} the method
 *  {@link #newDocumentRecord(int)} must be called <strong>exactly</strong> <code>n</code>
 *  times.
 * </UL>
 * The {@link #newDocumentRecord(int)} writes the (suitable coding of) the pointer,
 * and returns an {@link OutputBitStream} that may be used to write the document data. 
 * Note that there is no guarantee that the returned {@link OutputBitStream} coincides with the 
 * underlying bit stream. Moreover, there is no guarantee as to when the bits will be actually 
 * written on the underlying stream, except that when starting a new inverted list, the previous 
 * inverted list, if any, will be written onto the underlying stream.
 *
 * <P>The method {@link #writeDocumentPositions(OutputBitStream,int[],int,int,int)} is a facility
 * that writes, on the given {@link OutputBitStream}, the list of positions
 * using the specified compression scheme: the method should be called, if needed,
 * just after a {@link #newDocumentRecord(int)} call, on the
 * same output bit stream returned by the latter.
 *
 * <H2>Compression flags</H2>
 *
 * <P>As explained above, this class needs to know the compression method that it should use
 * to write the frequency, the document pointer, and the positions; moreover, for positions,
 * the method writes the number of positions first (called the <em>count</em>) followed
 * by the sequence of positions. Hence, there are four different compression methods that one
 * should choose (the method used for frequencies, pointers, counts and positions).
 *
 * <P>Such methods can be chosen when creating an instance of {@link IndexWriter} using
 * the available compression flags specified in {@link CompressionFlags}.
 *
 * <H2>Offset bit stream</H2>
 *
 * <P>Every inverted index may have an associated {@link OutputBitStream} of
 * offsets: this file contains <code>T+1</code> integers, where <code>T</code>
 * is the number of inverted lists (i.e., the number of terms), and the
 * <code>i</code>-th entry is a suitable coding of the position in bytes where
 * the <code>i</code>-th inverted list starts (the last entry is actually the
 * length, in bytes, of the inverted index file itself). The coding used for
 * the offset stream is a &gamma; code of the difference between the current
 * position and the last position.
 *
 * <H2>Statistics</H2>
 *
 * <P>This class performs some bookkeeping of the number of bits written
 * the output stream for several purposes: see, for instance, {@link #bitsForPositions},
 * {@link #bitsForAlignment}, etc.
 *
 * @author Paolo Boldi 
 * @author Sebastiano Vigna 
 * @since 0.6
 */


public class IndexWriter implements CompressionFlags{
	/** This value of {@link #state} means that we should call {@link #newInvertedList()}.*/
	protected static final int BEFORE_INVERTED_LIST = 0;

	/** This value of {@link #state} means that we are positioned at the start of an inverted list,
	 * and we should call {@link #writeFrequency(int)}.*/
	protected static final int BEFORE_FREQUENCY = 1;

	/** This value of {@link #state} means that we are ready to call {@link #newDocumentRecord()}. */
	protected static final int BEFORE_DOCUMENT_RECORD = 2;

	/** This value of {@link #state} means that we just started a new document record, and we
	 * should call {@link #writeDocumentPointer(OutputBitStream, int)}. */
	protected static final int BEFORE_POINTER = 3;

	/** This value of {@link #state} can be assumed only in indices that contain counts; it
	 * means that we are positioned just before the count for the current document record. */
	protected static final int BEFORE_COUNT = 4;

	/** This value of {@link #state} can be assumed only in indices that contain document positions; 
	 * it means that we are positioned just before the position list of the current document record. */
	protected static final int BEFORE_POSITIONS = 5;

	/** This is the first unused state. Subclasses may start from this value to define new states. */
	protected static final int FIRST_UNUSED_STATE = 6;

	/** The underlying {@link OutputBitStream}. */
	protected OutputBitStream obs;
	/** The offset {@link OutputBitStream}. */
	private OutputBitStream offset;
	/** The current state of the writer. */
	protected int state;
	/** The number of documents of the collection to be indexed. */
	protected int numberOfDocuments;
	/** The number of document records that the current inverted list will contain. */
	protected int frequency;
	/** The number of document records that the current inverted list will contain, divided by the number of documents. */
	protected double relativeFrequency;
	/** The number of document records already written for the current inverted list. */
	protected int writtenDocuments;
	/** The current document pointer. */
	protected int currentDocument;
	/** The current term. */
	protected int currentTerm;
	/** The last document pointer in the current list. */
	protected int lastDocument;
	/** The position (in bytes) where the last inverted list started. */
	private long lastInvertedListPos;
	/** The parameter <code>b</code> for Golomb coding of pointers. */
	protected int b;
	/** The parameter <code>log2b</code> for Golomb coding of pointers; it is the most significant bit of {@link #b}. */
	protected int log2b;
	/** The maximum number of positions in a document record so far. */
	protected int maxDocPos;
	/** The number of bits written for alignment. */
	public long bitsForAlignment;
	/** The number of bits written for frequencies. */
	public long bitsForFrequencies;
	/** The number of bits written for document pointers. */
	public long bitsForPointers;
	/** The number of bits written for counts. */
	public long bitsForCounts;
	/** The number of bits written for positions. */
	public long bitsForPositions;
	/** The coding flags. */
	public long flags;
	/** The coding for frequencies. */
	private int frequencyCoding;
	/** The coding for pointers. */
	protected int pointerCoding;
	/** The coding for counts. */
	private int countCoding;
	/** The coding for positions. */
	private int positionCoding;
	/** Whether this index contains counts. */
	public final boolean hasCounts;
	/** Whether this index contains positions. */
	public final boolean hasPositions;
	/** A temporary array for skewed Golomb coding. */
	private int sortedDeltas[] = IntArrays.EMPTY_ARRAY;
	/** The basename, if this object was built using  {@link #IndexWriter(CharSequence, int, boolean, long)}. */
	private CharSequence basename = null;


	/** Creates a new index writer, with the specified basename. The index will be written on a file (stemmed with <samp>.index</samp>).
	 *  If <code>writeOffsets</code>, also an offset file will be produced (stemmed with <samp>.offsets</samp>). 
	 *  When {@link #close()} will be called, the property file will also be produced (stemmed with <samp>.properties</samp>),
	 *  or enriched if it already exists.
	 * 
	 * @param basename the basename.
	 * @param numberOfDocuments the number of documents in the collection to be indexed.
	 * @param writeOffsets if <code>true</code>, the offset file will also be produced.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexWriter} introduction).
	 */
	public IndexWriter( CharSequence basename, int numberOfDocuments, boolean writeOffsets, long flags ) throws IOException {
		this( 
			new OutputBitStream( new FileOutputStream( basename + ".index" ) ),
			writeOffsets? new OutputBitStream( new FileOutputStream( basename + ".offsets") ) : null,
			numberOfDocuments,
			flags
		 );
		 this.basename = basename;
	}

	/** Creates a new index writer, with the specified underlying {@link OutputBitStream}.
	 *
	 * @param obs the underlying output bit stream.
	 * @param offset the offset bit stream.
	 * @param numberOfDocuments the number of documents in the collection to be indexed.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexWriter} introduction).
	 */
	public IndexWriter( OutputBitStream obs, OutputBitStream offset, int numberOfDocuments, long flags ) {
		this.obs = obs;
		this.offset = offset;
		this.numberOfDocuments = numberOfDocuments;
		this.frequency = -1;
		this.currentTerm = -1;
		this.maxDocPos = 0;
		this.flags = flags;

		frequencyCoding = (int) (( ( flags >>> FREQUENCIES_SHIFT & 0xFF ) != 0 ? flags : FREQUENCIES_DEFAULT ) >>> FREQUENCIES_SHIFT & 0xFF);
		pointerCoding = (int) (( ( flags >>> POINTERS_SHIFT & 0xFF ) != 0 ? flags  : POINTERS_DEFAULT ) >>> POINTERS_SHIFT & 0xFF);
		countCoding = (int) (( ( flags >>> COUNTS_SHIFT & 0xFF ) != 0 ? flags : COUNTS_DEFAULT ) >>> COUNTS_SHIFT & 0xFF);
		hasCounts = countCoding != NONE;

		if ( ! hasCounts  )	{
			positionCoding = (int) (( ( flags >>> POSITIONS_SHIFT & 0xFF ) != 0 ? flags : NO_POSITIONS ) >>> POSITIONS_SHIFT & 0xFF);
			if ( positionCoding != NONE ) throw new IllegalArgumentException( "Index has positions but no counts (this can't happen)" );
		}
		else positionCoding = (int) (( ( flags >>> POSITIONS_SHIFT & 0xFF ) != 0 ? flags : POSITIONS_DEFAULT ) >>> POSITIONS_SHIFT & 0xFF);
		
		hasPositions = positionCoding != NONE;
		if ( ! hasCounts && hasPositions ) throw new IllegalArgumentException( "Index has positions but no counts (this can't happen)" );
	}

	/** Creates a new index writer, with the specified underlying {@link OutputBitStream},
	 *  without an associated offset bit stream.
	 *
	 * @param obs the underlying output bit stream.
	 * @param numberOfDocuments the number of documents in the collection to be indexed.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexWriter} introduction).
	 */
	public IndexWriter( OutputBitStream obs, int numberOfDocuments, int flags ) {
		this( obs, null, numberOfDocuments, flags );
	}

	/** Starts a new inverted list. The previous inverted list, if any, is actually written
	 *  to the underlying bit stream; then, the stream is byte-aligned.
	 *  
	 *  @return the position (in bytes) of the underlying bit stream where the new inverted
	 *  list starts.
	 *  @throws IOException if an exception is thrown by the underlying stream.
	 *  @throws IllegalStateException if too few records were written for the previous inverted
	 *  list.
	 */
	public long newInvertedList() throws java.io.IOException, IllegalStateException {
		if ( this.frequency >= 0 && this.frequency < this.writtenDocuments ) throw new IllegalStateException( "Document record underflow (written " + this.writtenDocuments + "/" + this.frequency + ")" );
		if ( state != BEFORE_INVERTED_LIST && state != BEFORE_DOCUMENT_RECORD ) throw new IllegalStateException( "Trying to start new inverted list in state " + state );

		// Align!
		bitsForAlignment += obs.align();

		// The position (in bytes) where the new inverted list starts
		long pos = obs.writtenBits() / 8;
		// Reset variables
		this.writtenDocuments = 0;
		this.currentTerm++;
		this.currentDocument = -1;

		// If needed, write the offset
		if ( offset != null ) offset.writeGamma( (int)( pos - lastInvertedListPos ) );
		this.lastInvertedListPos = pos;
		state = BEFORE_FREQUENCY;
		return pos;
	}

	/** Writes the frequency.
	 *  
	 *  @param frequency the (positive) number of document records that this inverted list will contain.
	 *  @return the number of bits written.
	 *  @throws IOException if an exception is thrown by the underlying stream.
	 */
	public int writeFrequency( int frequency ) throws IOException {
		if ( state != BEFORE_FREQUENCY ) throw new IllegalStateException( "Trying to write frequency in state " + state );

		int bitCount;
		// Write the frequency
		switch( frequencyCoding ) {
		case GAMMA:
			bitCount = obs.writeGamma( frequency - 1 ); // frequency cannot be 0
			break;
		case DELTA:
			bitCount = obs.writeDelta( frequency - 1 ); // frequency cannot be 0
			break;
		default:
			throw new IllegalStateException( "The required frequency coding (" + frequencyCoding + ") is not supported." );
		}

		this.frequency = frequency;
		this.relativeFrequency = (double)frequency / numberOfDocuments;

		// We compute the modulus used for pointer Golomb coding 
		if ( pointerCoding == GOLOMB ) {
			this.b = Fast.golombModulus( relativeFrequency ); 
			this.log2b = Fast.mostSignificantBit( b );
		}

		state = BEFORE_DOCUMENT_RECORD;
		bitsForFrequencies += bitCount;
		return bitCount;
	}

	/** Starts a new inverted list. The previous inverted list, if any, is actually written
	 *  to the underlying bit stream; then, the stream is byte-aligned, and <code>frequency</code> 
	 *  is written onto the underlying stream.
	 *  
	 *  @param frequency the number of document records that this inverted list will contain.
	 *  @return the position (in bytes) of the underlying bit stream where the new inverted
	 *  list starts.
	 *  @throws IOException if an exception is thrown by the underlying stream.
	 *  @throws IllegalStateException if too few records were written for the previous inverted
	 *  list.
	 *  @deprecated Use {@link #newInvertedList()} followed by {@link #writeFrequency(int)}.
	 */
	public long newInvertedList( int frequency ) throws java.io.IOException, IllegalStateException {
		long pos = newInvertedList();
		writeFrequency( frequency );
		return pos;
	}

	/** Starts a new document record. 
	 * 
	 * <P>This method must be called exactly <code>n</code> times
	 * after a call to {@link #newInvertedList(int) newInvertedList(n)}. 
	 *
	 *  @return the output bit stream where the next document record data
	 *  should be written.
	 *  @throws IOException if an exception is thrown by the underlying stream.
	 *  @throws IllegalStateException if too many records were written for the current inverted list,
	 *  or if there is no current inverted list.
	 */
	public OutputBitStream newDocumentRecord() throws java.io.IOException, IllegalStateException {
		if ( this.frequency == this.writtenDocuments ) throw new IllegalStateException( "Document record overflow (written " + this.frequency + " already)" );
		if ( state != BEFORE_DOCUMENT_RECORD ) throw new IllegalStateException( "Trying to start new document record in state " + state );

		this.writtenDocuments++;
		this.lastDocument = this.currentDocument;
		state = BEFORE_POINTER;
		return obs;
	}

	/** Writes a document pointer.
	 * 
	 * <P>This method must be called immediately after {@link #newDocumentRecord()}.
	 * 
	 * @param out the output bit stream where the pointer will be written.
	 * @param pointer the document pointer.
	 * @return the number of bits written.
	 * @throws IOException if an exception is thrown by the underlying stream.
	 */
	public int writeDocumentPointer( OutputBitStream out, int pointer ) throws java.io.IOException, IllegalStateException {
		if ( state != BEFORE_POINTER ) throw new IllegalStateException( "Trying to write pointer in state " + state );

		this.currentDocument = pointer;
		int bitCount = 0;

		if ( frequency != numberOfDocuments ) { // We do not write pointers for everywhere occurring documents.
			switch( pointerCoding ) {
				case GAMMA:
					bitCount = out.writeGamma( pointer - lastDocument - 1 );
					break;
				case DELTA:
					bitCount = out.writeDelta( pointer - lastDocument - 1 );
					break;
				case GOLOMB:
					bitCount = out.writeGolomb( pointer - lastDocument - 1, b, log2b );
					break;
				default:
					throw new IllegalStateException( "The required pointer coding (" + pointerCoding + ") is not supported." );
			}
		}

		state = hasCounts ? BEFORE_COUNT : BEFORE_DOCUMENT_RECORD;
		bitsForPointers += bitCount;
		return bitCount;
	}



	/** Starts a new document record. This method should be called exactly <code>n</code> times
	 *  after a call to {@link #newInvertedList(int) newInvertedList(n)}. 
	 *
	 *  @param pointer the document pointer (<em>will</em> be written to the bit stream).
	 *  @return the <code>OutputBitStream</code> where the next document record
	 *  should be written.
	 *
	 *  @throws IOException if an exception is thrown by the underlying stream.
	 *  @throws IllegalStateException if too many records were written for the current inverted list,
	 *  or if there is no current inverted list.
	 *  @deprecated Use {@link #newDocumentRecord()} followed by {@link #writeDocumentPointer(OutputBitStream, int)}.
	 */
	public OutputBitStream newDocumentRecord( int pointer ) throws java.io.IOException, IllegalStateException {
		OutputBitStream resultObs = newDocumentRecord();
		writeDocumentPointer( resultObs, pointer );
		return resultObs;
	}

	/** Closes the underlying bit stream.
	 *
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if too few records were written for the previous inverted
	 *  list.
	 */
	public void close() throws java.io.IOException, IllegalStateException {
		if ( state != BEFORE_DOCUMENT_RECORD && state != BEFORE_INVERTED_LIST ) throw new IllegalStateException( "Trying to close index in state " + state );
		if ( this.frequency >= 0 && this.frequency < this.writtenDocuments ) throw new IllegalStateException( "Document record underflow during close (written " + this.writtenDocuments + "/" + this.frequency + ")" );

		bitsForAlignment += obs.align();

		if ( writtenBits() != obs.writtenBits() ) 
			throw new IllegalStateException( "Written bits count mismatch: we say " + writtenBits() + ", the stream says " + obs.writtenBits() );

		if ( offset != null ) {
			offset.writeGamma( (int)( obs.writtenBits() / 8 - lastInvertedListPos  ) );
			offset.close();
		}
		obs.close();
		
		if ( basename != null ) { 
			String propertyFilename = basename + ".properties";
			Properties properties = new Properties(); 
			try  {
				FileInputStream propertyFile = new FileInputStream( propertyFilename );
				properties.load( propertyFile );
				propertyFile.close();
			} catch ( FileNotFoundException e ) {}
			properties.putAll( properties() );
			FileOutputStream propertyFile = new FileOutputStream( propertyFilename );
			properties.store( propertyFile, "Last modified by IndexWriter" );
		}
	}
	

	/** Writes the count of the occurrences of the current term in the current document to the given {@link OutputBitStream}. 
	 *  @param out the output stream where the occurrences should be written.
	 *  @param count the count.
	 *  @return the number of bits written.
	 */
	public int writePositionCount( final OutputBitStream out, final int count ) throws java.io.IOException, IllegalStateException {
		if ( frequency < 0 ) throw new IllegalStateException( "Trying to write count without calling newInvertedList" );
		if ( state != BEFORE_COUNT ) throw new IllegalStateException( "Trying to write count in state " + state );
		final int bitCount;

		switch( countCoding ) {
			case GAMMA:
				bitCount = out.writeGamma( count - 1 );
				break;
			case DELTA:
				bitCount = out.writeDelta( count - 1 );
				break;
			default:
				throw new IllegalStateException( "The required count coding (" + countCoding + ") is not supported." );
		}
		
		state = hasPositions ? BEFORE_POSITIONS : BEFORE_DOCUMENT_RECORD;
		bitsForCounts += bitCount;
		return bitCount;
	}

	/** Writes the positions of the occurrences of the current term in the current document to the given {@link OutputBitStream}. 
	 *  First the count (number of positions) is written using the specified coding. Then,
	 *  the specified vector of occurrences is written using the specified coding.
	 *
	 *  @param out the output stream where the occurrences should be written.
	 *  @param occ the position vector (a sequence of strictly increasing natural numbers).
	 *  @param offset the first valid entry in <code>occ</code>.
	 *  @param len the number of valid entries in <code>occ</code>.
	 *  @param docSize the size of the current document (only for Golomb and interpolative coding; you can safely pass 0 otherwise).
	 *  @return the number of bits written.
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if there is no current inverted list.
	 */
	public int writeDocumentPositions( OutputBitStream out, int occ[], int offset, int len, int docSize ) throws java.io.IOException, IllegalStateException {
		if ( frequency < 0 ) throw new IllegalStateException( "Trying to write occurrences without calling newInvertedList" );
		if ( state != BEFORE_POSITIONS ) throw new IllegalStateException( "Trying to write positions in state " + state );

		int i;
		int prev = -1;
		int bitCount = 0;
		final int end = offset + len;

		switch( positionCoding ) {
			case GAMMA:
				for( i = offset; i < end; i++ ) {
					bitCount += out.writeGamma( occ[ i ] - prev - 1 );
					prev = occ[ i ];
				}
				break;
			case DELTA:
				for( i = offset; i < end; i++ ) {
					bitCount += out.writeDelta( occ[ i ] - prev - 1 );
					prev = occ[ i ];
				}
				break;
			case GOLOMB:
				if ( len < 3 ) {
					for( i = 0; i < len; i++ ) bitCount += out.writeMinimalBinary( occ[ i ], docSize );
					break;
				}

				// We compute b and log2b for positions
				final int positionB = Fast.golombModulus( len / (double)docSize );
				final int positionLog2b = Fast.mostSignificantBit( positionB );

				for( i = offset; i < end; i++ ) {
					bitCount += out.writeGolomb( occ[ i ] - prev - 1, positionB, positionLog2b );
					prev = occ[ i ];
				}
				break;
			case SKEWED_GOLOMB:
				if ( len < 3 ) {
					for( i = 0; i < len; i++ ) bitCount += out.writeMinimalBinary( occ[ i ], docSize );
					break;
				}

				if ( sortedDeltas.length < len ) sortedDeltas = new int[ len ];
				System.arraycopy( occ, offset, sortedDeltas, 0, len );
				i = len - 1;
				while( i-- != 0 ) sortedDeltas[ i + 1 ] -= sortedDeltas[ i ] + 1;
				Sorting.nth_element( sortedDeltas, 0, len / 2, len );
				final int sb = sortedDeltas[ len / 2 ] + 1; // The median of deltas

				bitCount = out.writeMinimalBinary( sb - 1, docSize );

				for( i = offset; i < end; i++ ) {
					bitCount += out.writeSkewedGolomb( occ[ i ] - prev - 1, sb );
					prev = occ[ i ];
				}
				break;
			case ARITH:
				bitCount = out.writeMinimalBinary( occ[ 0 ], docSize );

				if ( len > 1 ) {
					prev = 0;
					for( i = offset + 1; i < end; i++ ) if ( occ[ i ] - occ[ i - 1 ] > prev ) prev = occ[ i ] - occ[ i - 1 ];
					
					bitCount += out.writeMinimalBinary( prev, docSize );
					
					ArithmeticCoder coder = new ArithmeticCoder( prev ); // prev is strictly larger than any delta.
					for( i = offset + 1; i < end; i++ ) bitCount += coder.encode( occ[ i ] - occ[ i - 1 ] - 1, out );
					bitCount += coder.flush( out );
				}
				break;
			case INTERP:
				bitCount = InterpolativeCoding.write( out, occ, 0, len, 0, docSize - 1 );
				break;
			default:
				throw new IllegalStateException( "The required position coding (" + positionCoding + ") is not supported." );
		}

		state = BEFORE_DOCUMENT_RECORD;
		bitsForPositions += bitCount;
		if ( len > maxDocPos ) maxDocPos = len;
		return bitCount;	
	}

	/** This method returns the overall number of bits written onto the underlying stream.
	 *
	 *  @return the number of bits written, according to the variables keeping statistical records.
	 */
	public long writtenBits() {
		return bitsForAlignment + bitsForFrequencies + bitsForPointers + bitsForCounts + bitsForPositions;
	}

	/** This method should only be called after {@link #close()}. It returns the values for all the properties
	 *  specified in {@link IndexProperties}.
	 * 
	 * @return the properties specified in {@link IndexProperties} along with their values.
	 */
	public Properties properties() {
		Properties result = new Properties();
		result.setProperty( IndexProperties.DOCUMENTS, "" + numberOfDocuments );
		result.setProperty( IndexProperties.TERMS, "" + ( currentTerm + 1 ) );
		result.setProperty( IndexProperties.MAXDOCPOS, "" + maxDocPos );
		result.setProperty( IndexProperties.INDEXCLASS, Index.class.getName() );
		result.setProperty( IndexProperties.COMPRESSIONFLAGS, flags2String( flags ) );
		return result;	
	}

	/** This method unpacks the flags, expressed as a long, and returns a string with the compression
	 *  flags. The string is a pipe-separated sequence of compression flags, as described in
	 *  {@link it.unimi.dsi.mg4j.index.CompressionFlags}.
	 * 
	 *  @param flags the flags.
	 *  @return a string corresponding to the given flags.
	 */
	public static String flags2String( long flags ) {
		Field fields[] = CompressionFlags.class.getFields();
		String fieldName, subfieldName;
		MutableString result = new MutableString();
		
		try {
			for ( int i = 0; i < fields.length; i++ ) {
				// For all ***_SHIFT (e.g. FREQUENCIES_SHIFT)
				if ( ( fieldName = fields[ i ].getName() ).endsWith( "_SHIFT" ) ) {
					// Shift (e.g. the value of FREQUENCIES_SHIFT)
					int shift = fields[ i ].getInt( CompressionFlags.class );
					// Type (e.g. FREQUENCIES)
					String type = fieldName.substring( 0, fieldName.length() - "_SHIFT".length() );
					// The flags value, for the FREQUENCIES part only
					long fieldPart = (int) (( flags >>> shift ) & 0xFF) << shift;
					if ( fieldPart == 0 ) continue; // Default
					int j;
					for ( j = 0; j < fields.length; j++ ) {
						subfieldName = fields[ j ].getName();
						if ( i != j && subfieldName.indexOf( type ) >= 0 && subfieldName.indexOf( "DEFAULT" ) < 0 )
							if ( fieldPart == fields[ j ].getLong( CompressionFlags.class ) ) {
								if ( result.length() > 0 ) result.append( " | " );
								result.append( subfieldName );
								break;
							}
					}
					if ( j == fields.length ) 
						throw new RuntimeException( "Could not unpack the " + fieldName + " compression flag" );
				}
			}
			return result.toString();
		} catch (IllegalAccessException e) {
			throw new RuntimeException( e );
		}
	}

}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
