package it.unimi.dsi.mg4j.index;

import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.longs.LongList;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.File;
import java.io.IOException;
import java.util.Properties;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** An abstract representation of an index with skips.
 *
 * <P>An instance of this class stores, besides the data contained in 
 * an {@link it.unimi.dsi.mg4j.index.Index}, data related to
 * skipping (quantum, height, etc.).
 * 
 * @author Paolo Boldi
 * @author Sebastiano Vigna
 * @since 0.9
 */

public class SkipIndex extends Index {

	/** The parameter <code>h</code> (the maximum height of a skip tower). */
	public final int height;
	/** The quantum. */
	public final int quantum;

	/** Creates a new skip index using the given basename.
	 * 
	 *
	 * 
	 * @param basename the basename of the index.
	 * @param termMap the term list for this index, or <code>null</code> if there is no term list.
	 * @param loadOffsets offsets are loaded only if this parameter is <code>true</code>.
	 * @param pm an optional progress meter. If <code>null</code>, no progress information will be displayed.
	 */
	protected SkipIndex( final CharSequence basename, final TermMap termMap, final boolean loadOffsets, final ProgressMeter pm ) throws IOException {
		super( basename, termMap, loadOffsets, pm );
		quantum = Integer.parseInt( properties.getProperty( SkipIndexProperties.SKIPQUANTUM ) );
		height = Integer.parseInt( properties.getProperty( SkipIndexProperties.SKIPHEIGHT ) );
	}


	/** Creates a new skip index using the given data.
	 * 
	 * <P>This constructor provides an index that is initialised exactly to the provided data.
	 * It is mainly useful for debugging and testing purposes, in the case you are creating
	 * an index file (for instance, as a memory-based bit stream) but you have no property
	 * file to use with {@link #getInstance(CharSequence)}.
	 *
	 * <P>It is usually safe to provide <code>null</code> for <code>basename</code>, <code>indexFile</code>
	 * and <code>properties</code>, but the responsibility for data consistence is up to the caller.
	 *  
	 * @param basename the basename of this index, or <code>null</code>.
	 * @param indexFile the file containing this index, or <code>null</code>.
	 * @param properties the properties of this index, or <code>null</code>.
	 * @param numberOfDocuments the number of documents in this index.
	 * @param numberOfTerms the number of terms in this index.
	 * @param maxDocPos the maximum length of an occurrence list, or -1 if it is not known.
	 * @param flags a bit mask setting the coding techniques to be used (see {@link CompressionFlags}).
	 * @param termMap the term list for this index, or <code>null</code> if there is no term list.
	 * @param offsets the offset list; may be <code>null</code> if you do not plan using {@link IndexReader#position(int)}.
	 * @param sizes the size list; may be <code>null</code> if your code does not require it.
	 * @param quantum the quantum to be used if this index has skips (0 means that this index has no skips).
	 * @param height the height (meaningless if quantum is zero).
	 */
	public SkipIndex( String basename, File indexFile, Properties properties, int numberOfDocuments,
		int numberOfTerms, int maxDocPos, boolean isCaseSensitive, long flags, TermMap termMap, LongList offsets, IntList sizes, int quantum, int height ) throws java.io.IOException {
		super( basename, indexFile, properties, numberOfDocuments,
			numberOfTerms, maxDocPos, isCaseSensitive, flags, termMap, offsets, sizes );
		this.quantum = quantum;
		this.height = height;
	}

	/** Creates and returns a new {@link IndexReader} based on this index. After that, you
	 *  can use the reader to read this index.
	 * 
	 * @return a new {@link SkipIndexReader} to read this index.
	 */
	public IndexReader getReader() throws IOException {
		return getReader( InputBitStream.DEFAULT_BUFFER_SIZE );
	}

	/** Creates and returns a new {@link IndexReader} based on this index. After that, you
	 *  can use the reader to read this index.
	 * 
	 * @param bufferSize the size of the buffer to be used when opening the {@link InputBitStream} underlying the
	 * {@link SkipIndexReader} that is going to be returned.
	 * @return a new {@link SkipIndexReader} to read this index.
	 */
	public IndexReader getReader( int bufferSize ) throws IOException {
		return new SkipIndexReader( this, bufferSize );
	}
}
