package it.unimi.dsi.mg4j.index;

import java.io.IOException;
import it.unimi.dsi.fastutil.ints.IntList;
import it.unimi.dsi.fastutil.longs.LongList;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.util.Fast;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** Provides facilities to read a skip inverted index from an {@link InputBitStream}.
 *  Please consult the {@link SkipIndexWriter} documentation for more information about
 *  the structure of a skip inverted index.
 *
 *  <P>This class is used to read inverted lists with skips. Its
 * functioning is completely analogous to {@link it.unimi.dsi.mg4j.index.IndexReader}.
 * 
 *  <P>Note that {@link it.unimi.dsi.mg4j.index.Index} provides a much easier interface to
 * index access. Instead of creating an {@link IndexReader} explicitly, you can use {@link Index#getReader()}.
 *
 * @author Paolo Boldi 
 * @author Sebastiano Vigna 
 * @since 0.6
 */


public class SkipIndexReader extends IndexReader {
	/** The reference skip index. This variable shadows (and copies) {@link IndexReader#index} to
	 * avoid excessive type casting. */
	final protected SkipIndex skipIndex;
	
	/** The maximum height of a skip tower in the current block. May be less than {@link #h} if the block is defective. */
	private int maxh;

	/** The maximum valid index of the current skip tower, if any. */
	private int s;

	private double pointerQuantumSigma;

	/** This value of {@link #state} means that we are positioned 
	 * just before a tower. */
	protected static final int BEFORE_TOWER = IndexReader.FIRST_UNUSED_STATE + 0;

	/** This is the first unused state. Subclasses may start from this value to define new states. */
	protected static final int FIRST_UNUSED_STATE = IndexReader.FIRST_UNUSED_STATE + 1;

	/** We have <var>w</var> = <var>Hq</var>. */
	private int w;

	/** An array to decode the skip amounts. */
	private long skipAmount[];

	/** An array to decode the skip pointers. */
	private int skipPointer[];

	/** The number of bits read just after reading the last skip tower. */
	private long readBitsAtLastSkipTower;

	/** The current quantum bit length, as provided by the index. */
	private int quantumBitLength;

	/** The current entry bit length, as provided by the index. */
	private int entryBitLength;

	private static final boolean ASSERTS = true;
	private static final boolean DEBUG = false;

	static { if ( DEBUG && ASSERTS ); }

	/** Initialises internal skip buffers. */
	private void init() {
		w = ( 1 << skipIndex.height ) * skipIndex.quantum;
		skipAmount = new long[ skipIndex.height + 1 ];
		skipPointer = new int[ skipIndex.height + 1 ];
	}

	/** Creates a new skip index reader, with the specified underlying {@link Index} and default buffer size.
	 *
	 * @param index the index.
	 */
	public SkipIndexReader( final Index index ) throws IOException {
		super( index );
		this.skipIndex = (SkipIndex)index;
		init();
	}

	/** Creates a new skip index reader, with the specified underlying {@link Index}.
	 *
	 * @param index the index.
 	 * @param bufferSize the size of the buffer of the underlying input bit stream.
	 */
	public SkipIndexReader( final Index index, final int bufferSize ) throws IOException {
		super( index, bufferSize );
		this.skipIndex = (SkipIndex)index;
		init();
	}

	/** Creates a new skip index reader, with the specified underlying {@link Index} and input bit stream.
	 *
	 * @param index the index.
	 * @param ibs the underlying bit stream.
	 */
	public SkipIndexReader( final Index index, final InputBitStream ibs ) throws IOException {
		super( index, ibs );
		this.skipIndex = (SkipIndex)index;
		init();
	}


	/** Creates a new skip index reader, with the specified underlying {@link InputBitStream}.
	 *
	 * @param ibs the underlying input bit stream.
	 * @param offsets the offset list; may be <code>null</code> if you do not plan using {@link #position(int)}.
	 * @param sizes the size list; may be <code>null</code> if your code does not require it.
	 * @param N the number of documents in the collection.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexReader} introduction).
	 * @param q the quantum.
	 * @param h the maximum height of a skip tower.
	 * @deprecated Please use {@link Index#getReader()} or {@link IndexReader#IndexReader(Index)}.
	 */
	public SkipIndexReader( InputBitStream ibs, LongList offsets, IntList sizes, int N, long flags, int q, int h ) throws java.io.IOException {
		this( new SkipIndex( null, null, null, N, -1, -1, true, flags, null, offsets, sizes, q, h ) );
		this.ibs = ibs;
		this.frequency = -1;
		this.state = BEFORE_FREQUENCY;
		init();

		if ( skipIndex.positionCoding == POSITIONS_ARITH ) ibs.overflow( true ); // Sets overflow if positions must be coded using an arithmetic coder
	}


	/** Creates a new index reader, with the specified underlying {@link InputBitStream}.
	 *
	 * @param ibs the underlying input bit stream.
	 * @param offsets the offset list; may be <code>null</code> if you do not plan using {@link #position(int)}.
	 * @param N the number of documents in the collection.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexReader} introduction).
	 * @param q the quantum.
	 * @param h the maximum height of a skip tower.
	 * @deprecated Please use {@link Index#getReader()} or {@link IndexReader#IndexReader(Index)}.
	 */
	public SkipIndexReader( InputBitStream ibs, LongList offsets, int N, long flags, int q, int h ) throws IOException {
		this( ibs, offsets, null, N, flags, q, h);
	}

	/** Creates a new index reader, with the specified underlying {@link InputBitStream}.
	 *
	 * @param ibs the underlying input bit stream.
	 * @param N the number of documents in the collection.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexReader} introduction).
	 * @param q the quantum.
	 * @param h the maximum height of a skip tower.
	 * @deprecated Please use {@link Index#getReader()} or {@link IndexReader#IndexReader(Index)}.
	 */
	public SkipIndexReader( InputBitStream ibs, int N, long flags, int q, int h ) throws IOException {
		this( ibs, null, N, flags, q, h );
	}

	public int readFrequency() throws java.io.IOException, IllegalStateException {
		if ( state == BEFORE_TOWER ) readTower();
		
		int res = super.readFrequency();
		pointerQuantumSigma = Math.sqrt( skipIndex.quantum * ( 1 - relativeFrequency ) ) / relativeFrequency;
		quantumBitLength = entryBitLength = -1;
		return res;	
	}

	public int readDocumentPointer() throws IOException {
		if ( state == BEFORE_TOWER ) readTower();
		final int pointer = super.readDocumentPointer();
		if ( numberOfDocumentRecord % skipIndex.quantum == 0 ) state = BEFORE_TOWER;
		return pointer;
	}

	public int readPositionCount() throws java.io.IOException {
		if ( state == BEFORE_TOWER ) readTower();
		return super.readPositionCount();
	}


	/** Reads the entire skip tower for the current position. 
	 */
	private void readTower() throws java.io.IOException {
		readTower( -1 );
	}

	/** Reads the skip tower for the current position, possibly skipping part of the tower. 
	 * 
	 * <P>Note that this method will update {@link #state} only if it reads the entire tower,
	 * otherwise the state remains {@link #BEFORE_TOWER}.
	 * 
	 * @param pointer the tower will be read up to the first entry smaller than or equal to this pointer; use
	 * -1 to guarantee that the entire tower will be read. 
	 */

	private void readTower( int pointer ) throws java.io.IOException {
		int prevPointerDelta, i, j, k, cacheOffset, cache, towerLength = 0;
		long bitsAtTowerStart = 0;

		if ( ASSERTS ) assert numberOfDocumentRecord % skipIndex.quantum == 0;
		if ( state != BEFORE_TOWER ) throw new IllegalStateException( "readTower() called in state " + state );
		
		cacheOffset = numberOfDocumentRecord % w;
		k = cacheOffset / skipIndex.quantum;

		// Compute the height of the current skip tower.
		s = ( k == 0 )? skipIndex.height : Fast.leastSignificantBit( k );
		
		if ( ( cache = frequency - w * ( numberOfDocumentRecord / w ) ) < w ) {
			// If the block is defective, the actual maximum height may be smaller than h.
			maxh = Fast.mostSignificantBit( cache / skipIndex.quantum - k );
			if ( maxh < s ) s = maxh;
		}
		else maxh = skipIndex.height;
		
		if ( DEBUG ) System.err.println( "{" + this + "} pointer=" + pointer + " maxh=" + maxh + " s=" + s );
		
		i = s;

		if ( s >= 0 ) {
			if ( k == 0 ) {
				if ( quantumBitLength < 0 ) { 
					quantumBitLength = ibs.readDelta();
					entryBitLength = ibs.readDelta();
				}
				else {
					quantumBitLength += Fast.nat2int( ibs.readDelta() ); 
					entryBitLength += Fast.nat2int( ibs.readDelta() );
				}
				if ( DEBUG ) System.err.println( "{" + this + "} quantum bit length=" + quantumBitLength + " entry bit length=" + entryBitLength );
			}

			if ( s > 0 ) towerLength = entryBitLength * ( s + 1 ) + Fast.nat2int( ibs.readDelta() );
		
			bitsAtTowerStart = ibs.readBits();

			prevPointerDelta = Fast.nat2int( ibs.readGolomb(
				Fast.gaussianGolombModulus( pointerQuantumSigma * Math.sqrt( 1 << i ) ) 
			) ) + (int)Math.round( skipIndex.quantum * ( 1 << i ) / relativeFrequency );

			//System.err.println("Read pointer delta " + prevPointerDelta);

			skipPointer[ i ] = prevPointerDelta + currentDocument;
			skipAmount[ i ] = quantumBitLength * ( 1 << s ) + entryBitLength * ( ( 1 << s + 1 ) - s - 2 ) + Fast.nat2int( ibs.readDelta() );			

			//System.err.println( "Read bit delta " + A[ i ] );

			if ( skipPointer[ i ] > pointer ) {
				for( i = s - 1; i >= 0; i-- ) {
					prevPointerDelta = - Fast.nat2int( 
						ibs.readGolomb( 
							Fast.gaussianGolombModulus( pointerQuantumSigma * Math.sqrt( ( 1 << i ) / 2.0 ) ) ) 
					) + prevPointerDelta / 2;
				 	
					//System.err.println("Read pointer delta " + prevPointerDelta);

					skipPointer[ i ] = prevPointerDelta + currentDocument;
					skipAmount[ i ] = ( skipAmount[ i + 1 ] - entryBitLength * i ) / 2 - Fast.nat2int( ibs.readDelta() );
					
					//System.err.println( "Read bit delta " + A[ i ]);

					if ( skipPointer[ i ] <= pointer ) break;
				}
			}
		}

		/* If we did not read the entire tower, we need to fix the skips we read (as they
		 * are offsets from the *end* of the tower) and moreover we must make unusable the
		 * rest of the tower (for asserts).
		 */

		if ( i > 0 ) {
			for( j = s; j >= i; j-- ) skipAmount[ j ] += towerLength - ( ibs.readBits() -  bitsAtTowerStart ); 
			for( ; j >= 0; j-- ) skipPointer[ j ] = Integer.MAX_VALUE;
		}
		else state = skipIndex.hasCounts ? BEFORE_COUNT : BEFORE_POINTER;

		// We update the top of the tower.
		long delta = ibs.readBits() - readBitsAtLastSkipTower;
		for( j = s + 1; j <= maxh; j++ ) skipAmount[ j ] -= delta;

		readBitsAtLastSkipTower = ibs.readBits();

		if ( DEBUG ) {
			System.err.println( "{" + this + "} " + "Computed skip tower for document record number " + numberOfDocumentRecord + " (pointer " + currentDocument + ") from " + Math.max( i , 0 ) + ": " );
			System.err.print( "% " );
			for( j = Math.max( i, 0 ); j <= s; j++ ) System.err.print( skipPointer[ j ] + ":" + skipAmount[ j ] + " " );
			System.err.println();
		}

	}
	
	
	/** Returns true if we are positioned just after a pointer. 
	 * 
	 * @return true if we are positioned just after a pointer.
	 */

	protected boolean justAfterPointer() {
		return state == BEFORE_TOWER || numberOfDocumentRecord % skipIndex.quantum != 0 && super.justAfterPointer();
	}

	public boolean skipTo( int p ) throws java.io.IOException {
		int i, k, cacheOffset;

		if ( DEBUG ) System.err.println( "{" + this + "} " + "Going to enter skip code with lastDoc=" + currentDocument + ", numberOfDocumentRecord=" + numberOfDocumentRecord + ", p=" + p + ", endOfList()=" + endOfList() );
		
		if ( endOfList() ) return false;

		// Let us make sure that lastDoc contains the current document pointer.
		if ( ! justAfterPointer() ) readDocumentPointer();

		if ( currentDocument >= p ) {
			if ( DEBUG ) System.err.println( "{" + this + "} " + "No skip necessary, returning true" );
			return true; 
		}

		if ( state == BEFORE_TOWER ) readTower( p );

		if ( DEBUG ) System.err.println( "{" + this + "} " + "Entering skip code with lastDoc=" + currentDocument + ", numberOfDocumentRecord=" + numberOfDocumentRecord + ", p=" + p + ", endOfList()=" + endOfList() );

		for(;;) {
		
			if ( DEBUG ) System.err.println( "{" + this + "} " + "In for loop, lastDoc=" + currentDocument + ", maxh=" + maxh + ", numberOfDocumentRecord=" + numberOfDocumentRecord + ", p=" + p );

			cacheOffset = numberOfDocumentRecord % w;
			k = cacheOffset / skipIndex.quantum;

			if ( maxh < 0 ) break; // Defective quantum--no tower.

			/* This is the highest index in the current tower we should look at.
			   Higher indices cannot point farther than the i-th. */
			i = Fast.mostSignificantBit( ~ k  & ( 1 << maxh ) - 1 ) + 1;

			if ( DEBUG ) System.err.println( "{" + this + "} " + "Cannot look further than i=" + i );

			
			if ( ASSERTS ) for( int j = i; j >= 0; j-- ) assert skipPointer[ i ] != Integer.MAX_VALUE;
			for( ; i >= 0; i-- ) if ( skipPointer[ i ] <= p ) break;
			
			if ( DEBUG && i >=0 ) System.err.println( "{" + this + "} " + "Safely after for with i=" + i + ", P[i]=" + skipPointer[i] + ", A[i]=" + skipAmount[i] );
			if ( DEBUG && i < 0 ) System.err.println( "{" + this + "} " + "Safely after for with negative i=" + i );

			if ( i >= 0 ) {
				if ( DEBUG ) System.err.println( "{" + this + "} " + "[" + ibs.readBits() + "] Skipping " + ( skipAmount[ i ] - ( ibs.readBits() - readBitsAtLastSkipTower ) ) + " bits (" + ( ( ( k & - ( 1 << i ) ) + ( 1 << i ) ) * skipIndex.quantum - cacheOffset ) + " records) to get to document pointer " + skipPointer[ i ] );

				ibs.skip( skipAmount[ i ] - ( ibs.readBits() - readBitsAtLastSkipTower ) );
				state = BEFORE_TOWER;
				currentDocument = skipPointer[ i ];
				numberOfDocumentRecord += ( ( k & - ( 1 << i ) ) + ( 1 << i ) ) * skipIndex.quantum - cacheOffset;
				if ( endOfList() ) {
					if ( DEBUG ) System.err.println( "{" + this + "} " + "end-of-list, returning " + ( currentDocument == p ) );
					return currentDocument == p;
				}
				readTower( p );
			}
			else break;
		}

		if ( DEBUG ) System.err.println( "{" + this + "} " + "Completing sequentially, lastDoc=" + currentDocument + ", numberOfDocumentRecord=" + numberOfDocumentRecord + ", p=" + p );

		while( currentDocument < p ) {
			if ( DEBUG ) System.err.println( "{" + this + "} " + "Skipping sequentially (second), lastDoc=" + currentDocument + ", numberOfDocumentRecord=" + numberOfDocumentRecord + ", p=" + p );

			if ( endOfList() ) {
				if ( DEBUG ) System.err.println( "{" + this + "} " + "end-of-list, returning false" );
				return false;
			}
			readDocumentPointer();
		}

		if ( DEBUG ) System.err.println( "{" + this + "} " + "Returning true" );
		return true;
	}
	
	public String toString() {
		return "SkipIndexReader term=" + term + " docNum=" + numberOfDocumentRecord;
	}

}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
