package it.unimi.dsi.mg4j.index;

import it.unimi.dsi.fastutil.ints.Int2IntRBTreeMap;
import it.unimi.dsi.mg4j.io.FastByteArrayOutputStream;
import it.unimi.dsi.mg4j.io.NullOutputStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Properties;


/*
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/** Provides facilities to write skip inverted indices,
 *  that is, inverted indices with an additional skip structure. A skip inverted index
 *  allows one to skip ahead when reading inverted lists. More specifically,
 *  when reading the inverted list relative to a certain term, one may want to
 *  decide to skip all document records that concern documents with pointer
 *  less than a given integer. In a normal inverted index this is impossible:
 *  one would have to read all document records sequentially (by the way,
 *  this is what {@link IndexReader#skipTo(int)} does).
 *
 *  <H2>Skip structure</H2>
 *
 *  <P>Before reading this section, please consult {@link it.unimi.dsi.mg4j.index}
 *  to understand what is the usual structure of an inverted index.
 *
 *  <P>A skip inverted index is just like a usual inverted index, with the only
 *  difference that some of the document records contain additional
 *  information that allow to skip document records without actually reading them. 
 * 
 *  <P>The classical approach to skipping is to embed pointers that allow
 *  to skip a fixed amount of document records. MG4J, instead, embeds a perfect, static skip list in every
 *  inverted list using a new method to represent and compress skip towers. As a result, MG4J can embed
 *  tall towers with small quanta (e.g., quantum 64 and height 8) with a very small growth
 *  of the index size. The details of the method will be presented in a forthcoming paper.
 * 
 *  <P>The document records that contain a skip tower are called <em>skip document records</em>; 
 *  skip towers are stored within a document record just after the document pointer and
 *  just before the document data.
 *
 *  <P>The construction of a skip inverted index depends on two positive
 *  integers, <var>q</var> and <var>h</var>, called the <em>quantum</em> and
 *  the <em>height</em>. For sake of simplicity, let
 *  <var>H</var>=2<sup><var>h</var></sup> and
 *  <var>w</var>=<var>H</var><var>q</var>.
 *
 *  <P>Consider the records in a certain inverted list; these records are (logically) grouped
 *  from the left into <em>blocks</em> of width <var>w</var> (the last block, of
 *  course, may contain less than <var>w</var> records; see below).
 *  Hence, the records within a certain block are indexed from 0 to <var>w</var>-1.
 *  The skip records are only those whose index (within the block) is of the form
 *  <var>kq</var>, where <var>k</var> ranges from 0 to <var>H</var>-1; the number
 *  <var>k</var> is called the <em>skip index</em> of the skip record (within the
 *  block that contains it).
 *
 *  <P>The skip tower of the skip record with skip index <var>k</var> contains
 *  LSB(<var>k</var>)+1 entries, where LSB(<var>k</var>) is the index of the least significant bit
 *  of <var>k</var>, a.k.a. the number of
 *  trailing zeroes in the <var>h</var>-long binary expansion of
 *  <var>k</var>. Hence, the 0-th skip tower has <var>h</var>+1 entries, the
 *  <var>H</var>/2-th skip tower has <var>h</var> entries, and so on. Skip
 *  entries within each skip tower have an index, called the <em>skip tower
 *  index</em>: if a skip tower contains <var>s</var>+1 entries, their skip
 *  tower index will range from 0 to <var>s</var> (inclusive).
 *
 *  <P>Each entry within a skip tower is really made of two numbers: the
 *  <em>skip pointer</em> and the <em>skip amount</em>. Consider the
 *  <var>i</var>-th entry of the <var>k</var>-th skip tower (the one contained in
 *  the record <var>kq</var> within the block); let <var>P</var> be the skip
 *  pointer and <var>A</var> be the skip amount.
 *
 *  <P>The entry we are considering points to another skip record: the one
 *  whose skip index is <var>k</var>+2<sup><var>i</var></sup>. For example, the
 *  last entry (the entry with skip tower index <var>h</var>) in the skip tower
 *  of skip index 0 points to the <em>H</em>-th skip record within the block
 *  (that is: the first record in the following block, if any). The
 *  next-to-last entry (skip tower index <var>h</var>-1) in the same skip tower
 *  points to the <var>H</var>/2-th skip record within the block (that is: the
 *  record in the middle of the block).  The entry number 0 in the same skip
 *  tower points to the skip record with skip index 1.
 *
 *  <P>Now, <var>P</var> is the document pointer of the pointed skip document
 *  record, and <var>A</var> is the number of bits to skip from the end of the
 *  skip tower to the start of the pointed (the
 *  <var>k</var>+2<sup><var>i</var></sup>-th) tower.
 *
 *  <H2>Defective blocks</H2>
 *
 *  <P>The last block within an inverted list may contain less than
 *  <var>w</var> records, and in this case we say that it is defective. If a
 *  defective block contains only <var>C</var> &lt; <var>w</var> records, the
 *  number of entries in the <var>k</var>-th skip tower is not
 *  LSB(<var>k</var>)+1, but rather
 *  1+min(LSB(<var>k</var>),MSB(&lfloor;<var>C</var>/<var>q</var>&rfloor; - <var>k</var>)), where
 * MSB(<var>k</var>) is the most significant bit of <var>k</var>, and moreover MSB(0) = -1 by definition.
 *  In particular, defective quanta have an empty tower.
 *
 *  <H2>Class usage and statistics</H2>
 *
 *  <P>This class is used in the same way as {@link IndexWriter}. It further collects
 *  more statistics concerning the {@link #towerData compression of skips}.
 *
 * @author Paolo Boldi
 * @author Sebastiano Vigna
 * @since 0.6
 */
public class SkipIndexWriter extends IndexWriter {

	/** Maximum number of trials when optimising the entry bit length. */
	private final static int MAX_TRY = 32;

	/** The parameter <code>h</code> (the maximum height of a skip tower). */
	private int h;

	/** The parameter <code>q</code> (2<var><sup>h</sup>q</var> documents record are kept in the cache). */
	private int q;

	/** We have <var>w</var>=2<sup><var>h</var></sup><var>q</var>. */
	private int w;

	/** The number of document records written in the cache containing the current block. */
	private int cache;

	private double pointerQuantumSigma;

	/** The <var>k</var>-th entry of this array contains the document pointer of the <var>k</var>-th
	 *  skip document record within the current block. For sake of simplicity, <code>pointer[cache]</code>
	 *  contains the first document pointer within the next block. */
	private int[] skipPointer;

	/** The {@link OutputBitStream}s where cached document pointers are written. */
	private OutputBitStream[] cachePointer;

	/** The {@link FastByteArrayOutputStream}s underlying <code>cachePointer</code> . */
	private FastByteArrayOutputStream[] cachePointerByte;

	/** The {@link OutputBitStream}s where cached skip towers are written. Indices are skip
	 *  indices. */
	private OutputBitStream[] cacheSkip;

	/** An array whose entries (as many as those of {@link #cacheSkip} are all {@link #bitCount}. */
	private OutputBitStream[] cacheSkipBitCount;

	/** The {@link FastByteArrayOutputStream}s underlying <code>cacheSkip</code> . Indices are skip
	*  indices. */
	private FastByteArrayOutputStream[] cacheSkipByte;

	/** The {@link OutputBitStream}s where cached document data are written. */
	private OutputBitStream[] cacheData;

	/** The {@link FastByteArrayOutputStream}s underlying <code>cacheData</code> . */
	private FastByteArrayOutputStream[] cacheDataByte;

	/** An {@link OutputBitStream} wrapping a {@link NullOutputStream} for code-length preview. */
	private OutputBitStream bitCount;

	/** The sum of all tower data computed so far. */
	public TowerData towerData;

	/** The number of bits written for quantum lengths. */
	public long bitsForQuantumBitLengths;

	/** The number of bits written for entry lenghts. */
	public long bitsForEntryBitLengths;

	/** The number of written blocks. */
	public long numberOfBlocks;

	/** An estimate on the number of bits occupied per tower entry in the last written cache, or -1 if no cache has been
	 * written for the current inverted list. */
	public int prevEntryBitLength;

	/** An estimate on the number of bits occupied per quantum in the last written cache, or -1 if no cache has been
	 * written for the current inverted list. */
	public int prevQuantumBitLength;

	private static final boolean ASSERTS = true;
	private static final boolean DEBUG = false;
	private final static boolean STATS = false;
	private java.io.PrintWriter pointerSkipStats;
	private java.io.PrintWriter pointerNewSkipStats;
	private java.io.PrintWriter amountSkipStats;
	private java.io.PrintWriter amountNewSkipStats;
	private java.io.PrintWriter amountNewLengthStats;

	{ if ( DEBUG && ASSERTS && STATS ); }

	/** Creates a new skip index writer, with the specified underlying {@link OutputBitStream}.
	 *
	 * @param obs the underlying output bit stream.
	 * @param offset the offset bit stream.
	 * @param N the number of documents in the collection to be indexed.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexWriter} introduction).
	 * @param q the cache contains at most <var>2<sup>h</sup></var> document records.
	 * @param h the maximum height of a skip tower.
	 */
	public SkipIndexWriter( OutputBitStream obs, OutputBitStream offset, int N,
		int flags, int q, int h ) {
		super( obs, offset, N, flags );
		this.h = h;
		this.q = q;

		int two2h = 1 << h;
		w = two2h * q;

		if ( DEBUG ) {
			System.err.println( "Cache will contain at most " + w + " records (q=" + q + ",h=" + h + ")" );
			System.err.print( "Skip records will be " );
			for ( int i = 0; i < two2h; i++ ) System.err.print( ( i * q ) + " " );
			System.err.println();
		}

		towerData = new TowerData();
		cacheData = new OutputBitStream[two2h];
		cacheDataByte = new FastByteArrayOutputStream[two2h];
		cachePointer = new OutputBitStream[two2h];
		cachePointerByte = new FastByteArrayOutputStream[two2h];

		for ( int i = 0; i < two2h; i++ ) {
			cacheData[i] = new OutputBitStream( cacheDataByte[i] = new FastByteArrayOutputStream(), 0 );
			cachePointer[i] = new OutputBitStream( cachePointerByte[i] = new FastByteArrayOutputStream(), 0 );
		}

		cacheSkip = new OutputBitStream[two2h];
		cacheSkipBitCount = new OutputBitStream[two2h];
		cacheSkipByte = new FastByteArrayOutputStream[two2h];

		for ( int i = 0; i < two2h; i++ ) {
			cacheSkip[ i ] = new OutputBitStream( cacheSkipByte[i] = new FastByteArrayOutputStream() );
			cacheSkipBitCount[ i ] = new OutputBitStream( NullOutputStream.getInstance(), 0 );
		}
	
		skipPointer = new int[ two2h + 1 ];
		distance = new long[ two2h + 1 ];

		bitCount = new OutputBitStream( NullOutputStream.getInstance(), 0 );

		if ( STATS ) {
			try {
				pointerSkipStats = new PrintWriter( new java.io.FileWriter( "pointerSkip.stats" ) );
				pointerNewSkipStats = new PrintWriter( new FileWriter( "pointerNewSkip.stats" ) );
				amountSkipStats = new PrintWriter( new java.io.FileWriter( "amountSkip.stats" ) );
				amountNewSkipStats = new PrintWriter( new FileWriter( "amountNewSkip.stats" ) );
				amountNewLengthStats = new PrintWriter( new FileWriter( "amountNewLength.stats" ) );
			} catch ( IOException e ) {
				System.err.println( "Can't open stat files: " + e );
			}
		}
	}

	/** Creates a new skip index writer, with the specified underlying {@link OutputBitStream},
	 *  without an associated offset bit stream.
	 *
	 * @param obs the underlying output bit stream.
	 * @param N the number of documents in the collection to be indexed.
	 * @param flags a bit mask setting the coding techniques to be used (see the {@link IndexWriter} introduction).
	 * @param q the cache contains at most <var>2<sup>h</sup></var> document records.
	 * @param h the maximum height of a skip tower.
	 */
	public SkipIndexWriter( OutputBitStream obs, int N, int flags, int q, int h ) {
		this( obs, null, N, flags, q, h );
	}

	public long newInvertedList() throws java.io.IOException, IllegalStateException {
		if ( cache != 0 ) writeOutCache( -1 );
		return super.newInvertedList();
	}

	public int writeFrequency( int frequency ) throws java.io.IOException, IllegalStateException {
		int res = super.writeFrequency( frequency );
		pointerQuantumSigma = Math.sqrt( q * ( 1 - relativeFrequency ) ) / relativeFrequency;
		prevQuantumBitLength = prevEntryBitLength = -1;	
		if ( ASSERTS ) assert pointerQuantumSigma != 0 || relativeFrequency == 1;
		if ( DEBUG ) System.out.println( "----------- " + currentTerm );

		return res;
	}
	/** Starts a new inverted list. The previous inverted list, if any, is actually written
	 *  to the underlying bit stream; then, the stream is byte-aligned, and <code>frequency</code>
	 *  is written onto the underlying stream
	 *  using the specified coding (the number is decremented by one first, as it cannot be 0).
	 *
	 *  @param frequency the number of document records that this inverted list will contain.
	 *  @return the position (in bytes) of the underlying bit stream where the new inverted
	 *  list starts.
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if too few records were written for the previous inverted
	 *  list.
	 *  @deprecated Use {@link #newInvertedList()} followed by {@link #writeFrequency(int)}.
	 */
	public long newInvertedList( int frequency ) throws java.io.IOException, IllegalStateException {
		if ( DEBUG ) System.err.println( "New inverted list with frequency " + frequency );

		if ( cache != 0 ) writeOutCache( -1 );

		final long res = super.newInvertedList();
		super.writeFrequency( frequency );
		pointerQuantumSigma = Math.sqrt( q * ( 1 - relativeFrequency ) ) / relativeFrequency;
		prevQuantumBitLength = prevEntryBitLength = -1;	
		if ( ASSERTS ) assert pointerQuantumSigma != 0 || relativeFrequency == 1;
		if ( DEBUG ) System.out.println( "----------- " + currentTerm );

		return res;
	}

	public OutputBitStream newDocumentRecord() throws java.io.IOException, IllegalStateException {
		super.newDocumentRecord();
		return cacheData[ ( cache % w ) / q ];
	}

	public int writeDocumentPointer( OutputBitStream out, int pointer ) throws java.io.IOException, IllegalStateException {
		// If the previous block is over, write it out!
		if ( cache == w ) writeOutCache( pointer );

		// Record data pointer if we are on a skip; otherwise, write it to the cache.
		if ( cache % q == 0 ) {
			skipPointer[ cache / q ] = pointer;
			return super.writeDocumentPointer( cachePointer[ cache++ / q ], pointer );
		} 
		else return super.writeDocumentPointer( cacheData[ cache++ / q ], pointer );
	}

	/**
	 *  @deprecated Use {@link #newDocumentRecord()} followed by {@link #writeDocumentPointer(OutputBitStream, int)}.
	 */
	public OutputBitStream newDocumentRecord( int pointer ) throws java.io.IOException, IllegalStateException {
		if ( this.frequency < 0 ) throw new IllegalStateException( "Trying to write a new document record without calling newInvertedList" );
		if ( this.frequency == this.writtenDocuments ) throw new IllegalStateException( "Document record overflow (written " + this.frequency + " already)" );

		if ( DEBUG ) System.err.println( "New document record with pointer " + pointer );

		// Record that a new document is starting
		super.newDocumentRecord();

		// If the previous block is over, write it out!
		if ( cache == w ) writeOutCache( pointer );

		// Record data pointer if we are on a skip; otherwise, write it to the cache.
		if ( cache % q == 0 ) {
			skipPointer[cache / q] = pointer;
			bitsForPointers += super.writeDocumentPointer( cachePointer[cache / q ], pointer );
		} else
			bitsForPointers += super.writeDocumentPointer( cacheData[cache / q ], pointer );

		return cacheData[cache++ / q];
	}

	private int writeOutPointer( OutputBitStream out, int pointer ) throws java.io.IOException {
		if ( frequency == numberOfDocuments ) return 0; // We do not write pointers for everywhere occurring terms.

		switch ( pointerCoding ) {
			case GAMMA:
				return out.writeGamma( pointer - lastDocument - 1 );
			case DELTA:
				return out.writeDelta( pointer - lastDocument - 1 );
			case GOLOMB:
				return out.writeGolomb( pointer - lastDocument - 1, b, log2b );
			default:
				throw new IllegalStateException( "The required pointer coding (" + pointerCoding + ") is not supported." );
		}
	}

	/** Closes the underlying bit stream.
	 *
	 *  @throws IOException if an exception is thrown by the underlying stream
	 *  @throws IllegalStateException if too few records were written for the previous inverted
	 *  list.
	 */
	public void close() throws java.io.IOException, IllegalStateException {

		if ( cache != 0 ) writeOutCache( -1 );

		super.close();

		if ( STATS ) {
			pointerSkipStats.close();
			pointerNewSkipStats.close();
			amountSkipStats.close();
			amountNewSkipStats.close();
			amountNewLengthStats.close();
		}
	}

	/** A structure maintaining statistical data about tower construction. */

	public static class TowerData {
		/** The number of bits written for skip amounts at the top of a tower. */
		public long bitsForTopSkipAmounts;

		/** The number of bits written for skip pointers at the top of a tower. */
		public long bitsForTopSkipPointers;

		/** The number of bits written for skip amounts in the lower part of a tower. */
		public long bitsForLowerSkipAmounts;

		/** The number of bits written for skip pointers in the lower part of a tower. */
		public long bitsForLowerSkipPointers;

		/** The number of bits written for tower lengths. */
		public long bitsForTowerLengths;

		/** The number of written skip towers. */
		public long numberOfSkipTowers;

		/** The number of written skip entries. */
		public long numberOfSkipEntries;

		/** Clear all fields of this tower data. */

		void clear() {
			bitsForTopSkipAmounts = 0;
			bitsForTopSkipPointers = 0;
			bitsForLowerSkipAmounts = 0;
			bitsForLowerSkipPointers = 0;
			bitsForTowerLengths = 0;
			numberOfSkipTowers = 0;
			numberOfSkipEntries = 0;
		}


		/** Returns the overall number of bits used for skip pointers.
		 * @return the overall number of bits used for skip pointers.
		 */
		public long bitsForSkipPointers() { return bitsForTopSkipPointers + bitsForLowerSkipPointers; }

		/** Returns the overall number of bits used for skip amounts. 
		 * @return the overall number of bits used for skip amounts.
		 */
		public long bitsForSkipAmounts() { return bitsForTopSkipAmounts + bitsForLowerSkipAmounts; }

		/** Returns the overall number of bits used for tower entries (bits for tower lengths are not included).
		 * @return the overall number of bits used for tower entries.
		 */
		public long bitsForEntries() { return bitsForSkipPointers() + bitsForSkipAmounts(); }

		/** Returns the overall number of bits used for towers.
		 * @return the overall number of bits used for towers.
		 */
		public long bitsForTowers() { return bitsForTowerLengths + bitsForEntries(); }

		/** Returns the number of lower entries, that is, the number of entries minus the number of towers.
		 * @return the number of lower entries.
		 */
		public long numberOfLowerEntries() { return numberOfSkipEntries - numberOfSkipTowers; }
	}
	


	/** The <var>k</var>-th entry of this array contains the number of bits from the start of
	 * the <var>k</var>-th skip tower up to the end of the current block (more precisely,
	 * to the point that should be reached via skipping, which is just after the document pointer).
	 * Indices are skip indices. It is used just by {@link tryTower()}, but it is declared here for efficiency.
	 */
	private long[] distance;

	

	private void tryTower( final int quantumBitLength, final int entryBitLength, long toTheEnd, final OutputBitStream[] skip, final TowerData towerData ) throws java.io.IOException {
		int i, k, s;
		long d;
		int prevPointerDelta, basePointer;

		for ( k = ( cache - 1 ) / q; k >= 0; k-- ) {
			// Where are we? At the end of the k-th quantum.
			toTheEnd += cacheData[k].writtenBits();

			// Where are we? Just after the k-th skip tower.
			// Compute the maximum valid index of the skip tower (*MUST* be kept in sync with the subsequent loop).
			s = ( k == 0 ) ? h : Fast.leastSignificantBit( k );

			// This test handles defective blocks. In particular, defective quanta set s to -1.
			if ( cache < w ) s = Math.min( s, Fast.mostSignificantBit( ( cache / q ) - k ) );

			skip[ k ].writtenBits( 0 );

			if ( s >= 0 ) {
				if ( DEBUG ) System.out.print( "% (" + k + ") " );

				basePointer = skipPointer[k];
				prevPointerDelta = skipPointer[ k + ( 1 << s ) ] - basePointer;

				/* We write the top of a pointer skip tower as a difference from the average skip, which is
				 * q 2^s / relativeFrequency. */
				towerData.bitsForTopSkipPointers += skip[k].writeGolomb( 
					Fast.int2nat( prevPointerDelta - (int) Math.round( ( q * ( 1 << s ) ) / relativeFrequency ) ),
					Fast.gaussianGolombModulus( pointerQuantumSigma * Math.sqrt( 1 << s ) ) 
				);
				
				/* Then we write the skip from the end to the tower to q 2^s quanta farther. */
				towerData.bitsForTopSkipAmounts += skip[k].writeDelta( Fast.int2nat( 
					(int) ( ( toTheEnd - distance[k + ( 1 << s )] ) -
					( quantumBitLength * ( 1 << s ) + entryBitLength * ( ( 1 << s + 1 ) - s - 2 ) ) ) 
				) );

				if ( DEBUG ) System.out.print( skipPointer[k + ( 1 << s )] + ":" + ( toTheEnd - distance[k + ( 1 << s )] ) + " " );

				// Produce a (single) tower of height s
				for ( i = s - 1; i >= 0; i-- ) {
					towerData.bitsForLowerSkipPointers += skip[k].writeGolomb( 
						Fast.int2nat( ( prevPointerDelta / 2 ) - ( skipPointer[k + ( 1 << i )] - basePointer ) ),
						Fast.gaussianGolombModulus( pointerQuantumSigma * Math.sqrt( ( 1 << i ) / 2.0 ) ) 
					);

					prevPointerDelta = skipPointer[k + ( 1 << i )] - basePointer;

					towerData.bitsForLowerSkipAmounts += skip[k].writeDelta( 
						Fast.int2nat( (int) ( ( ( toTheEnd - distance[k + ( 1 << ( i + 1 ) )] - entryBitLength * i ) / 2 ) -
								( toTheEnd - distance[k + ( 1 << i )] ) ) ) 
					);

					if ( DEBUG ) System.out.print( skipPointer[ k + ( 1 << i ) ] + ":" + ( toTheEnd - distance[k + ( 1 << i )] ) + " " );
				}

				if ( s > 0 ) { // No length for single-entry towers.
					d = bitCount.writeDelta( Fast.int2nat( (int) skip[k].writtenBits() - ( s + 1 ) * entryBitLength ) );
					towerData.bitsForTowerLengths += d;
					toTheEnd += d;
				}

				toTheEnd += skip[k].writtenBits();

				if ( DEBUG ) System.out.print( " (" + (int) skip[k].writtenBits() + " bits)" );

				towerData.numberOfSkipEntries += ( s + 1 );
				towerData.numberOfSkipTowers++;

				if ( DEBUG ) System.out.println();
			}

			distance[ k ] = toTheEnd;

			// Where are we? Just before the beginning of the k-th skip tower
			toTheEnd += cachePointer[ k ].writtenBits();

			// Where are we? Just before the beginning of the k-th document record
		}
	}

	/** Write out the cache content.
	 * 
	 * @param nextPointer the first pointer of the next block, or -1 if this is the last block.
	 */
	private void writeOutCache( int nextPointer ) throws java.io.IOException {
		if ( DEBUG ) System.err.println( "Entered writeOutCache() with cache=" + cache + " (H is " + ( 1 << h ) + ", B is " + w + ")" );

		/* Number of bits to go after the first pointer of the first record of the next block (or, if there
		   is no other block in the current list, to go to the end of the list). */
		long toTheEnd;

		// Record the new document pointer for the highest tower
		int nextAfter = ( ( cache + q ) - 1 ) / q; // This is ceil( cache / q )

		if ( nextPointer >= 0 ) {
			skipPointer[nextAfter] = nextPointer;
			toTheEnd = writeOutPointer( bitCount, nextPointer );
		} else {
			skipPointer[nextAfter] = currentDocument + 1; // Fake: just for the last block
			toTheEnd = 0;
		}

		distance[nextAfter] = 0;

		int k, s;
		long d;

		// Compute quantum length in bits (without towers)
		int quantumBitLength = 0, entryBitLength = 0;

		for ( d = k = 0; k <= ( ( cache - 1 ) / q ); k++ ) d += ( cachePointer[k].writtenBits() + cacheData[k].writtenBits() );
		quantumBitLength = (int)( ( ( d * q ) + ( cache - 1 ) ) / cache );

		final TowerData td = new TowerData();
		final Int2IntRBTreeMap candidates = new Int2IntRBTreeMap(); 

		/* As a first try, we compute the tower costs using 0 as average entry bit length. */
		tryTower( quantumBitLength, 0, toTheEnd, cacheSkipBitCount, td );
		
		if ( td.numberOfSkipTowers > 0 ) { // There actually is at least a tower.
			/* Now we repeat this operation, trying to obtain the best value for the
			 * average entry bit length. 
			 */

			while( candidates.size() < MAX_TRY && ! candidates.containsValue( entryBitLength = (int)( td.bitsForTowers() / td.numberOfSkipEntries ) ) ) {
				td.clear();
				tryTower( quantumBitLength, entryBitLength, toTheEnd, cacheSkipBitCount, td );
				//System.out.print( "[" + entryBitLength + "->" + td.bitsForTowers() / td.numberOfSkipEntries + "]" );
				candidates.put( (int)( td.bitsForTowers() / td.numberOfSkipEntries ), entryBitLength );
			}

			if ( ASSERTS ) assert candidates.size() < MAX_TRY;

			entryBitLength = candidates.get( candidates.firstIntKey() );
			//System.out.println( "(" + entryBitLength + ")" );
			tryTower( quantumBitLength, entryBitLength, toTheEnd, cacheSkip, towerData );
		}

		// Ready to write out cache
		for ( k = 0; k <= ( ( cache - 1 ) / q ); k++ ) {
			if ( DEBUG ) System.err.println( "Before writing document pointer, position is " + obs.writtenBits() );

			/* See comments above. */
			s = ( k == 0 ) ? h : Fast.leastSignificantBit( k );

			if ( cache < w ) s = Math.min( s, Fast.mostSignificantBit( ( cache / q ) - k ) );

			d = cachePointer[k].writtenBits();
			cachePointer[k].flush();
			obs.write( cachePointerByte[k].array, (int) d );

			if ( DEBUG ) System.err.println( "After writing document pointer, position is " + obs.writtenBits() );

			d = cacheSkip[k].writtenBits();
			cacheSkip[k].flush();

			if ( s >= 0 ) {
				if ( k == 0 ) {
					if ( prevQuantumBitLength < 0 ) {
						bitsForQuantumBitLengths += obs.writeLongDelta( quantumBitLength );
						bitsForEntryBitLengths += obs.writeLongDelta( entryBitLength );
					}
					else {
						bitsForQuantumBitLengths += obs.writeLongDelta( Fast.int2nat( quantumBitLength - prevQuantumBitLength ) );
						bitsForEntryBitLengths += obs.writeLongDelta( Fast.int2nat( entryBitLength - prevEntryBitLength ) );
					}

					prevQuantumBitLength = quantumBitLength;
					prevEntryBitLength = entryBitLength;

					numberOfBlocks++;

					if ( DEBUG ) System.err.println( "Writing quantum bit length " + quantumBitLength + ", entry bit length " + entryBitLength );
				}

				if ( s > 0 ) obs.writeDelta( Fast.int2nat( (int)d - entryBitLength * ( s + 1 ) ) ); // No length for single-entry towers.
			} else if ( ASSERTS ) assert d == 0;

			obs.write( cacheSkipByte[k].array, (int) d );

			if ( DEBUG ) System.err.println( "After skip tower, position is " +	obs.writtenBits() );

			d = cacheData[k].writtenBits();
			cacheData[k].flush();
			obs.write( cacheDataByte[k].array, (int) d );
		}

		// Clean used caches
		for ( k = 0; k <= ( ( cache - 1 ) / q ); k++ ) {
			cachePointerByte[k].reset();
			cachePointer[k].writtenBits( 0L );

			cacheSkipByte[k].reset();
			cacheSkip[k].writtenBits( 0L );

			cacheDataByte[k].reset();
			cacheData[k].writtenBits( 0L );
		}

		cache = 0;

		if ( ASSERTS ) assert obs.writtenBits() == writtenBits();
	}

	/** This method returns the overall number of bits written onto the underlying stream.
	 *
	 *  @return the number of bits written, according to the variables keeping statistical records.
	 */
	public long writtenBits() {
		return super.writtenBits() + towerData.bitsForTopSkipPointers +
		towerData.bitsForTopSkipAmounts + towerData.bitsForLowerSkipPointers +
		towerData.bitsForLowerSkipAmounts + towerData.bitsForTowerLengths +
		bitsForQuantumBitLengths + bitsForEntryBitLengths;
	}
	
	public Properties properties() {
		Properties result = super.properties();
		result.setProperty( IndexProperties.INDEXCLASS, SkipIndex.class.getName() );
		result.setProperty( SkipIndexProperties.SKIPQUANTUM, "" + q );
		result.setProperty( SkipIndexProperties.SKIPHEIGHT, "" + h );
		return result;	
	}

}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
