package it.unimi.dsi.mg4j.io;

import it.unimi.dsi.mg4j.util.MutableString;

import java.io.IOException;
import java.io.PrintStream;

/** A debugging wrapper for output bit streams.
 *
 * <P>This class can be used to wrap an output bit stream. The semantics of the
 * resulting write operations is unchanged, but each operation will be logged.
 *
 * <P>To simplify the output, some operations have a simplified representation. In particular:
 *
 * <dl>
 * <dt><samp>|</samp>
 * <dd>{@link OutputBitStream#flush() flush()};
 * <dt><samp>-></samp>
 * <dd>{@link OutputBitStream#position(long) position()};
 * <dt><samp>[</samp>
 * <dd>creation;
 * <dt><samp>]</samp>
 * <dd>{@link OutputBitStream#close() close()};
 * <dt><samp>{<var>x</var>}</samp>
 * <dd>explicit bits;
 * <dt><samp>{<var>x</var>:<var>b</var>}</samp>
 * <dd>minimal binary coding of <var>x</var> with bound <var>b</var>;
 * <dt><samp>{<var>M</var>:<var>x</var>}</samp>
 * <dd>write <var>x</var> with coding <var>M</var>; the latter can be U (unary), g (&gamma;), z (&zeta;), d (&delta;), G (Golomb), GS (skewed Golomb);
 * when appropriate, <var>x</var> is followed by an extra integer (modulus, etc.).
 * </dl>
 *
 * @author Paolo Boldi
 * @author Sebastiano Vigna
 * @since 0.7.1
 */

public class DebugOutputBitStream extends OutputBitStream {

	private PrintStream pw;
	private OutputBitStream obs;
	 
	/** Creates a new debug output bit stream wrapping a given output bit stream and logging on a given writer.
	 *
	 * @param obs the output bit stream to wrap.
	 * @param pw a print stream that will receive the logging data.
	 */
	public DebugOutputBitStream( OutputBitStream obs, PrintStream pw ) {
		this.obs = obs;
		this.pw = pw;
		pw.print( "[" );
	}

	/** Creates a new debug output bit stream wrapping a given output bit stream and logging on standard error.
	 *
	 * @param obs the output bit stream to wrap.
	 */
	public DebugOutputBitStream( OutputBitStream obs ) {
		this( obs, System.err );
	}

	public void flush() throws java.io.IOException {
		pw.print( " |" );
		obs.flush();
	}

	public void close() throws java.io.IOException {
		pw.print( " |]" );
		obs.close();
	}

	public long writtenBits() {
		return obs.writtenBits();
	}

	public void writtenBits( long writtenBits ) {
		obs.writtenBits( writtenBits );
	}

	public int align() throws IOException {
		pw.print( " |" );
		return obs.align();
	}

    public void position( final long position ) throws IOException {
		pw.print( " ->" + position );
		obs.position( position );
	}

	private static MutableString byte2Binary( int x ) {
		MutableString s = new MutableString();
		for( int i = 0 ; i < 8; i++ ) {
			s.append( (char)( '0' + ( x % 2 ) ) );
			x >>= 1;
		}
		return s.reverse();
	}

	private static MutableString int2Binary( long x, int len ) {
		MutableString s = new MutableString();
		for( int i = 0 ; i < 64; i++ ) {
			s.append( (char)( '0' + ( x % 2 ) ) );
			x >>= 1;
		}
		return s.length( len ).reverse();
	}

	public int write( byte bits[], int len ) throws java.io.IOException {
		MutableString s = new MutableString( " {" );
		for( int i = 0; i < bits.length; i++ ) s.append( byte2Binary( bits[ i ] ) );
		pw.print( s.length( len ).append( "}" ) );
		return obs.write( bits, len );
	}

	public int writeBit( boolean bit ) throws java.io.IOException {
		pw.print( " {" + ( bit ? '1' : '0' ) + "}" );
		return obs.writeBit( bit );
	}

	public int writeBit( int  bit ) throws java.io.IOException {
		pw.print( " {" + bit + "}" );
		return obs.writeBit( bit );
	}

	public int writeInt( int x, int len ) throws java.io.IOException {
		pw.print( " {" + int2Binary( x, len ) + "}" );
		return obs.writeInt( x, len );
	}

	public int writeLong( long x, int len ) throws java.io.IOException {
		pw.print( " {" + int2Binary( x, len ) + "}" );
		return obs.writeLong( x, len );
	}

	public int writeUnary( int x ) throws java.io.IOException {
		pw.print( " {U:" + x + "}" );
		return obs.writeUnary( x );
	}

	public long writeLongUnary( long x ) throws java.io.IOException {
		pw.print( " {U:" + x + "}" );
		return obs.writeLongUnary( x );
	}

	public int writeGamma( int x ) throws java.io.IOException {
		pw.print( " {g:" + x + "}" );
		return obs.writeGamma( x );
	}

	public int writeLongGamma( long x ) throws java.io.IOException {
		pw.print( " {g:" + x + "}" );
		return obs.writeLongGamma( x );
	}

	public int writeDelta( int x ) throws java.io.IOException {
		pw.print( " {d:" + x + "}" );
		return obs.writeDelta( x );
	}

	public int writeLongDelta( long x ) throws java.io.IOException {
		pw.print( " {d:" + x + "}" );
		return obs.writeLongDelta( x );
	}

	public int writeMinimalBinary( int x, int b ) throws java.io.IOException {
		pw.print( " {m:" + x + "<" + b + "}" );
		return obs.writeMinimalBinary( x, b );
	}

	public int writeMinimalBinary( int x, int b, int log2b ) throws java.io.IOException {
		pw.print( " {m:" + x + "<" + b + "}" );
		return obs.writeMinimalBinary( x, b, log2b );
	}

	public int writeLongMinimalBinary( long x, long b ) throws java.io.IOException {
		pw.print( " {m:" + x + "<" + b + "}" );
		return obs.writeLongMinimalBinary( x, b );
	}

	public int writeLongMinimalBinary( long x, long b, int log2b ) throws java.io.IOException {
		pw.print( " {m:" + x + "<" + b + "}" );
		return obs.writeLongMinimalBinary( x, b, log2b );
	}

	public int writeGolomb( int x, int b ) throws java.io.IOException {
		pw.print( " {G:" + x + ":" + b + "}" );
		return obs.writeGolomb( x, b );
	}

	public int writeGolomb( int x, int b, int log2b ) throws java.io.IOException {
		pw.print( " {G:" + x + ":" + b + "}" );
		return obs.writeGolomb( x, b, log2b );
	}

	public long writeLongGolomb( long x, long b ) throws java.io.IOException {
		pw.print( " {G:" + x + ":" + b + "}" );
		return obs.writeLongGolomb( x, b );
	}

	public long writeLongGolomb( long x, long b, int log2b ) throws java.io.IOException {
		pw.print( " {G:" + x + ":" + b + "}" );
		return obs.writeLongGolomb( x, b, log2b );
	}

	public int writeSkewedGolomb( int x, int b ) throws java.io.IOException {
		pw.print( " {SG:" + x + ":" + b + "}" );
		return obs.writeSkewedGolomb( x, b );
	}

	public long writeLongSkewedGolomb( long x, long b ) throws java.io.IOException {
		pw.print( " {SG:" + x + ":" + b + "}" );
		return obs.writeLongSkewedGolomb( x, b );
	}

	public int writeZeta( int x, int k ) throws java.io.IOException {
		pw.print( " {z" + k + ":" + x + "}" );
		return obs.writeZeta( x, k );
	}

	public long writeLongZeta( long x, int k ) throws java.io.IOException {
		pw.print( " {z" + k + ":" + x + "}" );
		return obs.writeLongZeta( x, k );
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
