package it.unimi.dsi.mg4j.io;

import it.unimi.dsi.mg4j.util.Fast;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.channels.FileChannel;


/** Bit-level input stream.
 *
 * <blockquote><strong>IMPORTANT</strong>: In MG4J 0.6, this class has been completely
 * rewritten. It is much faster, but there are also incompabilities. Mainly,
 * unary representations are one-terminated sequences of zeroes (instead of
 * zero-terminated sequences of ones), and {@link #position(long)} sets a
 * <em>bit</em>, not a byte offset.</blockquote>
 *
 * <P>This class wraps any {@link InputStream} so that you can treat it as
 * <em>bit</em> stream.  Constructors and methods closely resemble those of
 * {@link InputStream}. Data can be read from such a stream in several ways:
 * reading an integer or long in fixed-width, unary, &gamma;, &delta;, &zeta; and Golomb
 * coding, or reading a number of bits that will be stored in a vector of
 * bytes. There is limited support for {@link #mark(int)}/{@link #reset()}
 * operations.
 *
 * <P>This class can also {@linkplain #InputBitStream(byte[]) wrap a byte
 * array}; this is much more lightweight than wrapping a {@link
 * FastByteArrayInputStream} wrapping the array. Overflowing the array
 * will cause an {@link java.io.EOFException}.
 * 
 * <P>Note that when reading using a vector of bytes bits are read in the
 * stream format (see {@link OutputBitStream}): the first bit is bit 7 of the
 * first byte, the eighth bit is bit 0 of the first byte, the ninth bit is bit
 * 7 of the second byte and so on. When reading integers using some coding,
 * instead, they are stored in the standard way, that is, in the <strong>lower</strong>
 * bits.
 *
 * <P>Additional features:
 * 
 * <UL>
 *
 * <LI>This class provides an internal buffer. By setting a buffer of
 * length 0 at creation time, you can actually bypass the buffering system:
 * Note, however, that several classes providing buffering have synchronised
 * methods, so using a wrapper instead of the internal buffer is likely to lead
 * to a performance drop.
 *
 * <LI>To work around the schizophrenic relationship between streams and random
 * access files in {@link java.io}, this class provides a {@link #flush()}
 * method that resets the internal state. At this point, you can safely reposition
 * the underlying stream and read again afterwards. For instance, this is safe
 * and will perform as expected:
 * <PRE>
 * FileInputStream fis = new FileInputStream( ... );
 * InputBitStream ibs = new InputBitStream( fis );
 * ... read operations on ibs ...
 * ibs.flush();
 * fis.getChannel().position( ... );
 * ... other read operations on ibs ...
 * </PRE>
 *
 * <P>As a commodity, an instance of this class will try to cast the underlying byte
 * stream to a {@link RepositionableStream} and to fetch by reflection the {@link
 * java.nio.channels.FileChannel} underlying the given input stream, in this
 * order.  If either reference can be successfully fetched, you can use
 * directly the {@link #position(long) position()} method with argument
 * <code>pos</code> with the same semantics of a {@link #flush()}, followed by
 * a call to <code>position(pos / 8)</code> (where the latter method belongs
 * either to the underlying stream or to its underlying file channel), followed
 * by a {@link #skip(long) skip(pos % 8)}.
 *
 * <LI>You can unget bits, integers and longs. A buffer {@link
 * #UNGET_BUFFER_SIZE} bytes long is allocated to this purpose. The buffer is
 * cleared on {@link #reset()}/{@link #flush()}.
 *
 * <li>This class supports also overflowing: if the overflow flag is set when
 * the EOF of the underlying {@link InputStream} is met, then zeroes are
 * subsequently read, and no exception is thrown. 
 *
 * </ul>
 *
 * <P><STRONG>This class is not synchronised</STRONG>. If multiple threads
 * access an instance of this class concurrently, they must be synchronised externally.
 *
 * @see OutputBitStream
 * @see java.io
 * @author Sebastiano Vigna
 * @since 0.1
 */

public class InputBitStream {

	/** The size of the {@link #unget(boolean)} buffer in bytes. */
	public final static int UNGET_BUFFER_SIZE = 16;
	/** The default size of the byte buffer in bytes (16Ki). */
	public final static int DEFAULT_BUFFER_SIZE = 16 * 1024;
	/** The underlying {@link InputStream}. */
	private InputStream is;
	/** The number of bits actually read from this bit stream. */
	private long readBits;
	/** Current bit buffer: the lowest {@link #fill} bits represent the current content (the remaining bits are undefined). */
	private int current;
	/** The stream buffer. */
	private byte buffer[];
	/** Whether we should use the byte buffer. */
	private boolean noBuffer;
	/** Current number of bits in the bit buffer (stored low). */
	private int fill;
	/** Current position in the byte buffer. */
	private int pos;
	/** Current number of bytes available in the byte buffer. */
	private int avail;
	/** Current position of the first byte in the byte buffer. */
	private long position;
	/** Byte buffer for ungetting bits. It is allocated on demand. */
	private byte ungetBytes[];
	/** How many bytes does the unget buffer hold (highest index to go out first) ?. */
	private int ungetCount = 0;
	/** Whether we allow overflowing the stream (getting zeroes). */
	private boolean overflow = false;
	/** In case we allow overflow, whether we are past EOF. */
	private boolean pastEOF = false;
	/** The cached file channel underlying {@link #is}, if any. */
	private FileChannel fileChannel;
	/** {@link #is} cast to a positionable stream, if possible. */
	private RepositionableStream rs;
	/** True if we are wrapping an array. */
	private boolean wrapping;

	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = false;

	/** Precomputed byte parsing for &gamma; coding. bits 0-7 contain the value, bits 8-15 the length of the code. 0 means that parsing goes beyond 8 bits. */
	private final static int[] GAMMA = {
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		7<<8 | 7, 7<<8 | 7, 7<<8 | 8, 7<<8 | 8, 7<<8 | 9, 7<<8 | 9, 7<<8 | 10, 7<<8 | 10, 7<<8 | 11, 7<<8 | 11, 7<<8 | 12, 7<<8 | 12, 7<<8 | 13, 7<<8 | 13, 7<<8 | 14, 7<<8 | 14,
		5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4,
		5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6,
		3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1,
		3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1, 3<<8 | 1,
		3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2,
		3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2, 3<<8 | 2,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
	};

	/** Precomputed byte parsing for &delta; coding. bits 0-7 contain the value, bits 8-15 the length of the code. 0 means that parsing goes beyond 8 bits. */
	private final static int[] DELTA = {
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		8<<8 | 7, 8<<8 | 8, 8<<8 | 9, 8<<8 | 10, 8<<8 | 11, 8<<8 | 12, 8<<8 | 13, 8<<8 | 14, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1,
		4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2,
		5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 3, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4, 5<<8 | 4,
		5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 5, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6, 5<<8 | 6,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
		1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0, 1<<8 | 0,
	};

	/** Precomputed byte parsing for &zeta;<sub>3</sub> coding. bits 0-7 contain the value, bits 8-15 the length of the code. 0 means that parsing goes beyond 8 bits. */
	private final static int[] ZETA_3 = {
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		7<<8 | 7, 7<<8 | 7, 7<<8 | 8, 7<<8 | 8, 7<<8 | 9, 7<<8 | 9, 7<<8 | 10, 7<<8 | 10, 7<<8 | 11, 7<<8 | 11, 7<<8 | 12, 7<<8 | 12, 7<<8 | 13, 7<<8 | 13, 7<<8 | 14, 7<<8 | 14,
		8<<8 | 15, 8<<8 | 16, 8<<8 | 17, 8<<8 | 18, 8<<8 | 19, 8<<8 | 20, 8<<8 | 21, 8<<8 | 22, 8<<8 | 23, 8<<8 | 24, 8<<8 | 25, 8<<8 | 26, 8<<8 | 27, 8<<8 | 28, 8<<8 | 29, 8<<8 | 30,
		8<<8 | 31, 8<<8 | 32, 8<<8 | 33, 8<<8 | 34, 8<<8 | 35, 8<<8 | 36, 8<<8 | 37, 8<<8 | 38, 8<<8 | 39, 8<<8 | 40, 8<<8 | 41, 8<<8 | 42, 8<<8 | 43, 8<<8 | 44, 8<<8 | 45, 8<<8 | 46,
		8<<8 | 47, 8<<8 | 48, 8<<8 | 49, 8<<8 | 50, 8<<8 | 51, 8<<8 | 52, 8<<8 | 53, 8<<8 | 54, 8<<8 | 55, 8<<8 | 56, 8<<8 | 57, 8<<8 | 58, 8<<8 | 59, 8<<8 | 60, 8<<8 | 61, 8<<8 | 62,
		3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0,
		3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0, 3<<8 | 0,
		4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1, 4<<8 | 1,
		4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2, 4<<8 | 2,
		4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3, 4<<8 | 3,
		4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4, 4<<8 | 4,
		4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5, 4<<8 | 5,
		4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6, 4<<8 | 6,
	};


	/** This (non-public) constructor exists just to provide fake initialisation.
	 */
	protected InputBitStream() {}

	
	/** Creates a new input bit stream wrapping a given input stream using a buffer of size {@link #DEFAULT_BUFFER_SIZE}.
	 *
	 * @param is the input stream to wrap.
	 */
	public InputBitStream( InputStream is ) {
		this( is, DEFAULT_BUFFER_SIZE );
	}


	/** Creates a new input bit stream wrapping a given input stream with a specified buffer size.
	 *
	 * @param is the input stream to wrap.
	 * @param bufSize the size in byte of the buffer; it may be 0, denoting no buffering.
	 */
	public InputBitStream( InputStream is, int bufSize ) {
		this.is = is;
		if ( ! ( this.noBuffer = bufSize == 0 ) ) this.buffer = new byte[ bufSize ];

		if ( is instanceof RepositionableStream ) rs = (RepositionableStream)is;
			
		if ( rs == null ) {
			try {
				fileChannel = (FileChannel)( is.getClass().getMethod( "getChannel", new Class[] {} ) ).invoke( is, new Object[] {} );
			}
			catch( IllegalAccessException e ) {}
			catch( IllegalArgumentException e ) {}
			catch( NoSuchMethodException e ) {}
			catch( java.lang.reflect.InvocationTargetException e ) {}
			catch( ClassCastException e ) {}
		}
	}

	/** Creates a new input bit stream wrapping a given byte array.
	 *
	 * @param a the byte array to wrap.
	 */
	public InputBitStream( byte a[] ) {
		is = NullInputStream.getInstance();
		buffer = a;
		avail = a.length;
		wrapping = true;
	}

	/** Creates a new input bit stream reading from a file.
	 *
	 * @param name the name of the file.
	 * @param bufSize the size in byte of the buffer; it may be 0, denoting no buffering.
	 */
	public InputBitStream( final String name, final int bufSize ) throws FileNotFoundException {
		this( new FileInputStream( name ), bufSize );
	}

	/** Creates a new input bit stream reading from a file.
	 *
	 * @param name the name of the file.
	 */
	public InputBitStream( final String name ) throws FileNotFoundException {
		this( new FileInputStream( name ), DEFAULT_BUFFER_SIZE );
	}

	/** Creates a new input bit stream reading from a file.
	 *
	 * @param file the file.
	 */
	public InputBitStream( final File file ) throws FileNotFoundException {
		this( new FileInputStream( file ), DEFAULT_BUFFER_SIZE );
	}

	/** Creates a new input bit stream reading from a file.
	 *
	 * @param file the file.
	 * @param bufSize the size in byte of the buffer; it may be 0, denoting no buffering.
	 */
	public InputBitStream( final File file, int bufSize ) throws FileNotFoundException {
		this( new FileInputStream( file ), bufSize );
	}


	/** Flushes the bit stream. All state information associated to the stream is reset. This
	 * includes bytes prefetched from the stream, bits in the bit buffer and unget'd bits.
	 *
	 * <P>This method is provided so that users of this class can easily wrap repositionable
	 * streams (for instance, file-based streams, which can be repositioned using
	 * the underlying {@link java.nio.channels.FileChannel}). It is guaranteed that after calling
	 * this method the underlying stream can be repositioned, and that the next read
	 * will draw data from the stream.
	 */

	public void flush() {
		if ( ! wrapping ) {
			position += pos;
			avail = 0;
			pos = 0;
		}
		fill = 0;
		ungetCount = 0;
	}

	/** Closes the bit stream. All resources associated to the stream are released.
	 */

	public void close() throws java.io.IOException {
		if ( is == null ) return;
		if ( is != System.in ) is.close();
		is = null;
		buffer = null;
		ungetBytes = null;
	}

	/** Sets the overflow flag.
	 *
	 * <P>If this flag is true, then after the end of the underlying {@link
	 * InputStream} we will continue returning zeroes (in particular, no
	 * exception will be thrown).
	 *
	 * @param overflow the new value of the flag.
	 * @see #pastEOF()
	 */

	public void overflow( boolean overflow ) {
		this.overflow = overflow;
	}

	/** Returns the number of bits that can be read (or skipped over) from this
	 * bit stream without blocking by the next caller of a method.
	 *
	 * @return the number of bits that can be read from this bit stream without blocking.
	 */

	public long available() throws java.io.IOException {
		return ( is.available() + ungetCount + avail ) * 8 + fill;
	}

	/** Gets the overflow flag.
	 *
	 * @return the current value of the flag.
	 * @see #overflow(boolean)
	 */
	public boolean overflow() {
		return overflow;
	}

	/** Checks whether we are past EOF.
	 *
	 * @return <code>true</code> if we are reading zeroes past the EOF.
	 * @see #overflow(boolean)
	 */
	public boolean pastEOF() {
		return pastEOF;
	}

	/** Returns the number of bits read from this bit stream.
	 *
	 * <P>Note that {@link #unget(boolean) ungetting bits} from this stream
	 * will result in the number of read bits to be <em>decreased</em>.
	 *
	 * @return the number of bits read so far.
	 */
	public long readBits() {
		return readBits;
	}

	/** Sets the number of bits read from this bit stream.
	 *
	 * <P>This method is provided so that, for instance, the 
	 * user can reset via <code>readBits(0)</code> the read-bits count
	 * after a {@link #flush()}.
	 *
	 * @param readBits the new value for the number of bits read so far.
	 */
	public void readBits( final long readBits ) {
		this.readBits = readBits;
	}

	/** Reads the next byte from the stream.
	 *
	 * <P>This method takes care of managing the buffering logic
	 * and the unget buffer transparently.
	 *
	 * <P>However, this method does <em>not</em> update {@link #readBits}.
	 * The caller should increment {@link #readBits} by 8 at each call, unless
	 * the bit are used to load {@link #current}.
	 */

	private int read() throws IOException {
		if ( ungetCount > 0 ) return ungetBytes[ --ungetCount ] & 0xFF;
		if ( pastEOF ) return 0;

		if ( noBuffer ) {
			final int t = is.read();
			if ( t == -1 ) {
				if ( overflow ) {
					pastEOF = true;
					return 0;
				}
				throw new EOFException();
			}
			else position++;
			return t;
		}

		if ( avail == 0 ) {
			avail = is.read( buffer );

			if ( avail == -1 ) {
				avail = 0;
				if ( overflow ) {
					pastEOF = true;
					return 0;
				}
				throw new EOFException();
			}
			else {
				position += pos;
				pos = 0;
			}
		}

		avail--;
		return buffer[ pos++ ] & 0xFF;
	}

	/** Refills {@link #current}. 
	 * 
	 * <P>This method must be <em>only</em> when {@link #fill} &lt; 8. It first tries to
	 * fill the lower byte in case {@link #fill} = 0. Then, it tries to refill another
	 * byte, but in this case EOF is not considered an error, and it is not thrown.
	 */

	private void refill() throws IOException {
		if ( ASSERTS ) assert fill < 8;
		
		if ( fill == 0 ) {
			current = read();
			fill = 8;
		}

		if ( avail > 0 ) {
			// If there is a byte in the buffer, we use it directly.
			current = ( current << 8 ) | read();
			fill += 8;
		}
		else {
			/* Otherwise, we are going to read a file, but we do not want to throw exceptions.
			 * This case is handled separately because try/catch blocks are time expensive. */
			try{
				current = ( current << 8 ) | read();
				fill += 8;
			}
			catch( EOFException dontCare ) {}
		}
	}

	/** Reads bits from the bit buffer, possibly refilling it.
	 *
	 * <P>This method is the basic mean for extracting bits from the underlying stream.
	 * 
	 * <P>You cannot read more than {@link #fill} bits with this method (unless {@link #fill} is 0,
	 * and <code>len</code> is nonzero, in which case the buffer will be refilled for you with 8 bits), and if you
	 * read exactly {@link #fill} bits the buffer will be empty afterwards. In particular,
	 * there will never be 8 bits in the buffer.
	 *
	 * <P>The bit buffer stores its content in the lower {@link #fill} bits. The content
	 * of the remaining bits is undefined.
	 *
	 * <P>This method updates {@link #readBits}.
	 *
	 * @param len the number of bits to read.
	 * @return the bits read (in the <strong>lower</strong> positions).
	 * @throws AssertionError if one tries to read more bits than available in the buffer and assertions are enabled.
	 */

	private int readFromCurrent( final int len ) throws IOException {

		if ( len == fill ) {
			// We just empty everything.
			readBits += len;
			fill = 0;
			return current & ( 1 << len ) - 1;
		}

		if ( fill == 0 ) {
			current = read();
			fill = 8;
		}

		if ( ASSERTS ) assert len <= fill : len + " bit(s) requested, " + fill + " available";

		readBits += len;

		return current >>> ( fill -= len ) & ( 1 << len ) - 1;
	}

	/** Aligns the stream.
	 *
	 * After a call to this function, the stream is byte aligned. Bits that have been
	 * read to align are discarded.
	 */

	public void align() throws IOException {
		if ( ( fill & 7 ) == 0 ) return;
		readBits += fill & 7;
		fill &= ~7;
	}

	/** Ungets a bit.
	 *
	 * The provided bit will be the next one read from the stream.
	 *
	 * @param bit a bit.
	 */

	public void ungetBit( boolean bit ) throws IOException {
		unget( bit ? 1 : 0 );
	}

	/** Ungets a bit.
	 *
	 * @param bit a bit.
	 * @deprecated As of <code>MG4J</code> 0.6, replaced by {@link #ungetBit(boolean)}.
	 */

	public void unget( boolean bit ) throws IOException {
		ungetBit( bit );
	}

	/** Ungets a bit.
	 * @param bit a bit.
	 * @see #unget(boolean)
	 */

	public void ungetBit( int bit ) throws IOException {
		if ( bit < 0 || bit > 1 ) throw new IllegalArgumentException( "The argument " + bit + " is not a bit." );

		if ( ASSERTS ) assert fill < 16;

		if ( bit != 0 ) current |= 1 << fill;
		else current &= ( 1 << fill ) - 1;
		fill++;
		readBits--;

		if ( fill == 16 ) {
			if ( ungetCount == UNGET_BUFFER_SIZE ) throw new IOException( "Unget buffer overflow." );
			if ( ungetBytes == null ) ungetBytes = new byte[ UNGET_BUFFER_SIZE ];
			ungetBytes[ ungetCount++ ] = (byte)current;
			fill = 8;
			current >>= 8;
		}
	}

	/** Ungets a bit.
	 * @param bit a bit.
	 * @deprecated As of <code>MG4J</code> 0.6, replaced by {@link #ungetBit(int)}.
	 */

	public void unget( int bit ) throws IOException {
		ungetBit( bit );
	}

	/** Ungets a long.
	 *
	 * The provided long will be the next one read from the stream, if the
	 * same length is used. This means that the bit of index <code>len</code>-1
	 * of <code>x</code> will be the next one to be read from the stream.
	 *
	 * @param x a long.
	 * @param len the number of (lower) bits to unget from <code>x</code>.
	 */

	public void ungetLong( long x, int len ) throws IOException {
		if ( len < 0 || len > 64 ) throw new IllegalArgumentException( "You cannot unget " + len + " bits from a long." );
		int i = len;
		while( i-- != 0 ) {
			unget( (int)( x & 1 ) );
			x >>>= 1;
		}
	}

	/** Ungets an integer.
	 *
	 * The provided integer will be the next one read from the stream, if the
	 * same length is used. This means that the bit of index <code>len</code>-1
	 * of <code>x</code> will be the next one to be read from the stream.
	 *
	 * @param x an integer.
	 * @param len the number of (lower) bits to unget from <code>x</code>.
	 */

	public void ungetInt( int x, int len ) throws IOException {
		if ( len < 0 || len > 32 ) throw new IllegalArgumentException( "You cannot unget " + len + " bits from an integer." );
		int i = len;
		while( i-- != 0 ) {
			unget( (int)( x & 1 ) );
			x >>>= 1;
		}
	}

	 
	/** Reads a sequence of bits. 
	 *
	 * Bits will be read in the natural way: the first bit is bit 7 of the
	 * first byte, the eightth bit is bit 0 of the first byte, the ninth bit is
	 * bit 7 of the second byte and so on.
	 *
	 * @param bits an array of bytes to store the result.
	 * @param len the number of bits to read.
	 */

	public void read( byte bits[], int len ) throws java.io.IOException {

		if ( fill < 8 ) refill();

		if ( len <= fill ) {
			if ( len <= 8 ) {
				bits[ 0 ] = (byte)( readFromCurrent( len ) << 8 - len );
				return;
			}
			else {
				bits[ 0 ] = (byte)( readFromCurrent( 8 ) );
				bits[ 1 ] = (byte)( readFromCurrent( len - 8 ) << 16 - len );
				return;
			}
		}
		else {
			int i, j = 0, b;

			if ( fill >= 8 ) {
				bits[ j++ ] = (byte)( readFromCurrent( 8 ) );
				len -= 8;
			} 

			final int shift = fill;

			bits[ j ] = (byte)( readFromCurrent( shift ) << 8 - shift );
			len -= shift;

			i = len >> 3;
			while( i-- != 0 ) {
				b = read();
				bits[ j ] |= ( b & 0xFF ) >>> shift;
				bits[ ++j ] = (byte)( b << 8 - shift );
			}
			readBits += len & ~7;

			len &= 7;
			if ( len != 0 ) {
				if ( len <= 8 - shift ) {
					bits[ j ] |= (byte)( readFromCurrent( len ) << 8 - shift - len );
				}
				else {
					bits[ j ] |= (byte)( readFromCurrent( 8 - shift ) );
					bits[ j + 1 ] = (byte)( readFromCurrent( len + shift - 8 ) << 16 - shift - len );
				}
			}
		}
	}



	/** Reads a bit.
	 *
	 * @return the next bit from the stream.
	 */

	public int readBit() throws java.io.IOException {
		return readFromCurrent( 1 );
	}


	/** Reads a fixed number of bits into an integer.
	 *
	 * @param len a bit length.
	 * @return an integer whose lower <code>len</code> bits are taken from the stream; the rest is zeroed.
	 */

	public int readInt( int len ) throws java.io.IOException {
		int i, x = 0;

		if ( len < 0 || len > 32 ) throw new IllegalArgumentException( "You cannot read " + len + " bits into an integer." );

		if ( len <= fill ) return readFromCurrent( len );

		len -= fill;
		x = readFromCurrent( fill );

		i = len >> 3;
		while( i-- != 0 ) x = x << 8 | read();
		readBits += len & ~7;

		len &= 7;

		return ( x << len ) | readFromCurrent( len );
	}


	/** Reads a fixed number of bits into a long.
	 *
	 * @param len a bit length.
	 * @return a long whose lower <code>len</code> bits are taken from the stream; the rest is zeroed.
	 */

	public long readLong( int len ) throws java.io.IOException {
		int i;
		long x = 0;

		if ( len < 0 || len > 64 ) throw new IllegalArgumentException( "You cannot read " + len + " bits into a long." );

		if ( len <= fill ) return readFromCurrent( len );

		len -= fill;
		x = readFromCurrent( fill );

		i = len >> 3;
		while( i-- != 0 ) x = x << 8 | read();
		readBits += len & ~7;

		len &= 7;

		return ( x << len ) | readFromCurrent( len );
	}


	/** Skips the given number of bits. 
	 *
	 * @param n the number of bits to skip.
	 * @return the actual number of skipped bits.
	 */

	public long skip( long n ) throws java.io.IOException {
		if ( n <= fill ) {
			if ( n < 0 ) throw new IllegalArgumentException( "Negative bit skip value: " + n );
			fill -= n;
			readBits += n;
			return n;
		}
		else {
			final long prevReadBits = readBits;

			n -= fill;
			readBits += fill;
			fill = 0;

			final long nb = n >> 3;

			if ( nb <= avail ) {
				// We skip bytes directly inside the buffer.
				pos += (int)nb;
				avail -= (int)nb;
				readBits += n & ~7;
			}
			else {
				// No way, we have to pass the byte skip to the underlying stream.
				n -= avail << 3;
				readBits += avail << 3;

				final long toSkip = nb - avail;
				final long skipped = is.skip( toSkip );

				position += ( avail + pos ) + skipped;
				pos = 0;
				avail = 0;

				readBits += skipped << 3;

				if ( skipped != toSkip ) return readBits - prevReadBits;
			}
			
			final int residual = (int)( n & 7 );
			if ( residual != 0 ) {
				current = read();
				fill = 8 - residual;
				readBits += residual;
			}
			return readBits - prevReadBits;
		}
	}

	/** Sets this stream bit position, if it is based on a {@link RepositionableStream} or on a {@link java.nio.channels.FileChannel}. 
	 *
	 * <P>Given an underlying stream that implements {@link
	 * RepositionableStream} or that can provide a {@link
	 * java.nio.channels.FileChannel} via the <code>getChannel()</code> method,
	 * a call to this method has the same semantics of a {@link #flush()},
	 * followed by a call to {@link
	 * java.nio.channels.FileChannel#position(long) position(position / 8)} on
	 * the byte stream, followed by a {@link #skip(long) skip(position % 8)}.
	 *
	 * @param position the new position expressed as a bit offset.
	 * @throws UnsupportedOperationException if the underlying byte stream does not implement
	 * {@link RepositionableStream} and if the channel it returns is not a {@link java.nio.channels.FileChannel}.
	 * @see FileChannel#position(long)
	 */

    public void position( final long position ) throws IOException {

		if ( position < 0 ) throw new IllegalArgumentException( "Illegal position: " + position );

		final long delta = ( position >> 3 ) - ( this.position + pos );

		if ( DEBUG ) System.err.println( this + ".position(" + position + "); curr: " + this.position + " delta: " + delta + " pos: " + pos + " avail: " + avail );

		if ( delta <= avail && delta >= - pos ) {
			// We can reposition just by moving into the buffer.
			avail -= delta;
			pos += delta;
			fill = ungetCount = 0;
			if ( DEBUG ) System.err.println( this + ": moved internally; pos: " + pos + " avail: " + avail );
		}
		else if ( rs != null ) {
			flush();
			rs.position( this.position = position >> 3 );
		}
		else if ( fileChannel != null ) {
			flush();
			fileChannel.position( this.position = position >> 3 );
		}
		else throw new UnsupportedOperationException( "position() can only be called if the underlying byte stream implements the RepositionableStream interface or if the getChannel() method of the underlying byte stream exists and returns a FileChannel" );

		final int residual = (int)( position & 7 );

		if ( DEBUG ) System.err.println( this + ": residual=" + residual );

		if ( residual != 0 ) {
			current = read();
			fill = 8 - residual;
		}
	}



	/** Tests if this stream supports the {@link #mark(int)} and {@link #reset()} methods.
	 *
	 * <P>This method will just delegate the test to the underlying {@link InputStream}.
	 * @return whether this stream supports {@link #mark(int)}/{@link #reset()}.
	 */

	public boolean markSupported() {
		return is.markSupported();
	}
	 
	/** Marks the current position in this input stream. A subsequent call to
	 * the {@link #reset()} method repositions this stream at the last marked position so
	 * that subsequent reads re-read the same bits.
	 *
	 * <P>This method will just delegate the mark to the underlying {@link InputStream}.
	 * Moreover, it will throw an exception if you try to mark outsite byte boundaries.
	 *
	 * @param readLimit the maximum limit of bytes that can be read before the mark position becomes invalid.
	 * @throws IOException if you try to mark outside byte boundaries.
	 */

	public void mark( int readLimit ) throws java.io.IOException {
		if ( fill != 0 ) throw new IOException( "You cannot mark a bit stream outside of byte boundaries." );
		is.mark( readLimit );
	}

	/** Repositions this bit stream to the position at the time the {@link #mark(int)} method was last called.
	 *
	 * <P>This method will just {@link #flush() flush the stream} and delegate
	 * the reset to the underlying {@link InputStream}.
	 */

	public void reset() throws java.io.IOException {
		flush();
		is.reset();
	}

	/** Reads a natural number in unary coding.
	 *
	 * Note that by unary coding we mean that 1 encodes 0, 01 encodes 1 and so on.
	 *
	 * @return the next unary-encoded natural number.
	 */

	public int readUnary() throws java.io.IOException {

		int x;

		if ( fill < 8 ) refill();
		
		final int currentLeftAligned = current << ( 16 - fill ) & 0xFFFF;
		if ( currentLeftAligned != 0 ) {
			if ( ( currentLeftAligned & 0xFF00 ) != 0 ) x = 7 - Fast.BYTEMSB[ currentLeftAligned >>> 8 ];
			else x = 15 - Fast.BYTEMSB[ currentLeftAligned & 0xFF ];
			readBits += x + 1;
			fill -= x + 1;
			return x;
		}

		x = fill;
		while( ( current = read() ) == 0 ) x += 8;
		x += 7 - ( fill = Fast.BYTEMSB[ current ] );
		readBits += x + 1;
		return x;
	}

	/** Reads a long natural number in unary coding.
	 *
	 * Note that by unary coding we mean that 1 encodes 0, 01 encodes 1 and so on.
	 *
	 * @return the next unary-encoded long natural number.
	 */

	public long readLongUnary() throws java.io.IOException {

		if ( ( current & ( 1 << fill ) - 1 ) != 0 ) return readUnary();

		long x = fill;
		while( ( current = read() ) == 0 ) x += 8;
		x += 7 - ( fill = Fast.BYTEMSB[ current ] );
		readBits += x + 1;
		return x;
	}

	/** Reads a natural number in &gamma; coding.
	 *
	 * @return the next &gamma;-encoded natural number.
	 */

	public int readGamma() throws java.io.IOException {
		if ( fill < 8 ) refill();
		if ( fill >= 8 ) {
			final int preComp = GAMMA[ current >> ( fill == 16 ? 8 : ( fill & 7 ) ) & 0xFF ];
			if ( preComp != 0 ) {
				readBits += preComp >> 8;
				fill -= preComp >> 8;
				return preComp & 0xFF;
			}
		}

		final int msb = readUnary();
		return ( ( 1 << msb ) | readInt( msb ) ) - 1;
	}

	/** Reads a long natural number in &gamma; coding.
	 *
	 * @return the next &gamma;-encoded long natural number.
	 */

	public long readLongGamma() throws java.io.IOException {
		if ( fill < 8 ) refill();
		if ( fill >= 8 ) {
			final int preComp = GAMMA[ current >> ( fill == 16 ? 8 : ( fill & 7 ) ) & 0xFF ];
			if ( preComp != 0 ) {
				readBits += preComp >> 8;
				fill -= preComp >> 8;
				return preComp & 0xFF;
			}
		}

		final int msb = readUnary();
		return ( ( 1L << msb ) | readLong( msb ) ) - 1;
	}

	/** Reads a natural number in &delta; coding.
	 *
	 * @return the next &delta;-encoded natural number.
	 */

	public int readDelta() throws java.io.IOException {
		if ( fill < 8 ) refill();
		if ( fill >= 8 ) {
			final int preComp = DELTA[ current >> ( fill == 16 ? 8 : ( fill & 7 ) ) & 0xFF ];
			if ( preComp != 0 ) {
				readBits += preComp >> 8;
				fill -= preComp >> 8;
				return preComp & 0xFF;
			}
		}

		final int msb = readGamma();
		return ( ( 1 << msb ) | readInt( msb ) ) - 1;
	}


	/** Reads a long natural number in &delta; coding.
	 *
	 * @return the next &delta;-encoded long natural number.
	 */

	public long readLongDelta() throws java.io.IOException {
		if ( fill < 8 ) refill();
		if ( fill >= 8 ) {
			final int preComp = DELTA[ current >> ( fill == 16 ? 8 : ( fill & 7 ) ) & 0xFF ];
			if ( preComp != 0 ) {
				readBits += preComp >> 8;
				fill -= preComp >> 8;
				return preComp & 0xFF;
			}
		}

		final int msb = readGamma();
		return ( ( 1L << msb ) | readLong( msb ) ) - 1;
	}


	/** Reads a natural number in a limited range using a minimal binary coding.
	 *
	 * @param b a strict upper bound.
	 * @return the next minimally binary encoded natural number.
	 * @throws IllegalArgumentException if you try to read a negative number or use a nonpositive base.
	 */

	public int readMinimalBinary( int b ) throws java.io.IOException {
		return readMinimalBinary( b, Fast.mostSignificantBit( b ) );
	}

	/** Reads a natural number in a limited range using a minimal binary coding.
	 *
	 * This method is faster than {@link #readMinimalBinary(int)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * @param b a strict upper bound.
	 * @param log2b the floor of the base-2 logarithm of the bound.
	 * @return the next minimally binary encoded natural number.
	 * @throws IllegalArgumentException if you try to read a negative number or use a nonpositive base.
	 */

	public int readMinimalBinary( int b, int log2b ) throws java.io.IOException {
		if ( b < 1 ) throw new IllegalArgumentException( "The bound " + b + " is not positive" );

		final int m = ( 1 << log2b + 1 ) - b; 
		final int x = readInt( log2b );

		if ( x < m ) return x;
		else return ( ( x << 1 ) + readBit() - m );
	}

	/** Reads a long natural number in a limited range using a minimal binary coding.
	 *
	 * @param b a strict upper bound.
	 * @return the next minimally binary encoded long natural number.
	 * @throws IllegalArgumentException if you try to read a negative number or use a nonpositive base.
	 */

	public long readLongMinimalBinary( long b ) throws java.io.IOException {
		return readLongMinimalBinary( b, Fast.mostSignificantBit( b ) );
	}


	/** Reads a long natural number in a limited range using a minimal binary coding.
	 *
	 * This method is faster than {@link #readLongMinimalBinary(long)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * @param b a strict upper bound.
	 * @param log2b the floor of the base-2 logarithm of the bound.
	 * @return the next minimally binary encoded long natural number.
	 * @throws IllegalArgumentException if you try to read a negative number or use a nonpositive base.
	 */

	public long readLongMinimalBinary( long b, int log2b ) throws java.io.IOException {
		if ( b < 1 ) throw new IllegalArgumentException( "The bound " + b + " is not positive" );

		final long m = ( 1L << log2b + 1 ) - b; 
		final long x = readLong( log2b );

		if ( x < m ) return x;
		else return ( ( x << 1 ) + readBit() - m );
	}

	/** Reads a natural number in Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * nothing will be read, and 0 will be returned. 
	 *
	 * @param b the modulus for the coding.
	 * @return the next Golomb-encoded natural number.
	 * @throws IllegalArgumentException if you use a nonpositive modulus.
	 */

	public int readGolomb( int b ) throws java.io.IOException {
		return readGolomb( b, Fast.mostSignificantBit( b ) );
	}

	/** Reads a natural number in Golomb coding.
	 *
	 * This method is faster than {@link #readGolomb(int)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * nothing will be read, and 0 will be returned. 
	 *
	 * @param b the modulus for the coding.
	 * @param log2b the floor of the base-2 logarithm of the coding modulus.
	 * @return the next Golomb-encoded natural number.
	 * @throws IllegalArgumentException if you use a nonpositive modulus.
	 */

	public int readGolomb( int b, int log2b ) throws java.io.IOException {
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) return 0;

		final int q = readUnary() * b;
		return q + readMinimalBinary( b, log2b );
	}



	/** Reads a long natural number in Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * nothing will be read, and 0 will be returned. 
	 *
	 * @param b the modulus for the coding.
	 * @return the next Golomb-encoded long natural number.
	 * @throws IllegalArgumentException if you use a nonpositive modulus.
	 */

	public long readLongGolomb( long b ) throws java.io.IOException {
		return readLongGolomb( b, Fast.mostSignificantBit( b ) );
	}

	/** Reads a long natural number in Golomb coding.
	 *
	 * This method is faster than {@link #readLongGolomb(long)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * nothing will be read, and 0 will be returned. 
	 *
	 * @param b the modulus for the coding.
	 * @param log2b the floor of the base-2 logarithm of the coding modulus.
	 * @return the next Golomb-encoded long natural number.
	 * @throws IllegalArgumentException if you use a nonpositive modulus.
	 */

	public long readLongGolomb( long b, int log2b ) throws java.io.IOException {
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) return 0;

		final long q = readUnary() * b;
		return q + readLongMinimalBinary( b, log2b );
	}


	/** Reads a natural number in skewed Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * nothing will be read, and 0 will be returned. 
	 *
	 * @param b the modulus for the coding.
	 * @return the next skewed Golomb-encoded natural number.
	 * @throws IllegalArgumentException if you use a negative modulus.
	 */

	public int readSkewedGolomb( int b ) throws java.io.IOException {
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) return 0;

		final int M = ( ( 1 << readUnary() + 1 ) - 1 ) * b;
		final int m = ( M / ( 2 * b ) ) * b;
		return m + readMinimalBinary( M - m );
	}


	/** Reads a long natural number in skewed Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * nothing will be read, and 0 will be returned. 
	 *
	 * @param b the modulus for the coding.
	 * @return the next skewed Golomb-encoded long natural number.
	 * @throws IllegalArgumentException if you use a negative modulus.
	 */

	public long readLongSkewedGolomb( long b ) throws java.io.IOException {
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) return 0;

		final long M = ( ( 1 << readUnary() + 1 ) - 1 ) * b;
		final long m = ( M / ( 2 * b ) ) * b;
		return m + readLongMinimalBinary( M - m );
	}


	/** Reads a natural number in &zeta; coding.
	 *
	 * @param k the shrinking factor.
	 * @return the next &zeta;-encoded natural number.
	 * @throws IllegalArgumentException if you use a nonpositive shrinking factor.
	 */

	public int readZeta( int k ) throws java.io.IOException {
		if ( k < 1 ) throw new IllegalArgumentException( "The shrinking factor " + k + " is not positive" );

		if ( k == 3 ) {
			if ( fill < 8 ) refill();
			if ( fill >= 8 ) {
				final int preComp = ZETA_3[ current >> ( fill == 16 ? 8 : ( fill & 7 ) ) & 0xFF ];
				if ( preComp != 0 ) {
					readBits += preComp >> 8;
					fill -= preComp >> 8;
					return preComp & 0xFF;
				}
			}
		}

		final int h = readUnary();
		final int left = 1 << h * k;
		final int m = readInt( h * k + k - 1 );
		if ( m < left ) return m + left - 1;
		return ( m << 1 ) + readBit() - 1;
	}

	/** Reads a long natural number in &zeta; coding.
	 *
	 * @param k the shrinking factor.
	 * @return the next &zeta;-encoded long natural number.
	 * @throws IllegalArgumentException if you use a nonpositive shrinking factor.
	 */

	public long readLongZeta( int k ) throws java.io.IOException {
		if ( k < 1 ) throw new IllegalArgumentException( "The shrinking factor " + k + " is not positive" );

		if ( k == 3 ) {
			if ( fill < 8 ) refill();
			if ( fill >= 8  ) {
				final int preComp = ZETA_3[ current >> ( fill == 16 ? 8 : ( fill & 7 ) ) & 0xFF ];
				if ( preComp != 0 ) {
					readBits += preComp >> 8;
					fill -= preComp >> 8;
					return preComp & 0xFF;
				}
			}
		}

		final int h = readUnary();
		final long left = 1 << h * k;
		final long m = readLong( h * k + k - 1 );
		if ( m < left ) return m + left - 1;
		return ( m << 1 ) + readBit() - 1;
	}


	/** Reads a natural number in variable-length nibble coding.
	 *
	 * @return the next variable-length nibble-encoded natural number.
	 */

	public int readNibble() throws java.io.IOException {
		int b;
		int x = 0;

		do {
			x <<= 3;
			b = readBit();
			x |= readInt( 3 );
		} while( b == 0 );
		
		return x;
	}
	
	/** Reads a long natural number in variable-length nibble coding.
	 *
	 * @return the next variable-length nibble-encoded long natural number.
	 */

	public long readLongNibble() throws java.io.IOException {
		int b;
		long x = 0;

		do {
			x <<= 3;
			b = readBit();
			x |= readInt( 3 );
		} while( b == 0 );
		
		return x;
	}
	
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
