package it.unimi.dsi.mg4j.io;

import it.unimi.dsi.mg4j.util.Fast;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.channels.FileChannel;


/** Bit-level output stream.
 *
 * <blockquote><strong>IMPORTANT</strong>: In MG4J 0.6, this class has been completely
 * rewritten. It is much faster, but there are also incompabilities. Mainly,
 * unary representations are one-terminated sequences of zeroes.</blockquote>
 *
 * <P>This class wraps any {@link OutputStream} so that you can treat it as
 * <em>bit</em> stream. Constructors and methods closely resemble those of
 * {@link OutputStream}. Data can be added to such a stream in several ways:
 * writing an integer or long in fixed-width, unary, &gamma;, &delta;, &zeta; and Golomb
 * coding, or providing a vector of bytes.
 * 
 * <P>This class can also {@linkplain #OutputBitStream(byte[]) wrap a byte
 * array}; this is much more lightweight than wrapping a {@link
 * FastByteArrayOutputStream} wrapping the array, but overflowing the array
 * will cause an {@link java.io.IOException}.
 * 
 * <P>Note that when writing using a vector of bytes bits are written in the natural
 * way: the first bit is bit 7 of the first byte, the eightth bit is bit 0 of
 * the first byte, the ninth bit is bit 7 of the second byte and so on. When
 * writing integers using some coding, instead, the <em>lower</em> bits are considered
 * for coding (in the fixed-width case, the given number of bits, otherwise
 * the lower bits starting from the most significant one).
 *
 * <h3>The bit stream format</h3>
 *
 * <P>The bit streams written by this class are <em>big endian</em>. That is, 
 * the first bit of the stream is bit 7 of the first byte, the eightth bit
 * is bit 0 of the first byte, the ninth bit is bit 7 of the second byte and so on.
 *
 * <P>Blocks of bits (such as coded integers) are written <em>starting from the
 * most significant bit</em>. In other words, if you take the first bytes of a stream
 * and print them in binary you will see exactly the sequence of bits you have
 * written. In particular, if you write 32-bit integers you will get a stream
 * which is identical to the one produced by a {@link java.io.DataOutput}.
 *
 * <P>Additional features:
 *
 * <ul>
 *
 * <LI>This class provides an internal buffer. By setting a buffer of
 * length 0 at creation time, you can actually bypass the buffering system:
 * Note, however, that several classes providing buffering have synchronised
 * methods, so using a wrapper instead of the internal buffer is likely to lead
 * to a performance drop.
 *
 * <LI>To work around the schizophrenic relationship between streams and random
 * access files in {@link java.io}, this class provides a {@link #flush()}
 * method that byte-aligns the streams, flushes to the underlying byte stream
 * all data and resets the internal state. At this point, you can safely reposition
 * the underlying stream and write again afterwards. For instance, this is safe
 * and will perform as expected:
 * <PRE>
 * FileOutputStream fos = new FileOutputStream( ... );
 * OutputBitStream obs = new OutputBitStream( fos );
 * ... write operations on obs ...
 * obs.flush();
 * fos.getChannel().position( ... );
 * ... other write operations on obs ...
 * </PRE>
 *
 * <P>As a commodity, an instance of this class will try to cast the underlying
 * byte stream to a {@link RepositionableStream} and to fetch by reflection the
 * {@link java.nio.channels.FileChannel} underlying the given output stream, in
 * this order.  If either reference can be successfully fetched, you can use
 * directly the {@link #position(long) position()} method with argument
 * <code>pos</code> with the same semantics of a {@link #flush()}, followed by
 * a call to <code>position(pos / 8)</code> (where the latter method belongs
 * either to the underlying stream or to its underlying file channel).  The
 * specified position must be byte aligned, as there is no clean way of reading
 * a fraction of a byte with the current APIs.
 *
 * </ul>
 *
 * <P><STRONG>This class is not synchronised</STRONG>. If multiple threads
 * access an instance of this class concurrently, they must be synchronised externally.
 *
 * @author Sebastiano Vigna
 * @since 0.1
 */

public class OutputBitStream {
	 
	/** The default size of the byte buffer in bytes (16Ki). */
	public final static int DEFAULT_BUFFER_SIZE = 16 * 1024;
	/** Are we debugging? */
	private final static boolean DEBUG = false;
	/** The underlying {@link OutputStream}. */
	private OutputStream os;
	/** The number of bits written to this bit stream. */
	private long writtenBits;
	/** Current bit buffer. */
	private int current;
	/** The stream buffer. */
	private byte buffer[];
	/** Current number of free bits in the bit buffer (the bits in the buffer are stored high). */
	private int free;
	/** Current position in the byte buffer. */
	private int pos;
	/** Current number of bytes available in the byte buffer. */
	private int avail;
	/** Size of the small buffer for temporary usage. */
	final static int TEMP_BUFFER_SIZE = 128;
	/** Small buffer for temporary usage. */
	private byte tempBuffer[] = new byte[ TEMP_BUFFER_SIZE ];
	/** The cached file channel underlying {@link #os}. */
	private FileChannel fileChannel;
	/** {@link #is} cast to a positionable stream. */
	private RepositionableStream rs;
	/** True if we are wrapping an array. */
	private boolean wrapping;


	/** This (non-public) constructor exists just to provide fake initialisation for classes such as {@link DebugOutputBitStream}.
	 */
	protected OutputBitStream() {}


	/** Creates a new output bit stream wrapping a given output stream using a buffer of size {@link #DEFAULT_BUFFER_SIZE}.
	 *
	 * @param os the output stream to wrap.
	 */
	public OutputBitStream( OutputStream os ) {
		this( os, DEFAULT_BUFFER_SIZE );
	}



	/** Creates a new output bit stream wrapping a given output stream with a specified buffer size.
	 *
	 * @param os the output stream to wrap.
	 * @param bufSize the size in byte of the buffer; it may be 0, denoting no buffering.
	 */
	public OutputBitStream( OutputStream os, int bufSize ) {
		this.os = os;
		if ( bufSize != 0 ) {
			this.buffer = new byte[ bufSize ];
			avail = bufSize;
		}
		free = 8;

		if ( os instanceof RepositionableStream ) rs = (RepositionableStream)os;
		
		if ( rs == null ) {
			try {
				fileChannel = (FileChannel)( os.getClass().getMethod( "getChannel", new Class[] {} ) ).invoke( os, new Object[] {} );
			}
			catch( IllegalAccessException e ) {}
			catch( IllegalArgumentException e ) {}
			catch( NoSuchMethodException e ) {}
			catch( java.lang.reflect.InvocationTargetException e ) {}
			catch( ClassCastException e ) {}
		}
	}


	/** Creates a new output bit stream wrapping a given byte array.
	 *
	 * @param a the byte array to wrap.
	 */
	public OutputBitStream( byte a[] ) {
		free = 8;
		buffer = a;
		avail = a.length;
		wrapping = true;
	}

	/** Creates a new output bit stream writing to file.
	 *
	 * @param name the name of the file.
	 * @param bufSize the size in byte of the buffer; it may be 0, denoting no buffering.
	 */
	public OutputBitStream( final String name, final int bufSize ) throws FileNotFoundException {
		this( new FileOutputStream( name ), bufSize );
	}

	/** Creates a new output bit stream writing to a file.
	 *
	 * @param name the name of the file.
	 */
	public OutputBitStream( final String name ) throws FileNotFoundException {
		this( new FileOutputStream( name ), DEFAULT_BUFFER_SIZE );
	}


	/** Creates a new output bit stream writing to file.
	 *
	 * @param file the file.
	 * @param bufSize the size in byte of the buffer; it may be 0, denoting no buffering.
	 */
	public OutputBitStream( final File file, final int bufSize ) throws FileNotFoundException {
		this( new FileOutputStream( file ), bufSize );
	}

	/** Creates a new output bit stream writing to a file.
	 *
	 * @param file the file.
	 */
	public OutputBitStream( final File file ) throws FileNotFoundException {
		this( new FileOutputStream( file ), DEFAULT_BUFFER_SIZE );
	}


	/** Flushes the bit stream.
	 *
	 * <P>This method will align the stream, write the bit buffer, empty the
	 * byte buffer and delegate to the {@link OutputStream#flush()} method of
	 * the underlying output stream.
	 *
	 * <P>This method is provided so that users of this class can easily wrap
	 * repositionable streams (for instance, file-based streams, which can be
	 * repositioned using the underlying {@link
	 * java.nio.channels.FileChannel}). <P> It is guaranteed that after calling
	 * this method the underlying stream can be repositioned, and that the next
	 * write to the underlying output stream will start with the content of the
	 * first write method called afterwards.
	 */

	public void flush() throws java.io.IOException {
		align();
		if ( os != null ) {
			if ( buffer != null ) {
				os.write( buffer, 0, pos );
				pos = 0;
				avail = buffer.length;
			}
			os.flush();
		}
	}


	/** Closes the bit stream. All resources associated to the stream are released.
	 */

	public void close() throws java.io.IOException {
		if ( os == null ) return;
		flush();
		if ( os != System.out && os != System.err ) os.close();
		os = null;
		buffer = null;
		tempBuffer = null;
	}

	/** Returns the number of bits written to this bit stream.
	 *
	 * @return the number of bits written so far.
	 */
	public long writtenBits() {
		return writtenBits;
	}

	/** Sets the number of bits written to this bit stream.
	 *
	 * <P>This method is provided so that, for instance, the 
	 * user can reset via <code>writtenBits(0)</code> the written-bits count
	 * after a {@link #flush()}.
	 *
	 * @return writtenBits the new value for the number of bits written so far.
	 */
	public void writtenBits( long writtenBits ) {
		this.writtenBits = writtenBits;
	}

	/** Writes a byte to the stream.
	 *
	 * <P>This method takes care of managing the buffering logic transparently.
	 *
	 * <P>However, this method does <em>not</em> update {@link #writtenBits}.
	 * The caller should increment {@link #writtenBits} by 8 at each call.
	 */

	private void write( int b ) throws IOException {
		if ( avail-- == 0 ) {
			if ( os == null ) {
				avail = 0;
				throw new IOException( "Array full" );
			}

			if ( buffer == null ) {
				os.write( b );
				avail = 0;
				return;
			}
			os.write( buffer );
			avail = buffer.length - 1;
			pos = 0;
		}

		buffer[ pos++ ] = (byte)b;
	}


	/** Writes bits in the bit buffer, possibly flushing it.
	 *
	 * You cannot write more than {@link #free} bits with this method. However,
	 * after having written {@link #free} bits the bit buffer will be empty. In
	 * particular, there should never be 0 free bits in the buffer.
	 *
	 * @param b the bits to write in the <strong>lower</strong> positions; the remaining positions must be zero.
	 * @param len the number of bits to write (0 is safe and causes no action).
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if one tries to write more bits than available in the buffer and debug is enabled.
	 */

	private int writeInCurrent( final int b, final int len ) throws java.io.IOException {

		//System.err.println("Writing " + len + " bits out of " + Fast.binary( b ) );

		if ( DEBUG ) if ( len > free ) throw new IllegalArgumentException( "" + len + " bit(s) to write, " + free + " available." );

		current |= ( b & ( ( 1 << len ) - 1 ) ) << ( free -= len );
		if ( free == 0 ) {
			write( current );
			free = 8;
			current = 0;
		}

		writtenBits += len;
		return len;
	}



	/** Aligns the stream.
	 *
	 * After a call to this function, the stream is byte aligned. Zeroes
	 * are used to pad it if necessary.
	 *
	 * @return the number of padding bits.
	 */

	public int align() throws IOException {
		if ( free != 8 ) return writeInCurrent( 0, free );
		else return 0;
	}

	/** Sets this stream bit position, if it is based  on a {@link RepositionableStream} or on a {@link java.nio.channels.FileChannel}. 
	 *
	 * <P>Given an underlying stream that implements {@link
	 * RepositionableStream} or that can provide a {@link
	 * java.nio.channels.FileChannel} via the <code>getChannel()</code> method,
	 * a call to this method has the same semantics of a {@link #flush()},
	 * followed by a call to {@link
	 * java.nio.channels.FileChannel#position(long) position(position / 8)} on
	 * the byte stream. Currently there is no clean, working way of supporting
	 * out-of-byte-boundary positioning.
	 *
	 * @param position the new position expressed as a bit offset; it must be byte-aligned.
	 * @throws IllegalArgumentException when trying to position outside of byte boundaries.
	 * @throws UnsupportedOperationException if the underlying byte stream does not implement
	 * {@link RepositionableStream} and if the channel it returns is not a {@link java.nio.channels.FileChannel}.
	 * @see FileChannel#position(long)
	 */

    public void position( final long position ) throws IOException {

		if ( position < 0 ) throw new IllegalArgumentException( "Illegal position: " + position );
		if ( ( position & 7 ) != 0 ) throw new IllegalArgumentException( "Not a byte-aligned position: " + position );

		if ( wrapping ) {
			if ( position > buffer.length ) throw new IllegalArgumentException( "Illegal position: " + position );
			flush();
			free = 8;
			pos = (int)position;
			avail = buffer.length - pos;
		}
		else if ( rs != null ) {
			flush();
			rs.position( position >> 3 );
		}
		else if ( fileChannel != null ) {
			flush();
			fileChannel.position( position >> 3 );
		}
		else throw new UnsupportedOperationException( "position() can only be called if the underlying byte stream implements the RepositionableStream interface or if the getChannel() method of the underlying byte stream exists and returns a FileChannel" );	
	}

	/** Writes a sequence of bits. 
	 *
	 * Bits will be written in the natural way: the first bit is bit 7 of the
	 * first byte, the eightth bit is bit 0 of the first byte, the ninth bit is
	 * bit 7 of the second byte and so on.
	 *
	 * @param bits a vector containing the bits to be written.
	 * @param len a bit length.
	 * @return the number of bits written (<code>len</code>).
	 */


	public int write( byte bits[], int len ) throws java.io.IOException {
		//System.err.println( " Going to write " + len + " bit(s) from " + binary(bits[0]) + " " +binary(bits[1]) + " " +binary(bits[2])+ " " +binary(bits[3]));

		if ( len == 0 ) return 0;
		if ( len <= free ) {
			return writeInCurrent( bits[ 0 ] >>> 8 - len, len );
		}
		else {
			final int shift = free;
			int i, j;

			writeInCurrent( bits[ 0 ] >>> 8 - shift, shift );

			//System.err.println("Writing (" + ( shift ) + ") " + Fast.binary( bits[ 0 ] >>> 8 - shift ) );

			len -= shift;

			j = 0;
			i = len >> 3;
			while( i-- != 0 ) {
				//System.err.println("Writing (8) " + Fast.binary( bits[ j ] << shift | ( bits[ j + 1 ] & 0xFF ) >>> 8 - shift ) );
				write( bits[ j ] << shift | ( bits[ j + 1 ] & 0xFF ) >>> 8 - shift );
				writtenBits += 8;
				j++;
			}

			final int queue = len & 7;
			if ( queue != 0 ) {
				if ( queue <= 8 - shift ) {
					writeInCurrent( bits[ j ] >>> 8 - shift - queue, queue );
					//System.err.println("Writing (" + ( queue ) + ") " + Fast.binary( bits[ j ] >>> 8 - shift - queue ) );
				}
				else {
					writeInCurrent( bits[ j ], 8 - shift );
					//System.err.println("Writing (" + ( 8 - shift ) + ") " + Fast.binary( bits[ j ] ) );
					writeInCurrent( bits[ j + 1 ] >>> 16 - queue - shift, queue + shift - 8 );
					//System.err.println("Writing (" + ( queue - 8 + shift ) + ") " + Fast.binary( bits[ j + 1 ] >>> 16 - queue - shift ) );
				}
			}

			return len + shift;
		}

	}

	/** Writes a bit.
	 *
	 * @param bit a bit.
	 * @return the number of bits written.
	 */

	public int writeBit( boolean bit ) throws java.io.IOException {
		return writeInCurrent( bit ? 1 : 0, 1 ); 
	}

	/** Writes a bit.
	 *
	 * @param bit a bit.
	 * @return the number of bits written.
	 */

	public int writeBit( int  bit ) throws java.io.IOException {
		if ( bit < 0 || bit > 1 ) throw new IllegalArgumentException( "The argument " + bit + " is not a bit." );
		return writeInCurrent( bit, 1 ); 
	}

	/** Writes a fixed number of bits from an integer.
	 *
	 * @param x an integer.
	 * @param len a bit length; this many lower bits of the first argument will be written
	 * (the most significant bit first).
	 * @return the number of bits written (<code>len</code>).
	 */

	public int writeInt( int x, int len ) throws java.io.IOException {

		if ( len < 0 || len > 32 ) throw new IllegalArgumentException( "You cannot write " + len + " bits to an integer." );

		if ( len <= free ) return writeInCurrent( x, len );

		final int queue = ( len - free ) & 7, blocks = ( len - free ) >> 3;
		int i = blocks, j = 0;

		if ( queue != 0 ) {
			tempBuffer[ blocks ] = (byte)x;
			x >>= queue;
		}

		while( i-- != 0 ) {
			tempBuffer[ i ] = (byte)x;
			x >>>= 8;
		}

		writeInCurrent( x, free );
		i = blocks;
		while( i-- != 0 ) write( tempBuffer[ j++ ] );
		writtenBits += blocks << 3;

		if ( queue != 0 ) writeInCurrent( tempBuffer[ blocks ], queue );
		return len;
	}

	/** Writes a fixed number of bits from a long.
	 *
	 * @param x a long.
	 * @param len a bit length; this many lower bits of the first argument will be written
	 * (the most significant bit first).
	 * @return the number of bits written (<code>len</code>).
	 */

	public int writeLong( long x, int len ) throws java.io.IOException {

		if ( len < 0 || len > 64 ) throw new IllegalArgumentException( "You cannot write " + len + " bits to a long." );

		if ( len <= free ) return writeInCurrent( (int)x, len );

		final int queue = ( len - free ) & 7, blocks = ( len - free ) >> 3;
		int i = blocks, j = 0;

		if ( queue != 0 ) {
			tempBuffer[ blocks ] = (byte)x;
			x >>= queue;
		}

		while( i-- != 0 ) {
			tempBuffer[ i ] = (byte)x;
			x >>>= 8;
		}

		writeInCurrent( (int)x, free );
		i = blocks;
		while( i-- != 0 ) write( tempBuffer[ j++ ] );
		writtenBits += blocks << 3;

		if ( queue != 0 ) writeInCurrent( tempBuffer[ blocks ], queue );
		return len;
	}

	/** Writes a natural number in unary coding.
	 *
	 * Note that by unary coding we mean that 1 encodes 0, 01 encodes 1 and so on.
	 *
	 * @param x a natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number.
	 */

	public int writeUnary( int x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		if ( x < free ) return writeInCurrent( 1, x + 1 );

		final int shift = free;
		x -= shift;

		writtenBits += shift;
		write( current );
		free = 8;
		current = 0;

		int i = x >> 3;

		writtenBits += ( x & 0x7FFFFFF8 );

		while( i-- != 0 ) write( 0 );

		writeInCurrent( 1, ( x & 7 ) + 1 );

		return x + shift + 1;
	}

	/** Writes a long natural number in unary coding.
	 *
	 * @see #writeUnary(int)
	 *
	 * @param x a long natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you write to write a negative number.
	 */

	public long writeLongUnary( long x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		if ( x < free ) return writeInCurrent( 1, (int)x + 1 );

		final int shift = free;
		x -= shift;

		writtenBits += shift;
		write( current );
		free = 8;
		current = 0;

		long i = x >> 3;

		writtenBits += ( x & 0x7FFFFFFFFFFFFFF8L );

		while( i-- != 0 ) write( 0 );

		writeInCurrent( 1, (int)( x & 7 ) + 1 );

		return x + shift + 1;
	}

	/** Writes a natural number in &gamma; coding.
	 *
	 * The &gamma; coding of a positive number of <var>k</var> bits is
	 * obtained writing <var>k</var>-1 in unary, followed by the lower
	 * <var>k</var>-1 bits of the number. The coding of a natural number is
	 * obtained by adding one and coding.
	 *
	 * @param x a natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you write to write a negative number.
	 */

	public int writeGamma( int x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		final int msb = Fast.mostSignificantBit( ++x );
		final int l = writeUnary( msb );
		return l + ( msb != 0 ? writeInt( x, msb ) : 0 );
	}

	/** Writes a long natural number in &gamma; coding.
	 *
	 * @see #writeGamma(int)
	 *
	 * @param x a long natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you write to write a negative number.
	 */

	public int writeLongGamma( long x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		final int msb = Fast.mostSignificantBit( ++x );
		final int l = writeUnary( msb );
		return l + ( msb != 0 ? writeLong( x, msb ) : 0 );
	}

	/** Writes a natural number in &delta; coding.
	 *
	 * The &delta; coding of a positive number of <var>k</var> bits is
	 * obtained writing <var>k</var>-1 in &gamma; coding, followed by the
	 * lower <var>k</var>-1 bits of the number. The coding of a natural
	 * number is obtained by adding one and coding.
	 *
	 * @param x a natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you write to write a negative number.
	 */

	public int writeDelta( int x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		final int msb = Fast.mostSignificantBit( ++x );
		final int l = writeGamma( msb );
		return l + ( msb != 0 ? writeInt( x, msb ) : 0 );
	}

	/** Writes a long natural number in &delta; coding.
	 *
	 * @see #writeDelta(int)
	 *
	 * @param x a long natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you write to write a negative number.
	 * @deprecated As of <code>MG4J</code> 0.2, replaced by {@link #writeLongDelta(long)}.
	 */

	public int writeDelta( long x ) throws java.io.IOException {
		return writeLongDelta( x );
	}

	/** Writes a long natural number in &delta; coding.
	 *
	 * @see #writeDelta(int)
	 *
	 * @param x a long natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you write to write a negative number.
	 */

	public int writeLongDelta( long x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		final int msb = Fast.mostSignificantBit( ++x );
		final int l = writeGamma( msb );
		return l + ( msb != 0 ? writeLong( x, msb ) : 0 );
	}

	/** Writes a natural number in a limited range using a minimal binary coding.
	 *
	 * @param x a natural number.
	 * @param b a strict upper bound for <code>x</code>.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a nonpositive base.
	 */

	public int writeMinimalBinary( int x, int b ) throws java.io.IOException {
		if ( b < 1 ) throw new IllegalArgumentException( "The bound " + b + " is not positive" );

		return writeMinimalBinary( x, b, Fast.mostSignificantBit( b ) );
	}

	/** Writes a natural number in a limited range using a minimal binary coding.
	 *
	 * This method is faster than {@link #writeMinimalBinary(int,int)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * @param x a natural number.
	 * @param b a strict upper bound for <code>x</code>.
	 * @param log2b the floor of the base-2 logarithm of the bound.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a nonpositive base.
	 */

	public int writeMinimalBinary( int x, int b, int log2b ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( b < 1 ) throw new IllegalArgumentException( "The bound " + b + " is not positive" );
		if ( x >= b ) throw new IllegalArgumentException( "The argument " + x + " exceeds the bound " + b );

		// Numbers smaller than m are encoded in log2b bits.
		final int m = ( 1 << log2b + 1 ) - b; 

		if ( x < m ) return writeInt( x, log2b );
		else return writeInt( m + x, log2b + 1 );
	}


	/** Writes a long natural number in a limited range using a minimal binary coding.
	 *
	 * @param x a natural number.
	 * @param b a strict upper bound for <code>x</code>.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a nonpositive base.
	 */

	public int writeLongMinimalBinary( long x, long b ) throws java.io.IOException {
		if ( b < 1 ) throw new IllegalArgumentException( "The bound " + b + " is not positive" );

		return writeLongMinimalBinary( x, b, Fast.mostSignificantBit( b ) );
	}


	/** Writes a long natural number in a limited range using a minimal binary coding.
	 *
	 * This method is faster than {@link #writeLongMinimalBinary(long,long)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * @param x a long natural number.
	 * @param b a strict upper bound for <code>x</code>.
	 * @param log2b the floor of the base-2 logarithm of the bound.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a nonpositive base.
	 */

	public int writeLongMinimalBinary( long x, long b, int log2b ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( b < 1 ) throw new IllegalArgumentException( "The bound " + b + " is not positive" );
		if ( x >= b ) throw new IllegalArgumentException( "The argument " + x + " exceeds the bound " + b );

		// Numbers smaller than m are encoded in log2b bits.
		final long m = ( 1 << log2b + 1 ) - b; 

		if ( x < m ) return writeLong( x, log2b );
		else return writeLong( m + x, log2b + 1 );
	}



	/** Writes a natural number in Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * the argument <code>x</code> may only be zero, and nothing will be written. 
	 *
	 * @param x a natural number.
	 * @param b the modulus for the coding.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a negative modulus.
	 */

	public int writeGolomb( int x, int b ) throws java.io.IOException {
		return writeGolomb( x, b, Fast.mostSignificantBit( b ) );
	}

	/** Writes a natural number in Golomb coding.
	 *
	 * This method is faster than {@link #writeGolomb(int,int)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * the argument <code>x</code> may only be zero, and nothing will be written. 
	 *
	 * @param x a natural number.
	 * @param b the modulus for the coding.
	 * @param log2b the floor of the base-2 logarithm of the coding modulus (it is irrelevant when <code>b</code> is zero).
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a negative modulus.
	 */

	public int writeGolomb( int x, int b, int log2b ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) {
			if ( x != 0 ) throw new IllegalArgumentException( "The modulus is 0, but the argument is " + x );
			return 0;
		}

		final int l = writeUnary( x / b );

		// The remainder to be encoded.
		return l + writeMinimalBinary( x % b, b, log2b );
	}



	/** Writes a long natural number in Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * the argument <code>x</code> may only be zero, and nothing will be written. 
	 *
	 * @param x a long natural number.
	 * @param b the modulus for the coding.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a negative modulus.
	 */

	public long writeLongGolomb( long x, long b ) throws java.io.IOException {
		return writeLongGolomb( x, b, Fast.mostSignificantBit( b ) );
	}

	/** Writes a long natural number in Golomb coding.
	 *
	 * This method is faster than {@link #writeLongGolomb(long,long)} because it does not
	 * have to compute <code>log2b</code>.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * the argument <code>x</code> may only be zero, and nothing will be written. 
	 *
	 * @param x a long natural number.
	 * @param b the modulus for the coding.
	 * @param log2b the floor of the base-2 logarithm of the coding modulus (it is irrelevant when <code>b</code> is zero).
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a negative modulus.
	 */

	public long writeLongGolomb( long x, long b, int log2b ) throws java.io.IOException {

		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( b < 1 ) throw new IllegalArgumentException( "The modulus " + b + " is not positive" );
		if ( b == 0 ) {
			if ( x != 0 ) throw new IllegalArgumentException( "The modulus is 0, but the argument is " + x );
			return 0;
		}

		final long l = writeLongUnary( x / b );

		// The remainder to be encoded.
		return l + writeLongMinimalBinary( x % b, b, log2b );
	}


	/** Writes a natural number in skewed Golomb coding.
	 *
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * the argument <code>x</code> may only be zero, and nothing will be written. 
	 *
	 * @param x a natural number.
	 * @param b the modulus for the coding.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a negative modulus.
	 */

	public int writeSkewedGolomb( int x, int b ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) {
			if ( x != 0 ) throw new IllegalArgumentException( "The modulus is 0, but the argument is " + x );
			return 0;
		}

		final int i = Fast.mostSignificantBit( x / b + 1 );
		final int l = writeUnary( i );
		final int M = ( ( 1 << i + 1 ) - 1 ) * b;
		final int m = ( M / ( 2 * b ) ) * b;

		return l + writeMinimalBinary( x - m, M - m );
	}

	/** Writes a long natural number in skewed Golomb coding.
	 * 
	 * <P>This method implements also the case in which <code>b</code> is 0: in this case,
	 * the argument <code>x</code> may only be zero, and nothing will be written. 
	 *
	 * @param x a long natural number.
	 * @param b the modulus for the coding.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a negative modulus.
	 */

	public long writeLongSkewedGolomb( long x, long b ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( b < 0 ) throw new IllegalArgumentException( "The modulus " + b + " is negative" );
		if ( b == 0 ) {
			if ( x != 0 ) throw new IllegalArgumentException( "The modulus is 0, but the argument is " + x );
			return 0;
		}

		final long i = Fast.mostSignificantBit( x / b + 1 );
		final long l = writeLongUnary( i );
		final long M = ( ( 1 << i + 1 ) - 1 ) * b;
		final long m = ( M / ( 2 * b ) ) * b;

		return l + writeLongMinimalBinary( x - m, M - m );
	}


	/** Writes a natural number in &zeta; coding.
	 *
	 * <P>&zeta; coding (with modulo <var>k</var>) records positive numbers in
	 * the intervals
	 * [1,2<sup><var>k</var></sup>-1],[2<sup><var>k</var></sup>,2<sup><var>k</var>+1</sup>-1],&hellip;,[2<sup><var>hk</var></sup>,2<sup>(<var>h</var>+1)<var>k</var></sup>-1]
	 * by coding <var>h</var> in unary, followed by a minimal binary coding of
	 * the offset in the interval.  The coding of a natural number is obtained
	 * by adding one and coding.
	 * 
	 * <P>A detailed analysis of &zeta; codes is given in 
	 * <blockquote>Paolo
 	 * Boldi and Sebastiano Vigna, <a href="http://vigna.dsi.unimi.it/papers.php#BoVWFII"><em>The WebGraph Framework II: Codes for the World-Wide Web</em></a>.
	 * </blockquote>
	 *
	 * @param x a natural number.
	 * @param k the shrinking factor.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a nonpositive shrinking factor.
	 */

	public int writeZeta( int x, int k ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( k < 1 ) throw new IllegalArgumentException( "The shrinking factor " + k + " is not positive" );

		final int msb = Fast.mostSignificantBit( ++x );
		final int h = msb / k;
		final int l = writeUnary( h );
		final int left = 1 << h * k;
		return l + ( x - left < left 
			? writeInt( x - left, h * k + k - 1 ) 
			: writeInt( x, h * k + k ) );
	}

	/** Writes a long natural number in &zeta; coding.
	 *
	 * @param x a long natural number.
	 * @param k the shrinking factor.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number or use a nonpositive shrinking factor.
	 */

	public long writeLongZeta( long x, int k ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );
		if ( k < 1 ) throw new IllegalArgumentException( "The shrinking factor " + k + " is not positive" );

		final int msb = Fast.mostSignificantBit( ++x );
		final int h = msb / k;
		final int l = writeUnary( h );
		final long left = 1 << h * k;
		return l + ( x - left < left 
			? writeLong( x - left, h * k + k - 1 ) 
			: writeLong( x, h * k + k ) );
	}
	

	/** Writes a natural number in variable-length nibble coding.
	 *
	 * <P>Variable-length nibble coding records a natural number by padding its binary
	 * representation to the left using zeroes, until its length is a multiple of three.
	 * Then, the resulting string is
     * broken in blocks of 3 bits, and each block is prefixed with a bit, which is
     * zero for all blocks except for the last one. 
	 * @param x a natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number.
	 */

	public int writeNibble( int x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		if ( x == 0 ) return writeInt( 8, 4 );
		final int msb = Fast.mostSignificantBit( x );
		int h = msb / 3;
		do {
			writeBit( h == 0 );
			writeInt( x >> h * 3 , 3 );
		} while( h-- != 0 );
		return ( ( msb / 3 ) + 1 ) << 2;
	}
	
	/** Writes a long natural number in variable-length nibble coding.
	 *
	 * @param x a long natural number.
	 * @return the number of bits written.
	 * @throws IllegalArgumentException if you try to write a negative number.
	 */

	public int writeLongNibble( long x ) throws java.io.IOException {
		if ( x < 0 ) throw new IllegalArgumentException( "The argument " + x + " is negative" );

		if ( x == 0 ) return writeInt( 8, 4 );
		final int msb = Fast.mostSignificantBit( x );
		int h = msb / 3;
		do {
			writeBit( h == 0 );
			writeInt( (int)( x >> h * 3 ) , 3 );
		} while( h-- != 0 );
		return ( ( msb / 3 ) + 1 ) << 2;
	}
	
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
