package it.unimi.dsi.mg4j.query;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.mg4j.search.DocumentIterator;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.index.IndexIterator;
import it.unimi.dsi.mg4j.query.parser.QueryParser;
import it.unimi.dsi.mg4j.query.parser.TokenMgrError;
import it.unimi.dsi.mg4j.search.IntervalIterator;
import it.unimi.dsi.mg4j.search.IntervalIterators;
import it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator;
import it.unimi.dsi.mg4j.search.score.Scorer;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.ProgressMeter;
import it.unimi.dsi.mg4j.util.ScoredDocumentBoundedSizeQueue;

import java.io.EOFException;
import java.io.IOException;
import java.text.ParseException;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.gnu.readline.Readline;
import org.gnu.readline.ReadlineLibrary;


/** A command-line interpreter to query indices. */
public class Query {
	
	private Query() {}
	private final static String THIS_CLASS_NAME = new Query().getClass().getName();

	/** A mode used to display results (as intervals). */
	private static final int DISPLAY_MODE = 0;
	/** A mode used to display scores. */
	private static final int SCORE_MODE = 1;
	
	/** The current mode. */
	private static int mode = DISPLAY_MODE;
	/** The current scorer class used (if the mode is {@link #SCORE_MODE}. */
	private static Class scorerClass = null;
	/** The maximum number of output per query. */
	private static int maxOutput = 50;
	/** Whether to actually output anything. */
	private static boolean output = true;

	/** A map from index names to indices. */
	private static Map indexMap = new LinkedHashMap();

	/** Outputs the given string followed by a newline if {@link #output} is true.
	 * 
	 * @param s a string.
	 */
	private static void printlnIfOutput( final String s ) {
		if ( output ) System.out.println( s );
	}
	
	/** Outputs the a newline if {@link #output} is true.
	 */
	private static void printlnIfOutput() {
		if ( output ) System.out.println();
	}
	
	/** Outputs the given string if {@link #output} is true.
	 * 
	 * @param s a string.
	 */
	private static void printIfOutput( final String s ) {
		if ( output ) System.out.print( s );
	}

	/** Interpret the given command, changing the static variables.
	 *  Possible commands are
	 *  <pre>
	 *         !display    &lt;-- pass to display-mode
	 *         !score ScorerClassName   &lt;-- use the given scorer (default constructor)
	 *         !output n   &lt;-- output at most n lines (0=all)+
	 *  </pre>
	 *  
	 * @param command the command.
	 */
	public static void interpretCommand( String command ) {
		try {
			String parts[] = command.substring( 1 ).split( " " );
			if ( parts[ 0 ].equals( "display" ) ) mode = DISPLAY_MODE;
			else if ( parts[ 0 ].equals( "output" ) ) {
				output = ! "off".equals( parts[ 1 ] );
				if ( output ) {
					int out = Integer.parseInt( parts[ 1 ] );
					maxOutput = ( out == 0 )? Integer.MAX_VALUE : out;
				}
			}
			else if ( parts[ 0 ].equals( "score" ) ) {
				Class c = Class.forName( "it.unimi.dsi.mg4j.search.score." + parts[ 1 ] );
				if ( Scorer.class.isAssignableFrom( c ) ) scorerClass = c;
				else throw new IllegalStateException( "Class " + c + " is not an implementation of Scorer" );
				mode = SCORE_MODE;
			}
			else if ( parts[ 0 ].equals( "quit" ) ) {
				System.exit( 0 ); 
			}
			else if ( parts[ 0 ].equals( "help" ) ){	
				System.err.println( "!help\tprints this help"); 
				System.err.println( "!display\tdisplay intervals"); 
				System.err.println( "!output <max>\toutput at most <max> results per query"); 
				System.err.println( "!output off\tturns off output"); 
				System.err.println( "!quit\tquits the program" );
				System.err.println( "!score <ScorerClass>\tdisplay scores according to <ScorerClass>"); 
			}
			else throw new IllegalStateException( "Invalid command " + command );
		} catch ( Throwable t ) {
			t.printStackTrace();
		}
	}
	
	/** Produces the output for the given document iterator.
	 * 
	 * @param docIt the document iterator.
	 * @param titles an optional list of titles, or <code>null</code>.
	 */
	private static void scoreModeOutput( final DocumentIterator docIt, final List titles ) {
		Scorer s;
		try {
			s = (Scorer)scorerClass.newInstance();
		} catch ( InstantiationException e ) {
			throw new RuntimeException( e );
		} catch (IllegalAccessException e) {
			throw new RuntimeException( e );
		}
		ScoredDocumentIterator sdi = s.scoredDocumentIterator( docIt );
		
		ScoredDocumentBoundedSizeQueue top = new ScoredDocumentBoundedSizeQueue( maxOutput );
		while ( sdi.hasNext() ) top.enqueue( sdi.nextDocument(), sdi.score() );
		
		int i, index;
		for ( i = 0; i < maxOutput && ! top.isEmpty(); i++ ) {
			index = top.dequeue();
			printlnIfOutput( "Document #" + top.document[ index ] + " score " + top.score[ index ] + ( titles != null ? " " + titles.get( top.document[ index ] ) : "" ) );
		}
	}
	
	/** Scores the given document iterator and produces score output.
	 * 
	 * @param result the document iterator.
	 * @param titles an optional list of titles, or <code>null</code>.
	 */
	public static void displayModeOutput( final DocumentIterator result, final List titles ) {	 
		for ( int i = 0; i < maxOutput && result.hasNext(); i++ ) {
			final int document = result.nextInt();
			printlnIfOutput( "Document #" + document + ( titles != null ? " " + titles.get( document ) : "" ) );
			Iterator keyIterator = indexMap.keySet().iterator();
			while ( keyIterator.hasNext() ) {
				Index index = (Index)indexMap.get( keyIterator.next() );
				if ( ! result.indices().contains( index ) ) continue;
				if ( index.hasPositions ) {
					IntervalIterator intIt = result.intervalIterator( index );
					printlnIfOutput( "\t" + index + "[" + intIt.extent() + "]:" );
					
					if ( intIt == IntervalIterators.TRUE ) printlnIfOutput( "TRUE" );
					else if ( intIt == IntervalIterators.FALSE ) printlnIfOutput( "FALSE" );
					else {
						while ( intIt.hasNext() ) printIfOutput( intIt.next() + " " );
						printlnIfOutput();
					}
				}
				else if ( index.hasCounts && result instanceof IndexIterator ) printlnIfOutput( "\t" + ((IndexIterator)result).count() + " occurrences" );
			}
		}
	}
	

	/** The main method. Reads a query from standard input and parses it. */
	public static void main( String arg[] ) throws ParseException, TokenMgrError, IOException, ClassNotFoundException {

		int ch;
		String charset = "UTF-8";
		String titleFilename = null;

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "titles", LongOpt.REQUIRED_ARGUMENT, null, 't' ),
			new LongOpt( "charset", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
		};

		Getopt g = new Getopt( THIS_CLASS_NAME, arg, "c:t:h", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: " + THIS_CLASS_NAME + " [OPTIONS] INDEX..." );
			System.err.println( "Reads queries from the command lines, and interrogates the given indices." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  INDEX      the basename of one or more indices" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -c, --charset             the input charset encoding (default: " + charset + ")" );
			System.err.println( "  -t, --titles              a Java List stored object containing titles for the documents" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help        print this help screen" );
			System.err.println( "" );
			return;

		case 't':
			titleFilename = g.getOptarg();
			break;

		case 'c':
			charset = g.getOptarg();
			break;

		case '?':
			return;

		default:
			break;
		}

		if ( arg.length - g.getOptind() == 0 ) {
			System.err.println( "You must provide one or more index basenames" );
			return;
		}

		ProgressMeter pm = new ProgressMeter( 1000000 );
		for ( int i = g.getOptind(); i < arg.length; i++ ) indexMap.put( arg[ i ], Index.getInstance( arg[ i ], pm ) );
		QueryParser parser = new QueryParser( indexMap, arg[ g.getOptind() ] );
		
		final List titles;
		
		if ( titleFilename != null ) {
			System.err.print( "Loading titles..." );
			titles = (List)Fast.loadObject( titleFilename );
			System.err.println( " done." );
		}
		else titles = null;
				
		try {
			 Readline.load( ReadlineLibrary.GnuReadline );
		 }
		 catch ( UnsatisfiedLinkError noNativeLibrary ) {
			System.err.println( "Couldn't load readline native lib. Using pure-Java implementation." );
			Readline.load( ReadlineLibrary.PureJava );
		 }

		 Readline.initReadline( "Query" );

		 Runtime.getRuntime()                       // if your version supports
		   .addShutdownHook( new Thread() {          // addShutdownHook (since 1.3)
			  public void run() {
				Readline.cleanup();
			  }
			});
	    
		String query;
		DocumentIterator result;

		try {
			for (;;) {
				while ((query = Readline.readline("[!help]>")) == null);
				if (query.charAt(0) == '!') {
					interpretCommand(query);
					continue;
				}
				try {
					result = parser.parse(query);
				} catch (Throwable t) {
					System.err.println(t);
					continue;
				}
				final long start = System.currentTimeMillis();
				switch (mode) {
					case DISPLAY_MODE :
						displayModeOutput( result, titles );
						break;
					case SCORE_MODE :
						scoreModeOutput( result, titles );
						break;
				}
				System.err.println( System.currentTimeMillis() - start + " ms" );
			}
		} catch ( EOFException endOfInput ) {}
	}

}
