package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntHeapSesquiIndirectDoublePriorityQueue;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceOpenHashMap;
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.util.MutableString;

import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;


/** An abstract iterator on documents, generating the intersection of the documents returned by 
 * a number of document iterators.
 * 
 * <P>To be usable, this class must be subclassed so to provide also an iterator on intervals.
 * Such iterators must be instantiable using {@link #getComposedIntervalIterator(Index)}.
 * The latter is an example of a <em>non-static factory method</em>, that is, a factory method which
 * depends on the enclosing instance. This pattern allows to easily specialise this class to iterators
 * that do different things, such as {@link it.unimi.dsi.mg4j.search.AndDocumentIterator} and
 * {@link it.unimi.dsi.mg4j.search.ConsecutiveDocumentIterator}, but that have a similar semantics at the document
 * level (the semantics may in fact be slightly different: for instance, not all document belonging
 * to all components will actually appear in a consecutive iterator, as there may be documents filtered
 * at the interval level).
 * 
 *  <P>The important invariant is that <em>only</em> after a call to {@link #nextDocument()}, a call
 *  to {@link #intervalIterator(Index)} will return an interval iterator over the document
 *  just returned, and that for at least one index in {@link #indices()} the iterator will not be empty
 *  or {@link it.unimi.dsi.mg4j.search.IntervalIterators#TRUE TRUE}.
 * 
 */

public abstract class AbstractIntersectionDocumentIterator extends AbstractIntIterator implements DocumentIterator {

	/** The component document iterators. */
	final protected DocumentIterator it[];
	/** The number of component iterators. */
	final private int n;
	/** A heap-based sesqui-indirect double priority queue used to keep track of the currently scanned documents. */
	final private IntHeapSesquiIndirectDoublePriorityQueue queue;
	/** The reference array used for the queue. */
	final private int curr[];
	/** The set of indices involved in this iterator. */
	final private ReferenceOpenHashSet indices = new ReferenceOpenHashSet();
	/** A map from indices to interval iterators. */
	final private Reference2ReferenceOpenHashMap intervalIterators;
	/** A map from indices to the iterators returned for the current document. The key set may
	 * not contain an index because the related iterator has never been requested. Moreover,
	 * the iterator in this map for a given index may differ from the one in {@link #intervalIterators}
	 * because it could be {@link IntervalIterators#TRUE TRUE} (in fact, in that case it may even
	 * happen that {@link #intervalIterators} does not contain the index). */
	final private Reference2ReferenceOpenHashMap currentIterators;
	/** An unmodifiable wrapper around {@link #currentIterators}. */
	final private Map unmodifiableCurrentIterators;

	/** The next integer that will be returned by this iterator, or -1 if it is necessary to advance
	 *  before (possibly) returning a new document. */
	private int next;
	/** The last integer returned by this iterator. */
	private int last = -1;
	/** If {@link #next} is negative, this variable denotes whether the scan is over. */
	private boolean endOfProcess;
	

	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = true;

	static { if ( DEBUG && ASSERTS ); }

	/** Creates a new intersection iterator using a given array of iterators.
	 *  @param it the iterators to be insersected.
	 */
	public AbstractIntersectionDocumentIterator( DocumentIterator it[] ) {
		this.it = it;
		this.n = it.length;
		this.curr = new int[ n ];
		// The primary/secondary comparator gives the minimum/maximum among the currently scanned integers. 
		queue = new IntHeapSesquiIndirectDoublePriorityQueue( curr );

		// We gather the indices used in all iterators.
		for( int i = 0; i < n; i++ ) indices.addAll( it[ i ].indices() );		

		intervalIterators = new Reference2ReferenceOpenHashMap( indices.size(), .5f );
		currentIterators = new Reference2ReferenceOpenHashMap( indices.size(), .5f );
		unmodifiableCurrentIterators = Reference2ReferenceMaps.unmodifiable( currentIterators );

		// We now load the queue with the first document from each iterator.
	
		next = -1;
		for ( int i = 0; i < n; i++ ) {
			if ( !it[ i ].hasNext() ) {
				// If any of the iterator is empty, we're over.
				endOfProcess = true;
				return;
			}
			curr[ i ] = it[ i ].nextInt();
			queue.enqueue( i );
		}

		align();
		if ( ! endOfProcess ) next = curr[ queue.first() ];
	}

	public Set indices() { return indices; }

	/** Align all iterators on the same document. 
	 * 
	 * <P>After a call to this method, all component iterators are positioned
	 * on the same document, unless {@link #endOfProcess} is true. If they are
	 * already aligned, this method just exits. {@link #next} is
	 * not modified by this method. 
	 */
	private void align() {
		if ( DEBUG ) System.err.println( this + ".align()..." ); 
		int first, candidateNext, res;

		while ( ! endOfProcess && curr[ first = queue.first() ] < ( candidateNext = curr[ queue.secondaryFirst() ] ) ) {
			res = it[ first ].skipTo( candidateNext );
			if ( res < 0 ) endOfProcess = true;
			else {
				curr[ first ] = res;
				queue.changed();
			} 
		}

		if ( DEBUG ) 
			if ( endOfProcess ) System.err.println( this + "...exhausted." );
			else System.err.println( this + "...aligned on " + curr[ queue.first() ] );
	}

	/** Advances this iterator, assuming that it is aligned.  
	 * 
	 * <P>More precisely, if <code>n</code>&ge;0 it advances all the lists that correspond to
	 * the currently least integer until an element larger than or equal
	 * to <code>n</code> would be returned (it might set {@link #endOfProcess} to true).
	 * Otherwise, it moves to the next element (which, however, could be a false positive).
	 *
	 * <P>This method assumes that all iterators are aligned (or {@link #endOfProcess}, in which case
	 * this method does nothing); at exit, they will still be aligned (but not necessarily valid),
	 * unless  {@link #endOfProcess}. {@link #next} is
	 * not modified by this method.
	 *
	 * @param n a lower bound on the document pointer, or -1.
	 */
	private void advance( int n ) {
		if ( DEBUG ) System.err.println( this + ".advance(" + n + ")" ); 
		if ( ASSERTS ) assert endOfProcess || curr[ queue.first() ] == curr[ queue.secondaryFirst() ];
			
		int first, res;

		// BEWARE: the non-short-circuited OR is critical
		if ( ! endOfProcess && ( n < 0 | curr[ first = queue.first() ] < n ) ) {
			if ( n >= 0 ) res = it[ first ].skipTo( n );
			else res = it[ first ].hasNext() ? it[ first ].nextInt() : -1;

			if ( res >= 0 ) {
				curr[ first ] = res;
				queue.changed();
				align();
			}
			else endOfProcess = true;
		}

		currentIterators.clear(); 
	}
	
	public int skipTo( int n ) {
		if ( last >= n ) return last;
		align();
		advance( n );
		next = -1;
		if ( endOfProcess ) return -1;
		last = curr[ queue.first() ];
		return last;
	}

	/** Invalidates the current document, and forces
	 * {@link #hasNext()} to find whether there is another one.
	 * 
	 * <P>This method call may be used by subclasses to skip documents on the
	 * base of some further analysis (e.g., on the interval iterators, as it happens 
	 * for {@link ConsecutiveDocumentIterator}).
	 */

	protected void invalidate() {
		next = -1;
	}

	public int nextInt() {
		if ( DEBUG ) System.err.println( this + ".nextInt()..." ); 

		if ( ! hasNext() ) throw new NoSuchElementException();

		last = next;
		invalidate();
		
		if ( DEBUG ) System.err.println( this + "...returning " + last ); 
		return last;
	}
	
	public int document() { 
		if ( last == -1 ) throw new IllegalStateException(); 
		return last; 
	}

	public int nextDocument() { return nextInt(); }

	public boolean hasNext() {
		if ( DEBUG ) System.err.println( this + ".hasNext()" );

		// We already know what to return
		if ( next >= 0 ) return true;
		align();
		advance( -1 ); 
		// If we're over, we stop here
		if ( endOfProcess ) return false;

		next = curr[ queue.first() ];
		return true;
	}

	public Map intervalIterators() {
		final Iterator i = indices.iterator();
		while( i.hasNext() ) intervalIterator( (Index) i.next() );
		return unmodifiableCurrentIterators;
	}

	public IntervalIterator intervalIterator() {
		if ( indices.size() != 1 ) throw new IllegalStateException();
		return intervalIterator( (Index)indices().iterator().next() );
	}

	public IntervalIterator intervalIterator( final Index index ) {
		if ( DEBUG ) System.err.println( this + ".intervalIterator(" + index + ")" );

		IntervalIterator intervalIterator;

		// If the iterator has been created and it's ready, we just return it.		
		if ( ( intervalIterator = (IntervalIterator) currentIterators.get( index ) ) != null ) return intervalIterator;

		int i, c;
			
		/* None of the iterators may be FALSE. Otherwise, if all
		 * iterators are TRUE, we return TRUE. Otherwise, we return
		 * (possibly after creation) the underlying interval iterator.
		 */

		for( i = c = 0; i < n; i++ ) {
			intervalIterator = it[ i ].intervalIterator( index );
			// We cannot be on a document one of whose iterators if FALSE.
			if ( intervalIterator == IntervalIterators.FALSE ) break;
			if ( intervalIterator != IntervalIterators.TRUE ) c++;
		}
		
		if ( i < n ) intervalIterator = IntervalIterators.FALSE;
		else if ( c == 0 ) intervalIterator = IntervalIterators.TRUE;
		else {
			intervalIterator = (IntervalIterator) intervalIterators.get( index );
			if ( intervalIterator == null ) intervalIterators.put( index, intervalIterator = getComposedIntervalIterator( index ) );
		}

		currentIterators.put( index, intervalIterator );	
		intervalIterator.reset();
		return intervalIterator;
	}
	
	
	abstract protected IntervalIterator getComposedIntervalIterator( Index index );
	

	public String toString() {
	   MutableString res = new MutableString();
	   res.append( this.getClass().getName() ).append( "(" ).delete( 0, res.lastIndexOf( '.' ) + 1 );
	   for ( int i = 0; i < n; i++ ) res.append( i > 0 ? "," : "" ).append( it[ i ] );
	   return res.append( ")" ).toString();
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
