package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.fastutil.objects.ObjectArrays;
import it.unimi.dsi.fastutil.objects.ObjectHeapSesquiIndirectDoublePriorityQueue;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.util.MutableString;

import java.util.NoSuchElementException;


/** An iterator on documents that returns the AND of a number of document iterators.
 *
 * <P>This class adds to {@link it.unimi.dsi.mg4j.search.AbstractIntersectionDocumentIterator}
 * an interval interator generating the AND of the intervals returned for each of the documents involved. 
 */

public class AndDocumentIterator extends AbstractIntersectionDocumentIterator {

	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = true;

	static { if ( DEBUG && ASSERTS ); }
	
	public AndDocumentIterator( DocumentIterator it[] ) {
		super( it );
	}

	protected IntervalIterator getComposedIntervalIterator( final Index index ) {
		return new AndIntervalIterator( index );
	}

	private class AndIntervalIterator extends AbstractObjectIterator implements IntervalIterator {
		/** The index of this iterator. */
		final Index index;
		/** The underlying iterators. */
		final private IntervalIterator it[];
		/** The number of underlying iterators. */
		final private int n;
		/** The reference array used for the queue. */	 
		final private Interval curr[];
		/** A heap-based sesqui-indirect double priority queue used to keep
		 *  track of the currently scanned intervals. */
		final private ObjectHeapSesquiIndirectDoublePriorityQueue queue;

		/** Whether the scan is over. */
		private boolean endOfProcess;
		/** Whether the next interval is already loaded in {@link #next}. */
		private boolean nextIsLoaded;
		/** The left extreme of the next interval that will be returned. */
		private int nextLeft;
		/** The right extreme of the next interval that will be returned. */
		private int nextRight;

		/** Creates a new AND interval iterator.
		 * 
		 * @param index the index of the iterator.
		 */

		public AndIntervalIterator( final Index index ) {
			// We just set up some internal data, but we perform no initialisation.
			this.index = index;
			n = AndDocumentIterator.this.it.length;
			curr = new Interval[ n ];
			it = new IntervalIterator[ n ];
			queue = new ObjectHeapSesquiIndirectDoublePriorityQueue( curr, Intervals.LEFT_INTERVAL_COMPARATOR, Intervals.RIGHT_INTERVAL_COMPARATOR );
		}


		public void reset() {
			nextIsLoaded = false;
			
			ObjectArrays.fill( curr, null );
			queue.clear();
			
			for ( int i = 0; i < n; i++ ) {
				it[ i ] = AndDocumentIterator.this.it[ i ].intervalIterator( index );
				if ( it[ i ] != IntervalIterators.TRUE && it[ i ].hasNext() ) {
					curr[ i ] = it[ i ].nextInterval();
					queue.enqueue( i );
				}
			}

			endOfProcess = queue.isEmpty();
		}


		public boolean hasNext() {
			if ( DEBUG ) System.err.println( this + ".hasNext()" );
			if ( nextIsLoaded ) return true;
			if ( endOfProcess ) return false;
			nextIsLoaded = true;

			do {
				nextLeft = curr[ queue.first() ].left;
				nextRight = curr[ queue.secondaryFirst() ].right;
				if ( DEBUG ) System.err.println( this + " is saving interval " + Interval.getInstance( nextLeft, nextRight ) );
				
				if ( !it[ queue.first() ].hasNext() ) { 
					endOfProcess = true;
					return true;
				}
				
				curr[ queue.first() ] = it[ queue.first() ].nextInterval();
				queue.changed();
				if ( ASSERTS ) assert curr[ queue.secondaryFirst() ].right >= nextRight;
			} while ( curr[ queue.secondaryFirst() ].right == nextRight );
			
			while ( curr[ queue.first() ].left == nextLeft && !endOfProcess ) {
				if ( !it[ queue.first() ].hasNext() ) {
					endOfProcess = true;
					return true;
				}
				curr[ queue.first() ] = it[ queue.first() ].nextInterval();
				queue.changed();
			}
			return true;
		}

		public Interval nextInterval() {
			if ( ! hasNext() ) throw new NoSuchElementException();
			nextIsLoaded = false;
			return Interval.getInstance( nextLeft, nextRight );
		}
		
		public int extent() {
			int i, e, s;

			s = 0;
			for ( i = 0; i < n; i++ ) {
				e = it[ i ] != IntervalIterators.TRUE ? it[ i ].extent() : 0;
				if ( e > 0 ) s += e;
			}
			return s;
		}
		
		public Object next() {
			return nextInterval();
		}

		public String toString() {
		   MutableString res = new MutableString();
		   res.append( this.getClass().getName() ).append( "(" ).delete( 0, res.lastIndexOf( '.' ) + 1 );
		   for ( int i = 0; i < n; i++ ) res.append( i > 0 ? "," : "" ).append( it[ i ] );
		   return res.append( ")" ).toString();
		}
	}


}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
