package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.fastutil.objects.ObjectArrays;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.util.MutableString;

import java.util.NoSuchElementException;

/** An iterator returning documents containing consecutive intervals satisfying the underlying queries;
 * the intervals must be in query order.
 */

public class ConsecutiveDocumentIterator extends AbstractIntersectionDocumentIterator {

	public final static boolean DEBUG = false;
	public final static boolean ASSERTS = true;
	static { if ( DEBUG && ASSERTS ); }
	
	public ConsecutiveDocumentIterator( DocumentIterator it[] ) {
		super( it );
	}

	public boolean hasNext() {
		while( super.hasNext() && ! intervalIterator().hasNext() ) invalidate();
		return super.hasNext();
	}

	protected IntervalIterator getComposedIntervalIterator( final Index index ) {
		return new ConsecutiveIntervalIterator( index );
	}

	private class ConsecutiveIntervalIterator extends AbstractObjectIterator implements IntervalIterator {
		/** The index of this iterator. */
		final Index index;
		/** The underlying iterators. */
		final private IntervalIterator it[];
		/** The number of underlying iterators. */
		final private int n;
		/** The reference array used for the queue. */	 
		final private Interval curr[];
		/** Whether the scan is over. */
		private boolean endOfProcess;
		/** Whether the next interval is already loaded in {@link #next}. */
		private boolean nextIsLoaded;

		/** Creates a new consecutive interval iterator.
		 * 
		 * @param index the index of the iterator.
		 */

		public ConsecutiveIntervalIterator( final Index index ) {
			// We just set up some internal data, but we perform no initialisation.
			this.index = index;
			n = ConsecutiveDocumentIterator.this.it.length;
			curr = new Interval[n];
			it = new IntervalIterator[ n ];
		}

		public void reset() {
			int i;
			nextIsLoaded = false;
			
			ObjectArrays.fill( curr, null );

			for ( i = 0; i < n; i++ ) {
				it[ i ] = ConsecutiveDocumentIterator.this.it[ i ].intervalIterator( index );
				if ( it[ i ] == IntervalIterators.TRUE ) throw new IllegalStateException();
				if ( ! it[ i ].hasNext() ) break;
				curr[ i ] = it[ i ].nextInterval();
			}

			if ( endOfProcess = i < n ) return;

			align();
			if ( endOfProcess ) return;
			nextIsLoaded = true;
		}


		/** After calling this method, either {@link #endOfProcess} is true (in which case
		 *  there are no more documents to be returned), or all component iterators are positioned
		 *  so that they form a consecutive sequence. If this condition was true at start, the
		 *  method does nothing.
		 */
		private void align() {
			if ( DEBUG ) System.err.println( this + ".align()" );
			while ( ! endOfProcess ) {
				if ( DEBUG ) System.err.println( java.util.Arrays.asList( curr ) );
				int k;
				for ( k = 1; k < n; k++ ) {
					while ( curr[ k ].left < curr[ k - 1 ].right + 1 )
						if ( it[ k ].hasNext() ) curr[ k ] = it[ k ].nextInterval();
						else { 
							endOfProcess = true; 
							return; 
						}
					if ( curr[ k ].left > curr[ k - 1 ].right + 1 ) {
						if ( it[ 0 ].hasNext() ) 
							curr[ 0 ] = it[ 0 ].nextInterval();
						else {
							endOfProcess = true;
							return;
						}
						break;
					} 
				}
				if ( k == n ) return; 
			}
		}

		public boolean hasNext() {
			if ( nextIsLoaded ) return true;
			if ( endOfProcess ) return false;
			if ( it[ 0 ].hasNext() ) curr[ 0 ] = it[ 0 ].nextInterval();
			else {
				endOfProcess = true;
				return false;
			}
			align();
			if ( endOfProcess ) return false;			
			if ( DEBUG ) System.err.println( "In hasNext(): " + java.util.Arrays.asList( curr ) );
			return nextIsLoaded = true;
		}
		
		public Interval nextInterval() {
			if ( ! hasNext() ) throw new NoSuchElementException();
			nextIsLoaded = false;
			return Interval.getInstance( curr[ 0 ].left, curr[ n - 1 ].right );
		}

		public int extent() {
			int i, e, s;

			s = 0;
			for ( i = 0; i < n; i++ ) {
				e = it[ i ] != IntervalIterators.TRUE ? it[ i ].extent() : 0;
				if ( e > 0 ) s += e;
			}
			return s;
		}
				
		public Object next() {
			return nextInterval();
		}

		public String toString() {
		   MutableString res = new MutableString();
		   res.append( this.getClass().getName() ).append( "(" ).delete( 0, res.lastIndexOf( '.' ) + 1 );
		   for ( int i = 0; i < n; i++ ) res.append( i > 0 ? "," : "" ).append( it[ i ] );
		   return res.append( ")" ).toString();
		}
	}


}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
