package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.mg4j.index.Index;

import java.util.Map;
import java.util.Set;

/** An iterator over documents and their intervals.
 * 
 * <P>This class extends {@link IntIterator} to a full-fledged document iterator. Each call to {@link #nextDocument()}
 * (a synonym for {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}) will return a document pointer. Just
 * after the call to {@link #nextDocument()}, {@link #intervalIterator(Index)} will return an interval iterator
 * enumerating intervals in the last returned document for the specified index. The latter method may return, as a special result, a
 * special {@link it.unimi.dsi.mg4j.search.IntervalIterators#TRUE TRUE} value: this means that even you should never care about the intervals this document
 * iterator returns (see {@link it.unimi.dsi.mg4j.search.IntervalIterators#TRUE TRUE} for some elaboration). 
 * 
 * <p><strong>Warning:</strong> the interval enumeration can be carried out only just after a call
 * to {@link #nextDocument()}. Subsequent calls to {@link #nextDocument()} <em>or even {@link java.util.Iterator#hasNext()}</em>
 * will reset the internal state of the iterator. In particular, trying to enumerate intervals after a call
 * to {@link java.util.Iterator#hasNext()} will throw an {@link java.lang.IllegalStateException}.
 */

public interface DocumentIterator extends IntIterator {


	/** Returns the interval iterator of this document iterator for single-index queries.
	 * 
	 * <P>This is a commodity method that can be used only for queries
	 * built over a single index.
	 * 
	 * @return an interval iterator.
	 * @return the next interval in the sequence.
	 * @see #intervalIterator(Index) 
	 * @throws IllegalStateException if this document iterator is not built on a single index.
	 */
	public IntervalIterator intervalIterator();


	/** Returns the interval iterator of this document iterator for the given index.
	 * 
	 * <P>After a call to {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}, this iterator
	 * can be used to retrieve the intervals in the current document (the
	 * one returned by {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}) for 
	 * the index <code>index</code>.
	 *  
	 * <P>Note that it is guaranteed that at least one index will return an interval.
	 * However, for disjunctive queries it cannot be guaranteed that <em>all</em>
	 * indices will return an interval.
	 * 
	 * @param index an index (must be one over which the query was built).
	 * @return an interval iterator over the current document in <code>index</code>.
	 * @throws UnsupportedOperationException if this index does not contain positions.
	 */

	public IntervalIterator intervalIterator( Index index );

	/** Returns an unmodifiable map from indices to interval iterators.
	 * 
	 * <P>After a call to {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}, this map
	 * can be used to retrieve the intervals in the current document. An invocation of {@link Map#get(java.lang.Object)}
	 * on this map with argument <code>index</code> yields the same result as
	 * {@link #intervalIterator(Index) intervalIterator(index)}.
	 *  
	 * @return a map from indices to interval iterators over the current document.
	 * @throws UnsupportedOperationException if this index does not contain positions.
	 * @see #intervalIterator(Index)
	 */

	public Map intervalIterators();
	
	/** Returns the set of indices over which this iterator is built.
	 * 
	 * @return the set of indices over which this iterator is built.
	 */

	public Set indices();

	/** A synonym for {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}.
	 *
	 * @return the next document.
	 */
	public int nextDocument();

	/** Returns the last document returned by {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}.
	 * 
	 * @return the last document returned by {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}, or -1 if no document has been returned yet.
	 */

	public int document();


	/** Skips all documents smaller than <code>n</code>.
	 *
	 * <P>Let <code>k</code> be the last document returned by {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt()}. If
	 * <code>k</code> is larger than or equal to <code>n</code>, then
	 * this method does nothing and returns <code>k</code>. Otherwise, a 
	 * call to this method is equivalent to 
	 * <pre>
	 * while( ( found = hasNext() ) && ( k = nextInt() ) < n );
	 * return found ? k : -1;
	 * </pre>
	 *
	 * <P>Beware that the first document larger than or equal to <code>n</code> (which is returned
	 * by this method) will <em>not</em> be returned by the next call to {@link it.unimi.dsi.fastutil.ints.IntIterator#nextInt}.
	 *
	 * @param n a document index.
	 * @return -1, if all documents produced by this iterator are smaller than <code>n</code>; otherwise,
	 * an document larger than or equal to <code>n</code> as described above.
	 */

	int skipTo( int n );

}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
