package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
                                                                                                                                                            
import java.io.Serializable;
                                                                                                                                                             
/** An integral interval. An interval is defined
 *  by two integers, called its {@link #left} and {@link #right} 
 *  extremes. 
 *  
 * <P>This class has no constructor: use the static factories
 * {@link #getInstance(int, int)} and {@link #getInstance(int)}, instead.
 * 
 * <P>Instances of this class are immutable. To reduce garbage collection,
 * intervals made of one non-negative small points are precomputed and returned upon request. 
 */

public final class Interval implements Serializable {
    /** One-point intervals between 0 (inclusive) and this number (exclusive) are generated
     * from a pre-computed array of instances. */
	private final static int MAX_SINGLE_POINT = 1024;
	/** The precomputed array of one-point intervals. */
	private final static Interval[] POINT_INTERVAL = new Interval[ MAX_SINGLE_POINT ];
	                                  
	static {
		int i = MAX_SINGLE_POINT;
		while( i-- != 0 ) POINT_INTERVAL[ i ] = new Interval( i, i );
	}

	/** The left extreme of the interval. */
	public final int left;
	/** The right extreme of the interval. */
	public final int right;
                                                                                                                                                             
	/** Builds an interval with given extremes.
	 * 
	 *  @param left the left extreme.
	 *  @param right the right extreme (which must be greater than 
	 *  or equal to the left extreme).
	 */
	private Interval( final int left, final int right ) {
		if ( left > right ) throw new IllegalArgumentException( "The left extreme (" + left + ") is greater than the right extreme (" + right + ")" );
		this.left = left;
		this.right = right;
	}
                                                                                                                                                             
	/** Returns an interval with given extremes.
	 * 
	 *  @param left the left extreme.
	 *  @param right the right extreme (which must be greater than 
	 *  or equal to the left extreme).
	 *  @return an interval with the given extremes.
	 */
	public static Interval getInstance( final int left, final int right ) {
		if ( left == right ) return getInstance( left );
		return new Interval( left, right );		
	}
                                                                                                                                                             
	/** Returns a one-point interval.
	 * 
	 *  @param point a point.
	 *  @return a one-point interval 
	 */
	public static Interval getInstance( final int point ) {
		if ( point >= 0 && point < MAX_SINGLE_POINT ) return POINT_INTERVAL[ point ]; 
		return new Interval( point, point );
	}
                                                                                                                                                             
	/** Returns the interval length, that is, the number of integers 
	 *  contained in the interval.
	 * 
	 *  @return the interval length.
	 */
	public int length() {
		return right - left + 1;
	}
                                                                                                                                                             
	public String toString() {
		return "[" + left + "," + right + "]";
	}
                                            
	public int hashCode() {
		return left ^ right;
	}
                                                                                                                                                             
	public boolean equals( Object o ) {
		if ( o instanceof Interval )
			return ((Interval)o).left == left && ((Interval)o).right == right;
		else return false;
	}
}
                                                                                                                                                        

// Local Variables:
// mode: jde
// tab-width: 4
// End:
