package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


import it.unimi.dsi.fastutil.Iterators;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import java.util.NoSuchElementException;
                                                                                                                                                           
/** A class providing static methods and objects that do useful things with interval iterators. */

public class IntervalIterators {

	private IntervalIterators() {}

	/** An iterator returning no intervals.
	 *
	 * <P>This class may be useful to implement your own in case you subclass
	 * {@link IntervalIterator}.
	 */

	public static class EmptyIntervalIterator extends Iterators.EmptyIterator implements IntervalIterator {
		protected EmptyIntervalIterator() {}

		public Interval nextInterval() { throw new NoSuchElementException(); }
		public int extent() { return -1; }
		public void reset() {}
	}

	public static final EmptyIntervalIterator EMPTY_ITERATOR = new EmptyIntervalIterator();
	
	/** An iterator that throws an exception on all method calls, except for {@link #hasNext()},
	 * which has a settable value.
	 * 
	 * <P>The only purpose of this class is to instantiate "absolute value" iterators
	 * such as {@link #TRUE} and {@link #FALSE}.
	 */

	public static class FakeIterator extends AbstractObjectIterator implements IntervalIterator {
		final boolean hasNext;
		private FakeIterator( boolean hasNext ) { this.hasNext = hasNext; }
		public boolean hasNext() { return hasNext; }
		public Interval nextInterval() { throw new UnsupportedOperationException(); }
		public void reset() {}
		public Object next() { throw new UnsupportedOperationException(); }
		public int extent() { return -1; }
		public String toString() { return this.getClass().getName() + "(" + Boolean.toString( hasNext ) + ")"; }
	}
                                                                 
	/** An iterator representing absolute truth.
	 * 
	 * <P>When this iterator is returned by a subquery on a given index, this means
	 * that the index has no part in the construction of the result, but that, at the same time,
	 * the result is true.
	 * 
	 * <P>For example, consider the query <samp>(foo OR title:bar) AND fee</samp> evaluated on
	 * a document which contains <samp>bar</samp> in the title index, but not <samp>foo</samp> in the
	 * primary index, and which moreover contains <samp>fee</samp> in the primary index. The subquery
	 * <samp>(foo OR title:bar)</samp> has value {@link #TRUE} on the primary index, as there is nothing
	 * to be returned, but the subquery is nonetheless true.
	 */

	public final static IntervalIterator TRUE = new FakeIterator( true );
	
	/** An iterator representing absolute falseness.
	 * 
	 * <P>When this iterator is returned by a subquery on a given index, this means that
	 * the subquery is false.
	 *
	 * <P>For example, consider the query <samp>(foo AND title:bar) OR fee</samp> evaluated on
	 * a document which contains <samp>bar</samp> in the title index, but not <samp>foo</samp> in the
	 * primary index, and which moreover contains <samp>fee</samp> in the primary index. The subquery
	 * <samp>(foo AND title:bar)</samp> has value {@link #FALSE} on both the primary and the title index.
	 */
	public final static IntervalIterator FALSE = new FakeIterator( false );
	
	/** Computes the Clarke-Cormack score of a document.
	 * 
	 * @param it the iterator; note that the iterator gets exhausted by this call.
	 * @param alpha the parameter used in the computation of the score.
	 * @param h the parameter used in the computation of the score.
	 * @return the score.
	 * @see it.unimi.dsi.mg4j.search.score.ClarkeCormackScorer
	 */
	public final static double clarkeCormackScore( IntervalIterator it, double alpha, int h ) {
		double result = 0;
		int lt;
		while ( it.hasNext() ) {
			lt = it.nextInterval().length();
			if ( lt < h ) result += 1;
			else result += Math.pow( h / (double) lt, alpha );
		}
		return result;
	}

	/** Computes the {@linkplain #clarkeCormackScore(IntervalIterator, double, int) Clarke-Cormack score}
	 *  using the default values for parameter (<code>h</code>=16 and <code>alpha</code>=1).
	 * 
	 * @param it the iterator; note that the iterator gets exhausted by this call.
	 * @return the score.
	 */
	public final static double clarkeCormackScore( IntervalIterator it ) {
		return clarkeCormackScore( it, 1, 16 );
	}

                                               
}
                                                                                                                                                        
// Local Variables:
// mode: jde
// tab-width: 4
// End:
