package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceOpenHashMap;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.util.MutableString;

import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;




/** A document iterator that filters another document iterator, returning just intervals (and containing
 * documents) whose length does not exceed a given threshold.
 * 
 * @author Paolo Boldi
 * @author Sebastiano Vigna
 * @since 0.9
 */

public class LowPassDocumentIterator extends AbstractIntIterator implements DocumentIterator {

	/** The underlying iterator. */
	final private DocumentIterator it;
	/** The iterator threshold. */
	final protected int threshold; 
	/** A map from indices to interval iterators. */
	final private Reference2ReferenceOpenHashMap intervalIterators;
	/** A map from indices to the iterators returned for the current document. The key set may
	 * not contain an index because the related iterator has never been requested. Moreover,
	 * the iterator in this map for a given index may differ from the one in {@link #intervalIterators}
	 * because it could be {@link IntervalIterators#TRUE} (in fact, in that case it may even
	 * happen that {@link #intervalIterators} does not contain the index). */
	final private Reference2ReferenceOpenHashMap currentIterators;
	/** An unmodifiable wrapper around {@link #currentIterators}. */
	final private Map unmodifiableCurrentIterators;

	/** The next document that will be returned by this iterator, or -1 if it is necessary to advance
	 *  before (possibly) returning a document. */
	private int next;
	/** The last document returned by this iterator, or -1 if no document has been returned yet. */
	private int last = -1;

	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = true;

	static { if ( DEBUG && ASSERTS ); }

	/** Creates a new low-pass document iterator over a given iterator.
	 * @param it the iterator to be filtered.
	 * @param threshold the filter threshold.
	 */
	public LowPassDocumentIterator( final DocumentIterator it, final int threshold ) {
		this.it = it;
		this.threshold = threshold;

		intervalIterators = new Reference2ReferenceOpenHashMap( indices().size(), .5f );
		currentIterators = new Reference2ReferenceOpenHashMap( indices().size(), .5f );
		unmodifiableCurrentIterators = Reference2ReferenceMaps.unmodifiable( currentIterators );

		next = -1;
	}

	public Set indices() { return it.indices(); }

	public int skipTo( int n ) {
		boolean isValid;
		if ( last >= n ) return last;
		next = it.skipTo( n );
		currentIterators.clear();
		if ( next < 0 ) return -1;
		while( ! ( isValid = isValid() ) && it.hasNext() ) {
			next = it.nextInt();
			currentIterators.clear();
		} 
		if ( ! isValid ) return next = -1;
		last = next;
		next = -1;
		return last;
	}

	public int nextInt() {
		if ( DEBUG ) System.err.println( this + ".nextInt()" ); 

		if ( ! hasNext() ) throw new NoSuchElementException();
		last = next;
		next = -1;
		return last;
	}
	
	public int document() { 
		if ( last == -1 ) throw new IllegalStateException(); 
		return last; 
	}

	public int nextDocument() { return nextInt(); }

	public boolean hasNext() {
		if ( DEBUG ) System.err.println( this + ".hasNext()" );
		if ( next >= 0 ) return true;
		if ( ! it.hasNext() ) return false;

		boolean isValid;
		do {
			next = it.nextInt(); 
			currentIterators.clear();
		} while( ! ( isValid = isValid() ) && it.hasNext() );
				
		if ( ! isValid ) next = -1;
		return isValid;
	}

	private boolean isValid() {
		boolean isValid = false;

		/* The policy here is that a low-pass is valid is at least one of the underlying
		 * interval iterators, once filtered, would return at least one interval (of course,
		 * this statement does not take into account TRUE iterators). */

		for( Iterator i = indices().iterator(); i.hasNext(); ) {
			IntervalIterator intervalIterator = intervalIterator( (Index)i.next() );
			// TRUE iterators play no role.
			if ( intervalIterator == IntervalIterators.TRUE ) continue;
			if ( intervalIterator.hasNext() ) isValid = true;
		}
		return isValid;
	}

	public Map intervalIterators() {
		final Iterator i = indices().iterator();
		while( i.hasNext() ) intervalIterator( (Index) i.next() );
		return unmodifiableCurrentIterators;
	}

	public IntervalIterator intervalIterator() {
		if ( indices().size() != 1 ) throw new IllegalStateException();
		return intervalIterator( (Index)indices().iterator().next() );
	}

	public IntervalIterator intervalIterator( final Index index ) {
		if ( DEBUG ) System.err.println( this + ".intervalIterator(" + index + ")" );

		IntervalIterator intervalIterator;

		// If the iterator has been created and it's ready, we just return it.		
		if ( ( intervalIterator = (IntervalIterator) currentIterators.get( index ) ) != null ) return intervalIterator;

		intervalIterator = it.intervalIterator( index );
			
		/* If the underlying iterator is TRUE or FALSE, then our constribution to the result is not relevant,
		 * and we just pass this information upwards. E.g., consider the query (A OR title:B)~2 with
		 * a document containing A but not B in its title. When evaluating the query for the title index,
		 * the subquery before the low-pass operator evalutes to TRUE, meaning that its truth is independent
		 * of the title field. This fact is not changed by the low-pass operator. */
		if ( intervalIterator != IntervalIterators.FALSE && intervalIterator != IntervalIterators.TRUE ) {
			intervalIterator = (IntervalIterator) intervalIterators.get( index );
			if ( intervalIterator == null ) intervalIterators.put( index, intervalIterator = new LowPassIntervalIterator( index ) ); 
		}
		
		currentIterators.put( index, intervalIterator );	
		intervalIterator.reset();
		return intervalIterator;
	}
	
	public String toString() {
	   MutableString result = new MutableString();
	   return result.append( this.getClass().getName() ).append( "(" ).delete( 0, result.lastIndexOf( '.' ) + 1 ).append( it ).append( ", " ).append( threshold ).append( ")" ).toString();
	}
	

	private class LowPassIntervalIterator extends AbstractObjectIterator implements IntervalIterator {
		/** The index of this iterator. */
		final Index index;
		/** The underlying interal iterator. */
		private IntervalIterator it;

		/** Whether the next interval is already loaded in {@link #next}. */
		private boolean nextIsLoaded;
		/** The next interval that will be returned. */
		private Interval next;
		
		public LowPassIntervalIterator( final Index index ) {
			this.index = index;
		}

		public void reset( ) {
			nextIsLoaded = false;
			it = LowPassDocumentIterator.this.it.intervalIterator( index );
		}

		public boolean hasNext() {
			if ( nextIsLoaded ) return true;
		
			while( it.hasNext() ) {
				next = it.nextInterval();
				if ( next.length() <= threshold ) {
					nextIsLoaded = true;
					break;
				}
			}
			if ( ! nextIsLoaded ) next = null;
			return nextIsLoaded;
		}
		
		public Interval nextInterval() {
			if ( ! hasNext() ) throw new NoSuchElementException();
			nextIsLoaded = false;
			return next;
		}
		
		public int extent() {
			return Math.min( it.extent(), threshold );
		}
		
		public Object next() {
			return nextInterval();
		}

		public String toString() {
		   MutableString res = new MutableString();
		   return res.append( this.getClass().getName() ).append( "(" ).delete( 0, res.lastIndexOf( '.' ) + 1 ).append( it ).append( ", " ).append( threshold ).append( ")" ).toString();
		}
	}


}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
