package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMap;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceOpenHashMap;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.util.MutableString;

import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;


/** A document iterator that returns documents <em>not</em> returned by its underlying iterator,
 * and returns just {@link it.unimi.dsi.mg4j.search.IntervalIterators#TRUE} on all interval iterators.
 * 
 * @author Paolo Boldi
 * @author Sebastiano Vigna
 * @since 0.9
 */

public class NotDocumentIterator extends AbstractIntIterator implements DocumentIterator {

	/** The underlying iterator. */
	final private DocumentIterator it;
	/** The number of documents. */
	final private int numberOfDocuments;
	/** An unmodifiable map mapping all indices in {@link #indices()} to {@link IntervalIterators#TRUE}. */
	final private Map unmodifiableIntervalIterators;

	/** The next document that will be returned by this iterator. */
	private int next;
	/** The last document returned by this iterator. */
	private int last = -1;
	/** The next document that must <em>not</em> be returned, or {@link #numberOfDocuments}
	 * if the underlying iterator is exhausted. {@link #next} is always less than or equal to
	 * this field. */
	private int nextToSkip;

	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = false;

	static { if ( DEBUG && ASSERTS ); }

	/** Creates a new low-pass document iterator over a given iterator.
	 * @param it the iterator to be filtered.
	 * @param numberOfDocuments the number of documents.
	 */
	public NotDocumentIterator( DocumentIterator it, int numberOfDocuments ) {
		this.it = it;
		this.numberOfDocuments = numberOfDocuments;
		this.next = 0;
		this.nextToSkip = it.hasNext() ? it.nextInt() : numberOfDocuments;
		final Reference2ReferenceMap iterators = new Reference2ReferenceOpenHashMap( indices().size() );
		final Iterator i = indices().iterator();
		while( i.hasNext() ) iterators.put( i.next(), IntervalIterators.TRUE );
		unmodifiableIntervalIterators = Reference2ReferenceMaps.unmodifiable( iterators );	
	}

	public Set indices() { return it.indices(); }

	public int skipTo( int n ) {
		if ( last >= n ) return last;
		nextToSkip = it.skipTo( n );
		if ( nextToSkip == -1 ) nextToSkip = numberOfDocuments;
		next = n;
		if ( hasNext() ) return nextInt();
		else return -1;
	}

	public int nextInt() {
		if ( DEBUG ) System.err.println( this + ".nextInt()" ); 

		if ( ! hasNext() ) throw new NoSuchElementException();
		last = next;
		next++;
		return last;
	}
	
	public int document() { 
		if ( last == -1 ) throw new IllegalStateException(); 
		return last; 
	}

	public int nextDocument() { return nextInt(); }

	public boolean hasNext() {
		if ( DEBUG ) System.err.println( this + ".hasNext()" );

		for(;;) {
			if ( next >= numberOfDocuments ) return false;
			if ( next < nextToSkip ) return true;
			
			next++;
			nextToSkip = it.hasNext() ? it.nextInt() : numberOfDocuments;
		}
	}

	public String toString() {
	   MutableString result = new MutableString();
	   return result.append( this.getClass().getName() ).append( "(" ).delete( 0, result.lastIndexOf( '.' ) + 1 ).append( it ).append( ")" ).toString();
	}

	public Map intervalIterators() {
		return unmodifiableIntervalIterators;
	}

	public IntervalIterator intervalIterator() {
		if ( indices().size() != 1 ) throw new IllegalStateException();
		return IntervalIterators.TRUE;
	}

	public IntervalIterator intervalIterator( final Index index ) {
		return IntervalIterators.TRUE;
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
