package it.unimi.dsi.mg4j.search;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Paolo Boldi and Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntHeapSesquiIndirectDoublePriorityQueue;
import it.unimi.dsi.fastutil.objects.ObjectArrays;
import it.unimi.dsi.fastutil.objects.ObjectHeapSesquiIndirectDoublePriorityQueue;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceOpenHashMap;
import it.unimi.dsi.fastutil.objects.ReferenceOpenHashSet;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.util.MutableString;

import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

/**  A document iterator that ORs given component iterators.
 */

public class OrDocumentIterator extends AbstractIntIterator implements DocumentIterator {
	/** The component iterators. */
	final protected DocumentIterator it[];
	/** The number of component iterators. */
	final private int n;
	/** A heap-based sesqui-indirect double priority queue used to keep track of the currently scanned integers. */
	final private IntHeapSesquiIndirectDoublePriorityQueue queue;
	/** The reference array used for the queue. */
	final private int curr[];
	/** The set of indices involved in this iterator. */
	final private ReferenceOpenHashSet indices = new ReferenceOpenHashSet();
	/** A map from indices to interval iterators. */
	final private Reference2ReferenceOpenHashMap intervalIterators;
	/** A map from indices to the iterators returned for the current document. The key set may
	 * not contain an index because the related iterator has never been requested. Moreover,
	 * the iterator in this map for a given index may differ from the one in {@link #intervalIterators}
	 * because it could be {@link IntervalIterators#TRUE} (in fact, in that case it may even
	 * happen that {@link #intervalIterators} does not contain the index). */
	final private Reference2ReferenceOpenHashMap currentIterators;
	/** An unmodifiable wrapper around {@link #currentIterators}. */
	final private Map unmodifiableCurrentIterators;

	/** The next integer that will be returned by this iterator, or -1 if it is necessary to advance
	 *  before (possibly) returning a document. */
	private int next;
	/** The last integer returned by this iterator. */
	private int last = -1;
	/** Whether the scan is over. */
	private boolean endOfProcess = false;

	private final static boolean DEBUG = false;
	private final static boolean ASSERTS = true;

	static { if ( DEBUG && ASSERTS ); }

	/** Creates a new union iterator using a given array of iterators.
	 *  @param it the iterators to be joined.
	 */
	public OrDocumentIterator( DocumentIterator it[] ) {
		this.it = it;
		this.n = it.length;
		this.curr = new int[ n ];
		// The primary/secondary comparator gives the minimum/maximum among the currently scanned integers. 
		queue = new IntHeapSesquiIndirectDoublePriorityQueue( curr );
		
		/* Now, for each index involved we build a corresponding interval iterator.
		 * Note that the set indices() may contain indices from empty subqueries, too. */
		for( int i = 0; i < n; i++ ) indices.addAll( it[ i ].indices() );		

		intervalIterators = new Reference2ReferenceOpenHashMap( indices.size(), .5f );
		currentIterators = new Reference2ReferenceOpenHashMap( indices.size(), .5f );
		unmodifiableCurrentIterators = Reference2ReferenceMaps.unmodifiable( currentIterators );

		next = -1;

		// Only add to the queue nonempty iterators...
		for ( int i = 0; i < n; i++ ) 
			if ( it[i].hasNext() ) {
				curr[i] = it[i].nextInt();
				queue.enqueue( i );
			}
		// If queue is empty, the process is over
		if ( queue.size() == 0 ) endOfProcess = true;

		next = endOfProcess? -1 : curr[ queue.first() ];
	}

	public Set indices() { return indices; }


	/** Advances the iterator. 
	 * 
	 * <P>More precisely, if <code>n</code>&ge;0 it advances all the lists that correspond to
	 * the currently least integer until an element larger than or equal
	 * to <code>n</code> would be returned (it might set {@link #endOfProcess} to true).
	 * Otherwise, it moves to the next element (which, however, could be a false positive).
	 *
	 * @param n a lower bound on the document pointer, or -1.
	 */
	private void advance( int n ) {
		if ( DEBUG ) System.err.println( this + ".advance(" + n + ")" ); 

		int first, res;
		// The least element
		int c = curr[ queue.first() ];

		// Advance all elements equal to the least one
		while( ! endOfProcess && curr[ first = queue.first() ] == c ) {
			//firstItemHasIntervals = false;
			if ( n >= 0 ) res = it[ first ].skipTo( n );
			else res = it[ first ].hasNext() ? it[ first ].nextInt() : -1;
			
			// Cannot advance the minimum
			if ( res < 0 ) {
				// Remove it and null its entry
				it [ queue.dequeue() ] = null;
				// If nothing else remains, we are done
				if ( queue.isEmpty() ) endOfProcess = true;
			}
			else {
				// Advance the top element, and signal this fact to the queue
				curr[ first ] = res;
				queue.changed();
			}
		}
		
		currentIterators.clear(); 
	}


	public int skipTo( int n ) {
		if ( last >= n ) return last;
		advance( n );
		if ( endOfProcess ) return -1;
		next = curr[ queue.first() ];
		return nextInt();
	}

	public int nextInt() {
		if ( DEBUG ) System.err.println( this + ".nextInt()" ); 

		if ( ! hasNext() ) throw new NoSuchElementException();

		last = next;
		next = -1;

		if ( DEBUG ) System.err.println( this + "...returning " + last ); 
		return last;
	}

	public int document() {
		if ( last == -1 ) throw new IllegalStateException(); 
		return last; 
	}

	public int nextDocument() { return nextInt(); }


	public boolean hasNext() {
		if ( DEBUG ) System.err.println( this + ".hasNext()" );

		// We already know what to return
		if ( next >= 0 ) return true;
		// If we're over, we stop here
		if ( endOfProcess ) return false;

		advance( -1 );

		// We check again whether we're over
		if ( endOfProcess ) return false;

		next = curr[ queue.first() ];
		return true;
	}

	public Map intervalIterators() {
		final Iterator i = indices.iterator();
		while( i.hasNext() ) intervalIterator( (Index) i.next() );
		return unmodifiableCurrentIterators;
	}

	public IntervalIterator intervalIterator() {
		if ( indices().size() != 1 ) throw new IllegalStateException();
		return intervalIterator( (Index)indices().iterator().next() );
	}

	public IntervalIterator intervalIterator( final Index index ) {
		if ( DEBUG ) System.err.println( this + ".intervalIterator(" + index + ")" );

		IntervalIterator intervalIterator;

		// If the iterator has been created and it's ready, we just return it.		
		if ( ( intervalIterator = (IntervalIterator) currentIterators.get( index ) ) != null ) return intervalIterator;

		int c, i;
		boolean atLeastOneIsTrue = false;
	
		/* If all iterators are TRUE, we return TRUE. Otherwise,
		 * we skip all FALSE iterators and all iterators whose document
		 * is not last, and merge the remaining ones. If
		 * all iterators are FALSE, we return FALSE.
		 */

		for( i = c = 0; i < n; i++ ) {
			if ( it[ i ] == null || it[ i ].document() != last ) continue;
			intervalIterator = it[ i ].intervalIterator( index );
			if ( intervalIterator == IntervalIterators.TRUE ) atLeastOneIsTrue = true;
			else if ( intervalIterator != IntervalIterators.FALSE ) c++;
		}
			
		if ( c == 0 ) intervalIterator = atLeastOneIsTrue? IntervalIterators.TRUE : IntervalIterators.FALSE;
		else {
			intervalIterator = (IntervalIterator) intervalIterators.get( index );
			if ( intervalIterator == null ) intervalIterators.put( index, intervalIterator = new OrIntervalIterator( index ) );
		}
				
		currentIterators.put( index, intervalIterator );	
		intervalIterator.reset();
		return intervalIterator;
	}


	public String toString() {
	   MutableString res = new MutableString();
	   res.append( this.getClass().getName() ).append( "(" ).delete( 0, res.lastIndexOf( '.' ) + 1 );
	   for ( int i = 0; i < n; i++ ) res.append( i > 0 ? "," : "" ).append( it[ i ] );
	   return res.append( ")" ).toString();
	}

	/** An interval iterator of nondecreasing disjoint intervals 
	 *  obtained ORing the output of a set of iterators with
	 *  the same property.
	 */

	private class OrIntervalIterator extends AbstractIntIterator implements IntervalIterator {
		/** The index of this iterator. */
		final Index index;
		/** The component interval iterators. */
		final private IntervalIterator it[];
		/** The number of component iterators. */
		final private int n;
		/** The reference array used for the queue. */	 
		final private Interval curr[];
		/** A heap-based sesqui-indirect double priority queue used to keep
		 *  track of the currently scanned intervals. */
		private ObjectHeapSesquiIndirectDoublePriorityQueue queue;

		/** Whether the scan is over. */
		private boolean endOfProcess = false;
		/** Whether the next interval is already loaded in {@link #next}. */
		private boolean nextIsLoaded;
		/** The left extreme of the next interval that will be returned. */
		private int nextLeft;
		/** The right extreme of the next interval that will be returned. */
		private int nextRight;

		private final static boolean DEBUG = false;
		private final static boolean ASSERTS = false;
		{ if ( DEBUG && ASSERTS ); }

		/** Creates a new OR interval iterator.
		 * 
		 * @param index the index of the iterator.
		 */
		public OrIntervalIterator( final Index index ) {
			// We just set up some internal data, but we perform no initialisation.
			this.index = index;
			n = OrDocumentIterator.this.it.length;
			curr = new Interval[n];
			it = new IntervalIterator[ n ];
			queue = new ObjectHeapSesquiIndirectDoublePriorityQueue( curr, Intervals.LEFT_INTERVAL_COMPARATOR, Intervals.RIGHT_INTERVAL_COMPARATOR );		
		}

		public void reset() {
			nextIsLoaded = false;
			
			ObjectArrays.fill( curr, null );
			queue.clear();

			DocumentIterator documentIterator;

			for ( int i = 0; i < n; i++ ) {
				documentIterator = OrDocumentIterator.this.it[ i ];
				
				if ( documentIterator == null || documentIterator.document() != last ) continue;
				it[ i ] = documentIterator.intervalIterator( index );
				if ( it[ i ] != IntervalIterators.TRUE && it[ i ].hasNext() ) {
					curr[ i ] = it[ i ].nextInterval();
					queue.enqueue( i );
				}
			}
			
			endOfProcess = queue.isEmpty();
		}

		public boolean hasNext() {
			if ( nextIsLoaded ) return true;
			if ( endOfProcess ) return false;
			nextIsLoaded = true;

			int resultIndex;

			do {
				resultIndex = queue.first();
				nextLeft = curr[ resultIndex ].left;
				nextRight = curr[ resultIndex ].right;
				if ( it[ resultIndex ].hasNext() ) {
					curr[ resultIndex ] = it[ resultIndex ].nextInterval();
					queue.changed();
				}
				else if ( !queue.isEmpty() )  queue.dequeue();
					  
				if ( queue.isEmpty() ) endOfProcess = true;
				if ( ASSERTS ) assert !queue.isEmpty() || endOfProcess;
			} while ( !endOfProcess && ( curr[ queue.first() ].left == nextLeft || curr[ queue.secondaryFirst() ].right <= nextRight ) );

			return true;
		}

		public Interval nextInterval() {
			if ( ! hasNext() ) throw new NoSuchElementException();
			nextIsLoaded = false;
			return Interval.getInstance( nextLeft, nextRight );
		}

		public int extent() {
			int i, e, s;
			
			s = Integer.MAX_VALUE;
			for ( i = 0; i < n; i++ ) {
				e = it[ i ] != null && it[ i ] != IntervalIterators.TRUE ? it[ i ].extent() : Integer.MAX_VALUE;
				if ( e < s ) s = e;
			}
			return s;
		}
		
		public Object next() {
			return nextInterval();
		}
 
		public String toString() {
		   MutableString res = new MutableString();
		   res.append( this.getClass().getName() ).append( "(" ).delete( 0, res.lastIndexOf( '.' ) + 1 );
		   for ( int i = 0; i < n; i++ ) res.append( i > 0 ? "," : "" ).append( it[ i ] );
		   return res.append( ")" ).toString();
		}
	}

																																							
//	   Local Variables:
//	   mode: jde
//	   tab-width: 4
//	   End:

}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
