package it.unimi.dsi.mg4j.search.score;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Paolo Boldi
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITfNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.mg4j.search.DocumentIterator;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.search.IntervalIterator;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * An abstract implementation of {@link it.unimi.dsi.mg4j.search.score.Scorer}
 * that scores a document iterator in such a way that one can make calls to
 * the score methods and to the intervalIterator methods in any order and
 * for any number of times. Implementing classes must provide: 
 * <ul>
 *    <li>a {@link #score(IntervalIterator)} method that tells how an interval
 *     iterator should be scored;
 *    <li>a {@link #score(Map)} method that tells how a document should be
 *     scored based on a map that sends {@link it.unimi.dsi.mg4j.index.Index} to
 *     {@link it.unimi.dsi.mg4j.search.IntervalIterator}.
 * </ul>
 * Instead of implementing the latter, one may implement a method {@link #weight(Index)}
 * that assigns a weight to each index: in that case, the overall score will be computed
 * as weighted sum.
 */
public abstract class AbstractScorer implements Scorer {
	
	abstract protected double score( IntervalIterator it );

	protected double weight( Index index ) { if ( false && index != null ); return 1; }

	protected double score( Map m ) {
		Iterator it = m.entrySet().iterator();
		double result = 0;
		Index index;
		IntervalIterator intervalIterator;
		Map.Entry entry;
		while ( it.hasNext() ) {
			entry = (Map.Entry) it.next();
			index = (Index) entry.getKey();
			intervalIterator = (IntervalIterator) entry.getValue();
			result += weight( index ) * score( intervalIterator );
		}
		return result;
	}


	public ScoredDocumentIterator scoredDocumentIterator( final DocumentIterator d ) {
		
		return new ScoredDocumentIterator() {

			public double score(Index index) {
				return AbstractScorer.this.score( d.intervalIterator( index ) );				
			}

			public double score() {
				return AbstractScorer.this.score( intervalIterators() );
			}

			public IntervalIterator intervalIterator() { return d.intervalIterator(); }
			public Map intervalIterators() { return d.intervalIterators(); }
			public IntervalIterator intervalIterator( Index index ) { return d.intervalIterator( index ); }
			public Set indices() { return d.indices(); }
			public int nextDocument() { return d.nextDocument(); }
			public int document() { return d.document(); }
			public int skipTo( int n ) { return d.skipTo( n ); }
			public int nextInt() { return d.nextInt(); }
			public int skip( int k ) { return d.skip( k ); }
			public boolean hasNext() { return d.hasNext(); }
			public Object next() { return d.next(); }
			public void remove() { d.remove(); }
		};
		
		
	}

}
