package it.unimi.dsi.mg4j.search.score;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Paolo Boldi
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITfNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.mg4j.search.IntervalIterator;
import it.unimi.dsi.mg4j.search.IntervalIterators;

/** Computes the Clarke-Cormack score of all interval iterators of a document.
 *  This score function is defined in <i>Shortest-Substring
 *  Retrieval and Ranking</i> (ACM Transactions on Information Systems, vol. 18, no. 1, Jan 2000,
 *  pages 44-78), at page 65; the weight assigned to different indices is the same for all indices
 *  (but it may be rewritten by subclasses, rewriting the {@link #weight(Index)} method).
 * 
 *  <p>The score depends on two parameters: an integer <code>h</code> and a double <code>alpha</code>.
 *  The score is obtained summing up a certain score assigned to all intervals in the interval iterator
 *  produced on that document for the given query. The score assigned to an interval is 1 if the interval
 *  has length smaller than <code>h</code>; otherwise, it is obtained by dividing <code>h</code> by
 *  the interval length, and raising the result to the power of <code>alpha</code>.
 * 
 *  <p>Note that the score assigned to each interval is between 0 and 1 (highest scores corresponding
 *  to best intervals). The score assigned to an interval iterator is thus bounded from above by the
 *  number of intervals.
 * 
 *  <p>Typically, one sets <code>h</code>=16 (or a bit larger) and <code>alpha</code>=1 (or a bit smaller),
 *  but the authors say that the method is rather stable w.r.t. changes in the values of parameters.  
 *  
 */
public class ClarkeCormackScorer extends AbstractScorer {
	
	/** The parameter h. */
	private int h;
	/** The parameter alpha. */
	private double alpha;

	/** Creates a scorer with given parameter.
	 * 
	 * @param h the parameter h.
	 * @param alpha the parameter alpha.
	 */
	public ClarkeCormackScorer( int h, double alpha ) {
		this.h = h;
		this.alpha = alpha;
	}
	
	/** Default constructor, assigning the default values (<code>h</code>=16, <code>alpha</code>=1) to the
	 *  parameter.
	 */
	public ClarkeCormackScorer() {
		h = 16;
		alpha = 1;
	}
	
	/** Computes the Clarke-Cormack score of an interval iterator.  
	 * 
	 * @param it the iterator.
	 * @return the score.
	 */
	protected double score( IntervalIterator it ) {
		if ( it == IntervalIterators.TRUE || it == IntervalIterators.FALSE ) return 0;
		double result = 0;
		int lt;
		while ( it.hasNext() ) {
			lt = it.nextInterval().length();
			if ( lt < h ) result += 1;
			else result += Math.pow( h / (double) lt, alpha );
		}
		return result;
	}
	
}
