package it.unimi.dsi.mg4j.search.score;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Paolo Boldi
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITfNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.mg4j.search.DocumentIterator;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.io.FastBufferedInputStream;
import it.unimi.dsi.mg4j.search.IntervalIterator;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Map;
import java.util.Set;

/** Compute scores that do not depend on intervals, but that
 *  just assign a fixed score to each document; scores are read
 *  from a file whose name is passed to the constructor.
 */
public class DocumentRankScorer {
	
	/** The array of scores. */
	private double score[];
	
	/** Builds a document scorer that assigns the score 0 to all documents. */
	public DocumentRankScorer() {}
	
	/** Builds a document scorer by first reading the ranks from a file.
	 *  Ranks are saved as doubles (the first double is the rank of document 0
	 *  and so on).
	 * 
	 *  @param filename the name of the rank file.
	 *  @param pm the progress meter to be used while loading ranks, or <code>null</code>.
	 */
	public DocumentRankScorer( String filename, ProgressMeter pm ) throws IOException {
		final FileInputStream fis = new FileInputStream( filename );
		final DataInputStream dos = new DataInputStream( new FastBufferedInputStream( fis , 64 * 1024 ) );

		if ( pm != null ) {
			pm.start();
			pm.itemsName( "doubles" );
		} 

		final int length = (int)(fis.getChannel().size() / 8);
		score = new double[ length ];
		for( int i = 0; i < length; i++ ) score[ i ] = dos.readDouble();

		dos.close();
		if ( pm != null ) {
			pm.count( length );
			pm.stop();
		}
	}

	public ScoredDocumentIterator scoredDocumentIterator( final DocumentIterator d ) {
		
		return new ScoredDocumentIterator() {

			public double score(Index index) {
				if ( score == null ) return 0;
				return score[ d.document() ];				
			}

			public double score() {
				if ( score == null ) return 0;
				return score[ d.document() ];				
			}

			public Map intervalIterators() { return d.intervalIterators(); }
			public IntervalIterator intervalIterator() { return d.intervalIterator(); }
			public IntervalIterator intervalIterator( Index index ) { return d.intervalIterator( index ); }
			public Set indices() { return d.indices(); }
			public int nextDocument() { return d.nextDocument(); }
			public int document() { return d.document(); }
			public int skipTo( int n ) { return d.skipTo( n ); }
			public int nextInt() { return d.nextInt(); }
			public int skip( int k ) { return d.skip( k ); }
			public boolean hasNext() { return d.hasNext(); }
			public Object next() { return d.next(); }
			public void remove() { d.remove(); }
		};
		
		
	}

}
