package it.unimi.dsi.mg4j.search.score;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Paolo Boldi
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITfNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

import it.unimi.dsi.fastutil.doubles.DoubleArrayList;
import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.search.IntervalIterator;

/** Merges a number of {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}s,
 *  obtaining a new {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}.
 *  This is done in the following way: you provide an {@link LinearAggregator}
 *  with a number of {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}s,
 *  each with an associated weight and a boolean (called <em>relevance
 *  flag</em>).
 * 
 *  <P>A document is said to be <em>relevant</em> iff it is returned
 *  by at least one of the {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}s
 *  that have the relevance flag set. For every relevant document, its
 *  aggregate score is obtained by combining linearly (using the
 *  given weights) the scores returned by the {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}s
 *  for that index. The scores returned by each {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}
 *  are firstly <em>normalized</em> (i.e., they are divided by the greatest score they ever return, unless
 *  they are constantly zero).  
 * 
 *  <P>Note that the aggregate iterator returned by the aggregation method will only
 *  return documents and their scores (no interval iterators!).
 */
public class LinearAggregator {
	
	private final static boolean DEBUG = true;
	
	/** The number of underlying {@link it.unimi.dsi.mg4j.search.score.ScoredDocumentIterator}. */
	private int n;
	/** The relevance flag. */
	private boolean relevant[];
	/** The weight. */
	private double weight[];
	
	/** Creates a linear aggregator.
	 * 
	 * @param relevant the array of relevance flags.
	 * @param weight the array of weights.
	 */
	public LinearAggregator( boolean relevant[], double weight[] ) {
		if ( relevant.length != weight.length ) throw new IllegalArgumentException();
		this.n = relevant.length;
		this.relevant = relevant;
		this.weight = weight; 
	}
	
	/** Aggregates the given {@link ScoredDocumentIterator}s, that must be in the same
	 *  number as there are weights (and relevance flags).
	 * 
	 * @param underlying the underlying iterators to be merged.
	 * @return the aggregate iterator.
	 */
	public ScoredDocumentIterator aggregate( ScoredDocumentIterator underlying[] ) {
		int i;
		
		if ( underlying.length != n ) throw new IllegalArgumentException();
		
		// documents[ i ] = list of documents returned by the i-th iterator
		final IntArrayList documents[] = new IntArrayList[ n ];
		// scores[ i ] = list of scores returned by the i-th iterator
		final DoubleArrayList scores[] = new DoubleArrayList[ n ];
		
		// Build lists
		for ( i = 0; i < n; i++ ) {
			documents[ i ] = new IntArrayList();
			scores[ i ] = new DoubleArrayList();
		}
		
		boolean over;
		do {
			over = true;
			for ( i = 0; i < n; i++ )
				if ( underlying[ i ].hasNext() ) {
					documents[ i ].add( underlying[ i ].nextDocument() );
					scores[ i ].add( underlying[ i ].score() );
					over = false;
				}
		} while ( !over );
		
		if ( DEBUG ) for ( int j = 0; j < n; j++ ) System.err.println( "Documents returned by iterator " + j + ": " + documents[ j ] );
		
		class MergedIterator extends AbstractIntIterator implements ScoredDocumentIterator {
			private int nextDocument = -1; // The next document to be returned by nextDocument
			private int lastDocument = -1; // The last document returned 
			private double nextScore = 0; // The aggregate score of the next document to be returned
			private double lastScore = 0; // The aggregate score of the last document 

			// cursor[ i ] = points to the next element of documents[ i ] that should be returned
			private int cursor[] = new int[ n ];

			public boolean hasNext() {
				if ( nextDocument >= 0 ) return true;
				if ( nextDocument == Integer.MAX_VALUE ) return false;
				
				int minDoc;
				double aggregateScore = 0;
				boolean isMinRelevant = false;
				
				do {
						minDoc = Integer.MAX_VALUE;
						for ( int i = 0; i < n; i++ )
							if ( cursor[ i ] < documents[ i ].size() ) { 
								isMinRelevant |= relevant[ i ];
								int candidateMin = documents[ i ].getInt( cursor[ i ] );
								if ( candidateMin < minDoc ) {
									minDoc = candidateMin;
									aggregateScore = weight[ i ] * scores[ i ].getDouble( cursor[ i ] );
								} else if ( candidateMin == minDoc ) 
									aggregateScore += weight[ i ] * scores[ i ].getDouble( cursor[ i ] );
							}
						for ( int i = 0; i < n; i++ )
							if ( cursor[ i ] < documents[ i ].size() && documents[ i ].getInt( cursor[ i ] ) == minDoc ) cursor[ i ]++;
				} while ( !isMinRelevant && minDoc < Integer.MAX_VALUE );
				
				nextDocument = minDoc;
				nextScore = aggregateScore;
				return ( nextDocument < Integer.MAX_VALUE );
			}

			public int nextDocument() {
				if ( !hasNext() ) throw new NoSuchElementException();
				lastDocument = nextDocument;
				lastScore = nextScore;
				nextDocument = -1;
				return lastDocument;
			}

			public int document() {
				if ( lastDocument < 0 ) throw new IllegalStateException();
				return lastDocument;
			}

			public double score() {
				if ( lastDocument < 0 ) throw new IllegalStateException();
				return lastScore;
			}
			
			public int skipTo( int n ) {
				int k = 0;
				
				if ( lastDocument >= n ) return lastDocument;
				while( hasNext() && ( k = nextInt() ) < n );
				return hasNext() ? k : -1;				
			}

			public int nextInt() { return nextDocument(); }

			public double score(Index index) { throw new UnsupportedOperationException(); }
			public Map intervalIterators() { throw new UnsupportedOperationException(); }
			public IntervalIterator intervalIterator() { throw new UnsupportedOperationException(); }
			public IntervalIterator intervalIterator(Index index) { throw new UnsupportedOperationException(); }
			public Set indices() { throw new UnsupportedOperationException(); }
			
		};
		
		return new MergedIterator();
		
		
	}
	

}
