package it.unimi.dsi.mg4j.tool;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import it.unimi.dsi.mg4j.index.IndexProperties;
import it.unimi.dsi.mg4j.io.FastBufferedInputStream;
import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MinimalPerfectHash;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Properties;


/** Builds batched occurrence files from a list of documents read from standard input.
 *
 * <P>This class reads from standard input a sequence of documents and produces
 * corresponding compressed occurrence files, containing a settable number of
 * occurrences, which must be merged using {@link SecondPass}. Documents are
 * separated by a special Unicode character (default is 10, i.e., newline)
 * that can be set with an option. Every document is considered as a sequence
 * of words: a word is a maximal sequence of letters and digits (in the Java sense).
 *
 * <P>The only mandatory argument is a <em>basename</em>, which will be used to
 * stem the names of all files generated.
 *
 * <P>Since documents are read sequentially, every document has a <em>natural
 * index</em> starting from 0. If no permutation is specified, the <em>document
 * index</em> of each document corresponds to its natural index.  If, however,
 * a permutation is specified, the document index of a document is the image
 * through the permutation of its natural index. More precisely, a permutation
 * for <var>N</var> documents is a list of <var>N</var> distinct integers
 * between 0 and <var>N</var>-1 inclusive, and a document with natural index
 * <var>i</var> has document index given by the <var>i</var>-th element of the
 * list. This is useful when indexing ranked documents (e.g., if you are
 * indexing a part of the web and would like the index to return documents
 * with higher rank first). If the permutation file is provided, it must be
 * a sequence of integers, written using the {@link java.io.DataOutputStream#writeInt(int)} method;
 * if <var>N</var> is the number of documents, the file is to contain exactly
 * <var>N</var> distinct integers between 0 and <var>N</var>-1.
 * 
 * <P>Also every term has an index starting from 0. Increasing indices,
 * starting from 0, are assigned as new terms are met. If you prefer having
 * lexicographically ordered terms, you should call {@link MiddlePass} after
 * {@link FirstPass}.
 *
 * <P>These are the files currently generated:
 *
 * <dl>
 *
 * <dt><samp><var>basename</var>.terms</samp>
 *
 * <dd>For each indexed term, the corresponding literal string in UTF-8 encoding. More precisely,
 * the <var>i</var>-th line of the file (starting from 0) contains the literal
 * string corresponding to term index <var>i</var>.
 *
 * <dt><samp><var>basename</var>.frequencies</samp>
 *
 * <dd>This file contains, for each term, the number of documents in which the term appears in &gamma; coding.
 * More precisely, <var>i</var>-th integer of the file (starting from 0) is the
 * number of documents in which the term of index <var>i</var> appears.
 *
 * <dt><samp><var>basename</var>.sizes</samp>
 *
 * <dd>This file contains, for each indexed document, the corresponding size (=number of words) in &gamma;
 * coding. More precisely, <var>i</var>-th integer of the file (starting from
 * 0) is the size in words of the document of index <var>i</var>.
 *
 * <dt><samp><var>basename</var>.batch<var>i</var></samp>
 *
 * <dd>The <var>i</var>-th batch of occurrences. These files are used in the {@link SecondPass} to produce
 * the inverted index.
 *
 * <dt><samp><var>basename</var>.properties</samp>
 *
 * <dd>A Java {@link java.util.Properties property file} containing information
 * about the index. Currently, the following keys are generated:
 * 
 * <dl>
 * <dt>documents
 * <dd>number documents in the collection;
 * <dt>terms
 * <dd>number of indexed terms;
 * <dt>occurrences
 * <dd>number of words throughout the whole collection;
 * <dt>batches
 * <dd>number of generated batch files;
 * <dt>basename
 * <dd>the basename (see above);
 * <dt>maxdocsize
 * <dd>maximum size of a document in words;
 * <dt>iscasesensitive
 * <dd>a boolean denoting whether the index is case sensitive;
 * <dt>occsperbatch
 * <dd>maximum number of occurrences in each batch;
 * <dt>permutation
 * <dd>name of the permutation file used.
 * </dl>
 * 
 * @author Sebastiano Vigna
 * @since 0.6
 */

final public class FirstPass {

    /** Size in MiB for all buffered I/O. */
    private static int bufferSize;
    /** Temporary array to sort by distribution counting. */
    private static Occurrence temp[];
    /** Temporary array to sort by distribution counting. Will be manually enlarged when needed. */
    private static int count[];
    /** The prefix for the names of all generated files. */
    private static String baseName;
	/** The trimming threshold is checked at each batch of documents of this size. */
	private final static int TRIM_CHECK = 4 * 1024;
	/** The trimming threshold will be doubled if more than this fraction of documents were larger than the current threshold. */
	private final static double TRIM_HIGH_WATERMARK_FRACTION = .0025;
	private final static int TRIM_HIGH_WATERMARK = (int)( TRIM_HIGH_WATERMARK_FRACTION * TRIM_CHECK );
	/** The trimming threshold will be halved if less than this fraction of documents were larger than half of the current threshold. */
	private final static double TRIM_LOW_WATERMARK_FRACTION = .0005;
	private final static int TRIM_LOW_WATERMARK = (int)( TRIM_LOW_WATERMARK_FRACTION * TRIM_CHECK );

	private final static boolean ASSERTS = false;
	{ if ( ASSERTS ); }

	private FirstPass() {}
	
	/** Writes in compressed form a set of occurrences.
	 *
	 * @param occurrence a vector of occurrences.
	 * @param numOccurrences the number of valid occurrences in <code>occurrence</code>.
	 * @param numTerms the number of documents seen so far.
	 * @param numTerms the number of terms seen so far.
	 * @param batch the index of this batch.
	 * @param sortOnTermsOnly whether to sort just of terms.
	 */
	static private void writeOccurrences( Occurrence occurrence[], int numOccurrences, int numDocuments, int numTerms, int batch, boolean sortOnTermsOnly ) throws IOException {
		try {
			final OutputBitStream out = new OutputBitStream( new FileOutputStream( baseName + "batch" + batch ), bufferSize );
			
			if ( sortOnTermsOnly ) {
				if ( count.length < numTerms ) {
					count = null;
					count = new int[ numTerms ];
				} 
				Occurrence.countSortOnTerms( occurrence, temp, numOccurrences, count, numTerms );
				Occurrence.writeOccurrences( temp, numOccurrences, out );
			}
			else {
				if ( count.length < Math.max( numTerms, numDocuments ) ) {
					count = null;
					count = new int[ Math.max( numTerms, numDocuments ) ];
				}

				Occurrence.countSortOnDocuments( occurrence, temp, numOccurrences, count, numDocuments );
				Occurrence.countSortOnTerms( temp, occurrence, numOccurrences, count, numTerms );
				Occurrence.writeOccurrences( occurrence, numOccurrences, out );
			}

			out.close();
		}
		catch( IOException e ) { 
			System.err.println( "I/O Error on batch " + batch ); 
			throw e;
		}
	}

    public static void main( final String arg[] ) throws FileNotFoundException, IOException, ClassNotFoundException {

		int ch;
		boolean zeroth = false;
		char delimiter = 10;
		String permFile = null;
		String bufferSizeSpec = "1Mi", batchSizeSpec = "2Mi", charSet = "UTF-8";
		int occsPerBatch;
		int quantum = 10000;
		int perm[] = null;
		int numDocuments = -1;
		Object2IntOpenHashMap termMap = null;
		MinimalPerfectHash termMph = null;
		boolean isCaseSensitive = true;
		boolean downcase = false;

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "zeroth", LongOpt.NO_ARGUMENT, null, 'z' ),
			new LongOpt( "batch-size", LongOpt.REQUIRED_ARGUMENT, null, 's' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
			new LongOpt( "delimiter", LongOpt.REQUIRED_ARGUMENT, null, 'd' ),
			new LongOpt( "permutation", LongOpt.REQUIRED_ARGUMENT, null, 'p' ),
			new LongOpt( "charset", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
			new LongOpt( "dot-meter-quantum", LongOpt.REQUIRED_ARGUMENT, null, 'q' ),
			new LongOpt( "case-insensitive", LongOpt.NO_ARGUMENT, null, 'i' ),
			new LongOpt( "downcase", LongOpt.NO_ARGUMENT, null, 1 ),
		};

		Getopt g = new Getopt( "FirstPass", arg, "zq:c:p:d:b:s:h", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: FirstPass [OPTIONS] BASENAME" );
			System.err.println( "Builds a set of batches from a list of documents." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  BASENAME    all file names will be stemmed from this name" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -z, --zeroth              assume ZerothPass has been run and BASENAME.mph" );
			System.err.println( "                            has been built." );
			System.err.println( "  -c, --charset             the input charset encoding (default: " + charSet + ")" );
			System.err.println( "  -s, --batch-size          the size of a batch (default: " + batchSizeSpec + ")" );
			System.err.println( "  -b, --buffer-size         the size of an I/O buffer (default: " + bufferSizeSpec + ")" );
			System.err.println( "  -d, --delimiter           the Unicode index of the document delimiter (default: " + (int)delimiter + ")" );
			System.err.println( "  -p, --permutation         the name of a permutation file" );
			System.err.println( "  -q, --dot-meter-quantum   the quantum of the process meter (default: " + quantum + ")" );
			System.err.println( "  -i, --case-insensitive    create a case-insensitive index (input must be downcase)" );
			System.err.println( "  --downcase                downcase all words (and create a case-insensitive index)" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help        print this help screen" );
			System.err.println( "" );
			return;

		case 's':
			batchSizeSpec = g.getOptarg();
			break;

		case 'b':
			bufferSizeSpec = g.getOptarg();
			break;

		case 'c':
			charSet = g.getOptarg();
			break;

		case 'z':
			zeroth = true;
			break;

		case 'd':
			delimiter = (char)Integer.parseInt( g.getOptarg() );
			break;

		case 'p':
			permFile = g.getOptarg();
			break;
			
		case 'q':
			quantum = Integer.parseInt( g.getOptarg() );
			break;

		case 'i':
			isCaseSensitive = false;
			break;
			
		case 1:
			isCaseSensitive = false;
			downcase = true;
			break;
			
		case '?':
			return;


		default:
			break;
		}

		bufferSize = Fast.parseIntSize( bufferSizeSpec );
		occsPerBatch = Fast.parseIntSize( batchSizeSpec );

		if ( arg.length - g.getOptind() == 1 ) baseName = arg[ g.getOptind() ] + ".";
		else {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}

		if ( zeroth ) { // We load the property file generated by ZerothPass
			final Properties properties = new Properties();
			properties.load( new FileInputStream( baseName + "properties" ) );
			numDocuments = Integer.parseInt( properties.getProperty( "documents" ) );
		}

		System.err.println("Documents will be separated by the Unicode character " + (int)delimiter );

		temp = new Occurrence[ occsPerBatch ];

		// This set records which terms have been already seen in a document.
		final IntOpenHashSet termInDoc = new IntOpenHashSet();

		// We bufferize the standard input.
		final FastBufferedReader input = new FastBufferedReader( new InputStreamReader( System.in, charSet ), bufferSize );

		// The word we are scanning.
		final MutableString word = new MutableString();

		// A map from terms to 0-based indices.
		if ( zeroth ) termMph = (MinimalPerfectHash) ( new ObjectInputStream( new FastBufferedInputStream( new FileInputStream( baseName + "mph" ), bufferSize ) ) ).readObject();
		else {
			termMap = new Object2IntOpenHashMap();
			termMap.defaultReturnValue( -1 );
		}

		// Total number of occurrences.
		long totOccurrences = 0;
		int // Current number of documents.
			n = 0,
			// Total number of terms.
			numTerms = zeroth ? termMph.size() : 0, 
			// Current batch number.
			batch = 0, 
			// Total number of occurrences in this batch.
			numOccurrences = 0, 
			// Word position in the current document.
			pos = 0,
			// The threshold with which termInDoc is trimmed.
			trimThreshold = 256,
			// Number of documents in the last block that were above the threshold
			aboveThreshold = 0,
			// Number of documents in the last block that were above half the threshold
			aboveHalfThreshold = 0;

		// For each term index, the number of documents containing the term. This array is manually enlarged when needed.
		int frequency[] = new int[ zeroth ? numTerms : 1024 ];

		count = new int[ Math.max( numDocuments, numTerms ) ];

		int i, j, c, t, maxDocSize = 0;
		final ProgressMeter pm = new ProgressMeter( 1000000 );

		// We now read the permutation vector.

		if ( permFile != null ) {
			System.err.print( "Reading permutation..." );
			int m;
			perm = new int[ m = (int)( new File( permFile ).length() / 4  ) ];
			DataInputStream di = new DataInputStream( new FastBufferedInputStream( new FileInputStream( permFile ), bufferSize ) );
			for( i = 0; i < m; i++ ) {
				j = di.readInt();
				if ( j < 0 || j >= m ) {
					System.err.println( "The permutation file contains the illegal mapping " + i + " |-> " + j );
					return;
				}
				perm[ i ] = j;
			}
			di.close();
			System.err.println( " done." );

			if ( numDocuments < 0 ) numDocuments = perm.length;
			else if ( numDocuments != perm.length ) throw new IllegalStateException( "The permutation file contains " + perm.length + " integers, but the ZerothPass property file claims that there are " + numDocuments + " documents" );
		}


		// The occurrence vector. Should be as large as possible, avoiding swap.
		Occurrence occurrence[] = new Occurrence[ occsPerBatch ];
		i = occurrence.length;

		System.err.print( "Creating occurrences..." );
		pm.itemsName( "objects" );
		pm.start();
		while( i-- != 0 ) occurrence[i] = new Occurrence();
		pm.stop();
		pm.count( occurrence.length );
		pm.done();

		System.err.print( "Indexing documents..." );

		pm.quantum( quantum );
		pm.itemsName( "documents" );
		if ( zeroth ) pm.expectedUpdates( numDocuments );
		pm.start();

		try {
			final PrintWriter termFile = zeroth ? null : new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( baseName + "terms" ), "UTF-8" ) ) ); // Terms in index order.
			final OutputBitStream sizeFile = new OutputBitStream( new FileOutputStream( baseName + "sizes" ) ); // Size in words of each document.

			while( ( c = input.readWord( word ) ) != -1 ) {
					 
				if ( word.length() != 0 ) {
					if ( downcase ) word.toLowerCase();
					// We check whether we have already seen this term. If not, we add it to termIndices and store it in termFile.
					if ( zeroth ) t = termMph.get( word );
					else if ( ( t = termMap.getInt( word ) ) == -1 ) {
						termMap.put( word.copy(), t = numTerms++ );
						word.println( termFile );
						if ( numTerms % 1000000 == 0 ) System.err.print( "[" + Fast.format( numTerms ) + " term(s), " + Fast.format( totOccurrences ) + " occ(s)]" );
					}

					// We check whether we have already seen this word *in this document*.
					if ( termInDoc.add( t ) ) {
						if ( t >= frequency.length ) frequency = IntArrays.grow( frequency, t + 1 );
						frequency[ t ]++;
					}

					// We now record the occurrence. If a permutation has
					// been specified, we have to remap the document index
					// through it.
					if ( perm != null ) occurrence[ numOccurrences++ ].set( t, perm[ n ], pos );
					else occurrence[ numOccurrences++ ].set( t, n, pos );

					totOccurrences++;

					if ( numOccurrences == occsPerBatch ) {
						writeOccurrences( occurrence, numOccurrences, numDocuments < 0 ? n + 1 : numDocuments, numTerms, batch++, perm == null );
						numOccurrences = 0;
					}

					pos++;
				}
						  
 				if ( c == delimiter ) { // New document
					n++;
					pm.update();
					sizeFile.writeGamma( pos );
					if ( pos > maxDocSize ) maxDocSize = pos;
					pos = 0;
					
					/* The following code implements a simple strategy to avoid keeping around
					 * an excessively large table in termInDoc (as clear() empties the map, but
					 * does not deallocate the underlying table; this implies that each clear() will
					 * zero a uselessly large hash-table state array). At each document, the map
					 * is trimmed using trimThreshold as argument (which won't actually trim anything
					 * if the map is already small enough). If more than TRIM_HIGH_WATERMARK
					 * documents in the last TRIM_CHECK were larger than trimThreshold, we double it.
					 * However, if less then TRIM_LOW_WATERMARK documents were larger than half
					 * of trimThreshold, we half it. */

					if ( ( t = termInDoc.size() ) > trimThreshold / 2 ) {
						aboveHalfThreshold++;
						if ( t > trimThreshold ) aboveThreshold++;
					}
					if ( n % TRIM_CHECK == 0 ) {
						if ( aboveThreshold > TRIM_HIGH_WATERMARK ) trimThreshold *= 2;
						else if ( aboveHalfThreshold <= TRIM_LOW_WATERMARK && trimThreshold > 1 ) trimThreshold /= 2; 
						aboveHalfThreshold = aboveThreshold = 0;
					}
					termInDoc.clear();
					termInDoc.trim( trimThreshold );
				}
			}

			if ( pos != 0 ) { // Missing last document
				n++;
				pm.update();
				sizeFile.writeGamma( pos );
			}

			if ( numOccurrences > 0 ) writeOccurrences( occurrence, numOccurrences, numDocuments < 0 ? n + 1 : numDocuments, numTerms, batch++, perm == null );

			pm.done();
			
			if ( numDocuments >= 0  && n != numDocuments ) System.err.println( "WARNING: The input stream contains " + n + " documents, but the ZerothPass property file claims that there are " + numDocuments + " documents" );
			if ( perm != null && n != perm.length ) System.err.println( "WARNING: The input stream contains " + n + " documents, but the permutation contains " + perm.length + " integers" );
				
			if ( ! zeroth ) termFile.close();
			sizeFile.close();
			termMap = null;
			termMph = null;
			occurrence = temp = null;
			count = null;
			
			final OutputBitStream frequencies = new OutputBitStream( new FileOutputStream( baseName + "frequencies" ) );
			for( i = 0; i < numTerms; i++ ) frequencies.writeGamma( frequency[ i ] );
			frequencies.close();

			frequency = null;

			// We now free the term indices map and convert the size file to
			// document index, if necessary. 

			if ( perm != null ) {
				int size[] = new int[ n ];
				final InputBitStream size2File = new InputBitStream( new FileInputStream( baseName + "sizes" ), bufferSize );
				for( i = 0; i < n; i++ ) size[ i ] = size2File.readGamma();
				size2File.close();
				int invPerm[] = new int[ perm.length ];
				for( i = 0; i < n; i++ ) invPerm[ perm[ i ] ] = i;
				final OutputBitStream permSizeFile = new OutputBitStream( new FileOutputStream( baseName + "sizes" ), bufferSize );
				for( i = 0; i < n; i++ ) permSizeFile.writeGamma( size[ invPerm[ i ] ] );
				permSizeFile.close();
			}

			// Finally, we save all data related to this index in a property file.
			final Properties properties = new Properties();
			properties.setProperty( IndexProperties.DOCUMENTS, String.valueOf( n ) );
			properties.setProperty( IndexProperties.TERMS, String.valueOf( numTerms ) );
			properties.setProperty( IndexProperties.MAXDOCSIZE, String.valueOf( maxDocSize ) );
			properties.setProperty( IndexProperties.ISCASESENSITIVE, Boolean.toString( isCaseSensitive ) );

			properties.setProperty( "batches", String.valueOf( batch ) );
			properties.setProperty( "occurrences", String.valueOf( totOccurrences ) );
			properties.setProperty( "occsperbatch", String.valueOf( occsPerBatch ) );

			if ( perm != null ) properties.setProperty( "permutation", permFile );
			final FileOutputStream propertyFile = new FileOutputStream( baseName + "properties" );
			properties.store( propertyFile, "FirstPass properties" );
			propertyFile.close();
		}
		catch( IOException dontCare ) { dontCare.printStackTrace(); }
    }
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
