package it.unimi.dsi.mg4j.tool;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.objects.ObjectHeapSemiIndirectPriorityQueue;
import it.unimi.dsi.mg4j.index.CompressionFlags;
import it.unimi.dsi.mg4j.index.Index;
import it.unimi.dsi.mg4j.index.IndexProperties;
import it.unimi.dsi.mg4j.index.IndexReader;
import it.unimi.dsi.mg4j.index.IndexWriter;
import it.unimi.dsi.mg4j.index.SkipIndexProperties;
import it.unimi.dsi.mg4j.index.SkipIndexWriter;
import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Properties;

/** Merges several indices.
 * 
 * <P>This class merges indices, by renumbering suitably document
 * numbers and coalescing the term list. Note that it can be used
 * also with a single index, making it possible to recompress easily
 * an index using different compression flags.
 * 
 * @author Sebastiano Vigna
 * @since 0.9.1
 * 
 */

final public class Merge implements CompressionFlags {

	final private static String THIS_CLASS_NAME = new Merge().getClass().getName();

	private Merge() {}

    public static void main( String arg[] ) throws IOException {

		int ch;
		int writerFlags = 0, quantum = 64, height = 8;
		MutableString writerFlagsString = new MutableString();
		boolean skips = false;
		int bufferSize;
		String bufferSizeSpec = "1Mi";

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
			new LongOpt( "comp", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
			new LongOpt( "quantum", LongOpt.REQUIRED_ARGUMENT, null, 'Q' ),
			new LongOpt( "height", LongOpt.REQUIRED_ARGUMENT, null, 'H' ),
		};

		Getopt g = new Getopt( THIS_CLASS_NAME, arg, "Q:H:b:hc:", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: " + THIS_CLASS_NAME + " [OPTIONS] OUTPUTBASENAME INPUTBASENAME ..." );
			System.err.println( "Merges several indices." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  OUTPUTBASENAME      basename of the merged index" );
			System.err.println( "  INPUTBASENAME ...   basenames of the indices to be merged" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -b, --buffer-size          the size of an I/O buffer (default: " + bufferSizeSpec + ")" );
			System.err.println( "  -c, --comp                 a compression flags for the index (may be specified several times)." );
			System.err.println( "  -Q, --quantum              enable skips with given quantum (default: " + quantum + ")." );
			System.err.println( "  -H, --height               enable skips with given height (default: " + height + ")." );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help        print this help screen" );
			System.err.println( "" );
			System.exit( 0 );

		case 'b':
			bufferSizeSpec = g.getOptarg();
			break;

		case 'Q':
			quantum = Integer.parseInt( g.getOptarg() );
			if ( quantum <= 0 ) {
				System.err.println( "Invalid quantum: " + quantum );
				return;
			}
			skips = true;
			break;

		case 'H':
			height = Integer.parseInt( g.getOptarg() );
			if ( height < 0 ) {
				System.err.println( "Invalid height: " + height );
				return;
			}
			skips = true;
			break;

		case 'c':
			try {
				writerFlags |= IndexWriter.class.getField( g.getOptarg() ).getLong( IndexWriter.class );
				if ( writerFlagsString.length() != 0 ) writerFlagsString.append( " | " );
				writerFlagsString.append( g.getOptarg() );
			}
			catch ( Exception notFound ) {
				System.err.println( "Compression method " + g.getOptarg() + " unknown." );
				return;
			}
			break;

		case '?':
			return;

		default:
			break;
		}

		bufferSize = Fast.parseIntSize( bufferSizeSpec );

		if ( arg.length - g.getOptind() < 2 ) {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}
		
		int i, j, k;

		// The number of indices to be merged
		final int numIndices = arg.length - g.getOptind() - 1;
		// The basename of the output index
		final String outputBasename = arg[ g.getOptind() ];
		// The basename of the indices to be merged
		final String[] basename = new String[ numIndices ];
		// An index for each index to be merged
		final Index[] index = new Index[ numIndices ];
		// An index reader for each index to be merged (will scan sequentially the whole index)
		final IndexReader[] indexReader = new IndexReader[ numIndices ];
		// An input bit stream for each index, scanning the global counts
		final InputBitStream[] globCounts = new InputBitStream[ numIndices ];
		// The term-queue reference array
		final MutableString[] term = new MutableString[ numIndices ];
		// A buffered reader for each index (will be used to read terms)
		final FastBufferedReader termReader[] = new FastBufferedReader[ numIndices ];
		// The term queue: it contains the last term read by each reader
		final ObjectHeapSemiIndirectPriorityQueue queue = new ObjectHeapSemiIndirectPriorityQueue( term, numIndices ); 

		final ProgressMeter pm = new ProgressMeter( 1000000 );
		// Global values for the merged index
		int numberOfDocuments = 0, maxDocPos = 0, numberOfLists = 0;
		// This will remain set if *all* indices to be merged agree
		boolean hasCounts = true, hasPositions = true;

		for( i = 0; i < numIndices; i++ ) {
			basename[ i ] = arg[ g.getOptind() + 1 + i ];
			index[ i ] = Index.getInstance( basename[ i ] );
			if ( ! index[ i ].hasCounts ) hasCounts = false;
			if ( ! index[ i ].hasPositions ) hasPositions = false;
			maxDocPos = Math.max( maxDocPos, index[ i ].maxDocPos );
			indexReader[ i ] = index[ i ].getReader( bufferSize );
			numberOfDocuments += index[ i ].numberOfDocuments;
			numberOfLists += index[ i ].numberOfTerms;
			globCounts[ i ] = new InputBitStream( basename[ i ] + ".globcounts" );
			term[ i ] = new MutableString();
			termReader[ i ] = new FastBufferedReader( new InputStreamReader( new FileInputStream( basename[ i ] + ".terms" ), "UTF-8" ) );
			termReader[ i ].readLine( term[ i ] );
			// If the term list is nonempty, we enqueue it
			if ( term[ i ].length() != 0 ) queue.enqueue( i );
		}

		try {
			// We concatenate all size lists
			final int size[] = new int[ numberOfDocuments ];
			double averageSize = 0;

			final OutputBitStream outputSizes = new OutputBitStream( outputBasename + ".sizes", bufferSize );

			pm.quantum( 1000000 );
			pm.itemsName( "ints" );
			pm.start( "Merging sizes..." );

			int maxDocSize = 0;

			for( i = k = 0; i < numIndices; i++ ) {
				final InputBitStream sizes = new InputBitStream( basename[ i ] + ".sizes", bufferSize );
				int s = 0;
				j = index[ i ].numberOfDocuments;
				while( j-- != 0 ) {
					size[ k++ ] = s = sizes.readGamma();
					if ( s > maxDocSize ) maxDocSize = s;
					outputSizes.writeGamma( s );
					averageSize += s;
				} 
				sizes.close();
			}

			outputSizes.close();
			pm.done();
			averageSize /= numberOfDocuments;

			// To write the new index
			final OutputBitStream mergedIndex = new OutputBitStream( outputBasename + ".index", bufferSize );
			// To write the new index offsets
			final OutputBitStream offsets = new OutputBitStream( outputBasename + ".offsets" );
			// To write the global count of each term
			final OutputBitStream outputGlobCounts = new OutputBitStream( outputBasename + ".globcounts" );
			// To write the frequency of each term
			final OutputBitStream frequencies = new OutputBitStream( outputBasename + ".frequencies" );
			// To write the new term list
			final PrintWriter termFile = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( outputBasename + ".terms" ), "UTF-8" ), bufferSize ) );

			final IndexWriter indexWriter = skips ? new SkipIndexWriter( mergedIndex, offsets, numberOfDocuments, writerFlags, quantum, height ) :  new IndexWriter( mergedIndex, offsets, numberOfDocuments, writerFlags );

			// The current term
			MutableString currTerm;
			// If inIndex[i] is true, the i-th index contains the current term
			boolean[] inIndex = new boolean[ numIndices ];
			// The first k elements list (in order) the indices containing the current term
			int[] currIndex = new int[ numIndices ];
			// For each index, the frequency of the current term (if it appears there)
			int[] frequency = new int[ numIndices ];
			// To read positions 
			int[] position = new int[ maxDocPos ];

			// Counters (just for bookkeeping).
			long 
				bitsForPointers[] = new long[ 10 ], 
				bitsForPositions[] = new long[ 10 ],
				// Total number of pointers and occurrences
				numPointers = 0, numOccurrences = 0;

			// Number of pointers for terms with frequency <10%, <20%,...
			long 
				pointersByFrequency[] = new long[10], 
				// Number of occurrences for terms with frequency <10%, <20%,...
				occurrencesByFrequency[] = new long[10];

			pm.expectedUpdates( numberOfLists );
			pm.itemsName( "lists" );
			pm.start( "Merging lists..." );

			int count, currDoc, numPrevDocs, totalFrequency, numTerms = 0, freqClass;
			long totalGlobCount;
			OutputBitStream obs;

			while( ! queue.isEmpty() ) {
				k = 0;
				// We read a new word from the queue, copy it and write it to the term file
				currTerm = term[ j = currIndex[ k++ ] = queue.first() ].copy();
				inIndex[ j ] = true;
				currTerm.println( termFile );
				termReader[ j ].readWord( term[ j ] );
				if ( term[ j ].length() == 0 ) queue.dequeue();
				else queue.changed();
				pm.update();

				// Then, we extract all equal words from the queue, accumulating the set of indices in inIndex and currIndex
				while( ! queue.isEmpty() && term[ queue.first() ].equals( currTerm ) ) {
					j = currIndex[ k++ ] = queue.first();
					inIndex[ j ] = true;
					termReader[ j ].readWord( term[ j ] );
					if ( term[ j ].length() == 0 ) queue.dequeue();
					else queue.changed();
					pm.update();
				}
							
				numTerms++;
					
				totalFrequency = 0;
				totalGlobCount = 0;
				for( i = 0; i < k; i++ ) {
					totalFrequency += ( frequency[ currIndex[ i ] ] = indexReader[ currIndex[ i ] ].readFrequency() );
					totalGlobCount += globCounts[ currIndex[ i ] ].readGamma();
				}
				
				indexWriter.newInvertedList();
				indexWriter.writeFrequency( totalFrequency );

				frequencies.writeGamma( totalFrequency );
				outputGlobCounts.writeLongGamma( totalGlobCount );

				freqClass = (int)( 10 * totalFrequency / (double)( numberOfDocuments + 1 ) );

				occurrencesByFrequency[ freqClass ] += totalGlobCount;

				numPointers += totalFrequency;
				pointersByFrequency[ freqClass ] += totalFrequency;

				numPrevDocs = 0;
				for( i = 0; i < numIndices; i++ ) {
					if ( inIndex[ i ] ) {
						j = frequency[ i ];
						while( j-- != 0 ) {
							obs = indexWriter.newDocumentRecord();
							currDoc = indexReader[ i ].readDocumentPointer() + numPrevDocs;
							bitsForPointers[ freqClass ] += indexWriter.writeDocumentPointer( obs, currDoc );
					
							if ( hasCounts ) {
								count = indexReader[ i ].readPositionCount();
								indexWriter.writePositionCount( obs, count );
								numOccurrences += count;
								if ( hasPositions ) {
									indexReader[ i ].readDocumentPositions( position );
									bitsForPositions[ freqClass ] += indexWriter.writeDocumentPositions( obs, position, 0, count, size[ currDoc ] );
								} 
							}					
						}
						inIndex[ i ] = false;
					}
					numPrevDocs += index[ i ].numberOfDocuments;
				}
			}
			pm.done();

			indexWriter.close();
			offsets.close();
			frequencies.close();
			outputGlobCounts.close();
			termFile.close();
			
			final Properties properties = new Properties();
			properties.setProperty( IndexProperties.DOCUMENTS, String.valueOf( numberOfDocuments ) );
			properties.setProperty( IndexProperties.MAXDOCPOS, String.valueOf( maxDocPos ) );
			properties.setProperty( IndexProperties.MAXDOCSIZE, String.valueOf( maxDocSize ) );
			properties.setProperty( IndexProperties.COMPRESSIONFLAGS, writerFlagsString.toString() );

			if ( skips ) {
				properties.setProperty( SkipIndexProperties.SKIPQUANTUM, String.valueOf( quantum ) );
				properties.setProperty( SkipIndexProperties.SKIPHEIGHT, String.valueOf( height ) );
			}
			
			properties.setProperty( IndexProperties.TERMS, String.valueOf( numTerms ) );
			properties.setProperty( IndexProperties.INDEXCLASS, "it.unimi.dsi.mg4j.index." + ( skips ? "SkipIndex" : "Index" ) );

			properties.setProperty( "occurrences", String.valueOf( numOccurrences ) );

			final FileOutputStream propertyFile = new FileOutputStream( outputBasename + ".properties" );
			properties.store( propertyFile, "Merge properties" );
			propertyFile.close();

			System.out.println( "Number of documents: " + Fast.format( numberOfDocuments ) );
			System.out.println( "Average size: " + Fast.format( averageSize ) );
			System.out.println( "Number of terms: " + Fast.format( numTerms ) );
			System.out.println( "Frequencies: " + Fast.format( indexWriter.bitsForFrequencies ) + " bits, " + Fast.format( indexWriter.bitsForFrequencies/ (double)numTerms ) + " bits/frequency." );
			System.out.println( "Document pointers: " + Fast.format( numPointers ) + " (" + Fast.format( indexWriter.bitsForPointers ) + " bits, " + Fast.format( indexWriter.bitsForPointers/ (double)numPointers ) + " bits/pointer).");
			if ( indexWriter.hasCounts ) System.out.println( "Counts: " + Fast.format( numPointers ) + " (" + Fast.format( indexWriter.bitsForCounts ) + " bits, " + Fast.format( indexWriter.bitsForCounts/ (double)numPointers ) + " bits/count).");
			if ( indexWriter.hasPositions ) System.out.println( "Occurrences: " + Fast.format( numOccurrences ) + " (" +  Fast.format( indexWriter.bitsForPositions ) + " bits, " + Fast.format( indexWriter.bitsForPositions/ (double)numOccurrences ) + " bits/occurrence).");
	
			if ( skips ) {
				final SkipIndexWriter s = (SkipIndexWriter)indexWriter;
				System.out.println( "Skip towers: " + Fast.format( s.towerData.numberOfSkipTowers ) + " (" + 
					Fast.format( s.towerData.bitsForTowers() ) + " bits [" + 
					Fast.format( s.towerData.bitsForTowers()*100.0/s.writtenBits() ) + "%], " + 
					Fast.format( s.towerData.bitsForTowers()/ (double)s.towerData.numberOfSkipTowers ) + 
					" bits/tower)" );
				System.out.println( "Skip entries: " + Fast.format( s.towerData.numberOfSkipEntries ) + " (" + 
					Fast.format( s.towerData.bitsForEntries()/ (double)s.towerData.numberOfSkipEntries ) + " bits/entry)" );
				System.out.println( "Skip tower lengths: " + Fast.format( s.towerData.bitsForTowerLengths ) + " bits (" + Fast.format( s.towerData.bitsForTowerLengths/ (double)s.towerData.numberOfSkipTowers ) + " bits/tower)" );
				System.out.println( "Quantum bit lengths: " + Fast.format( s.bitsForQuantumBitLengths ) + " bits (" + Fast.format( s.bitsForQuantumBitLengths/ (double)s.numberOfBlocks ) + " bits/block)" );
				System.out.println( "Entry bit lengths: " + Fast.format( s.bitsForEntryBitLengths ) + " bits (" + Fast.format( s.bitsForEntryBitLengths/ (double)s.numberOfBlocks ) + " bits/block)" );

				System.out.println( "Top skip amounts: " + Fast.format( s.towerData.bitsForTopSkipAmounts ) + " bits (" + Fast.format( s.towerData.bitsForTopSkipAmounts/ (double)s.towerData.numberOfSkipTowers ) + " bits/amount)" );
				System.out.println( "Top skip pointers: " + Fast.format( s.towerData.bitsForTopSkipPointers ) + " bits (" + Fast.format( s.towerData.bitsForTopSkipPointers/ (double)s.towerData.numberOfSkipTowers ) + " bits/pointer)" );
				System.out.println( "Lower skip amounts: " + Fast.format( s.towerData.bitsForLowerSkipAmounts ) + " bits (" + Fast.format( s.towerData.bitsForLowerSkipAmounts/ (double)s.towerData.numberOfLowerEntries() ) + " bits/amount)" );
				System.out.println( "Lower skip pointers: " + Fast.format( s.towerData.bitsForLowerSkipPointers ) + " bits (" + Fast.format( s.towerData.bitsForLowerSkipPointers/ (double)s.towerData.numberOfLowerEntries() ) + " bits/pointer)" );
				System.out.println( "Skip amounts: " + Fast.format( s.towerData.bitsForSkipAmounts() ) + " bits (" + Fast.format( s.towerData.bitsForSkipAmounts()/ (double)s.towerData.numberOfSkipEntries ) + " bits/amount)" );
				System.out.println( "Skip pointers: " + Fast.format( s.towerData.bitsForSkipPointers() ) + " bits (" + Fast.format( s.towerData.bitsForSkipPointers()/ (double)s.towerData.numberOfSkipEntries ) + " bits/pointer)" );
			}
			System.out.println( "Alignment: " + Fast.format( indexWriter.bitsForAlignment ) + " bits, "  + Fast.format( indexWriter.bitsForAlignment/ (double)numTerms ) + " bits/alignment." );

			if ( indexWriter.hasPositions ) System.out.println( "Total: " + Fast.format( indexWriter.writtenBits() ) + " bits, " + Fast.format( indexWriter.writtenBits()/ (double)numOccurrences ) + " bits/occurrence" );
			else System.out.println( "Total: " + Fast.format( indexWriter.writtenBits() ) + " bits, " + Fast.format( indexWriter.writtenBits()/ (double)numOccurrences ) + " bits/pointer" );

			System.out.println();
			System.out.println("Number of pointers by frequency:");
			for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * pointersByFrequency[ i ] / numPointers ) + "%");
			System.out.println();

			System.out.println("Number of occurrences by frequency:");
			for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * occurrencesByFrequency[ i ] / numOccurrences ) + "%" ); 

			System.out.println();
			System.out.println("Number of bits for pointers by frequency:");
			for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * bitsForPointers[ i ] / indexWriter.bitsForPointers ) + "%");
			System.out.println();

			if ( indexWriter.hasPositions ) {
				System.out.println("Number of bits for occurrences by frequency:");
				for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * bitsForPositions[ i ] / indexWriter.bitsForPositions ) + "%" ); 
			}				
		} catch( IOException dontCare ) { dontCare.printStackTrace(); }
    }
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
