package it.unimi.dsi.mg4j.tool;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

import jal.ints.BinaryPredicate;
import jal.ints.Sorting;

import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.ProgressMeter;
import it.unimi.dsi.mg4j.util.Fast;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Properties;

/** Permutes a set of batches and related files so that the resulting index has lexicographically ordered terms.
 *
 * <P>After {@link FirstPass}, the term index follows the order in which the different
 * terms have been met. By preprocessing a set of batches with this class,
 * terms are permuted so that the resulting term index coincides with
 * lexicographical order. All data files previously produced by {@link FirstPass} are permuted
 * accordingly.
 *
 * @author Sebastiano Vigna
 * @since 0.6
 */

final public class MiddlePass {

	/** The term array. */
	static String term[];

	private MiddlePass() {}

	private static class TermPredicate implements BinaryPredicate {
		public boolean apply( int x, int y ) { return term[ x ].compareTo( term[ y ] ) < 0; }
	}

	public static void main( String arg[] ) throws IOException {
		int bufferSize;
		String bufferSizeSpec = "1Mi", baseName;	
		int ch;

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
		};

		Getopt g = new Getopt( "MiddlePass", arg, "b:h", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: MiddlePass [OPTIONS] BASENAME" );
			System.err.println( "Permutes a set of batches so that terms are ordered lexicographically." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  BASENAME	all file names will be stemmed from this name" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -b, --buffer-size the size of an I/O buffer (default: " + bufferSizeSpec + ")" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help		print this help screen" );
			System.err.println( "" );
			return;

		case 'b':
			bufferSizeSpec = g.getOptarg();
			break;

		case '?':
			return;

		default:
			break;
		}

		bufferSize = Fast.parseIntSize( bufferSizeSpec );

		if ( arg.length - g.getOptind() == 1 ) baseName = arg[ g.getOptind() ] + ".";
		else {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}

		int i, j;

		final FileInputStream propertyFile = new FileInputStream( baseName + "properties" );
		final Properties properties = new Properties();
		properties.load( propertyFile );
		propertyFile.close();

		// Number of terms.
		final int numTerms = Integer.parseInt( properties.getProperty( "terms" ) );
		// Number of batches.
		int numBatches = Integer.parseInt( properties.getProperty( "batches" ) );
		// Occurrences per batch.
		int occsPerBatch = Integer.parseInt( properties.getProperty( "occsperbatch" ) );

		// The occurrence vector.
		final Occurrence occurrence[] = new Occurrence[ occsPerBatch ];
		final Occurrence sortedOccurrence[] = new Occurrence[ occsPerBatch ];
		// A buffered input stream for each batch.

		// First phase: we build the term ordering array.

		System.err.print( "Reading terms... " );

		term = new String[ numTerms ];
		final BufferedReader terms = new BufferedReader( new InputStreamReader( new FileInputStream( baseName + "terms" ), "UTF-8" ), bufferSize );
		for( i = 0; i < numTerms; i++ ) term[i] = terms.readLine();
		terms.close();

		final int perm[] = new int[ numTerms ];
		i = numTerms;
		while( i-- != 0 ) perm[ i ] = i;

		System.err.print( "sorting terms... " );

		Sorting.sort( perm, 0, perm.length, new TermPredicate() );

		System.err.print( "saving sorted terms... " );

		final PrintWriter sortedTerms = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( baseName + "terms" ), "UTF-8" ), bufferSize ) ); // Terms in index order.
		for( i = 0; i < numTerms; i++ ) sortedTerms.println( term[ perm[ i ] ] );
		sortedTerms.close();

		term = null;

		final int count[] = new int[ numTerms + 1 ];

		if ( new File( baseName + "frequencies" ).exists() ) {

			System.err.print( "reading frequencies... " );

			final InputBitStream frequencies = new InputBitStream( new FileInputStream ( baseName + "frequencies" ), bufferSize );
			for( i = 0; i < numTerms; i++ ) count[ i ] = frequencies.readGamma();
			frequencies.close();

			System.err.print( "writing frequencies... " );

			final OutputBitStream sortedFrequencies = new OutputBitStream( new FileOutputStream ( baseName + "frequencies" ), bufferSize );
			for( i = 0; i < numTerms; i++ ) sortedFrequencies.writeGamma( count[ perm[ i ] ] );
			sortedFrequencies.close();
		}
		else System.err.print( "(no frequencies)..." );

		System.err.print( "creating occurrences... " );

		i = occsPerBatch;
		while( i-- != 0 ) occurrence[ i ] = new Occurrence();

		System.err.println( " done." );

		int l;
		InputBitStream in;
		OutputBitStream out;

		// We now invert the sorting permutation.
		int temp[] = new int[ numTerms ];
		i = numTerms;
		while( i-- != 0 ) temp[ perm[ i ] ] = i;
		System.arraycopy( temp, 0, perm, 0, numTerms );
		temp = null;

		ProgressMeter pm = new ProgressMeter( 10, "batches" );

		System.err.print( "Permuting batches..." );
		pm.start();

		for( i = 0; i < numBatches; i++ ) {

			in = new InputBitStream( new FileInputStream( baseName + "batch" + i ), bufferSize );
			l = Occurrence.readOccurrences( occurrence, occsPerBatch, in );
			in.close();

			j = l;
			while( j-- != 0 ) occurrence[ j ].termIndex = perm[ occurrence[ j ].termIndex ];
				
			Occurrence.countSortOnTerms( occurrence, sortedOccurrence, l, count, numTerms );
		  
			out = new OutputBitStream( new FileOutputStream( baseName + "batch" + i ), bufferSize );
			Occurrence.writeOccurrences( sortedOccurrence, l, out );
			out.close();

			pm.update();
		}

		pm.done();
	}
}
// Local Variables:
// mode: jde
// tab-width: 4
// End:
