package it.unimi.dsi.mg4j.tool;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;

/** A class denoting an occurrence.
 * 
 * <P>An instance of this class is used to store an occurrence. The class
 * stores the index of the term, the index of the document and the position (starting
 * from 0) of the occurrence.
 *
 * @author Sebastiano Vigna
 * @since 0.6
 */

public class Occurrence implements Comparable {
	/**  The term index. */
	public int termIndex;
	/**  The document index. */
	public int docIndex;
	/**  The position of this occurrence of term {@link #termIndex} in document {@link #docIndex}. */
	public int docPosition;
	
	/** Creates a new occurrence with all fields initialised to zero.
	 */
	public Occurrence() {}

	/** Creates a new occurrence with given indices.
	 *
	 * @param termIndex the term index.
	 * @param docIndex the document index.
	 * @param docPosition the position of the occurrences in the document.
	 */
	public Occurrence( final int termIndex, final int docIndex, final int docPosition ) {
		set( termIndex, docIndex, docPosition );
	}
	
	/** Sets the fields of this occurrence.
	 *
	 * @param termIndex the term index.
	 * @param docIndex the document index.
	 * @param docPosition the position of the occurrences in the document.
	 */
	public void set( final int termIndex, final int docIndex, final int docPosition ) {
		this.termIndex = termIndex;
		this.docIndex = docIndex;
		this.docPosition = docPosition;
	}
	
	public int hashCode() {
		return termIndex ^ docIndex ^ docPosition;
	}
	
	public boolean equals( Object o ) {
		if ( o instanceof Occurrence ) {
			final Occurrence occ = (Occurrence)o;
			return termIndex == occ.termIndex && docIndex == occ.docIndex && docPosition == occ.docPosition;
		}
		return false;
	}
	
	/** Compares this occurrence with another object.
	 *
	 * <P>Comparison between occurrences is lexicographical w.r.t. {@link #termIndex},
	 * {@link #docIndex} and {@link #docPosition}, in this order.
	 *
	 * @param o an occurrence.
	 * @return a negative integer, zero, or a positive integer as this
	 * occurence is less than, equal to, or greater than the specified occurrence.
	 * @throws ClassCastException if the argument is not an {@link Occurrence}.
	 * @see #compareTo(Object)
	 */
	
	public int compareTo( final Object o ) {
		final Occurrence occ = (Occurrence)o;
		int t;
		if ( ( t = termIndex - occ.termIndex ) != 0 ) return t;
		if ( ( t = docIndex - occ.docIndex ) != 0 ) return t;
		return docPosition - occ.docPosition;
	}
	
	public String toString() {
		return "[termIndex: " + termIndex + " docIndex: " + docIndex + " docPosition: " + docPosition + "]";
	}


	/** Performs a distribution-counting sort over a vector of occurrences using only {@link #termIndex} as key.
	 *
	 * @param in a vector of occurrences to be sorted.
	 * @param out a vector to store the sorted permutation of <code>in</code>; its length must be at least <code>len</code>.
	 * @param len the number of valid occurrences in <code>occurrence</code>.
	 * @param count a vector to perform the counting; its length must be greater than <em>n</em>.
	 * @param n the number of keys (terms).
	 */

    public static void countSortOnTerms( final Occurrence in[],  final Occurrence out[], final int len, final int[] count, final int n ) {
		int i;
		final Occurrence s[] = in;
		final Occurrence t[] = out;
		final int c[] = count;

		IntArrays.fill( c, 0, n, 0 );

		i = len;
		while( i-- != 0 ) c[ s[ i ].termIndex ]++;
		  
		for( i = 1; i < n; i++ ) c[ i ] += c[ i - 1 ];

		i = len;
		while( i-- != 0 ) t[ --c[ s[ i ].termIndex ] ] = s[ i ];
    }


	/** Performs a distribution-counting sort over a vector of occurrences using only {@link #docIndex} as key.
	 *
	 * @param in a vector of occurrences to be sorted.
	 * @param out a vector to store the sorted permutation of <code>in</code>; its length must be at least <code>len</code>.
	 * @param len the number of valid occurrences in <code>occurrence</code>.
	 * @param count a vector to perform the counting; its length must be greater than <em>n</em>.
	 * @param n the number of keys (documents).
	 */

	public static void countSortOnDocuments( final Occurrence in[],  final Occurrence out[], final int len, final int[] count, final int n ) {
		int i;
		final Occurrence s[] = in;
		final Occurrence t[] = out;
		final int c[] = count;

		IntArrays.fill( c, 0, n, 0 );

		i = len;
		while( i-- != 0 ) c[ s[ i ].docIndex ]++;
		  
		for( i = 1; i < n; i++ ) c[ i ] += c[ i - 1 ];

		i = len;
		while( i-- != 0 ) t[ --c[ s[ i ].docIndex ] ] = s[ i ];
	}


	/** Writes in compressed form a vector of sorted occurrences.
	 *
	 * @param occurrence a vector of occurrences.
	 * @param len the number of valid occurrences in <code>occurrence</code>.
	 * @param out an already opened bit stream where the output will be sent.
	 */

    public static void writeOccurrences( final Occurrence occurrence[], final int len, final OutputBitStream out ) throws java.io.IOException {
		int i = 0;
		Occurrence curr, prev;
	  
		out.writeDelta( len );
		
		out.writeDelta( occurrence[ 0 ].termIndex );
		out.writeDelta( occurrence[ 0 ].docIndex );
		out.writeDelta( occurrence[ 0 ].docPosition );

		curr = occurrence[ 0 ];
		
		for( i = 1; i < len; i++ ) {
			prev = curr;
			curr = occurrence[ i ];

			out.writeUnary( curr.termIndex - prev.termIndex );
			
			if ( curr.termIndex == prev.termIndex ) {
				out.writeDelta( curr.docIndex - prev.docIndex );
				
				if ( curr.docIndex == prev.docIndex ) {
					out.writeDelta( curr.docPosition - prev.docPosition - 1 );
				}
				else {
					out.writeDelta( curr.docPosition );
				}
			}
			else {
				out.writeDelta( curr.docIndex );
				out.writeDelta( curr.docPosition );
			}
		}
    }

	/** Reads a compressed stream of occurrences into a vector.
	 *
	 * @param occurrence a vector of occurrences.
	 * @param len the maximum number of occurrences to be read.
	 * @param in an already opened input bit stream.
	 * @return the number of occurrences actually read.
	 */

    public static int readOccurrences( final Occurrence occurrence[], final int len, final InputBitStream in ) throws java.io.IOException {

		final int l = Math.min( Math.min( len, in.readDelta() ), occurrence.length );

		if ( l == 0 ) return l;
		
		int i,
			// Current term index.
			curTermIndex, 
			// Current document index.
			curDocIndex, 
			// Current document position.
			curDocPosition, 
			// Temporary variables to hold the next differences in a in.
			termIndexDelta, docIndexDelta, docPositionDelta;

		curTermIndex = in.readDelta();
		curDocIndex = in.readDelta();
		curDocPosition = in.readDelta();

		occurrence[ 0 ].set( curTermIndex, curDocIndex, curDocPosition );
		
		for( i = 1; i < l; i++ ) {
			
			termIndexDelta = in.readUnary();
			docIndexDelta = in.readDelta();
			docPositionDelta = in.readDelta();
			
			if ( termIndexDelta != 0 ) {
				curDocIndex = 0;
				curDocPosition = -1;
			}
			
			if ( docIndexDelta != 0 ) {
				curDocPosition = -1;
			}
			
			occurrence[ i ].set( curTermIndex += termIndexDelta, curDocIndex += docIndexDelta, curDocPosition += docPositionDelta + 1 );
		}

		return l;
	}
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
