package it.unimi.dsi.mg4j.tool;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.objects.ObjectHeapSemiIndirectPriorityQueue;
import it.unimi.dsi.mg4j.index.CompressionFlags;
import it.unimi.dsi.mg4j.index.IndexWriter;
import it.unimi.dsi.mg4j.index.SkipIndexWriter;
import it.unimi.dsi.mg4j.io.InputBitStream;
import it.unimi.dsi.mg4j.io.OutputBitStream;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Properties;

/** Builds an inverted index by merging occurrence batches produced by {@link FirstPass}.
 * Some statistics will be printed to standard output at the end of the indexing process.
 *
 * <P>This class merges the occurrence files produced by {@link FirstPass} (and
 * possibly reordered by {@link MiddlePass}) into an inverted index.
 *
 * <P>These are the files currently generated:
 * <dl>
 *
 * <dt><samp><var>basename</var>.index</samp>
 *
 * <dd>The inverted index.
 *
 * <dt><samp><var>basename</var>.offsets</samp>
 *
 * <dd>For each term, the byte offset in
 * <samp><var>basename</var>.index</samp> at which the inverted lists start.
 * More precisely, the first integer is the offset for term 0 in &gamma; coding, and then the
 * <var>i</var>-th integer is the difference between the <var>i</var>-th and
 * the <var>i</var>&minus;1-th offset in &gamma; coding. If <var>T</var>
 * terms were indexed, this file will contain <var>T</var>+1 integers,
 * the last being the difference (in bytes) between the length of the entire
 * inverted index and the offset of the last inverted list.
 *
 * <dt><samp><var>basename</var>.globcounts</samp>
 *
 * <dd>For each term, the number of its occurrences 
 * throughout the whole document collection, in &gamma; coding. More
 * precisely, the <var>i</var>-th integer of the file (starting from 0) is the
 * number of occurrences of the term of index <var>i</var>.
 *
 * <dt><samp><var>basename</var>.properties</samp>
 *
 * <dd>This class adds some information to the {@link java.util.Properties property file} produced
 * by {@link FirstPass}. Currently, the following keys are generated:
 * 
 * <dl>
 * <dt>compressionflags
 * <dd>the mask of compression flags used when generating the index;
 * <dt>maxcount 
 * <dd>the maximum count in the collection, that is, the maximum count of
 * a term maximised on all terms and documents.
 * </dl>
 *
 * </dl>
 *
 * @author Sebastiano Vigna
 * @since 0.6
 * 
 */

final public class SecondPass implements CompressionFlags {

	final private static boolean ASSERTS = false;

	private SecondPass() {}

    public static void main( String arg[] ) {

		int ch;
		int writerFlags = 0, quantum = 64, height = 8;
		MutableString writerFlagsString = new MutableString();
		boolean computeFrequencies = false, skips = false;
		int bufferSize;
		String bufferSizeSpec = "1Mi", baseName;

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
			new LongOpt( "comp", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
			new LongOpt( "quantum", LongOpt.REQUIRED_ARGUMENT, null, 'Q' ),
			new LongOpt( "height", LongOpt.REQUIRED_ARGUMENT, null, 'H' ),
			new LongOpt( "compute-frequencies", LongOpt.NO_ARGUMENT, null, 'f' ),
		};

		Getopt g = new Getopt( "SecondPass", arg, "fQ:H:b:hc:", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: SecondPass [OPTIONS] BASENAME" );
			System.err.println( "Builds an inverted index from a set of batches." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  BASENAME    all file names will be stemmed from this name" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -b, --buffer-size          the size of an I/O buffer (default: " + bufferSizeSpec + ")" );
			System.err.println( "  -c, --comp                 a compression flag for the index (may be specified several times)." );
			System.err.println( "  -f, --compute-frequencies  compute frequencies from batches." );
			System.err.println( "  -Q, --quantum              enable skips with given quantum (default: " + quantum + ")." );
			System.err.println( "  -H, --height               enable skips with given height (default: " + height + ")." );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help        print this help screen" );
			System.err.println( "" );
			return;

		case 'b':
			bufferSizeSpec = g.getOptarg();
			break;

		case 'f':
			computeFrequencies = true;
			break;

		case 'Q':
			quantum = Integer.parseInt( g.getOptarg() );
			if ( quantum <= 0 ) {
				System.err.println( "Invalid quantum: " + quantum );
				return;
			}
			skips = true;
			break;

		case 'H':
			height = Integer.parseInt( g.getOptarg() );
			if ( height < 0 ) {
				System.err.println( "Invalid height: " + height );
				return;
			}
			skips = true;
			break;

		case 'c':
			try {
				writerFlags |= IndexWriter.class.getField( g.getOptarg() ).getLong( IndexWriter.class );
				if ( writerFlagsString.length() != 0 ) writerFlagsString.append( " | " );
				writerFlagsString.append( g.getOptarg() );
			}
			catch ( Exception notFound ) {
				System.err.println( "Compression method " + g.getOptarg() + " unknown." );
				return;
			}
			break;

		case '?':
			return;

		default:
			break;
		}

		bufferSize = Fast.parseIntSize( bufferSizeSpec );

		if ( arg.length - g.getOptind() == 1 ) baseName = arg[ g.getOptind() ] + ".";
		else {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}

		int i;
		long j;

		try {
			final FileInputStream propertyFile = new FileInputStream( baseName + "properties" );
			final Properties properties = new Properties();
			properties.load( propertyFile );
			propertyFile.close();

			// Number of documents.
			final int N = Integer.parseInt( properties.getProperty( "documents" ) );
			// Number of terms.
			final int numTerms = Integer.parseInt( properties.getProperty( "terms" ) );
			// Number of batches.
			int numBatches = Integer.parseInt( properties.getProperty( "batches" ) );
			// Number of occurrences.
			final long numOccurrences = Long.parseLong( properties.getProperty( "occurrences" ) );

			// Size in terms of each document.
			final int size[] = new int[ N ];
			double averageSize = 0;
			final InputBitStream sizes = new InputBitStream( new FileInputStream ( baseName + "sizes" ), bufferSize );
			for( i = 0; i < N; i++ ) averageSize += ( size[i] = sizes.readGamma() );
			sizes.close();
			averageSize /= N;
				
			// The occurrence heap.
			final Occurrence[] occurrence = new Occurrence[ numBatches ];
			// A buffered input stream for each batch.
			final InputBitStream fibs[] = new InputBitStream[ numBatches ];

			Occurrence curr;

			int occInDoc[] = new int[ 16 * 1024 ],
				// Number of occurrences left in a batch.
				numOcc[] = new int[ numBatches ], 
				// Current term index in a batch.
				currTerm[] = new int[ numBatches ], 
				// Current document index in a batch.
				currDocument[] = new int[ numBatches ], 
				// Current document position in a batch.
				currPosition[] = new int[ numBatches ],
				// The batch from which the current occurrence is coming
				currBatch, 
				// Number of occurrences in the current document.
				numPosInDoc = 0,
				// The previous term index in the output stream, or -1 at start.
				prevTermIndex = -1,
				// The previous document index, or -1 if we just started a new term.
				prevDocIndex = -1,
				// The frequency class (0-9) of the previous term.
				freqClass = 0,
				// Temporary variables to hold the next differences in a batch.
				termIndexDelta, docIndexDelta, docPositionDelta,
				// Frequency of the current term
				frequency, 
				// Maximum number of positions of a term in a document
				maxDocPos = 0;

			// Counters (just for bookkeeping).
			long 
				bitsForPointers[] = new long[ 10 ], 
				bitsForPositions[] = new long[ 10 ],
				// Number of occurrences of the current term
				termOccurrences = 0,
				// Total number of pointers
				numPointers = 0;
				
			// Number of pointers for terms with frequency <10%, <20%,...
			long 
				pointersByFrequency[] = new long[10], 
				// Number of occurrences for terms with frequency <10%, <20%,...
				occurrencesByFrequency[] = new long[10];


			final OutputBitStream index = new OutputBitStream( new FileOutputStream( baseName + "index" ), bufferSize );
			final OutputBitStream offsets = new OutputBitStream( new FileOutputStream( baseName + "offsets" ) );
			final IndexWriter indexWriter = skips ? new SkipIndexWriter( index, offsets, N, writerFlags, quantum, height ) :  new IndexWriter( index, offsets, N, writerFlags );

			InputBitStream frequencies = null; 

			properties.setProperty( "compressionflags", writerFlagsString.toString() );
			if ( skips ) {
				properties.setProperty( "skipquantum", String.valueOf( quantum ) );
				properties.setProperty( "skipheight", String.valueOf( height ) );
			}
			else {
				// This is essential to avoid that multiple runs of this class get confused.
				properties.remove( "skipquantum" );
				properties.remove( "skipheight" );
			}
			
			final OutputBitStream termOccs = new OutputBitStream( new FileOutputStream( baseName + "globcounts" ) );
			final ProgressMeter pm = new ProgressMeter( 10, "batches" );

			pm.start();
			
			for(;;) {
				int batch = numBatches;

				System.err.print( "Initialising batches..." );

				final ObjectHeapSemiIndirectPriorityQueue queue = new ObjectHeapSemiIndirectPriorityQueue( occurrence, numBatches ); 

				//We initialise the heap by opening all batches and reading their first triple.
				for( i = 0; i < batch; i++ ) {
					fibs[i] = new InputBitStream( new FileInputStream( baseName + "batch" + i ), bufferSize );
					numOcc[i] = fibs[i].readDelta();
					numOcc[i]--; // We immediately read the first entry.
					occurrence[ i ] = 
						new Occurrence( 
											currTerm[i] =  fibs[i].readDelta(), 
											currDocument[i] = fibs[i].readDelta(), 
											currPosition[i] = fibs[i].readDelta()
											);
					queue.enqueue( i );
					pm.update();
				}
				
				pm.done();

				if ( ! computeFrequencies ) frequencies = new InputBitStream( new FileInputStream ( baseName + "frequencies" ) );
				
				if ( computeFrequencies ) System.err.print( "Computing frequencies..." );
				else System.err.print( "Merging batches..." );

				j = numOccurrences + 1;

				int freq[] = null;

				if ( computeFrequencies ) freq = new int[ numTerms ];

				pm.quantum( 1000000 );
				pm.itemsName( "Koccurrences" );
				pm.start();
				pm.expectedUpdates( numOccurrences / 1000 ); 
				OutputBitStream obs;

				if ( numOccurrences > 0 ) while( j-- != 0 ) { // When j = 0, there is no occurrence, but we flush what we have.

					curr = occurrence[ currBatch = queue.first() ];
					// System.err.println( curr );

					if ( j % 1000 == 0 ) pm.update();

					if ( computeFrequencies ) {
						if ( ( curr.docIndex != prevDocIndex || curr.termIndex != prevTermIndex ) && prevTermIndex != -1 ) freq[ prevTermIndex ]++;
						prevTermIndex = curr.termIndex;
						prevDocIndex = curr.docIndex;
					}
					else {
						if ( curr.docIndex != prevDocIndex // Change of document.
							 || curr.termIndex != prevTermIndex ) { // Change of term.
							
							if ( numPosInDoc != 0 ) { // Not the first round, so we must flush the previous record
								
								// We record the number of occurrences of the term in the current document.
								obs = indexWriter.newDocumentRecord();
								bitsForPointers[ freqClass ] += indexWriter.writeDocumentPointer( obs, prevDocIndex );
								//obs = indexWriter.newDocumentRecord( prevDocIndex );
								
								termOccurrences += numPosInDoc;
								if ( numPosInDoc > maxDocPos ) maxDocPos = numPosInDoc;
								
								if ( indexWriter.hasCounts ) {
									indexWriter.writePositionCount( obs, numPosInDoc );
									if ( indexWriter.hasPositions ) bitsForPositions[ freqClass ] += indexWriter.writeDocumentPositions( obs, occInDoc, 0, numPosInDoc, size[ prevDocIndex ] );
								} 
							}
							
							if ( curr.termIndex != prevTermIndex ) { // A new term.
								
								if ( prevTermIndex != - 1 ) { // If it is not the first round, we emit the number of occurrences of the last term.
									termOccs.writeLongGamma( termOccurrences );
									occurrencesByFrequency[ freqClass ] += termOccurrences;
								}
								
								if ( curr.termIndex != -1 ) { // If it is not the last round, we open a new inverted list.
									numPointers += ( frequency = frequencies.readGamma() );
									freqClass = (int)( 10 * frequency / (double)( N + 1 ) );
									pointersByFrequency[ freqClass ] += frequency;
									
									indexWriter.newInvertedList();
									indexWriter.writeFrequency( frequency );
									
									termOccurrences = 0;
								}
							}
							
							numPosInDoc = 0;
						}
						
						if ( numPosInDoc >= occInDoc.length ) occInDoc = IntArrays.grow( occInDoc, numPosInDoc + 1 );
						occInDoc[ numPosInDoc++ ] = curr.docPosition;
					}

					prevTermIndex = curr.termIndex;
					prevDocIndex = curr.docIndex;

					if ( numOcc[ currBatch ] != 0 ) { // There are other occurrences in the i-th batch.
						numOcc[ currBatch ]--;
						termIndexDelta = fibs[ currBatch ].readUnary();
						docIndexDelta = fibs[ currBatch ].readDelta();
						docPositionDelta = fibs[ currBatch ].readDelta();

						if ( termIndexDelta != 0 ) {
							currDocument[ currBatch ] = 0;
							currPosition[ currBatch ] = -1;
						}
						  
						if ( docIndexDelta != 0 ) {
							currPosition[ currBatch ] = -1;
						}

						curr.termIndex = ( currTerm[ currBatch ] += termIndexDelta );
						curr.docIndex = ( currDocument[ currBatch ] += docIndexDelta );
						curr.docPosition = ( currPosition[ currBatch ] += docPositionDelta + 1 );

						if ( ASSERTS ) {
							assert termIndexDelta >= 0;
							assert docIndexDelta >= 0;
							assert docPositionDelta >= 0;
							assert curr.termIndex >= 0;
							assert curr.docPosition >= 0; 
							assert curr.docIndex >= 0; 
						}

						queue.changed();
					}
					else {
						// We reduce by one the heap, if it is nonempty. Otherwise,
						// we insert a fake occurrence acting as a sentinel.
						queue.dequeue();
						if ( queue.isEmpty() ) {
							occurrence[ 0 ].termIndex = -1;
							queue.enqueue( 0 );
						} 
					}
				}

				pm.done();

				if ( computeFrequencies ) {
					computeFrequencies = false;

					System.err.print( "Writing frequencies..." );
					final OutputBitStream freqs = new OutputBitStream( new FileOutputStream( baseName + "frequencies" ) );
					for( i = 0; i < numTerms; i++ ) freqs.writeGamma( freq[ i ] );
					freqs.close();
					System.err.println( " done." );
				}
				else {
					indexWriter.close();
					offsets.close();
					frequencies.close();
					
					termOccs.writeLongGamma( termOccurrences ); // No one wrote the number of occs of the last term.
					termOccs.close();
					break;
				}
			}

			properties.setProperty( "maxdocpos", String.valueOf( maxDocPos ) );
			properties.setProperty( "indexclass", "it.unimi.dsi.mg4j.index." + ( skips ? "SkipIndex" : "Index" ) );
			final FileOutputStream propertyFile2 = new FileOutputStream( baseName + "properties" );
			properties.store( propertyFile2, "SecondPass properties" );
			propertyFile2.close();

			System.out.println( "Number of documents: " + Fast.format( N ) );
			System.out.println( "Average size: " + Fast.format( averageSize ) );
			System.out.println( "Number of terms: " + Fast.format( numTerms ) );
			System.out.println( "Number of occurrences: " + Fast.format( numOccurrences ) );
			System.out.println( "Frequencies: " + Fast.format( indexWriter.bitsForFrequencies ) + " bits, " + Fast.format( indexWriter.bitsForFrequencies/ (double)numTerms ) + " bits/frequency." );
			System.out.println( "Document pointers: " + Fast.format( numPointers ) + " (" + Fast.format( indexWriter.bitsForPointers ) + " bits, " + Fast.format( indexWriter.bitsForPointers/ (double)numPointers ) + " bits/pointer).");
			if ( indexWriter.hasCounts ) System.out.println( "Counts: " + Fast.format( numPointers ) + " (" + Fast.format( indexWriter.bitsForCounts ) + " bits, " + Fast.format( indexWriter.bitsForCounts/ (double)numPointers ) + " bits/count).");
			if ( indexWriter.hasPositions ) System.out.println( "Occurrences: " + Fast.format( numOccurrences ) + " (" +  Fast.format( indexWriter.bitsForPositions ) + " bits, " + Fast.format( indexWriter.bitsForPositions/ (double)numOccurrences ) + " bits/occurrence).");
	
			if ( skips ) {
				final SkipIndexWriter s = (SkipIndexWriter)indexWriter;
				System.out.println( "Skip towers: " + Fast.format( s.towerData.numberOfSkipTowers ) + " (" + 
					Fast.format( s.towerData.bitsForTowers() ) + " bits [" + 
					Fast.format( s.towerData.bitsForTowers()*100.0/s.writtenBits() ) + "%], " + 
					Fast.format( s.towerData.bitsForTowers()/ (double)s.towerData.numberOfSkipTowers ) + 
					" bits/tower)" );
				System.out.println( "Skip entries: " + Fast.format( s.towerData.numberOfSkipEntries ) + " (" + 
					Fast.format( s.towerData.bitsForEntries()/ (double)s.towerData.numberOfSkipEntries ) + " bits/entry)" );
				System.out.println( "Skip tower lengths: " + Fast.format( s.towerData.bitsForTowerLengths ) + " bits (" + Fast.format( s.towerData.bitsForTowerLengths/ (double)s.towerData.numberOfSkipTowers ) + " bits/tower)" );
				System.out.println( "Quantum bit lengths: " + Fast.format( s.bitsForQuantumBitLengths ) + " bits (" + Fast.format( s.bitsForQuantumBitLengths/ (double)s.numberOfBlocks ) + " bits/block)" );
				System.out.println( "Entry bit lengths: " + Fast.format( s.bitsForEntryBitLengths ) + " bits (" + Fast.format( s.bitsForEntryBitLengths/ (double)s.numberOfBlocks ) + " bits/block)" );

				System.out.println( "Top skip amounts: " + Fast.format( s.towerData.bitsForTopSkipAmounts ) + " bits (" + Fast.format( s.towerData.bitsForTopSkipAmounts/ (double)s.towerData.numberOfSkipTowers ) + " bits/amount)" );
				System.out.println( "Top skip pointers: " + Fast.format( s.towerData.bitsForTopSkipPointers ) + " bits (" + Fast.format( s.towerData.bitsForTopSkipPointers/ (double)s.towerData.numberOfSkipTowers ) + " bits/pointer)" );
				System.out.println( "Lower skip amounts: " + Fast.format( s.towerData.bitsForLowerSkipAmounts ) + " bits (" + Fast.format( s.towerData.bitsForLowerSkipAmounts/ (double)s.towerData.numberOfLowerEntries() ) + " bits/amount)" );
				System.out.println( "Lower skip pointers: " + Fast.format( s.towerData.bitsForLowerSkipPointers ) + " bits (" + Fast.format( s.towerData.bitsForLowerSkipPointers/ (double)s.towerData.numberOfLowerEntries() ) + " bits/pointer)" );
				System.out.println( "Skip amounts: " + Fast.format( s.towerData.bitsForSkipAmounts() ) + " bits (" + Fast.format( s.towerData.bitsForSkipAmounts()/ (double)s.towerData.numberOfSkipEntries ) + " bits/amount)" );
				System.out.println( "Skip pointers: " + Fast.format( s.towerData.bitsForSkipPointers() ) + " bits (" + Fast.format( s.towerData.bitsForSkipPointers()/ (double)s.towerData.numberOfSkipEntries ) + " bits/pointer)" );
			}
			System.out.println( "Alignment: " + Fast.format( indexWriter.bitsForAlignment ) + " bits, "  + Fast.format( indexWriter.bitsForAlignment/ (double)numTerms ) + " bits/alignment." );

			if ( indexWriter.hasPositions ) System.out.println( "Total: " + Fast.format( indexWriter.writtenBits() ) + " bits, " + Fast.format( indexWriter.writtenBits()/ (double)numOccurrences ) + " bits/occurrence" );
			else System.out.println( "Total: " + Fast.format( indexWriter.writtenBits() ) + " bits, " + Fast.format( indexWriter.writtenBits()/ (double)numOccurrences ) + " bits/pointer" );

			System.out.println();
			System.out.println("Number of pointers by frequency:");
			for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * pointersByFrequency[ i ] / numPointers ) + "%");
			System.out.println();

			System.out.println("Number of occurrences by frequency:");
			for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * occurrencesByFrequency[ i ] / numOccurrences ) + "%" ); 

			System.out.println();
			System.out.println("Number of bits for pointers by frequency:");
			for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * bitsForPointers[ i ] / indexWriter.bitsForPointers ) + "%");
			System.out.println();

			if ( indexWriter.hasPositions ) {
				System.out.println("Number of bits for occurrences by frequency:");
				for( i = 0; i < 10; i++ ) System.out.println( "<" + (i+1)*10 + "%: " + Fast.format( 100.0 * bitsForPositions[ i ] / indexWriter.bitsForPositions ) + "%" ); 
			}				
		} catch( IOException dontCare ) { dontCare.printStackTrace(); }
    }
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
