package it.unimi.dsi.mg4j.tool;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.objects.ObjectArrays;
import it.unimi.dsi.fastutil.objects.ObjectHeapSemiIndirectPriorityQueue;
import it.unimi.dsi.fastutil.objects.ObjectIterators;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;
import it.unimi.dsi.fastutil.objects.ObjectSet;
import it.unimi.dsi.mg4j.index.IndexProperties;
import it.unimi.dsi.mg4j.io.FastBufferedReader;
import it.unimi.dsi.mg4j.util.Fast;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.Properties;


/** Builds the list of terms appearing in a sequence of documents read from standard input.
 *
 * <P>This class reads from standard input a sequence of documents and produces
 * two files containing the terms appearing in the documents, both in appearance and sorted order.
 * The only mandatory argument is a <em>basename</em>, which will be used to
 * stem the names of all files generated.
 * 
 * <P>By default, the whole set of terms contained in the document collection is accumulated in a
 * set, and then dumped. For large collections, it is possible to limit the amount of memory used
 * by dumping at regular intervals the set of terms accumulated so far. The sets are then merged.
 * Note that in this case it is possible that the unsorted list of terms contains duplicates.
 * 
 * <P>These are the files currently generated:
 *
 * <dl>
 *
 * <dt><samp><var>basename</var>.terms.unsorted[.dups]</samp>
 *
 * <dd>For each indexed term, the corresponding literal string in UTF-8 encoding. More precisely,
 * the <var>i</var>-th line of the file (starting from 0) contains the literal
 * string corresponding to term index <var>i</var>. Terms are in appearance order.
 * If several dumps were necessary to build the index, a suffix <samp>.dups</samp> is added
 * to remember that the list may contain duplicates.
 *
 * <dt><samp><var>basename</var>.terms</samp>
 *
 * <dd>For each indexed term, the corresponding literal string in UTF-8 encoding. More precisely,
 * the <var>i</var>-th line of the file (starting from 0) contains the literal
 * string corresponding to term index <var>i</var>. Terms are sorted lexicographically.
 * 
 * <dt><samp><var>basename</var>.properties</samp>
 *
 * <dd>A Java {@link java.util.Properties property file} containing information
 * about the index. Currently, the following keys are generated:
 * 
 * <dl>
 * <dt>basename
 * <dd>the basename (see above);
 * <dt>documents
 * <dd>number documents in the collection;
 * <dt>terms
 * <dd>number of indexed terms.
 * </dl>

 * </dl>
 * @author Sebastiano Vigna
 * @since 0.9
 */

final public class ZerothPass {

	private ZerothPass() {}
	private final static String THIS_CLASS_NAME = new ZerothPass().getClass().getName();

	/** Merges the given files, which must contain UTF-8 sorted lists.
	 * 
	 * @param inputFilename an array of filenames to be merged.
	 * @param outputFilename the output filename
	 * @param bufferSize the buffer size.
	 * @return the overall number of lines in the output file.
	 */

	private static int mergeSortedFiles( final String[] inputFilename, final String outputFilename, final int bufferSize, final ProgressMeter pm ) throws IOException {
		final int n = inputFilename.length;
		final MutableString[] string = new MutableString[ n ];
		final FastBufferedReader[] reader = new FastBufferedReader[ n ];  
		
		final ObjectHeapSemiIndirectPriorityQueue queue = new ObjectHeapSemiIndirectPriorityQueue( string, n ); 

		//We initialise the heap by opening all batches and reading their first triple.
		for( int i = 0; i < n; i++ ) {
			reader[ i ] = new FastBufferedReader( new InputStreamReader( new FileInputStream( inputFilename[ i ] ), "UTF-8" ), bufferSize );
			reader[ i ].readWord( string[ i ] = new MutableString() );
			if ( string[ i ].length() != 0 ) queue.enqueue( i );
		}

		final PrintWriter writer = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( outputFilename ), "UTF-8" ), bufferSize ) );

		int curr, numTerms = 0;
		final MutableString prev = new MutableString();

		if ( pm != null ) {
			pm.start( "Merging..." );
			pm.itemsName( "items" );
		} 

		while( ! queue.isEmpty() ) {
			curr = queue.first();
			if ( ! string[ curr ].equals( prev ) ) {
				numTerms++;
				string[ curr ].println( writer );
				prev.replace( string[ curr ] );
			}
			reader[ curr ].readWord( string[ curr ] );
			if ( string[ curr ].length() == 0 ) queue.dequeue();
			else queue.changed();
			if ( pm != null ) pm.update();
		}
		
		if ( pm != null ) pm.done();
		
		writer.close();
		return numTerms;
	}

	/** Dumps the terms contained in the given set after sorting them. 
	 * 
	 * @param terms a set of terms.
	 * @param filename an output filename.
	 * @param bufferSize the buffer size.
	 * @param support a support array (will be enlarged, if necessary, but it must not be <code>null</code>).
	 */

	private static void dumpSortedTerms( final ObjectSet terms, final String filename, final int bufferSize, MutableString[] support ) throws UnsupportedEncodingException, FileNotFoundException {
		final int n = terms.size();
		support = (MutableString[]) ObjectArrays.ensureCapacity( support, n, 0 );
		ObjectIterators.unwrap( terms.iterator(), support );

		System.err.print( "Sorting terms..." );
		Arrays.sort( support, 0, n );

		System.err.print( " saving..." );

		PrintWriter termFile = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( filename ), "UTF-8" ), bufferSize ) );
		for( int i = 0; i < n; i++ ) support[ i ].println( termFile );
		termFile.close();
		
		ObjectArrays.fill( support, null );

		System.err.print( " done." );
	}


    public static void main( String arg[] ) throws UnsupportedEncodingException, FileNotFoundException {

		int ch;
		char delimiter = 10;
		String charSet = "UTF-8";
		int quantum = 10000;
		int bufferSize;
		int maxBatchTerms = Integer.MAX_VALUE;
		String bufferSizespec = "64Ki", basename;
		boolean doUnsorted = true;
		boolean downcase = false;

		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
			new LongOpt( "delimiter", LongOpt.REQUIRED_ARGUMENT, null, 'd' ),
			new LongOpt( "charset", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
			new LongOpt( "max-batch-terms", LongOpt.REQUIRED_ARGUMENT, null, 'm' ),
			new LongOpt( "dot-meter-quantum", LongOpt.REQUIRED_ARGUMENT, null, 'q' ),
			new LongOpt( "no-unsorted", LongOpt.NO_ARGUMENT, null, 'n' ),
			new LongOpt( "downcase", LongOpt.NO_ARGUMENT, null, 1 ),
		};

		Getopt getopt = new Getopt( THIS_CLASS_NAME, arg, "nq:c:d:b:m:h", longopts );
		getopt.setOpterr( true );

		while ( ( ch = getopt.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: " + THIS_CLASS_NAME + " [OPTIONS] BASENAME" );
			System.err.println( "Compute terms from a list of documents." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  BASENAME    all file names will be stemmed from this name" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -c, --charset             the input charset encoding (default: " + charSet + ")" );
			System.err.println( "  -b, --buffer-size         the size of an I/O buffer (default: " + bufferSizespec + ")" );
			System.err.println( "  -d, --delimiter           the Unicode index of the document delimiter (default: " + (int)delimiter + ")" );
			System.err.println( "  -n, --no-unsorted         do not generate BASENAME.terms.unsorted" );
			System.err.println( "  -m, --max-batch-terms     maximum number of terms in a batch (default: unlimited)" );
			System.err.println( "  -q, --dot-meter-quantum   the quantum of the process meter (default: " + quantum + ")" );
			System.err.println( "  --downcase                downcase all words" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help        print this help screen" );
			System.err.println( "" );
			return;

		case 'b':
			bufferSizespec = getopt.getOptarg();
			break;

		case 'c':
			charSet = getopt.getOptarg();
			break;

		case 'd':
			delimiter = (char)Integer.parseInt( getopt.getOptarg() );
			break;

		case 'q':
			quantum = Integer.parseInt( getopt.getOptarg() );
			break;

		case 'm':
			maxBatchTerms = Integer.parseInt( getopt.getOptarg() );
			break;

		case 'n':
			doUnsorted = false;
			break;

		case 1:
			downcase = true;
			break;

		case '?':
			return;

		default:
			break;
		}

		bufferSize = Fast.parseIntSize( bufferSizespec );

		if ( arg.length - getopt.getOptind() == 1 ) basename = arg[ getopt.getOptind() ] + ".";
		else {
			System.err.println( "Wrong number (" + ( arg.length - getopt.getOptind() ) + ") of arguments." );
			return;
		}

		System.err.println("Documents will be separated by the Unicode character " + (int)delimiter );

		// We bufferize the standard input.
		final FastBufferedReader input = new FastBufferedReader( new InputStreamReader( System.in, charSet ), bufferSize );
		
		// The word we are scanning.
		final MutableString word = new MutableString();

		// The set of terms (for this batch). We want it to be *fast*, so we set a low load factor.
		ObjectOpenHashSet terms = new ObjectOpenHashSet( maxBatchTerms != Integer.MAX_VALUE ? maxBatchTerms : Hash.DEFAULT_INITIAL_SIZE, .5f );
		MutableString[] support = maxBatchTerms != Integer.MAX_VALUE ? new MutableString[ maxBatchTerms ] : new MutableString[ 0 ]; 

		int i, t, numTerms = 0, numBatches = 0;
		boolean firstWordUnseen = true;
		final ProgressMeter pm = new ProgressMeter( quantum );
		pm.itemsName( "documents" );

		pm.start( "Indexing terms..." );

		PrintWriter termFile = null;
		if ( doUnsorted ) termFile = new PrintWriter( new BufferedWriter( new OutputStreamWriter( new FileOutputStream( basename + "terms." + ( maxBatchTerms != Integer.MAX_VALUE ? "unsorted.dups" : "unsorted" ) ), "UTF-8" ), bufferSize ) );

		try {
			while( ( t = input.readWord( word ) ) != -1 ) {
				
				if ( downcase ) word.toLowerCase();
				firstWordUnseen = false;
				
				if ( word.length() != 0 ) {
					// We check whether we have already seen this term. If not, we add it to termIndices.
					if ( ! terms.contains( word ) ) {
						terms.add( word.copy() );
						numTerms++;
						
						if ( numTerms % 1000000 == 0 ) System.err.print( "[" + Fast.format( numTerms ) + " terms]" );

						if ( numTerms == maxBatchTerms ) {
							System.err.print( '[' );
							dumpSortedTerms( terms, basename + "terms." + numBatches++, bufferSize, support );
							System.err.print( ']' );
							terms.clear();
							numTerms = 0;
						}

						if ( doUnsorted ) word.println( termFile );
					}
				}

				if ( t == delimiter ) {// New document
					pm.update();
					firstWordUnseen = true;
				} 
				
			}

			if ( ! firstWordUnseen ) pm.update();

			if ( maxBatchTerms != Integer.MAX_VALUE ) {
				System.err.print( '[' );
				dumpSortedTerms( terms, basename + "terms." + numBatches++, bufferSize, support );
				System.err.print( ']' );
				terms = null;
			}

			if ( doUnsorted ) termFile.close();

			pm.done();

			if ( maxBatchTerms == Integer.MAX_VALUE ) {
				dumpSortedTerms( terms, basename + "terms", bufferSize, support );
				System.err.println();
				terms = null;
			}
			else {
				String[] inputFilename = new String[ numBatches ];
				for( i = 0; i < numBatches; i++ ) inputFilename[ i ] = basename + "terms." + i;
				pm.quantum( quantum * 100 );
				numTerms = mergeSortedFiles( inputFilename, basename + "terms", bufferSize, pm );							
				for( i = 0; i < numBatches; i++ ) new File( inputFilename[ i ] ).delete();
			}

			// Finally, we save all data related to this index in a property file.
			final Properties properties = new Properties();
			properties.setProperty( "basename", basename.substring( 0, basename.length() -1 ) );
			properties.setProperty( IndexProperties.DOCUMENTS, String.valueOf( pm.count() ) );
			properties.setProperty( IndexProperties.TERMS, String.valueOf( numTerms ) );
			final FileOutputStream propertyFile = new FileOutputStream( basename + "properties" );
			properties.store( propertyFile, "ZerothPass properties" );
			propertyFile.close();

		}
		catch( IOException dontCare ) { dontCare.printStackTrace(); }
    }
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
