package it.unimi.dsi.mg4j.util;

import java.io.Serializable;

import cern.colt.bitvector.QuickBitVector;
import cern.jet.random.engine.MersenneTwister;

/** A Bloom filter.
 * 
 * <P>Instances of this class represent a set of character sequences (with false positives)
 * using a Bloom filter. Because of the way Bloom filters work,
 * you cannot remove elements.
 *
 * <P>Bloom filters have an expected error rate, depending on the number
 * of hash functions used, on the filter size and on the number of elements in the filter. This implementation
 * uses a variable optimal number of hash functions, depending on the expected
 * number of elements. More precisely, a Bloom
 * filter for <var>n</var> character sequences with <var>d</var> hash functions will use 
 * ln 2 <var>d</var><var>n</var> &#8776; 1.44 <var>d</var><var>n</var> bits;
 * false positives will happen with probability 2<sup>-<var>d</var></sup>.  
 *
 * <P>Hash functions are generated at creation time using universal hashing. Each hash function
 * uses {@link #NUMBER_OF_WEIGHTS} random integers, which are cyclically multiplied by
 * the character codes in a character sequence. The resulting integers are XOR-ed together.
 *
 * <P>This class exports access methods that are very similar to those of {@link java.util.Set},
 * but it does not implement that interface, as too many non-optional methods
 * would be unimplementable (e.g., iterators).
 *
 * @author Sebastiano Vigna
 */
public class BloomFilter implements Serializable {
	/** The number of weights used to create hash functions. */
	final public static int NUMBER_OF_WEIGHTS = 16;
	/** The number of bits in this filter. */
	final public int m;
	/** The number of hash functions used by this filter. */
	final public int d;
	/** The underlying bit vector. */
	final private long[] bits;
	/** The random integers used to generate the hash functions. */
	final private int[][] weight;

	/** The number of elements currently in the filter. It may be
	 * smaller than the actual number of additions of distinct character
	 * sequences because of false positives.
	 */
	private int size;

	/** The natural logarithm of 2, used in the computation of the number of bits. */
	private final static double NATURAL_LOG_OF_2 = Math.log( 2 );

	private final static boolean DEBUG = false;

	/** Creates a new Bloom filter with given number of hash functions and expected number of elements.
	 * 
	 * @param n the expected number of elements.
	 * @param d the number of hash functions; if the filter add not more than <code>n</code> elements,
	 * false positives will happen with probability 2<sup>-<var>d</var></sup>.
	 */
	public BloomFilter( final int n, final int d ) {
		this.d = d;
		bits = new long[ (int)Math.ceil( ( n * d / NATURAL_LOG_OF_2 ) / 64 ) ];
		if ( bits.length > Integer.MAX_VALUE / 64 ) throw new IllegalArgumentException( "This filter would require " + bits.length * 64L + " bits" );
		m = bits.length * 64;

		if ( DEBUG ) System.err.println( "Number of bits: " + m );
		
		final MersenneTwister mersenneTwister = new MersenneTwister();
		weight = new int[ d ][];
		for( int i = 0; i < d; i++ ) {
			weight[ i ] = new int[ NUMBER_OF_WEIGHTS ];
			for( int j = 0; j < NUMBER_OF_WEIGHTS; j++ )
				 weight[ i ][ j ] = mersenneTwister.nextInt();
		}
	}

	/** The number of character sequences in the filter.
	 * 
	 * @return the number of character sequences in the filter (but see {@link #contains(CharSequence)}).
	 */

	public int size() {
		return size;
	}

	/** Hashes the given sequence with the given hash function.
	 * 
	 * @param s a character sequence.
	 * @param l the length of <code>s</code>.
	 * @param k a hash function index (smaller than {@link #d}).
	 * @return the position in the filter corresponding to <code>s</code> for the hash function <code>k</code>.
	 */

	private int hash( final CharSequence s, final int l, final int k ) {
		final int[] w = weight[ k ];
		int h = 0, i = l;
		while( i-- != 0 ) h ^= s.charAt( i ) * w[ i % NUMBER_OF_WEIGHTS ];
		return ( h & 0x7FFFFFFF ) % m; 
	}

	/** Checks whether the given character sequence is in this filter. 
	 * 
	 * <P>Note that this method may return true on a character sequence that is has
	 * not been added to the filter. This will happen with probability 2<sub>-<var>d</var></sub>,
	 * where <var>d</var> is the number of hash functions specified at creation time, if
	 * the number of the elements in the filter is less than <var>n</var>, the number
	 * of expected elements specified at creation time.
	 * 
	 * @param s a character sequence.
	 * @return true if the sequence is in the filter (or if a sequence with the
	 * same hash sequence is in the filter).
	 */

	public boolean contains( final CharSequence s ) {
		int i = d, l = s.length();
		while( i-- != 0 ) if ( ! QuickBitVector.get( bits, hash( s, l, i ) ) ) return false;
		return true;
	}

	/** Adds a character sequence to the filter.
	 * 
	 * @param s a character sequence.
	 * @return true if the character sequence was not in the filter (but see {@link #contains(CharSequence)}).
	 */

	public boolean add( final CharSequence s ) {
		boolean result = false;
		int i = d, l = s.length(), h;
		while( i-- != 0 ) {
			h = hash( s, l, i );
			if ( ! QuickBitVector.get( bits, h ) ) result = true;
			QuickBitVector.set( bits, h );
		}
		if ( result ) size++;
		return result;
	}
}
