package it.unimi.dsi.mg4j.util;

import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.zip.CRC32;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002-2003 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** CRC-32 signed order-preserving minimal perfect hash tables.
 *
 * <P>This class source exemplifies a signed minimal perfect hash table that
 * signes each word with a 32-bit CRC, thus avoiding false positives with high
 * probability.
 * 
 * @author Sebastiano Vigna
 * @author Marco Olivo
 * @since 0.4
 */

public class CRC32SignedMinimalPerfectHash extends SignedMinimalPerfectHash {

    transient java.util.zip.CRC32 crc32;
    long signature[];

	/** Creates a new CRC32-signed order-preserving minimal perfect hash table 
	 * for the given set of words using the given number of weights.
	 *
	 * @param words some words to hash; it is assumed that this {@link
	 * Collection} does not contain words with a common prefix of
	 * <code>weightLength</code> characters.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions.
	 * @see MinimalPerfectHash#MinimalPerfectHash(Collection,int)
	 */
	public CRC32SignedMinimalPerfectHash( Collection words, int weightLength ) {
		super( words, weightLength );
	}

	/** Creates a new CRC32-signed order-preserving minimal perfect hash table for the given
	 * set of words, using as many weights as the longest word in the collection.
     *
	 * @param words some words to hash; it is assumed that this {@link Collection} does not contain duplicates.
	 * @see MinimalPerfectHash#MinimalPerfectHash(Collection)
	 *	 
	 */
    public CRC32SignedMinimalPerfectHash( Collection words ) {
		super( words );
    }

	/** Creates a new CRC32-signed order-preserving minimal perfect hash table for the given file 
	 * of words using the given number of weights.
	 *
	 * @param wordFile an UTF-8 file containing one word on each line; it is assumed that 
	 * it does not contain words with a common prefix of
	 * <code>weightLength</code> characters.
	 * @param encoding the encoding of <code>wordFile</code>; if <code>null</code>, it
	 * is assumed to be the platform default encoding.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions.
	 * @see MinimalPerfectHash#MinimalPerfectHash(String,String,int)
	 */
	public CRC32SignedMinimalPerfectHash( String wordFile, String encoding, int weightLength ) {
		super( wordFile, encoding, weightLength );
	}

	/** Creates a new CRC32-signed order-preserving minimal perfect hash table for
	 * the given file of words, using as many weights as the longest word in the file.
	 *
	 * @param wordFile a file in the platform default encoding containing one word on each line; it is assumed that the file does
	 * not contain twice the same word.
	 * @see MinimalPerfectHash#MinimalPerfectHash(String)
	 */
	public CRC32SignedMinimalPerfectHash( String wordFile ) {
		super( wordFile );
	}



	public void initSignatures( Collection words ) {
		CharSequence s;
		int l;
		char c;
		int i = 0;

		crc32 = new CRC32();
		signature = new long[ words.size() ];

		for ( Iterator e = words.iterator(); e.hasNext(); i++ ) {
			s = (CharSequence) e.next();
			crc32.reset();
			l = s.length();
			while ( l-- != 0 ) {
				c = s.charAt( l );
				crc32.update( (byte) c );
				crc32.update( (byte) ( c >> 8 ) );
			}
			signature[ i ] = crc32.getValue();
		}
	}

	public boolean checkSignature( CharSequence word, int index ) {
		char c;
		int n;

		crc32.reset();
		n = word.length();
		while ( n-- != 0 ) {
			c = word.charAt( n );
			crc32.update( (byte) c );
			crc32.update( (byte) ( c >> 8 ) );
		}
		return signature[ index ] == crc32.getValue();
	}


	private void readObject(java.io.ObjectInputStream s) throws IOException, ClassNotFoundException {
		s.defaultReadObject();
		crc32 = new CRC32();
	}

}
