package it.unimi.dsi.mg4j.util;

import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.objects.Object2LongOpenHashMap;
import it.unimi.dsi.mg4j.io.FastBufferedInputStream;
import it.unimi.dsi.mg4j.io.FastBufferedOutputStream;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002, 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/** All-purpose optimised static-method container class.
 *
 * <P>This class contains static optimised utility methods that are used by all
 * MG4J classes.
 *
 * @author Sebastiano Vigna
 * @since 0.1
 */

public final class Fast {

	private Fast() {}

	/** Computes the most significant bit of an integer.
	 *
	 * @param x an integer.
	 * @return the most significant bit of the argument (-1 for 0).
	 */

	public static int mostSignificantBit( int x ) {
		int bit = 0;
		while ( ( x & 0xFFFFFF00 ) != 0 ) {
			bit += 8;
			x >>>= 8;
		}

		return bit + BYTEMSB[ x ];
	}

	/** Computes the most significant bit of a long integer.
	 *
	 * @param x a long integer.
	 * @return the most significant bit of the argument (-1 for 0 ).
	 */

	public static int mostSignificantBit( long x ) {
		int bit = 0;
		while ( ( x & 0xFFFFFFFFFFFFFF00L ) != 0 ) {
			bit += 8;
			x >>>= 8;
		}

		return bit + BYTEMSB[ (int)x ];
	}

	/** Computes the least significant bit of an integer.
	 *
	 * @param x an integer.
	 * @return the least significant bit of the argument (-1 for 0).
	 */

	public static int leastSignificantBit( int x ) {
		if ( x == 0 ) return -1;
		int bit = 0;
		while ( ( x & 0xFF ) == 0 ) {
			bit += 8;
			x >>>= 8;
		}

		return bit + BYTELSB[ x & 0xFF ];
	}

	/** Computes the least significant bit of a long integer.
	 *
	 * @param x a long integer.
	 * @return the least significant bit of the argument (-1 for 0).
	 */

	public static int leastSignificantBit( long x ) {
		if ( x == 0 ) return -1;
		int bit = 0;
		while ( ( x & 0xFF ) == 0 ) {
			bit += 8;
			x >>>= 8;
		}

		return bit + BYTELSB[ (int)x & 0xFF ];
	}

 
	public final static double[] GOLOMB_STEP = {
		.3819660112501052,
		 .245122333753307,
		.1808274866038356,
		.1433251161454971,
		.1187285383664304,
		.1013462873713007,
		.0884076465179451,
		.0784006803660170,
		.0704298717679771,
		.0639308889222416,
		.0585303826783648,
		.0539714717143864,
		.0500716000363801,
		.0466974625983358,
		.0437494423620109,
	};
		
	public final static int GOLOMB_STEP_LENGTH = GOLOMB_STEP.length;
	public final static double 
		GOLOMB_THRESHOLD = .0411515989924386,
		GOLOMB_ADD =  -( 1 + Math.log( 2 ) ) / 2,
		GOLOMB_MULT = Math.log( 2 );

	/** Computes the Golomb modulus for a given frequency using
	 * the formula {@link #GOLOMB_ADD} + {@link #GOLOMB_MULT} / <code>p</code>
	 * if <code>p</code> is smaller than {@link #GOLOMB_THRESHOLD},
	 * scanning {@link #GOLOMB_STEP} otherwise. This gives results that are
	 * extremely close to &lceil; log( 2 - <code>p</code> ) / log( 1 - <code>p</code> ) &rceil;,
	 * but order of magnitudes more quickly.
	 * 
	 * @param p the frequency divided by the overall number of documents.
	 * @return the Golomb modulus for the given ratio.
	 */
	
	
	public static int golombModulus( double p ) {
		if ( p < GOLOMB_THRESHOLD ) return (int)Math.ceil( GOLOMB_ADD + GOLOMB_MULT / p );
		int i = GOLOMB_STEP_LENGTH;
		while( i-- != 0 ) if ( p < GOLOMB_STEP[ i ] ) return i + 2;
		return 1;
	}

	protected final static double GOLOMB_GAUSSIAN = 2 * Math.sqrt( 2 / Math.PI ) * Math.log( 2 );

	/** Computes the Golomb modulus for (positive and negative) 
	 * integers normally distributed with given standard deviation using
	 * the formula &lceil; 2 sqrt( 2 / &pi; ) ln(2) &sigma; &rceil;.
	 *
	 * <P>The resulting Golomb modulus is near to optimal for coding such
	 * integers after they have been passed through {@link #int2nat(int)}. Note,
	 * however, that Golomb coding is <em>not</em> optimal for a normal distribution.
	 *
	 * @param sigma the standard deviation.
	 * @return the Golomb modulus for the given standard deviation.
	 */
	public static int gaussianGolombModulus( double sigma ) {
		return (int)Math.ceil( GOLOMB_GAUSSIAN * sigma );
	}

	/** Maps integers bijectively into natural numbers.
	 * 
	 * <P>This method will map a negative integer <var>x</var> to -2<var>x</var>-1 and
	 * a nonnegative integer <var>x</var> to 2<var>x</var>. It can be used to save arbitrary
	 * integers using the standard coding methods (which all work on natural numbers).
	 * 
	 * <P>The inverse of the above map is computed by {@link #nat2int(int)}
	 *
	 * @param x an integer.
	 * @return the argument mapped into natural numbers.
	 * @see #nat2int(int)
	 */

	public static int int2nat( int x ) {
		return x >= 0 ? x << 1 : ( -x << 1 ) - 1;
	}

	/** Maps natural numbers bijectively into integers.
	 * 
	 * <P>This method computes the inverse of {@link #int2nat(int)}
	 *
	 * @param x a natural  number.
	 * @return the argument mapped into an integer.
	 * @see #int2nat(int)
	 */

	public static int nat2int( int x ) {
		return x % 2 == 0 ? x >> 1 : -( ( x + 1 ) >> 1 );
	}

	/** A reasonable format for real numbers. */
	private static java.text.NumberFormat formatDouble = new java.text.DecimalFormat( "#,##0.00" );
	
	/** Formats a number.
	 *
	 * <P>This method formats a double separating thousands and printing just two fractional digits.
	 * @param d a number.
	 * @return a string containing a pretty print of the number.
	 */
	public static String format( final double d ) {
		final StringBuffer s = new StringBuffer();
		return formatDouble.format( d, s, new java.text.FieldPosition( 0 ) ).toString();
	}
	
	/** A reasonable format for integers. */
	private static java.text.NumberFormat formatLong = new java.text.DecimalFormat( "#,###" );
	
	/** Formats a number.
	 *
	 * <P>This method formats a double separating thousands and printing just two fractional digits.
	 * @param l a number.
	 * @return a string containing a pretty print of the number.
	 */
	public static String format( final long l ) {
		final StringBuffer s = new StringBuffer();
		return formatLong.format( l, s, new java.text.FieldPosition( 0 ) ).toString();
	}
	
	private static final Pattern PARSE_SIZE_REGEX = Pattern.compile( "(\\d+)([KMGT]i?)?" );

	private static final Object2LongOpenHashMap UNIT2SIZE = new Object2LongOpenHashMap();

	static {
		UNIT2SIZE.put( "K", (long)1E3 );
		UNIT2SIZE.put( "M", (long)1E6 );
		UNIT2SIZE.put( "G", (long)1E9 );
		UNIT2SIZE.put( "T", (long)1E12 );
		UNIT2SIZE.put( "Ki", 1L << 10 );
		UNIT2SIZE.put( "Mi", 1L << 20 );
		UNIT2SIZE.put( "Gi", 1L << 30 );
		UNIT2SIZE.put( "Ti", 1L << 40 );
	}

	/** Parses a size specified using units (e.g., K, Ki, M, Mi,&hellip;).
	 *
	 * The argument must be in the form <var>number</var> [<var>unit</var>] (with
	 * no space inbetween). Unit may be one of K (10<sup>3</sup>), Ki (2<sup>10</sup>), M (10<sup>6</sup>), Mi (2<sup>20</sup>), 
	 * G (10<sup>9</sup>), Gi (2<sup>30</sup>), T (10<sup>12</sup>), Ti (2<sup>40</sup>).
	 * 
	 * @param s a size specified as above.
	 * @return the corresponding number.
	 * @throws IllegalArgumentException if <var>number</var> is negative or <var>unit</var> is not
	 * among the abovementioned ones.
	 */
		
	public static long parseSize( CharSequence s ) {
		final Matcher m = PARSE_SIZE_REGEX.matcher( s );
		long unitSize = 1;
		long size;
		if ( ! m.matches() ) throw new IllegalArgumentException( "Invalid size specification " + s );

		if ( m.group( 2 ) != null ) unitSize = UNIT2SIZE.getLong( m.group( 2 ) );
		if ( unitSize == 0 ) throw new IllegalArgumentException( "Invalid unit specification " + m.group( 2 ) );

		try {
			size = Integer.parseInt( m.group( 1 ) );
			if ( size < 0 ) throw new NumberFormatException();
		}
		catch( NumberFormatException e ) {
			throw new IllegalArgumentException( "Invalid size specification " + m.group( 1 ) );
		}
		
		return size * unitSize;
	}

	/** Parses a size specified using units (e.g., K, Ki, M, Mi,&hellip;), ensuring
	 * that the results fits an integer.
	 *
	 * @param s a size specified as above.
	 * @return the corresponding number.
	 * @throws IllegalArgumentException {@link #parseSize(CharSequence)} would, or if
	 * the resulting size exceeds {@link Integer#MAX_VALUE}.
	 * @see #parseSize(CharSequence)
	 */
	public static int parseIntSize( CharSequence s ) {
		final long size = parseSize( s );
		if ( size > Integer.MAX_VALUE ) throw new IllegalArgumentException( "Size specification " + s + " is larger than " + Integer.MAX_VALUE );
		return (int)size;
	}


	/** Precomputed least significant bits for bytes (-1 for 0 ). */
	public static final int[] BYTELSB = {
		-1, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		7, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0
	};
	    

	/** Precomputed most significant bits for bytes (-1 for 0 ). */
	public static final int[] BYTEMSB = {
		-1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 
		4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 
		5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
		5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7
	};

	/** Stores an object in a given file.
	 *
	 * @param o an object.
	 * @param filename a filename.
	 */

	public static void storeObject( final Object o, final String filename ) throws FileNotFoundException, IOException {
		final ObjectOutputStream oos = new ObjectOutputStream( new FastBufferedOutputStream( new FileOutputStream( filename ), 64 * 1024 ) );
		oos.writeObject( o );
		oos.close();
	}

	/** Loads an object from a given file.
	 *
	 * @param filename a filename.
	 */
	public static Object loadObject( final String filename ) throws IOException, ClassNotFoundException {
		final ObjectInputStream ois = new ObjectInputStream( new FastBufferedInputStream( new FileInputStream( filename ), 64 * 1024 ) );
		final Object result = ois.readObject();
		ois.close();
		return result;
	}


	/** Stores an array of integers in a file.
	 *
	 * @param filename a filename.
	 * @param array an array of integers.
	 * @see #storeInts(int[], int, int, String, ProgressMeter)
	 */
	public static void storeInts( final int array[], final String filename ) throws IOException {
		storeInts( array, 0, array.length, filename, null );
	}

	/** Stores an integer-array fragment in a file.
	 * 
	 * <P>This method scans <code>length</code> integers contained in <code>array</code>, starting
	 * from the position <code>offset</code>, and write them to a {@link DataOutputStream} opened
	 * using the given filename. 
	 *
	 * @param filename the filename.
	 * @param array an array.
	 * @param offset the position of the first integer to store.
	 * @param length how many integers to store.
	 * @param pm a progress meter, or <code>null</code>.
	 * @see #loadInts(String, ProgressMeter)
	 */
	public static void storeInts( int array[], int offset, int length, String filename, ProgressMeter pm ) throws IOException {
		IntArrays.ensureOffsetLength( array, offset, length );
		DataOutputStream dos = new DataOutputStream( new FastBufferedOutputStream( new FileOutputStream( filename ), 64 * 1024 ) );

		if ( pm != null ) {
			pm.start();
			pm.itemsName( "ints" );
		} 

		for( int i = 0; i < length; i++ ) dos.writeInt( array[ offset + i ] );

		dos.close();
		
		if ( pm != null ) {
			pm.count( length );
			pm.stop();
		}
	}



	/** Loads a list of integers from a data-input file stream.
	 *
	 * <P>This method considers the given filename as a list of integers saved
	 * in {@link java.io.DataOutput} format.
	 *
	 * @param filename a filename.
	 * @return an array of integers, loaded from the given file opened as a {@link java.io.DataInput}.
	 */
	public static int[] loadInts( final String filename ) throws IOException {
		return loadInts( filename, null );
	}

	/** Loads a list of integers from a data-input file stream.
	 *
	 * <P>This method considers the given filename as a list of integers saved
	 * in {@link java.io.DataOutput} format.
	 *
	 * @param filename a filename.
	 * @param pm a progress meter, or <code>null</code>.
	 * @return an array of integers, loaded from the given file opened as a {@link java.io.DataInput}.
	 * @see #storeInts(int[], int, int, String, ProgressMeter)
	 * 
	 */
	public static int[] loadInts( final String filename, final ProgressMeter pm ) throws IOException {
		final FileInputStream fis = new FileInputStream( filename );
		final DataInputStream dos = new DataInputStream( new FastBufferedInputStream( fis, 64 * 1024 ) );

		if ( pm != null ) {
			pm.start();
			pm.itemsName( "ints" );
		} 

		final int length = (int)(fis.getChannel().size() / 4);
		final int[] vector = new int[ length ];
		for( int i = 0; i < length; i++ ) vector[ i ] = dos.readInt();

		dos.close();
		if ( pm != null ) {
			pm.count( length );
			pm.stop();
		}
		
		return vector;
	}
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
