/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002, 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
package it.unimi.dsi.mg4j.util;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.bytes.ByteArrayFrontCodedList;
import it.unimi.dsi.fastutil.chars.CharArrayFrontCodedList;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.fastutil.objects.AbstractObjectList;
import it.unimi.dsi.fastutil.objects.AbstractObjectListIterator;
import it.unimi.dsi.fastutil.objects.ObjectList;
import it.unimi.dsi.fastutil.objects.ObjectListIterator;
import it.unimi.dsi.mg4j.io.FastBufferedReader;

import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;

/** Compact storage of strings using front-coding compression.
 * 
 * <P>This class stores a list of strings using front-coding compression (of course,
 * the compression will be reasonable only if the list is sorted, but you could
 * also use instances of this class just as a handy way to manage a large
 * amount of strings). It implements an immutable {@link
 * it.unimi.dsi.fastutil.objects.ObjectList} that returns the <var>i</var>-th
 * string (as a {@link MutableString}) when the {@link #get(int)} method is
 * called with argument <var>i</var>. The returned mutable string may be freely
 * modified.
 *
 * <P>As a commodity, this class provides a main method that reads from
 * standard input a sequence of newline-separated words, and
 * writes a corresponding serialized front-coded string list.
 *
 * <H2>Implementation Details</H2>
 * 
 * <P>To store the list of strings, we use either a UTF-8 coded {@link
 * ByteArrayFrontCodedList}, or a {@link CharArrayFrontCodedList}, depending on
 * the value of the {@link #FrontCodedStringList(Iterator,int,boolean)
 * utf8} parameter at creation time. In the first case, if the
 * strings are ASCII-oriented the resulting array will be much smaller, but
 * access times will increase manifold, as each string must be UTF-8 encoded
 * before being returned. 
 */

public class FrontCodedStringList extends AbstractObjectList implements Serializable {

	public static final long serialVersionUID = -7046029254386353130L;

	/** The underlying front-coded list (either a {@link ByteArrayFrontCodedList}, 
	 * or a {@link CharArrayFrontCodedList}, depending on the value of {@link #utf8}). */
	protected ObjectList frontCodedList;

	/** The underlying {@link ByteArrayFrontCodedList}, or <code>null</code>.*/
	transient protected ByteArrayFrontCodedList byteFrontCodedList;

	/** The underlying {@link CharArrayFrontCodedList}, or <code>null</code>.*/
	transient protected CharArrayFrontCodedList charFrontCodedList;

	/** Whether this front-coded list is UTF-8 encoded. */
	protected boolean utf8;


	/** Creates a new front-coded string list containing the character sequences returned by the given iterator.
	 * 
	 * @param words an iterator returning {@linkplain CharSequence character sequences}.
	 * @param ratio the desired ratio.
	 * @param utf8 if true, the strings will be stored as UTF-8 byte arrays.
	 */

	public FrontCodedStringList( final Iterator words, final int ratio, boolean utf8 ) {

		this.utf8 = utf8;
			
			frontCodedList = utf8 ?
				(ObjectList)(byteFrontCodedList = new ByteArrayFrontCodedList( 
												   new AbstractObjectIterator() {
													   public boolean hasNext() { return words.hasNext(); }
													   public Object next() {
														   try {
															   return ( (CharSequence)words.next() ).toString().getBytes( "UTF-8" ); 
														   }
														   catch( java.io.UnsupportedEncodingException cantHappen ) { return null; }
													   }
												   },
												   ratio ) ) :
				(ObjectList)(charFrontCodedList = new CharArrayFrontCodedList(
												   new AbstractObjectIterator() {
													   public boolean hasNext() { return words.hasNext(); }
													   public Object next() { 
														   CharSequence s = (CharSequence)words.next();
														   int i = s.length();
														   final char a[] = new char[ i ];
														   while( i-- != 0 ) a[ i ] = s.charAt( i );
														   return a;
													   }
												   }, 
											   ratio ) );
		
	}

	/** Creates a new front-coded string list containing the character sequences contained in the given collection.
	 * 
	 * @param c a collection containing {@linkplain CharSequence character sequences}.
	 * @param ratio the desired ratio.
	 * @param utf8 if true, the strings will be stored as UTF-8 byte arrays.
	 */
	public FrontCodedStringList( final Collection c, final int ratio, boolean utf8 ) {
		this( c.iterator(), ratio, utf8 );
	}

	/** Returns whether this front-coded string list is storing its strings as UTF-8 encoded bytes.
	 *
	 * @return true if this front-coded string list is keeping its data as an array of UTF-8 encoded bytes.
	 */
	public boolean utf8() {
		return utf8;
	}

	/** Returns the ratio of the underlying front-coded list.
	 *
	 * @return the ratio of the underlying front-coded list.
	 */
	public int ratio() {
		return utf8 ? byteFrontCodedList.ratio() : charFrontCodedList.ratio();
	}

	/** Returns the element at the specified position in this front-coded as a mutable string.
	 *
	 * @param index an index in the list.
	 * @return a {@link MutableString} that will contain the string at the specified position. The string may be freely modified.
	 */
	public Object get( final int index ) { 
		return MutableString.wrap( utf8 
								   ? byte2Char( byteFrontCodedList.getArray( index ), null ) 
								   : charFrontCodedList.getArray( index ) );
	}

	/** Returns the element at the specified position in this front-coded list by storing it in a mutable string.
	 *
	 * @param index an index in the list.
	 * @param s a mutable string that will contain the string at the specified position.
	 */
	public void get( final int index, MutableString s ) { 
		if ( utf8 ) {
			final byte[] a = byteFrontCodedList.getArray( index );
			s.length( countUTF8Chars( a ) );
			byte2Char( a, s.array() );
		}
		else {
			int res = charFrontCodedList.get( index, s.array );
			if ( res < 0 ) {
				s.length( s.array.length - res );
				res = charFrontCodedList.get( index, s.array );
			}
			else s.length( res );
		}
	}

	/* The following methods are highly optimized UTF-8 converters exploiting
	the fact that since it was ourselves in the first place who created the
	coding, we can be sure it is correct. */

	private static int countUTF8Chars( final byte a[] ) {
		final int length = a.length;
		int result = 0, b;
		for( int i = 0; i < length; i++ ) {
			b = ( a[ i ] & 0xFF ) >> 4;
			if ( b < 8 ) result++;
			else if ( b < 14 ) {
				result++;
				i++;
			}
			else {
				result++;
				i += 2;
			}
		}

		return result;
	}

	private static char[] byte2Char( final byte[] a, char[] s ) {
		final int length = a.length;
		if ( s == null ) s = new char[ countUTF8Chars( a ) ];
		int b, c, d;

		for( int i = 0, j = 0; i < length; i++ ) {
			b = a[ i ] & 0xFF;

			if ( ( b >> 4 ) < 8 ) s[ j++ ] = (char)b;
			else if ( ( b >> 4 ) < 14 ) {
				c = a[ ++i ] & 0xFF;
				if ( ( c & 0xC0 ) != 0x80 ) throw new IllegalStateException( "Malformed internal UTF-8 encoding" );
				s[ j++ ] = (char)( ( ( b & 0x1F) << 6 ) | ( c & 0x3F ) );
			}
			else {
				c = a[ ++i ] & 0xFF;
				d = a[ ++i ];
				if ( ( c & 0xC0 ) != 0x80 || ( d & 0xC0 ) != 0x80 ) throw new IllegalStateException( "Malformed internal UTF-8 encoding" );
				s[ j++ ] = (char)( ( ( b & 0x0F ) << 12 ) | ( ( c & 0x3F ) << 6 ) | ( ( d & 0x3F ) << 0 ) );
			}
		}

		return s;
	}


	public ObjectListIterator objectListIterator( final int k ) { return new AbstractObjectListIterator() {
			ObjectListIterator i = frontCodedList.objectListIterator( k );
			
			public boolean hasNext() { return i.hasNext(); }
			public boolean hasPrevious() { return i.hasPrevious(); }
			public Object next() { return  MutableString.wrap( utf8 ? byte2Char( (byte[])i.next(), null ) : (char[])i.next() ); }
			public Object previous() { return MutableString.wrap( utf8 ? byte2Char( (byte[])i.next(), null ) :(char[])i.previous() ); }
			public int nextIndex() { return i.nextIndex(); }
			public int previousIndex() { return i.previousIndex(); }
		};
	}

	public int size() { return frontCodedList.size(); }

	private void readObject(java.io.ObjectInputStream s) throws java.io.IOException, ClassNotFoundException {
		s.defaultReadObject();
		if ( utf8 ) byteFrontCodedList = (ByteArrayFrontCodedList)frontCodedList;
		else charFrontCodedList = (CharArrayFrontCodedList)frontCodedList;
	}

	public static void main( String arg[] ) throws FileNotFoundException, IOException {

		String listName = null;
		String encoding = null;
		String bufferSizeSpec = "64Ki";
		int bufferSize;
		int ratio = 4;
		boolean utf8 = false;
		
		int ch;
		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
			new LongOpt( "encoding", LongOpt.REQUIRED_ARGUMENT, null, 'e' ),
			new LongOpt( "ratio", LongOpt.REQUIRED_ARGUMENT, null, 'r' ),
			new LongOpt( "utf8", LongOpt.NO_ARGUMENT, null, 'u' ),
		};

		Getopt g = new Getopt( "FrontCodedStringList", arg, "b:e:r:hu", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: FrontCodedStringList [OPTIONS] LISTNAME" );
			System.err.println( "Builds a front-coded list of strings reading from standard input a" );
			System.err.println( "newline-separated list of words." );
			System.err.println();
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  LISTNAME                a filename for the list" );
			System.err.println();
			System.err.println( "Optional arguments:" );
			System.err.println( "  -b, --buffer-size        the size of I/O buffer (default: " + bufferSizeSpec + ")" );
			System.err.println( "  -e, --encoding ENCNAME   a character encoding used to read words" );
			System.err.println( "                           (default: the platform default encoding)" );
			System.err.println( "  -r, --ratio RATIO        the compression ratio (default: " + ratio + ")" );
			System.err.println( "  -u, --utf8               keep the strings as UTF-8 byte arrays" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help               print this help screen" );
			System.err.println();
			return;

		case 'e':
			encoding = g.getOptarg();
			break;

		case 'u':
			utf8 = true;
			break;

		case 'r':
			ratio = Integer.parseInt( g.getOptarg() );
			break;

		case 'b':
			bufferSizeSpec = g.getOptarg();
			break;
			
		case '?':
			return;

		default:
			break;
		}
		
		bufferSize = Fast.parseIntSize( bufferSizeSpec );

		if ( arg.length - g.getOptind() == 1 ) listName = arg[ g.getOptind() ];
		else {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}
		
		final FastBufferedReader words = encoding != null
			? new FastBufferedReader( new InputStreamReader( System.in, encoding ), bufferSize )
			: new FastBufferedReader( new InputStreamReader( System.in ), bufferSize );

		System.err.print( "Reading words..." );
		final ProgressMeter pm = new ProgressMeter( 100000, "words" );
		pm.start();

		final FrontCodedStringList p = new FrontCodedStringList( new AbstractObjectIterator() {
			MutableString s = new MutableString(), next;
			boolean toAdvance = true;
					
			public boolean hasNext() {
				if ( toAdvance ) {
					try {
						next = words.readLine( s );
					} catch (IOException e) {
						throw new RuntimeException( e );
					}
					toAdvance = false;
				}
				return next != null;
			}
						
			public Object next() {
				if ( ! hasNext() ) throw new NoSuchElementException();
				toAdvance = true;
				pm.update();
				return new MutableString( s );
			}
			}, ratio, utf8
		);

		pm.done();

		System.err.print( "Writing to file..." );

		ObjectOutputStream out = new ObjectOutputStream( new BufferedOutputStream( new FileOutputStream( listName ), bufferSize ) );
		out.writeObject( p );
		out.close();

		System.err.println( " done." );
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
