package it.unimi.dsi.mg4j.util;

import java.util.Collection;
import java.util.Iterator;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002-2003 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** {@link java.lang.String#hashCode()}-signed order-preserving minimal perfect hash tables.
 *
 * <P>This class source exemplifies a signed minimal perfect hash table that
 * signes each word with the hash code that would be returned by 
 * {@link java.lang.String#hashCode()} if the word were a {@link String}, thus avoiding false positives with high
 * probability.
 * 
 * 
 * @author Sebastiano Vigna
 * @since 0.8
 */

public class HashCodeSignedMinimalPerfectHash extends SignedMinimalPerfectHash {
    int signature[];

	/** Creates a new {@link java.lang.String#hashCode()}-signed order-preserving minimal perfect hash table 
	 * for the given set of words using the given number of weights.
	 *
	 * @param words some words to hash; it is assumed that this {@link
	 * Collection} does not contain words with a common prefix of
	 * <code>weightLength</code> characters.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions.
	 * @see MinimalPerfectHash#MinimalPerfectHash(Collection,int)
	 */
	public HashCodeSignedMinimalPerfectHash( Collection words, int weightLength ) {
		super( words, weightLength );
	}

	/** Creates a new {@link java.lang.String#hashCode()}-signed order-preserving minimal perfect hash table for the given
	 * set of words, using as many weights as the longest word in the collection.
     *
	 * @param words some words to hash; it is assumed that this {@link Collection} does not contain duplicates.
	 * @see MinimalPerfectHash#MinimalPerfectHash(Collection)
	 *	 
	 */
    public HashCodeSignedMinimalPerfectHash( Collection words ) {
		super( words );
    }

	/** Creates a new {@link java.lang.String#hashCode()}-signed order-preserving minimal perfect hash table for the given file 
	 * of words using the given number of weights.
	 *
	 * @param wordFile an UTF-8 file containing one word on each line; it is assumed that 
	 * it does not contain words with a common prefix of
	 * <code>weightLength</code> characters.
	 * @param encoding the encoding of <code>wordFile</code>; if <code>null</code>, it
	 * is assumed to be the platform default encoding.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions.
	 * @see MinimalPerfectHash#MinimalPerfectHash(String,String,int)
	 */
	public HashCodeSignedMinimalPerfectHash( String wordFile, String encoding, int weightLength ) {
		super( wordFile, encoding, weightLength );
	}

	/** Creates a new {@link java.lang.String#hashCode()}-signed order-preserving minimal perfect hash table for
	 * the given file of words, using as many weights as the longest word in the file.
	 *
	 * @param wordFile a file in the platform default encoding containing one word on each line; it is assumed that the file does
	 * not contain twice the same word.
	 * @see MinimalPerfectHash#MinimalPerfectHash(String)
	 */
	public HashCodeSignedMinimalPerfectHash( String wordFile ) {
		super( wordFile );
	}


	public void initSignatures( Collection words ) {
		CharSequence s;
		int i, j, h, l;
	
		signature = new int[ words.size() ];
	
		j = 0;
		for ( Iterator e = words.iterator(); e.hasNext(); j++ ) {
			s = (CharSequence)e.next();
			l = s.length();
			for ( i = h = 0; i < l; i++ ) h = 31 * h + s.charAt( i );
			signature[ j ] = h;
		}
	}

	public boolean checkSignature( CharSequence word, int index ) {
		int i, h, l;
	
		l = word.length();
		for ( i = h = 0; i < l; i++ ) h = 31 * h + word.charAt( i );
		return signature[ index ] == h;
	}

}
