package it.unimi.dsi.mg4j.util;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002, 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.booleans.BooleanArrays;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.mg4j.index.AbstractTermMap;
import it.unimi.dsi.mg4j.io.FastBufferedReader;

import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.util.AbstractCollection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.NoSuchElementException;

import cern.colt.GenericPermuting;
import cern.colt.GenericSorting;
import cern.colt.Swapper;
import cern.colt.function.IntComparator;
import cern.jet.random.engine.MersenneTwister;



/** Order-preserving minimal perfect hash tables.
 *
 * <P>Given a collection of terms without duplicates (it is not forced to be a {@link java.util.Set} so that one can also
 * use arrays), the constructor of this class finds an order-preserving minimal perfect hash function for
 * the collection. Subsequent calls to the {@link #hash(CharSequence)} method will return the ordinal position of
 * the provided string (if it was in the collection; otherwise, the result is a random position). The class
 * can then be saved by serialisation and reused later.
 *
 * <P>If you also need to establish whether a given term was or not in the
 * collection, you should consider using a {@link SignedMinimalPerfectHash}
 * implementing class instead.
 *
 * <P>This class is very scalable, and if you have enough memory it will handle
 * efficiently hundreds of millions of terms: in particular, the 
 * {@linkplain #MinimalPerfectHash(String) offline constructor}
 * can build a map without loading the terms in memory.
 * 
 * <P>To do its job, the class must store three vectors of weights that are used to compute
 * certain hash functions. By default, the vectors are long as the longest term, but if
 * your collection is sorted you can ask (passing {@link #WEIGHT_UNKNOWN_SORTED_TERMS} to a constructor)
 * for the shortest possible vector length. This optimisation does not change the
 * memory footprint, but it can improve performance.
 * 
 * <P>As a commodity, this class provides a main method that reads from
 * standard input a sequence of newline-separated terms, and
 * writes a serialised minimal perfect hash table for the given list. You can
 * specify on the command line the kind of table (e.g., {@link
 * CRC32SignedMinimalPerfectHash}) and have it fetched by reflection.
 * 
 *
 * <h3>Implementation Details</h3> 
 *
 * <P>An object of this class uses about 1.25<var>n</var> integers between 0 and <var>n</var>-1 inclusive
 * to hash <var>n</var>. This is asymptotically optimal, but for small collections using an integer
 * wastes a large number of bits in each entry.
 *
 * <P>The technique used here is suggested (but not fully described) in the second edition of <em>Managing
 * Gigabytes</em>. It is due to Havas, Majewski, Wormald and Czech, and it is fully described in
 * the monograph by Czech, Havas and Majewski on perfect hashing (</i>Perfect Hashing</i>, 
 * Theoretical Computer Science 182(1997), pages 1-143).
 *
 * <P>First, a triple of intermediate hash functions (generated via universal hashing) define for each
 * term a 3-hyperedge in a hypergraph with 1.25<var>n</var> vertices. Each intermediate hash function
 * uses a vector of random weights; the length of the vector is by default the length of the longest
 * term, but if the collection is sorted it is possible to compute the minimum length of a prefix that will
 * distinguish any pair of term. 
 * 
 * <P>Then, by successively
 * stripping hyperedges with one vertex of degree one, we create an ordering on the hyperedges. 
 * Finally, we assign (in reverse order) to each vertex of a hyperedge a number in the range
 * 0 and <var>n</var>-1 in such a way that the sum of the three numbers modulo <var>n</var> is exactly
 * the original index of the term corresponding to the hyperedge. This is possible with high probability, 
 * so we try until we succeed.
 * 
 * <P>Note that the mathematical results guaranteeing that it is possible to find a
 * function in expected constant time are <em>asymptotic</em>. 
 * Thus, when the number of terms is less than {@link #TERM_THRESHOLD}, an instance
 * of this class just stores the terms in a vector and scans them linearly. This behaviour,
 * however, is completely transparent to the user.
 *
 * <h3>Rounds and Logging</h3>
 * 
 * <P>Building a minimal perfect hash map may take hours. As it happens with all probabilistic algorithms,
 * one can just give estimates of the expected time.
 * 
 * <P>There are two probabilistic sources of problems: degenerate hyperedges and non-acyclic hypergraphs.
 * The ratio between the number of vertices and the number of terms guarantee that acyclicity is true with high
 * probability. However, when generating the hypergraph we use three hash functions, and it must never happen
 * that the value of two of those functions coincide on a given term. Because of the birthday paradox, the
 * probability of getting a nondegerate edge is just
 * <blockquote>
 * (<var>m</var>-1)(<var>m</var>-2)/<var>m</var><sup>2</sup>,
 * </blockquote>
 * where <var>m</var>=1.25<var>n</var>. Since this must happen for <var>n</var> times, we must raise
 * this probability to <var>n</var>, and as <var>n</var> grows the value quickly (and monotonically) 
 * reaches <i>e</i><sup>-12/5</sup>.
 * We conclude that the expected number of trials to generate a random 3-hypergraph 
 * is bounded by <i>e</i><sup>12/5</sup>, which is about 11. Note that this bound 
 * <em>does not depend on <var>n</var></em>.
 *
 * <P>Once the hypergraph has been generated, the stripping procedure may fail. However, the expected number
 * of trials tends to 1 as <var>n</var>
 * approaches infinity (Czech, Havas and Majewski, for instance, report that on a set of 50,000 terms
 * they obtained consistently one trial for more than 5000 experiments). The number of expected trials, of course,
 * must be multiplied by 11, as every failure of the stripping procedure requires generating a new hypergraph.
 *  
 * <P>To help diagnosing problem with the generation process
 * class, a property named <samp>it.unimi.dsi.mg4j.util.minimalperfecthash.verbose</samp> is checked for: if set to true, then
 * the class will log what's happening to standard error. In particular, it will signal whenever a degenerate
 * hyperedge is generated, and the various construction phases.
 *
 * <P>Note that if during the generation process the log warns more than once about duplicate hyperedges, you should
 * suspect that there are duplicates in the term list, as duplicate hyperedges are <em>extremely</em> unlikely.
 *
 * @author Sebastiano Vigna
 * @author Paolo Boldi
 * @since 0.1
 */

public class MinimalPerfectHash extends AbstractTermMap implements Serializable {
	/** The minimum number of terms that will trigger the construction of a minimal perfect hash;
	 * overwise, terms are simply stored in a vector. */
	public static final int TERM_THRESHOLD = 16;
	/** A special value denoting that the weight length is unknown, and should be computed using
	 * the maximum length of a term. */
	public static final int WEIGHT_UNKNOWN = -1;
	/** A special value denoting that the weight length is unknown, and should be computed assuming
	 * that the terms appear in lexicographic order. */
	public static final int WEIGHT_UNKNOWN_SORTED_TERMS = -2;

	/** The number of buckets. */
	final protected int n;
	/** The number of vertices of the intermediate hypergraph. */
	final protected int m;
	/** The maximum amount of right shift to perform on a 32-bit number so to obtain something greater than or equal to {@link #m}. */
	final protected int rightShift;
	/** Vector of weights to compute the first intermediate hash function. */
	final protected int weight0[];
	/** Vector of weights to compute the second intermediate hash function. */
	final protected int weight1[];
	/** Vector of weights to compute the third intermediate hash function. */
	final protected int weight2[];
	/** The length of the components of the weight vectors (it's faster than asking the length of the vectors). */
	final protected int weightLength;
	/** The final magick&mdash;the function turning the intermediate hash values into the final bucket. */
	final protected int g[];
	/** Whether we should log what's happening (it is set at class load time 
	 * using the system property <samp>it.unimi.dsi.mg4j.util.minimalperfecthash.verbose</samp>). */
	final protected static boolean VERBOSE = Boolean.getBoolean( "it.unimi.dsi.mg4j.util.minimalperfecthash.verbose" );
	/** The term collection (it could be an anonymous class wrapping a file). */
	final protected transient Collection terms;
	/** Four times the number of buckets. */
	protected transient long n4;
	/** If {@link #n} is smaller than {@link #TERM_THRESHOLD}, a vector containing the terms. */
	protected transient CharSequence[] t;

    public static final long serialVersionUID = -7046029254386353130L;
    

	/* The following two methods MUST be kept synchronised. The reason why we duplicate code is 
	 * that we do not want the overhead of allocating an array when searching for a string. */

	/** Hashes a given term using the intermediate hash functions.
	 *
	 * @param term a term to hash.
	 * @param h a three-element vector that will be filled with the three intermediate hash values.
	 */

	protected void hash( final CharSequence term, final int h[] ) {
		int h0 = 0, h1 = 0xFFFFFFFF, h2 = 0x55555555; // We need three different value to map correctly the empty string
		char c;
		  
		int i = term.length();
		if ( weightLength < i ) i = weightLength;
		while( i-- != 0 ) {
			c = term.charAt(i);
			h0 ^= weight0[ i ] * c;
			h1 ^= weight1[ i ] * c;
			h2 ^= weight2[ i ] * c;
		}

		h[ 0 ] = ( h0 >>> rightShift ) % m;
		h[ 1 ] = ( h1 >>> rightShift ) % m;
		h[ 2 ] = ( h2 >>> rightShift ) % m;
	}

	/** Gets a term out of the stored array {@link #t}. 
	 * 
	 * <P>Note: This method does not check for {@link #t} being non-<code>null</code>.
	 * 
	 * @param term a term.
	 * @return the position of the given term in the generating collection, starting from 0. If the
	 * term was not in the original collection, the result is 0.
	 */
	protected int getFromT( final CharSequence term ) {
		int i = n;
		/* Note that we invoke equality *on the stored MutableString*. This
		 * is essential due to the known weaknesses in CharSequence's contract. */
		while( i-- != 0 ) if ( t[ i ].equals( term ) ) return i; 
		return 0;
	}



	/** Hashes a given term.
	 *
	 * @param term a term to be hashed.
	 * @return the position of the given term in the generating collection, starting from 0. If the
	 * term was not in the original collection, the result is a random position.
	 *
	 */
	public int get( final CharSequence term ) {
		
		if ( t != null ) return getFromT( term );
		
		int h0 = 0, h1 = 0xFFFFFFFF, h2 = 0x55555555;  // We need three different value to map correctly the empty string
		char c;

		int i = term.length();
		if ( weightLength < i ) i = weightLength;
		while( i-- != 0 ) {
			c = term.charAt( i );
			h0 ^= weight0[ i ] * c;
			h1 ^= weight1[ i ] * c;
			h2 ^= weight2[ i ] * c;
		}

		return (int) (( (long)g[ ( h0 >>> rightShift ) % m ] + (long)g[ ( h1 >>> rightShift ) % m ] + (long)g[ ( h2 >>> rightShift ) % m ] + n4 ) % n);
	}


	/** Hashes a given term.
	 *
	 * @param term a term to be hashed.
	 * @return the position of the given term in the generating collection, starting from 0. If the
	 * term was not in the original collection, the result is a random position.
	 */

	public int get( final MutableString term ) {

		if ( t != null ) return getFromT( term );

		int h0 = 0, h1 = 0xFFFFFFFF, h2 = 0x55555555;  // We need three different value to map correctly the empty string
		char c;

		int i = term.length();
		final char[] a = term.array;
		if ( weightLength < i ) i = weightLength;
		while( i-- != 0 ) {
			c = a[ i ];
			h0 ^= weight0[ i ] * c;
			h1 ^= weight1[ i ] * c;
			h2 ^= weight2[ i ] * c;
		}

		return (int) (( (long)g[ ( h0 >>> rightShift ) % m ] + (long)g[ ( h1 >>> rightShift ) % m ] + (long)g[ ( h2 >>> rightShift ) % m ] + n4 ) % n);
	}

	/** Hashes a given term.
	 * 
 	 * @param term a term to be hashed.
	 * @return the position of the given term in the generating collection, starting from 0. If the
	 * term was not in the original collection, the result is a random position.
	 * @deprecated As of <code>mg4j</code> 1.0, replaced by {@link #get(CharSequence)}.
	 */

	public int hash( final CharSequence term ) {
		return get( term );
	}


	/** Hashes a given term.
	 * 
	 * @param term a term to be hashed.
	 * @return the position of the given term in the generating collection, starting from 0. If the
	 * term was not in the original collection, the result is a random position.
	 * @deprecated As of <code>mg4j</code> 1.0, replaced by {@link #get(MutableString)}.
	 */

	public int hash( final MutableString term ) {
		return get( term );
	}

	/** Gets the length of the weight vectors.
	 *
	 * @return the length of weight vectors used in this table.
	 * @deprecated As of <code>mg4j</code> 0.8, replaced by {@link #weightLength()}.
	 */
	public int getWeightLength() {
		return weightLength;
	}

	/** Returns the length of the weight vectors.
	 *
	 * @return the length of weight vectors used in this table.
	 */

	public int weightLength() {
		return weightLength;
	}


	/** Returns the number of terms hashed.
	 *
	 * @return the number of terms hashed.
	 */

	public int size() {
		return n;
	}



	private void writeObject(java.io.ObjectOutputStream s) throws java.io.IOException, ClassNotFoundException {
		s.defaultWriteObject();
		if ( n < TERM_THRESHOLD ) s.writeObject( t );
	}

	private void readObject(java.io.ObjectInputStream s) throws java.io.IOException, ClassNotFoundException {
		s.defaultReadObject();
		n4 = n * 4;
		if ( n < TERM_THRESHOLD ) t = (CharSequence[])s.readObject();
	}


	/** Creates a new order-preserving minimal perfect hash table for the given
	 * set of terms, using as many weights as the longest term in the collection.
	 *
	 * <P><strong>Caution:</strong> if the collection contains twice the same
	 * term, this constructor will never return.
	 *
	 * @param terms some terms to hash; it is assumed that this {@link Collection} does not contain duplicates or the empty term.
	 */
	public MinimalPerfectHash( Collection terms ) {
		this( terms, WEIGHT_UNKNOWN );
	}

	/** Creates a new order-preserving minimal perfect hash table for the given set of terms using the given number of weights.
	 *
	 * <P>This constructor can be used to force the use of a reduced number of weights if you are sure that
	 * the first <code>weightLength</code> characters of each term in <code>terms</code> are
	 * distinct.
	 *
	 * <P>If you do not know your weight length in advance, you can pass two
	 * special values as <code>weightLength</code>: {@link #WEIGHT_UNKNOWN}
	 * will force the computation of <code>weightLength</code> as the maximum
	 * length of a term, whereas {@link #WEIGHT_UNKNOWN_SORTED_TERMS} forces
	 * the assumption that terms are sorted: in this case, we search for
	 * the minimum prefix that will disambiguate all terms in the collection. 
	 * 
	 * <P><strong>Caution:</strong> if two terms in the collection have a
	 * common prefix of length <code>weightLength</code> this constructor will
	 * never return.
	 * 
	 * @param terms some terms to hash; if <code>weightLength</code>
	 * is specified explicitly, it is assumed that this {@link
	 * Collection} does not contain terms with a common prefix of
	 * <code>weightLength</code> characters or the empty term.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions, {@link #WEIGHT_UNKNOWN} or {@link #WEIGHT_UNKNOWN_SORTED_TERMS}.
	 * @see #MinimalPerfectHash(Collection)
	 */

	public MinimalPerfectHash( Collection terms, int weightLength ) {
		if ( weightLength != WEIGHT_UNKNOWN && weightLength != WEIGHT_UNKNOWN_SORTED_TERMS && weightLength <= 0 ) throw new IllegalArgumentException( "Non-positive weight length: " + weightLength );

		this.terms = terms;
		n = terms.size();
		n4 = n * 4;
		m = (int)( n * 1.25f );
		rightShift = ( 32 - Fast.mostSignificantBit( m ) ) / 2;
		g = new int[ m ];


		if ( weightLength < 0 ) {
			if ( VERBOSE ) System.err.print( "Computing weight length..." );

			Iterator i = terms.iterator();
			int maxLength = Integer.MIN_VALUE, minPrefixLength = Integer.MIN_VALUE;
			
			if ( i.hasNext() ) {
				CharSequence currTerm, prevTerm = new MutableString( (CharSequence) i.next() );  // We cannot assume that the string will persist after a call to next().
				int k, prevLength = prevTerm.length(), currLength;	

				maxLength = prevLength;

				while( i.hasNext() ) {
					currTerm = (CharSequence) i.next();
					currLength = currTerm.length();

					if ( weightLength == WEIGHT_UNKNOWN_SORTED_TERMS ) {
						for( k = 0; k < prevLength && k < currLength && currTerm.charAt( k ) == prevTerm.charAt( k ); k++ );
		
						if ( k == currLength && k == prevLength ) throw new IllegalArgumentException( "The term list contains a duplicate (" + currTerm + ")" );
						minPrefixLength = Math.max( minPrefixLength, k + 1 );

						prevTerm = new MutableString( currTerm ); // We cannot assume that the string will persist after a call to next().
						prevLength = currLength;
					}

					maxLength = Math.max( maxLength, currLength );
				}
			}

			weightLength = weightLength == WEIGHT_UNKNOWN_SORTED_TERMS ? minPrefixLength : maxLength;

			if ( VERBOSE ) System.err.println( " done [max term length=" + maxLength + "; weight length=" + weightLength + "]." );
		}

		weight0 = new int[weightLength];
		weight1 = new int[weightLength];
		weight2 = new int[weightLength];
		this.weightLength = weightLength;

		if ( n < 16 ) {
			int j = 0;
			t = new MutableString[ n ];
			for( Iterator i = terms.iterator(); i.hasNext(); ) t[ j++ ] = new MutableString( (CharSequence)i.next() ); 
		}
		else {		
			t = null;
			new Visit( terms );
		}
	}

	/** Creates a new order-preserving minimal perfect hash table for the given file 
	 * of terms using the given number of weights.
	 *
	 * @param termFile an file containing one term on each line; it is assumed that 
	 * it does not contain terms with a common prefix of
	 * <code>weightLength</code> characters or the empty term.
	 * @param encoding the encoding of <code>termFile</code>; if <code>null</code>, it
	 * is assumed to be the platform default encoding.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions, {@link #WEIGHT_UNKNOWN} or {@link #WEIGHT_UNKNOWN_SORTED_TERMS}.
	 * @see #MinimalPerfectHash(Collection, int)
	 */


	public MinimalPerfectHash( final String termFile, final String encoding, int weightLength ) {
		this( new AbstractCollection() {
			String file = termFile; 
			int size = -1;
			
			public Iterator iterator() {
					return new Iterator() {
						final FastBufferedReader fbr;
						MutableString s = new MutableString(), next;
						boolean toAdvance = true;
					
						{
							try {
								fbr = encoding != null
									? new FastBufferedReader( new InputStreamReader( new FileInputStream( termFile ), encoding == null ? "UTF-8" : encoding ) )
									: new FastBufferedReader( new FileReader( termFile ) );
							} catch (FileNotFoundException e) {
								throw new RuntimeException( e );
							} catch (UnsupportedEncodingException e) {
								throw new RuntimeException( e );
							}
						}
						
						public boolean hasNext() {
							if ( toAdvance ) {
								try {
									next = fbr.readLine( s );
								} catch (IOException e) {
									throw new RuntimeException( e );
								}
								toAdvance = false;
							}
							return next != null;
						}
						
						public Object next() {
							if ( ! hasNext() ) throw new NoSuchElementException();
							toAdvance = true;
							return s;
						}
					
						public void remove() {
							throw new UnsupportedOperationException();
						}
						
					};
			}
			
			public synchronized int size() {
				if ( size == -1 ) {
					Iterator i = iterator();
					size = 0;
					while( i.hasNext() ) {
						size++;
						i.next();
					}
				}
				return size;
			}
			
		}, weightLength );
	}

	/** Creates a new order-preserving minimal perfect hash table for
	 * the given file of terms, using as many weights as the longest term in the file.
	 *
	 * @param termFile a file in the platform default encoding 
	 * containing one term on each line; it is assumed that the file does
	 * not contain duplicates or the empty term.
	 * @see #MinimalPerfectHash(String, String, int)
	 */

	public MinimalPerfectHash( final String termFile ) {
		this( termFile, null, WEIGHT_UNKNOWN );
	}

	private class Visit {

		MersenneTwister r = new MersenneTwister( new Date() );
		int d[] = new int[m]; // The degree of each vertex of the intermediate hypergraph.
		/* An 3&times;n array recording the triple of vertices involved in each hyperedge. It is *reversed*
		   w.r.t. you would expect to reduce object creation. */
		int edge[][] = new int[3][n];
		boolean removed[] = new boolean[n]; // Whether a hyperedge has been removed.
		int last[] = new int[m]; // The next position to fill in the respective incidence vector. Used also to store the hyperedge permutation and speed up permute().
		int inc[] = new int[m * 3]; // For each vertex of the intermediate hypergraph, the vector of incident hyperedges.
		int incOffset[] = new int[m]; // For each vertex of the intermediate hypergraph, the offset into the vector of incident hyperedges. Used also to speed up permute().
		int stack[] = new int[n]; // The hyperedge stack. Used also to invert the hyperedge permutation.
		int top; // Initial top of the hyperedge stack.

		public Visit( Collection terms ) {
			int i, j, k, s, v = -1, h[] = new int[3];
				
			do {
				
				if ( VERBOSE ) System.err.print( "Generating random hypergraph..." );
				top = 0;
				
				/* We choose the random weights for the three intermediate hash functions. */
				for( i = 0; i < weightLength; i++ ) {
					weight0[i] = r.nextInt();
					weight1[i] = r.nextInt();
					weight2[i] = r.nextInt();
				}
					 
					 
				/* We build the hyperedge list, checking that we do not create a degenerate hyperedge. */
				i = 0;
				CharSequence cs = null;
				IntArrays.fill( d, 0 );

				for( Iterator w = terms.iterator(); w.hasNext(); ) {
					hash( cs = (CharSequence)w.next(), h );
					if ( h[ 0 ] == h[ 1 ] || h[ 1 ] == h[ 2 ] || h[ 2 ] == h[ 0 ] ) break;

					edge[0][i] = h[0];
					edge[1][i] = h[1];
					edge[2][i] = h[2];

					i++;
				}
					 
				if ( i < n ) {
					if ( VERBOSE ) System.err.println( " hypergraph generation interrupted by degenerate hyperedge " + i + " (string: \"" + cs + "\")." );
					continue;
				} 

				/* We compute the degree of each vertex. */
				for( j = 0; j < 3; j++ ) {
					i = n;
					while( i-- != 0 ) d[ edge[ j ][ i ] ]++; 
				}

				if ( VERBOSE ) {
					System.err.println( " done." ); 
					System.err.print( "Checking for duplicate hyperedges..." );
				}
		
				/* Now we quicksort hyperedges lexicographically, keeping into last their permutation. */
				i = n;
				while( i-- != 0 ) last[ i ] = i;
				
				GenericSorting.quickSort( 0, n, new IntComparator() {
					public int compare( int x, int y ) {
						int r;
						if ( ( r = edge[ 0 ][ x ] - edge[ 0 ][ y ] ) != 0 ) return r;
						if ( ( r = edge[ 1 ][ x ] - edge[ 1 ][ y ] ) != 0 ) return r;
						return edge[ 2 ][ x ] - edge[ 2 ][ y ];
					}
				},
				new Swapper() {
					public void swap( int x, int y ) {
						int e0 = edge[ 0 ][ x ], e1 = edge[ 1 ][ x ], e2 = edge[ 2 ][ x ], p = last[ x ];
						edge[ 0 ][ x ] = edge[ 0 ][ y ];
						edge[ 1 ][ x ] = edge[ 1 ][ y ];
						edge[ 2 ][ x ] = edge[ 2 ][ y ];
						edge[ 0 ][ y ] = e0;
						edge[ 1 ][ y ] = e1;
						edge[ 2 ][ y ] = e2;
						last[ x ] = last[ y ];
						last[ y ] = p;
					}
				}
				);
				
				i = n - 1;
				while( i-- != 0 ) if ( edge[ 0 ][ i + 1 ] == edge[ 0 ][ i ] && edge[ 1 ][ i + 1 ] == edge[ 1 ][ i ] && edge[ 2 ][ i + 1 ] == edge[ 2 ][ i ]) break;

				if ( i != -1 ) {
					if ( VERBOSE ) System.err.println( " found double hyperedge for terms " + last[ i + 1 ] + " and " + last[ i ] + "." );
					continue;
				}
				
				/* We now invert last and permute all hyperedges back into their place. Note that
				 * we use last and incOffset to speed up the process. */
				i = n;
				while( i-- != 0 ) stack[ last[ i ] ] = i;
				
				GenericPermuting.permute( stack, new Swapper() {
					public void swap( int x, int y ) {
						int e0 = edge[ 0 ][ x ], e1 = edge[ 1 ][ x ], e2 = edge[ 2 ][ x ];
						edge[ 0 ][ x ] = edge[ 0 ][ y ];
						edge[ 1 ][ x ] = edge[ 1 ][ y ];
						edge[ 2 ][ x ] = edge[ 2 ][ y ];
						edge[ 0 ][ y ] = e0;
						edge[ 1 ][ y ] = e1;
						edge[ 2 ][ y ] = e2;
					}
				}, last, incOffset
				);
				
				if ( VERBOSE ) {
					System.err.println( " done." ); 
					System.err.print( "Visiting hypergraph..." );
				}
				
					 
				/* We set up the offset of each vertex in the incidence
				   vector. This is necessary to avoid creating m incidence vectors at
				   each round. */
				IntArrays.fill( last, 0 );
				incOffset[ 0 ] = 0;

				for( i = 1; i < m; i++ ) incOffset[ i ] = incOffset[ i - 1 ] + d[ i - 1 ];
					 
				/* We fill the vector. */
				for( i = 0; i < n; i++ ) 
					for( j = 0; j < 3; j++ ) {
						v = edge[j][i];
						inc[ incOffset[ v ] + last[ v ]++ ] = i;
					}
					 
				/* We visit the hypergraph. */
				BooleanArrays.fill( removed, false );

				for( i = 0; i < m; i++ )
					if ( d[ i ] == 1 )
						visit( i );
					 
				if ( VERBOSE ) {
					if ( top == n ) System.err.println( " done." );
					else System.err.println( " failed: stripped " + top + " hyperedges out of " + n + "." );
				} 
			} while ( top != n );
				
			if ( VERBOSE ) System.err.print( "Assigning values..." );

			/* We assign values. */
			IntArrays.fill( g, -1 );

			while( top > 0 ) {
				k = stack[--top];
					 
				s = 0;
					 
				for( j = 0; j < 3; j++ ) {
					if ( g[edge[j][k]] == -1 ) {
						g[edge[j][k]] = 0;
						v = edge[j][k];
					}
					else s += g[edge[j][k]];
				}
				
				g[v] = ( ( k - s ) % n + n ) % n;
			}

			if ( VERBOSE ) System.err.println( " done." );
		}

		/*

		This is the original recursive visit. It is here for documentation purposes. It
		cannot handle more than about 1,000,000 terms.

		private void visit( int x ) {
		int i, j, k = -1;
			
		for( i = 0; i < last[x]; i++ ) 
		if ( !removed[ k = ((int[])inc[x])[i] ] ) break; // The only hyperedge incident on x in the current configuration.
		stack[top++] = k;


		if ( top % 100000 == 0 ) System.err.println( "Processed " + top + " edges" );

		// We update the degrees and the incidence lists. 
		for( i = 0; i < 3; i++ ) {
		removed[k] = true;
		d[edge[i][k]]--;
		}
				
		// We follow recursively the other vertices of the hyperedge, if they have degree one in the current configuration.
		for( i = 0; i < 3; i++ ) 
		if ( edge[i][k] != x && d[edge[i][k]] == 1 ) 
		visit( edge[i][k] );
		}
		*/

		
		int recStackI[] = new int[ n ]; // The stack for i.
		int recStackK[] = new int[ n ]; // The stack for k.
		int recTop; // Initial top of the recursion stack.

		private void visit( int x  ) {
			int i, k = -1;
			boolean inside;
			
			// Stack initialization
			recTop = 0;
			inside = false;
			
			while ( true ) {
				if ( ! inside ) {
					for( i = 0; i < last[x]; i++ ) 
						if ( !removed[ k = inc[ incOffset[ x ] + i ] ] ) break; // The only hyperedge incident on x in the current configuration.
				
					stack[top++] = k;
					
					/* We update the degrees and the incidence lists. */
					for( i = 0; i < 3; i++ ) {
						removed[k] = true;
						d[edge[i][k]]--;
					}
				}
				
				/* We follow recursively the other vertices of the hyperedge, if they have degree one in the current configuration. */
				for( i = 0; i < 3; i++ ) 
					if ( edge[i][k] != x && d[edge[i][k]] == 1 ) {
						recStackI[ recTop ] = i+1;
						recStackK[ recTop ] = k;
						recTop++;
						x = edge[i][k];
						inside = false;
						break;
					}
				if ( i < 3 ) continue;
				
				if ( --recTop < 0 ) return;
				i = recStackI[ recTop ];
				k = recStackK[ recTop ];
				inside = true;
			}
		}
	}


    public static void main( String arg[] ) throws IllegalArgumentException, SecurityException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException, FileNotFoundException, IOException {

		String bufferSizeSpec = "64Ki";
		int bufferSize;
		String tableName = null;
		String tableClass = "it.unimi.dsi.mg4j.util.MinimalPerfectHash";
		String termFile = null;
		String encoding = "UTF-8";
		boolean sorted = false;
		
		int ch;
		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "class", LongOpt.REQUIRED_ARGUMENT, null, 'c' ),
			new LongOpt( "encoding", LongOpt.REQUIRED_ARGUMENT, null, 'e' ),
			new LongOpt( "offline", LongOpt.REQUIRED_ARGUMENT, null, 'o' ),
			new LongOpt( "buffer-size", LongOpt.REQUIRED_ARGUMENT, null, 'b' ),
			new LongOpt( "sorted", LongOpt.REQUIRED_ARGUMENT, null, 's' ),
		};

		Getopt getopt = new Getopt( "MinimalPerfectHash", arg, "b:o:e:c:sh", longopts );
		getopt.setOpterr( true );

		while ( ( ch = getopt.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: MinimalPerfectHash [OPTIONS] TABLENAME" );
			System.err.println( "Builds a minimal perfect hash table reading from standard input a" );
			System.err.println( "newline-separated list of terms." );
			System.err.println( "" );
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  TABLENAME                a filename for the table" );
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -b, --buffer-size        the size of I/O buffer (default: " + bufferSizeSpec + ")" );
			System.err.println( "  -c, --class MPHCLASS     a subclass of MinimalPerfectHash that will be used to" );
			System.err.println( "                           create the table (default: " + tableClass + ")" );
			System.err.println( "  -o, --offline FILENAME   terms will be read from the given file (and not" );
			System.err.println( "                           loaded in memory)" );
			System.err.println( "  -e, --encoding ENCNAME   a character encoding used to read terms" );
			System.err.println( "                           (default: " + encoding + ")" );
			System.err.println( "  -s, --sorted             the term list is sorted--optimise weight length" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help               print this help screen" );
			System.err.println( "" );
			return;

		case 'c':
			tableClass = getopt.getOptarg();
			break;
			
		case 'o':
			termFile = getopt.getOptarg();
			break;

		case 'e':
			encoding = getopt.getOptarg();
			break;

		case 'b':
			bufferSizeSpec = getopt.getOptarg();
			break;
			
		case 's':
			sorted = true;
			break;
		
		case '?':
			return;

		default:
			break;
		}

		bufferSize = Fast.parseIntSize( bufferSizeSpec );

		if ( arg.length - getopt.getOptind() == 1 ) tableName = arg[ getopt.getOptind() ];
		else {
			System.err.println( "Wrong number (" + ( arg.length - getopt.getOptind() ) + ") of arguments." );
			return;
		}

		MinimalPerfectHash minimalPerfectHash;
		
		if ( termFile == null ) {
			System.err.print( "Reading terms..." );
			java.util.ArrayList termList = new ArrayList();
			MutableString term = new MutableString();
			final ProgressMeter pm = new ProgressMeter( 100000, "terms" );
			final FastBufferedReader terms = new FastBufferedReader( new InputStreamReader( System.in, encoding ), bufferSize );
				
			pm.start();

			while( terms.readLine( term ) != null ) {
				pm.update();
				termList.add( term.copy() );
			}

			pm.stop();

			System.err.println( " done." );
			System.err.println( pm );

			System.err.println( "Building minimal perfect hash table..." );
			minimalPerfectHash = (MinimalPerfectHash)Class.forName( tableClass, true, ClassLoader.getSystemClassLoader() ).getConstructor( new Class[] { java.util.Collection.class, int.class } ).newInstance( new Object[] { termList, new Integer( sorted ? WEIGHT_UNKNOWN_SORTED_TERMS : WEIGHT_UNKNOWN ) } );
		}
		else minimalPerfectHash = (MinimalPerfectHash)Class.forName( tableClass, true, ClassLoader.getSystemClassLoader() ).getConstructor( new Class[] { String.class, String.class, int.class } ).newInstance( new Object[] { termFile, encoding, new Integer( sorted ? WEIGHT_UNKNOWN_SORTED_TERMS : WEIGHT_UNKNOWN ) } );

		System.err.print( "Writing to file..." );

		ObjectOutputStream out = new ObjectOutputStream( new BufferedOutputStream( new FileOutputStream( tableName ), bufferSize ) );
		out.writeObject( minimalPerfectHash );
		out.close();

		System.err.println( " done." );
    }
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
