/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002, 2003, 2004 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
package it.unimi.dsi.mg4j.util;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.ints.IntIterators;
import it.unimi.dsi.fastutil.ints.IntListIterator;
import it.unimi.dsi.fastutil.objects.AbstractObjectList;
import it.unimi.dsi.fastutil.objects.AbstractObjectListIterator;
import it.unimi.dsi.fastutil.objects.ObjectListIterator;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;

/** A {@link it.unimi.dsi.mg4j.util.FrontCodedStringList} whose indices are permuted.
 * 
 * <P>It may happen that a list of strings compresses very well 
 * using front coding, but unfortunately alphabetical order is <em>not</em>
 * the right order for the strings in the list. Instances of this class
 * wrap an instance of {@link it.unimi.dsi.mg4j.util.FrontCodedStringList}
 * together with a permutation &pi;: inquiries with index <var>i</var> will
 * actually return the string with index &pi;<sub><var>i</var></sub>.
 */

public class PermutedFrontCodedStringList extends AbstractObjectList implements Serializable {

	public static final long serialVersionUID = -7046029254386353130L;

	/** The underlying front-coded string list. */
	final protected FrontCodedStringList frontCodedStringList;
	/** The permutation. */
	final protected int[] permutation;

	/** Creates a new permuted front-coded string list using a given front-coded string list and permutation.
	 * 
	 * @param frontCodedStringList the underlying front-coded string list.
	 * @param permutation the underlying permutation.
	 */

	public PermutedFrontCodedStringList( final FrontCodedStringList frontCodedStringList, int[] permutation ) {
		this.frontCodedStringList = frontCodedStringList;
		this.permutation = permutation;
	}
	
	public Object get( final int index ) { 
		return frontCodedStringList.get( permutation[ index ] );
	}

	/** Returns the element at the specified position in this front-coded list by storing it in a mutable string.
	 *
	 * @param index an index in the list.
	 * @param s a mutable string that will contain the string at the specified position.
	 */
	public void get( final int index, MutableString s ) { 
		frontCodedStringList.get( permutation[ index ], s );
	}

	public int size() {
		return frontCodedStringList.size();
	}

	public ObjectListIterator objectListIterator( final int k ) { return new AbstractObjectListIterator() {
			final IntListIterator i = IntIterators.fromTo( 0, frontCodedStringList.size() );
			
			public boolean hasNext() { return i.hasNext(); }
			public boolean hasPrevious() { return i.hasPrevious(); }
			public Object next() { return frontCodedStringList.get( permutation[ i.nextInt() ] ); }
			public Object previous() { return frontCodedStringList.get( permutation[ i.previousInt() ] ); }
			public int nextIndex() { return i.nextIndex(); }
			public int previousIndex() { return i.previousIndex(); }
		};
	}

	public static void main( String arg[] ) throws FileNotFoundException, IOException, ClassNotFoundException {

		String listName = null;
		String permutationName = null;
		String permutedListName = null;
		boolean invert = false;
		
		int ch;
		LongOpt[] longopts = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "invert", LongOpt.NO_ARGUMENT, null, 'i' ),
		};

		Getopt g = new Getopt( "PermutedFrontCodedStringList", arg, "hi", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: PermutedFrontCodedStringList [OPTIONS] LIST PERMUTATION PERMUTEDLIST" );
			System.err.println( "Builds a permuted front-coded list of strings using a given front-coded" );
			System.err.println( "string list and permutation." );
			System.err.println();
			System.err.println( "Mandatory arguments:" );
			System.err.println( "  LIST                     a front-coded string list" );
			System.err.println( "  PERMUTATION              a permutation for the indices of the list" );
			System.err.println( "  PERMUTEDLIST             the filename for the resulting permuted list" );
			System.err.println();
			System.err.println( "Optional arguments:" );
			System.err.println( "  -i, --invert             invert PERMUTATION before creating the permuted list" );
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help               print this help screen" );
			System.err.println();
			return;

		case 'i':
			invert = true;
			break;

		case '?':
			return;

		default:
			break;
		}
		
		if ( arg.length - g.getOptind() == 3 ) {
			listName = arg[ g.getOptind() ];
			permutationName = arg[ g.getOptind() + 1 ];
			permutedListName = arg[ g.getOptind() + 2 ];
		}
		else {
			System.err.println( "Wrong number (" + ( arg.length - g.getOptind() ) + ") of arguments." );
			return;
		}
		
		int[] basePermutation =  Fast.loadInts( permutationName ), permutation;
		if ( invert ) {
			int i = basePermutation.length;
			permutation = new int[ i ];
			while( i-- != 0 ) permutation[ basePermutation[ i ] ] = i;
		}
		else permutation = basePermutation;
		
		basePermutation = null;
		
		Fast.storeObject( new PermutedFrontCodedStringList( (FrontCodedStringList)Fast.loadObject( listName ), permutation ), permutedListName );
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
