package it.unimi.dsi.mg4j.util;

import it.unimi.dsi.fastutil.doubles.DoubleComparator;
import it.unimi.dsi.fastutil.doubles.DoubleHeapSemiIndirectPriorityQueue;

/** A queue of scored documents with fixed maximum capacity.
 * 
 * <P>Instances of this class contain a queue in which it possible to {@linkplain #enqueue(int, double) enqueue a document with given score}.
 * The capacity of the queue is fixed at creation time: once the queue is filled, new elements are enqueued
 * by removing those that are first w.r.t. the secondary comparator. In particular, using the 
 * {@linkplain #ScoredDocumentBoundedSizeQueue(int) constructor with implicit comparators} will
 * give a queue that remembers just the documents with highest score.
 * 
 * <P>Note that the accessible arrays {@link #document} and {@link #score} contain the document and scores corresponding to
 * the indices returned by {@link it.unimi.dsi.fastutil.doubles.DoubleHeapSesquiIndirectDoublePriorityQueue#dequeue()}.  
 */

public class ScoredDocumentBoundedSizeQueue {
	/** The underlying queue. */
	private final DoubleHeapSemiIndirectPriorityQueue queue;
	/** The maximum number of documents to be ranked. */
	private int maxSize;

	/** An array containing document scores. */
	public final double[] score;
	/** An array containing documents. */
	public final int[] document;

	/** Creates a new empty bounded-size queue with a given capacity and natural order as primary comparator.
	 *
	 * <P>This constructor uses as secondary comparator the opposite of the natural order.
	 *
	 * @param capacity the initial capacity of this queue.
	 */
	public ScoredDocumentBoundedSizeQueue( final int capacity ) {
		score = new double[ capacity ];
		maxSize = capacity;
		document = new int[ maxSize ];
		queue = new DoubleHeapSemiIndirectPriorityQueue( score, capacity  );
	}

	/** Creates a new empty bounded-size queue with a given capacity.
	 *
	 * @param capacity the initial capacity of this queue.
	 * @param c the comparator used in this queue, or <code>null</code> for the natural order.
	 */
	public ScoredDocumentBoundedSizeQueue( final int capacity, final DoubleComparator c ) {
		score = new double[ capacity ];
		maxSize = capacity;
		document = new int[ maxSize ];
		queue = new DoubleHeapSemiIndirectPriorityQueue( score, capacity, c );
	}

	/** Enqueues a document with given score.
	 * 
	 * @param d the document to enqueue.
	 * @param s its score.
	 */

	public void enqueue( int d, double s ) {
		int currentSize = queue.size();
		if ( currentSize >= maxSize ) {
			final int smallestScoreIndex = queue.first();

			if ( d > score[ smallestScoreIndex ] ) {
				document[ smallestScoreIndex ] = d;
				score[ smallestScoreIndex ] = s;
				queue.changed();
			}
		} else {
			document[ currentSize ] = d;
			score[ currentSize ] = s;
			queue.enqueue( currentSize );
		}
	}

	public boolean isEmpty() {
		return queue.isEmpty();
	}

	public int size() {
		return queue.size();
	}

	/** Dequeues an index from the queue.
	 * 
	 * The actual document and score may be retrieved by peeking into
	 * {@link #document} and {@link #score} using the index returned by this method.
	 * Documents are dequeued in inverse score order.
	 * @return the next index to be dequeued.
	 */

	public int dequeue() {
		return queue.dequeue();
	}
}
