package it.unimi.dsi.mg4j.util;

import java.io.Serializable;
import java.util.Collection;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2002, 2003, 2004, 2004 Sebastiano Vigna and Marco Olivo 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 2.1 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/** Signed order-preserving minimal perfect hash tables.
 *
 * <blockquote><strong>IMPORTANT</strong>: In MG4J 1.0, the serialisation
 * format is incompatible with previous versions.</blockquote>
 * 
 * <P>Minimal perfect hash tables will always return a result, even for terms that were
 * not present in the collection indexed by the table. Sometimes you may prefer to single out
 * terms that were not present in the collection.
 *
 * <P>To this purpose, MG4J provides <em>signed</em> minimal perfect tables. In a signed
 * table, every term in the collection gets a <em>signature</em> that is used to 
 * tell false positives. Signature may go from the simple, CRC based signatures
 * provided by the sample {@link CRC32SignedMinimalPerfectHash} and 
 * {@link HashCodeSignedMinimalPerfectHash} classes, to sophisticated
 * cryptographic signatures, to (at the other extreme) a class that actually stores
 * the terms (and thus completely avoids false positives) such as {@link LiterallySignedMinimalPerfectHash}.
 *
 * <P>A signed table extends this class, and provides two methods: a {@link
 * #initSignatures(Collection)} method that sets up the necessary data
 * structures, and a {@link #checkSignature(CharSequence,int)} method that
 * checks a given character sequence against the signature stored for
 * a term having given index.
 *
 * <P>It is good practise, of course, to replicate the constructors of this
 * class in all implementing subclasses (by simply invoking <code>super</code>
 * with the same arguments). Moreover, to be useful classes implementing this
 * class <em>must</em> be serialisable.
 *
 * @author Sebastiano Vigna
 * @author Marco Olivo
 * @since 0.4
 */

public abstract class SignedMinimalPerfectHash extends MinimalPerfectHash implements Serializable {

	public static final long serialVersionUID = -7046029254386353130L;

	/** Creates a new signed order-preserving minimal perfect hash table for the given
	 * set of terms, using as many weights as the longest term in the collection.
	 *
	 * <P>After calling the corresponding constructor of {@link MinimalPerfectHash}, this
	 * constructor will invoke {@link #initSignatures(Collection)}.
	 *
	 * @param terms some terms to hash; it is assumed that this {@link Collection} does not contain duplicates.
	 * @see MinimalPerfectHash#MinimalPerfectHash(Collection)
	 */
	public SignedMinimalPerfectHash( Collection terms ) {
		super( terms );
		initSignatures( this.terms );
	}


	/** Creates a new signed order-preserving minimal perfect hash table for the given set of terms using the given number of weights.
	 *
	 * <P>After calling the corresponding constructor of {@link MinimalPerfectHash}, this
	 * constructor will invoke {@link #initSignatures(Collection)}.
	 *
	 * @param terms some terms to hash; it is assumed that this {@link
	 * Collection} does not contain terms with a common prefix of
	 * <code>weightLength</code> characters.
	 * @param weightLength the number of weights used generating the intermediate hash functions.
	 * @see MinimalPerfectHash#MinimalPerfectHash(Collection,int)
	 */
	public SignedMinimalPerfectHash( Collection terms, int weightLength ) {
	    super( terms, weightLength );
		initSignatures( this.terms );
	}

	/** Creates a new CRC32-signed order-preserving minimal perfect hash table for the given file 
	 * of terms using the given number of weights.
	 *
	 * <P>After calling the corresponding constructor of {@link MinimalPerfectHash}, this
	 * constructor will invoke {@link #initSignatures(Collection)}.
	 *
	 * @param termFile an UTF-8 file containing one term on each line; it is assumed that 
	 * it does not contain terms with a common prefix of
	 * <code>weightLength</code> characters.
	 * @param encoding the encoding of <code>termFile</code>; if <code>null</code>, it
	 * is assumed to be the platform default encoding.
	 * @param weightLength the number of weights used generating the
	 * intermediate hash functions.
	 * @see MinimalPerfectHash#MinimalPerfectHash(String,String,int)
	 */
	public SignedMinimalPerfectHash(String termFile, String encoding, int weightLength) {
		super(termFile, encoding, weightLength);
		initSignatures( terms );
	}


	/** Creates a new CRC32-signed order-preserving minimal perfect hash table for
	 * the given file of terms, using as many weights as the longest term in the file.
	 *
	 * <P>After calling the corresponding constructor of {@link MinimalPerfectHash}, this
	 * constructor will invoke {@link #initSignatures(Collection)}.
	 *
	 * @param termFile a file in the platform default encoding containing one term on each line; it is assumed that the file does
	 * not contain twice the same term.
	 * @see MinimalPerfectHash#MinimalPerfectHash(String)
	 */
	public SignedMinimalPerfectHash( String termFile ) {
		super( termFile );
		initSignatures( terms );
	}




	/** Hashes a given term.
	 *
	 * @param term a term to hash.
	 * @return the position of the given term in the generating collection, starting from 0, if the
	 * term was in the original collection; otherwise, -1.
	 */
	public int get( CharSequence term ) {
		final int i = super.get( term );
		if ( checkSignature( term, i ) ) return i;
		else return -1;
	}


	/** Sets up the signature system from a collection.
	 *
	 * <P>This abstract method must be overriden by implementing subclasses.
	 * It must set up all data structures that are necessary to handle
	 * signatures; in particular, it will usually compute signatures for
	 * all terms in the given collection.
	 *
	 * @param terms the collection of terms given to the constructor of this class.
	 * @see CRC32SignedMinimalPerfectHash#initSignatures(Collection)
	 * @see LiterallySignedMinimalPerfectHash#initSignatures(Collection)
	 */
	protected abstract void initSignatures( Collection terms ) ;


	
	/** Checks a signature.
	 *
	 * <P>This abstract method must be overriden by implementing subclasses.
	 * It must check whether the signature of the given character sequence matches
	 * the one stored for the <code>index</code>-th term.
	 *
	 * @param term a character sequence.
	 * @param index an integer denoting a term in the indexed collection.
	 * @return true iff the signature of the given character sequence matches
	 * the one stored for the <code>index</code>-th term.
	 * @see CRC32SignedMinimalPerfectHash#checkSignature(CharSequence,int)
	 * @see LiterallySignedMinimalPerfectHash#checkSignature(CharSequence,int)
	 */
	protected abstract boolean checkSignature( CharSequence term, int index );

}
